﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class CogForm : UserControl
    {
        public CogForm()
        {
            InitializeComponent();

            
        }

        public void UpdateCOG(int n)
        {
            if (n < 0 || n >= LVCOGs.Items.Count)
                return;

            ListViewItem lvi = LVCOGs.Items[n];
            JKCOG cog = GlobalVars.Level.cogs[n];

            lvi.Text = cog.name;
            string s = string.Empty;
            foreach (COGValue cv in cog.vals)
                s += " " + cv.name + "=" + cv.AsString;
            lvi.SubItems[1].Text = s;
        }

        public void RefreshList()
        {
            Updating = true;
            try
            {
                LVCOGs.BeginUpdate();
                LVCOGs.Items.Clear();
                for (int i = 0; i < GlobalVars.Level.cogs.Count; i++)
                {
                    JKCOG cog = GlobalVars.Level.cogs[i];

                    ListViewItem lvi = LVCOGs.Items.Add(i.ToString() + ": " + cog.name);
                    string s = string.Empty;
                    foreach (COGValue cv in cog.vals)
                        s += " " + cv.name + "=" + cv.AsString;
                    lvi.SubItems.Add(s);
                }
                LVCOGs.EndUpdate();


                coggroups.BeginUpdate();
                coggroups.Items.Clear();
                coggroups.Items.Add(new CogGroup("All"));
                foreach (CogGroup cg in GlobalVars.Level.coggroups)
                    coggroups.Items.Add(cg);
                coggroups.EndUpdate();
            }
            finally
            {
                Updating = false;
            }

            LVCOGs.SelectedIndices.Clear();
            if (LVCOGs.Items.Count > 0)
            {
                LVCOGs.SelectedIndices.Add(0);
            }
            else
            {
                LVCOGs_SelectedIndexChanged(LVCOGs, new EventArgs());
            }
        }

        public void SelectGroupAll()
        {
            SelectGroup("All");
        }

        public void SelectGroup(string nm)
        {
            int i = GroupIndex(nm);
            if(i != -1)
            {
                coggroups.SetItemChecked(i, true);
                return;
            }

            // default to all? or do nothing

        }

        bool Updating = false;

        public new void Show()
        {
            if (IsHandleCreated)//Visible)
            {
                //BringToFront();
                if(GlobalVars.NedMain != null)
                    GlobalVars.NedMain.tabControl3.SelectedTab = GlobalVars.NedMain.tabCogs;
                return;
            }

            base.Show();
        }

        private void LVCOGs_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (Updating)
                return;

            SGVals.Rows.Clear();

            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            if (n < 0 || n >= GlobalVars.Level.cogs.Count)
                return;

            JKCOG cog = GlobalVars.Level.cogs[n];

            Text = "Placed Cogs - COG " + n.ToString() + ": " + cog.name;
            for(int i=0; i<cog.vals.Count; i++)
            {
                COGValue cv = cog.vals[i];

                DataGridViewRow row = new DataGridViewRow();
                row.Tag = i;
                row.CreateCells(SGVals, new string[] { cv.name + " (" + Values.GetCogTypeName(cv.cogType) + ")", cv.AsJedString });
                SGVals.Rows.Add(row);
            }
        }

        private void BNRefresh_Click(object sender, EventArgs e)
        {
            try
            {
                foreach (JKCOG cog in GlobalVars.Level.cogs)
                {
                    List<string> vvals = new List<string>();
                    List<string> vnames = new List<string>();

                    for (int j = cog.vals.Count - 1; j >= 0; j--)
                    {
                        COGValue v = cog.vals[j];
                        vnames.Add(v.name);
                        vvals.Add(v.AsString);
                        cog.vals.RemoveAt(j);
                    }

                    CogFile cf = new CogFile();
                    cf.LoadNoLocals(cog.name);
                    for (int j = 0; j < cf.Count; j++)
                    {
                        COGValue v = new COGValue();
                        cog.vals.Add(v);
                        COGValue vnew = cf[j];
                        v.Assign(vnew);

                        int k = -1;
                        for(int x=0; x<vnames.Count; x++)
                            if (vnames[x].Equals(vnew.name, StringComparison.InvariantCultureIgnoreCase))
                            {
                                k = x;
                                break;
                            }

                        if (k != -1)
                            v.Val(vvals[k]);
                        else
                            v.Val(vnew.AsString);

                        v.Resolve();
                    }
                }
            }
            finally
            {
                RefreshList();
            }
        }

        private void SGVals_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            JKCOG cog = GlobalVars.Level.cogs[n];

            if (e.RowIndex < 0 || e.RowIndex >= SGVals.Rows.Count)
                return;

            DataGridViewRow row = SGVals.Rows[e.RowIndex];
            string val = row.Cells[1].Value as string;
            COGValue cv = cog.vals[(int)row.Tag];
            if (!cv.JedVal(val))
                Utils.PanMessage(Utils.MessageType.Error, "Invalid value for " + Values.GetCogTypeName(cv.cogType) + ": " + val);
        }

        public int AddCOG(string fname)
        {
            return AddCogWithVals(fname, null);
        }

        public int AddCogWithVals(string fname, JKCOG fromcog)
        {
            CogFile cf = new CogFile();
            cf.LoadNoLocals(fname);

            JKCOG cg = new JKCOG();
            cg.name = fname;
            for (int i = 0; i < cf.Count; i++)
            {
                COGValue cv = cf[i];
                COGValue v = new COGValue();
                v.Assign(cv);
                v.Val(cv.AsString);
                cg.vals.Add(v);
            }

            if (fromcog != null)
            {
                for (int i = 0; i < cg.vals.Count; i++)
                {
                    COGValue v = cg.vals[i];
                    int n = fromcog.vals.IndexOfName(v.name);
                    if (n != -1)
                        v.JedVal(fromcog.vals[n].AsJedString);
                }
            }

            int ret = GlobalVars.Level.cogs.Count;
            GlobalVars.Level.cogs.Add(cg);

            Updating = true;
            ListViewItem lvi = LVCOGs.Items.Add(cg.name);
            string s = string.Empty;
            foreach (COGValue cv in cg.vals)
                s += " " + cv.name + "=" + cv.AsString;
            lvi.SubItems.Add(s);
            Updating = false;

            BNRefresh_Click(BNRefresh, new EventArgs());

            return ret;
        }

        private void SGVals_CellDoubleClick(object sender, DataGridViewCellEventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            JKCOG cog = GlobalVars.Level.cogs[n];

            if (e.RowIndex < 0 || e.RowIndex >= SGVals.Rows.Count)
                return;

            DataGridViewRow row = SGVals.Rows[e.RowIndex];
            string val = row.Cells[1].Value as string;
            COGValue cv = cog.vals[(int)row.Tag];

            switch (cv.cogType)
            {
                case COGType.Unknown:
                    break;

                case COGType.AI:
                    ResPicker.PickAI(ref val);
                    break;

                case COGType.COG:
                    {
                        int i = int.Parse(val);
                        if (i < 0)
                        {
                            string fname = string.Empty;
                            ResPicker.PickCOG(ref fname);
                            if (!string.IsNullOrEmpty(fname))
                                val = AddCOG(fname).ToString();
                        }
                        else
                            GotoCOG(i);
                    }
                    break;

                case COGType.KEY:
                    ResPicker.PickKEY(ref val);
                    break;

                case COGType.MAT:
                    ResPicker.PickMAT(ref val, GlobalVars.Level.GetMasterCMP());
                    break;

                case COGType.Msg:
                    break;

                case COGType.Model:
                    ResPicker.Pick3DO(ref val);
                    break;

                case COGType.Sector:
                    {
                        int i = int.Parse(val);
                        if (n >= 0)
                            GlobalVars.NedMain.GotoSC(n);
                    }
                    break;

                case COGType.Wav:
                    ResPicker.PickSecSound(ref val);
                    break;

                case COGType.Surface:
                    {
                        string[] parts = val.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                        if (parts.Length == 2)
                        {
                            int i = int.Parse(parts[0]);
                            int p = int.Parse(parts[1]);
                            if (i >= 0)
                                GlobalVars.NedMain.GotoSF(i, p);
                        }
                    }
                    break;

                case COGType.Template:
                    ResPicker.PickTemplate(ref val);
                    break;

                case COGType.Thing:
                    {
                        int p = int.Parse(val);
                        if (p >= 0)
                            GlobalVars.NedMain.GotoTH(p);
                    }
                    break;
            }

            row.Cells[1].Value = val;// this should hopefully kick off the callback to actually update cogvalue
            SGVals.RefreshEdit();
        }

        private void BNAdd_Click(object sender, EventArgs e)
        {
            string fname = string.Empty;
            ResPicker.PickCOG(ref fname);
            if (string.IsNullOrEmpty(fname))
                return;

            GotoCOG(AddCOG(fname));
        }

        private void BNDelete_Click(object sender, EventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            Utils.DeleteCOG(GlobalVars.Level, n);
            if (n >= LVCOGs.Items.Count)
                n = LVCOGs.Items.Count - 1;
            GotoCOG(n);
        }

        public void GotoCOG(int n)
        {
            if (n < 0 || n >= GlobalVars.Level.cogs.Count)
                return;

            // check if cog is in current group, and if not.. find custom one that its in?  else select All
            Log.Info("GotoCOG NEED TO SELECT COG GROUP");

            LVCOGs.SelectedIndices.Clear();
            LVCOGs.SelectedIndices.Add(n);
            Show();
        }

        private void BNDuplicate_Click(object sender, EventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            GotoCOG(AddCogWithVals(GlobalVars.Level.cogs[n].name, GlobalVars.Level.cogs[n]));
        }

        private void BNGetSel_Click(object sender, EventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            JKCOG cog = GlobalVars.Level.cogs[n];

            if (SGVals.SelectedRows.Count != 1)
                return;

            int rowIndex = SGVals.SelectedRows[0].Index;

            if (rowIndex < 0 || rowIndex >= SGVals.Rows.Count)
                return;

            DataGridViewRow row = SGVals.Rows[rowIndex];
            //string val = row.Cells[1].Value as string;
            COGValue cv = cog.vals[(int)row.Tag];

            switch (cv.cogType)
            {
                case COGType.Thing:
                case COGType.Sector:
                case COGType.Surface:
                    {
                        object obj = GlobalVars.NedMain.GetCurObjForCog(cv.cogType);
                        if (obj != null)
                            cv.obj = obj;
                        UpdateCOG(n);
                    }
                    break;

                default:
                    MessageBox.Show("It only works for sector/surface and thing entries");
                    break;
            }

            row.Cells[1].Value = cv.AsString;
        }

        private void BNEdit_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                MessageBox.Show("Save project first");
                return;
            }

            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            JKCOG cog = GlobalVars.Level.cogs[n];

            string fname = string.Empty;
            string testPath = Path.Combine(GlobalVars.ProjectDir, cog.name);
            if (File.Exists(testPath))
                fname = testPath;
            else
            {
                testPath = Path.Combine(Path.Combine(GlobalVars.ProjectDir, "cog"), cog.name);
                if (File.Exists(testPath))
                    fname = testPath;
            }

            if (string.IsNullOrEmpty(fname))
            {
                if (MessageBox.Show("The COG is inside a GOB file. Extract it to edit?", "Question", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                    return;

                string dstPath = Path.Combine(GlobalVars.ProjectDir, "cog");
                if(!Directory.Exists(dstPath))
                    Directory.CreateDirectory(dstPath);

                string outname = cog.name;
                if (Utils.InputBox("Rename", "Want to rename file?", ref outname) != DialogResult.OK)
                    return;

                fname = Path.Combine(dstPath, outname);

                using (BaseFile src = FileOperations.OpenGameFile(cog.name))
                using (BaseFile dst = FileOperations.OpenFileWrite(fname))
                    dst.Write(src.Read(src.Length));


                // update cog ref
                cog.name = outname;

                RefreshList();
            }

            System.Diagnostics.ProcessStartInfo psi = new System.Diagnostics.ProcessStartInfo();
            psi.UseShellExecute = true;
            psi.FileName = fname;
            psi.WorkingDirectory = Path.GetDirectoryName(fname);
            psi.ErrorDialogParentHandle = GlobalVars.NedMain.Handle;
            System.Diagnostics.Process.Start(psi);
        }

        private void BNNew_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                MessageBox.Show("Save project first");
                return;
            }

            string s = string.Empty;
            Utils.InputBox("New COG", "Enter filename", ref s);
            s = Path.GetFileNameWithoutExtension(s);
            if (string.IsNullOrEmpty(s))
                return;

            s += ".cog";

            string cogfolder = Path.Combine(GlobalVars.ProjectDir, "cog");
            if (!Directory.Exists(cogfolder))
                Directory.CreateDirectory(cogfolder);

            string cogname = Path.Combine(cogfolder, s);

            using (StreamWriter sw = File.CreateText(cogname))
            {
                sw.WriteLine("symbols");
                sw.WriteLine();
                sw.WriteLine("end");
                sw.WriteLine();
                sw.WriteLine("code");
                sw.WriteLine();
                sw.WriteLine("end");
            }

            LVCOGs.SelectedIndices.Clear();
            LVCOGs.SelectedIndices.Add(AddCOG(s));

            System.Diagnostics.ProcessStartInfo psi = new System.Diagnostics.ProcessStartInfo();
            psi.UseShellExecute = true;
            psi.FileName = cogname;
            psi.WorkingDirectory = cogfolder;
            psi.ErrorDialogParentHandle = GlobalVars.NedMain.Handle;
            System.Diagnostics.Process.Start(psi);
        }

        private void CogForm_Load(object sender, EventArgs e)
        {

        }

        private void BNView_Click(object sender, EventArgs e)
        {
            if (LVCOGs.SelectedIndices.Count != 1)
                return;

            int n = LVCOGs.SelectedIndices[0];

            JKCOG cog = GlobalVars.Level.cogs[n];

            string fname = Path.Combine(Path.GetTempPath(), Path.GetFileName(cog.name));

            using (BaseFile src = FileOperations.OpenGameFile(cog.name))
            using (BaseFile dst = FileOperations.OpenFileWrite(fname))
                dst.Write(src.Read(src.Length));


            System.Diagnostics.ProcessStartInfo psi = new System.Diagnostics.ProcessStartInfo();
            psi.UseShellExecute = true;
            psi.FileName = fname;
            psi.WorkingDirectory = Path.GetDirectoryName(fname);
            psi.ErrorDialogParentHandle = GlobalVars.NedMain.Handle;
            System.Diagnostics.Process.Start(psi);
        }

        private void checkedListBox1_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (e.NewValue == CheckState.Checked)
            {
                for (int i = 0; i < coggroups.Items.Count; i++)
                    if(i != e.Index)
                        coggroups.SetItemChecked(i, false);
            }
        }

        private void newGroup_Click(object sender, EventArgs e)
        {
            string nm = "";
            if (Utils.InputBox("Group Name", "Enter name for new group", ref nm) != DialogResult.OK)
                return;

            if (string.IsNullOrEmpty(nm))
                return;

            if (GlobalVars.Level.coggroups.GetGroup(nm) != null)
                return;

            GlobalVars.Level.coggroups.Add(new CogGroup(nm));

            RefreshList();
        }

        private void addGroupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string nm = string.Empty;
            if (Utils.InputBox("Group Name", "Enter COG group name", ref nm) != DialogResult.OK)
                return;

            AddGroup(nm);
        }

        private void renameGroupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            CogGroup cg = contextMenuGroup;
            if (cg == null)
                return;

            string nm = cg.Name;
            if (Utils.InputBox("Group Name", "Enter COG group name", ref nm) != DialogResult.OK)
                return;

            cg.Name = nm;

            RefreshGroups();
        }

        int GroupIndex(string nm)
        {
            for (int i = 0; i < coggroups.Items.Count; i++)
            {
                CogGroup cg = coggroups.Items[i] as CogGroup;

                if (cg.Name == nm)
                    return i;
            }

            return -1;
        }

        CogGroup GroupByIndex(int i)
        {
            if (i < 0 || i >= coggroups.Items.Count)
                return null;

            return coggroups.Items[i] as CogGroup;
        }

        void AddGroup(string nm)
        {

        }

        void RefreshGroups()
        {
            // this prolly dont work.. gotta rebuild the list and keep selection and shit probably
            coggroups.Refresh();
        }

        int contextMenuIndex = -1;
        CogGroup contextMenuGroup
        {
            get
            {
                return GroupByIndex(contextMenuIndex);
            }
        }

        private void coggroups_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
                contextMenuIndex = coggroups.IndexFromPoint(e.Location);
        }
    }
}
