﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class EpisodeEdit : UserControl
    {
        public new void Show()
        {
            if(GlobalVars.NedMain != null)
                GlobalVars.NedMain.tabControl1.SelectedTab = GlobalVars.NedMain.tabEpisode;

            RefreshEpisode();
        }

        public class COGStrings : Dictionary<string, string>
        {
            public COGStrings()
                : base(StringComparer.InvariantCultureIgnoreCase)
            {

            }

            public void LoadFromFile(string name)
            {
                Stream stream = null;
                try
                {
                    using (BaseFile bf = FileOperations.OpenFileRead(name))
                        stream = new MemoryStream(bf.Read(bf.Length));
                }
                catch
                {

                }

                LoadFromStream(stream);
            }

            public void LoadFromStream(Stream stream)
            {
                if (stream == null)
                    return;

                try
                {
                    Tokenizer t = new Tokenizer(stream);

                    t.EatLine();

                    while (!t.EndOfStream)
                    {
                        string tok = t.ReadToken();
                        if (string.IsNullOrEmpty(tok) || tok.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                            break;

                        t.EatToken();

                        string txt = t.ReadLine(false, true);

                        if (tok.Length > 2)
                            tok = tok.Substring(1, tok.Length - 2);
                        else
                            continue;//need valid key

                        if (txt.Length > 2)
                            txt = txt.Substring(1, txt.Length - 2);
                        else
                            txt = string.Empty;

                        Add(tok, txt);
                    }
                }
                catch
                {

                }
            }

            public void SaveToFile(string name)
            {
                if (Count == 0)
                {
                    if (File.Exists(name))
                        File.Delete(name);
                    return;
                }

                Directory.CreateDirectory(Path.GetDirectoryName(name));

                using (StreamWriter sw = File.CreateText(name))
                {
                    sw.WriteLine("MSGS " + Count.ToString());
                    sw.WriteLine();
                    sw.WriteLine("#  \"<key>\"     <unused number>   \"<string>\"");
                    sw.WriteLine();
                    foreach (string key in Keys)
                        sw.WriteLine("\"" + key + "\" 0 \"" + this[key] + "\"");
                    sw.WriteLine();
                    sw.WriteLine("END");
                }
            }

            public void DeleteString(string key)
            {
                if (ContainsKey(key))
                    Remove(key);
            }

            public string GetString(string key)
            {
                if (!ContainsKey(key))
                    return string.Empty;

                return this[key];
            }

            public void AddString(string key, string str)
            {
                if (ContainsKey(key))
                {
                    this[key] = str;
                    return;
                }

                Add(key.ToUpper(), str);
            }
        }

        class JKRec
        {
            public int lnum;
            public int cdnum;
            public int levnum;
            public string ltype;
            public string fname;
            public int lpow;
            public int dpow;
            public int gotoA;
            public int gotoB;
            //public string comment;
        }

        COGStrings cstrings = new COGStrings();
        List<JKRec> Recs = new List<JKRec>();

        public void UpdateLevelFile(string lastName, string newName)
        {
            lastName = FileOperations.ChangeFileExt(FileOperations.ExtractName(lastName), ".jkl");
            newName = FileOperations.ChangeFileExt(FileOperations.ExtractName(newName), ".jkl");

            // if new and old are same, do nothing
            if (lastName.Equals(newName, StringComparison.InvariantCultureIgnoreCase))
            {
                // if our current only record is the default one.. then maybe lets swap it out for this guy and say all is good
                if (Recs.Count == 1 && Recs[0].fname == "Untitled.jkl")
                    Recs[0].fname = newName;

                return;
            }

            bool inRecs = false;
            foreach(JKRec rec in Recs)
                if (rec.fname.Equals(lastName, StringComparison.InvariantCultureIgnoreCase))
                {
                    rec.fname = newName;
                    inRecs = true;
                    break;
                }

            // if not in the list, add?
            if (!inRecs)
                AddLevelEntry(newName);

            RefreshEditor();
        }

        public EpisodeEdit()
        {
            InitializeComponent();
        }

        void RefreshEditor()
        {
            LVSeqs.Items.Clear();
            foreach (JKRec rec in Recs)
            {
                ListViewItem lvi = LVSeqs.Items.Add(rec.lnum.ToString() + ":");
                lvi.SubItems.Add(rec.fname);
                lvi.SubItems.Add(rec.levnum.ToString());
                lvi.SubItems.Add(rec.gotoA.ToString());
                lvi.SubItems.Add(rec.gotoB.ToString());
            }
        }

        void LoadEpJK(Stream stream)
        {
            Recs.Clear();

            try
            {
                Tokenizer t = new Tokenizer(stream);

                string ln = t.ReadLine();
                if (ln.Length <= 2)
                    ln = string.Empty;
                else
                    ln = ln.Substring(1, ln.Length - 2);
                EBEpName.Text = ln;

                t.EatToken();
                CBGameType.SelectedIndex = (int)(Math.Log((double)t.ReadInt()) / Math.Log(2.0));

                t.EatToken();
                int numseq = t.ReadInt();
                for (int x = 0; x < numseq; x++)
                {
                    JKRec rec = new JKRec();

                    string ltok = t.ReadToken();
                    ltok = ltok.Substring(0, ltok.Length - 1);
                    rec.lnum = int.Parse(ltok);

                    rec.cdnum = t.ReadInt();
                    rec.levnum = t.ReadInt();
                    rec.ltype = t.ReadToken();
                    rec.fname = t.ReadToken();
                    rec.lpow = t.ReadInt();
                    rec.dpow = t.ReadInt();
                    rec.gotoA = t.ReadInt();
                    rec.gotoB = t.ReadInt();

                    Recs.Add(rec);
                }
            }
            catch
            {
                CreateNewEpisode();
            }
        }

        void CreateNewEpisode()
        {
            Recs.Clear();

            EBEpName.Text = "New Level";
            CBGameType.SelectedIndex = 0;

            JKRec rec = new JKRec();

            rec.lnum = 10;
            rec.cdnum = 0;
            rec.levnum = 1;
            rec.ltype = "LEVEL";
            rec.fname = FileOperations.ChangeFileExt(FileOperations.ExtractName(GlobalVars.NedMain.LevelFile), ".jkl");
            rec.lpow = 0;
            rec.dpow = 0;
            rec.gotoA = -1;
            rec.gotoB = -1;

            Recs.Add(rec);


            RefreshEditor();
            LVSeqs.SelectedIndices.Clear();
            LVSeqs.SelectedIndices.Add(0);
        }

        void SaveEpJK(string name)
        {
            using (StreamWriter sw = File.CreateText(name))
            {
                sw.WriteLine("\"" + EBEpName.Text + "\"");
                sw.WriteLine();
                sw.WriteLine("TYPE " + (1<<CBGameType.SelectedIndex).ToString());
                sw.WriteLine();
                sw.WriteLine("SEQ " + Recs.Count.ToString());
                sw.WriteLine();
                sw.WriteLine("# <line> <cd>  <level>  <type>   <file>         <lightpow>  <darkpow>   <gotoA>  <gotoB>");
                sw.WriteLine();
                foreach (JKRec rec in Recs)
                    sw.WriteLine(rec.lnum.ToString() + ":\t" + rec.cdnum.ToString() + "\t" + rec.levnum.ToString() + "\t" + rec.ltype + "\t" + rec.fname + "\t" + rec.lpow.ToString() + "\t" + rec.dpow.ToString() + "\t" + rec.gotoA.ToString() + "\t" + rec.gotoB.ToString());
                sw.WriteLine();
                sw.WriteLine("end");
            }
        }

        private void EpisodeEdit_Load(object sender, EventArgs e)
        {
            RefreshEpisode();
        }

        public void RefreshEpisode()
        {
            if (DesignMode)
                return;

            if (string.IsNullOrEmpty(GlobalVars.ProjectDir) &&
                string.IsNullOrEmpty(GlobalVars.ProjectGob))
            {
                CreateNewEpisode();
                return;
            }

            using (BaseFile cuni = FileOperations.TryProjectFile(string.Empty, "cogstrings.uni") ??
                            FileOperations.TryProjectFile("misc", "cogstrings.uni") ??
                            FileOperations.TryGOBFile(GlobalVars.ProjectGob, "misc", "cogstrings.uni"))
            {
                cstrings.Clear();
                if (cuni != null)
                    cstrings.LoadFromStream(cuni.GetStream());
            }

            using (BaseFile jk = FileOperations.TryProjectFile(string.Empty, "episode.jk") ??
                                FileOperations.TryGOBFile(GlobalVars.ProjectGob, string.Empty, "episode.jk"))
            {
                Recs.Clear();
                if(jk != null)
                    LoadEpJK(jk.GetStream());
            }
            RefreshEditor();
            LVSeqs.SelectedIndices.Clear();
            if(LVSeqs.Items.Count > 0)
                LVSeqs.SelectedIndices.Add(0);
        }

        public void Save()
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
                return;

            SaveEpJK(Path.Combine(GlobalVars.ProjectDir, "episode.jk"));

            string olduni = Path.Combine(GlobalVars.ProjectDir, "cogstrings.uni"/*shouldnt this be in misc?*/);
            if (File.Exists(olduni))
                File.Delete(olduni);

            cstrings.SaveToFile(Path.Combine(GlobalVars.ProjectDir, @"misc\cogstrings.uni"));
        }

        private void BNSave_Click(object sender, EventArgs e)
        {
            Save();
        }

        int IndexOfLPow(int pow)
        {
            for (int x = 0; x < CBLPow.Items.Count; x++)
            {
                string s = CBLPow.Items[x] as string;
                if (int.Parse(s.Substring(0, s.IndexOf(' '))) == pow)
                    return x;
            }

            return -1;
        }

        int IndexOfDPow(int pow)
        {
            for (int x = 0; x < CBDPow.Items.Count; x++)
            {
                string s = CBDPow.Items[x] as string;
                if (int.Parse(s.Substring(0, s.IndexOf(' '))) == pow)
                    return x;
            }

            return -1;
        }

        private void LVSeqs_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (LVSeqs.SelectedIndices.Count != 1)
                return;

            int index = LVSeqs.SelectedIndices[0];
            if (index < 0 || index >= Recs.Count)
                return;

            JKRec rec = Recs[index];

            EBLnum.Text = rec.lnum.ToString();
            EBCDNum.Text = rec.cdnum.ToString();
            EBLevNum.Text = rec.levnum.ToString();
            EBGotoA.Text = rec.gotoA.ToString();
            EBGotoB.Text = rec.gotoB.ToString();
            CBLPow.SelectedIndex = IndexOfLPow(rec.lpow);
            CBDPow.SelectedIndex = IndexOfDPow(rec.dpow);
            CBType.Text = rec.ltype;
            EBFname.Text = rec.fname;

            string f = FileOperations.ChangeFileExt(rec.fname, string.Empty);
            EBLevName.Text = cstrings.GetString(f);

            MMLevText.Text = cstrings.GetString(f + "_TEXT_00") + "\n" + cstrings.GetString(f + "_TEXT_01");
        }

        void AddLevelEntry(string fname)
        {
            if (LVSeqs.SelectedIndices.Count != 1)
                return;

            int i = LVSeqs.SelectedIndices[0];

            JKRec jklr = Recs[i];

            JKRec rec = new JKRec();
            rec.lnum = jklr.lnum + 1;
            rec.cdnum = 0;
            rec.levnum = jklr.levnum + 1;
            rec.ltype = "LEVEL";
            rec.fname = FileOperations.ChangeFileExt(FileOperations.ExtractName(fname), ".jkl");
            rec.lpow = 0;
            rec.dpow = 0;
            rec.gotoA = -1;
            rec.gotoB = -1;

            Recs.Insert(i + 1, rec);

            ListViewItem lvi = LVSeqs.Items.Insert(i + 1, rec.lnum.ToString() + ":");
            lvi.SubItems.Add(rec.fname);
            lvi.SubItems.Add(rec.levnum.ToString());
            lvi.SubItems.Add(rec.gotoA.ToString());
            lvi.SubItems.Add(rec.gotoB.ToString());

            LVSeqs.SelectedIndices.Clear();
            LVSeqs.SelectedIndices.Add(i + 1);
        }

        private void BNAdd_Click(object sender, EventArgs e)
        {
            AddLevelEntry(GlobalVars.NedMain.LevelFile);
        }

        private void BNDelete_Click(object sender, EventArgs e)
        {
            if (LVSeqs.SelectedIndices.Count != 1)
                return;

            int i = LVSeqs.SelectedIndices[0];

            if (Recs.Count <= 1)
            {
                MessageBox.Show("Only one left!", "Episode Editor", MessageBoxButtons.OK);
                return;
            }

            Recs.RemoveAt(i);
            LVSeqs.Items.RemoveAt(i);
            if (i > 0)
            {
                LVSeqs.SelectedIndices.Clear();
                LVSeqs.SelectedIndices.Add(i - 1);
            }
        }

        private void BNUpdate_Click(object sender, EventArgs e)
        {
            if (LVSeqs.SelectedIndices.Count != 1)
                return;

            int i = LVSeqs.SelectedIndices[0];

            JKRec rec = Recs[i];

            int.TryParse(EBLnum.Text, out rec.lnum);
            int.TryParse(EBCDNum.Text, out rec.cdnum);
            int.TryParse(EBLevNum.Text, out rec.levnum);
            int.TryParse(EBGotoA.Text, out rec.gotoA);
            int.TryParse(EBGotoB.Text, out rec.gotoB);
            if (CBLPow.Text.IndexOf(' ') == -1)
                rec.lpow = 0;
            else
                int.TryParse(CBLPow.Text.Substring(0, CBLPow.Text.IndexOf(' ')), out rec.lpow);
            if (CBDPow.Text.IndexOf(' ') == -1)
                rec.dpow = 0;
            else
                int.TryParse(CBDPow.Text.Substring(0, CBDPow.Text.IndexOf(' ')), out rec.dpow);
            rec.ltype = CBType.Text;
            rec.fname = EBFname.Text;

            string f = FileOperations.ChangeFileExt(rec.fname, string.Empty);

            cstrings.AddString(f, EBLevName.Text);

            cstrings.AddString(f + "_TEXT_00", MMLevText.Lines[0]);
            string d = string.Empty;
            for (int x = 1; x < MMLevText.Lines.Length; x++)
            {
                d += MMLevText.Lines[x];
                if (x < (MMLevText.Lines.Length - 1))
                    d += " ";
            }
            cstrings.AddString(f + "_TEXT_01", d);


            ListViewItem lvi = LVSeqs.Items[i];
            lvi.SubItems.Clear();
            lvi.Text = rec.lnum.ToString() + ":";
            lvi.SubItems.Add(rec.fname);
            lvi.SubItems.Add(rec.levnum.ToString());
            lvi.SubItems.Add(rec.gotoA.ToString());
            lvi.SubItems.Add(rec.gotoB.ToString());
        }

        private void BNStrings_Click(object sender, EventArgs e)
        {
            StringsEditor.EditStrings(cstrings);
        }

        private void Button1_Click(object sender, EventArgs e)
        {
            if (LVSeqs.SelectedIndices.Count != 1)
                return;

            int i = LVSeqs.SelectedIndices[0];
            if (i == -1)
                return;

            JKRec jkr = Recs[i];
            if (!jkr.ltype.Equals("LEVEL", StringComparison.InvariantCultureIgnoreCase))
            {
                MessageBox.Show("Select a level first", "Episode editor", MessageBoxButtons.OK);
                return;
            }

            string s = FileOperations.ChangeFileExt(jkr.fname, "_start.cog");

            CogGenerator coggen = new CogGenerator();
            coggen.cstrs = cstrings;
            coggen.GoalBase = jkr.levnum * 1000;
            coggen.Name = Path.Combine(GlobalVars.ProjectDir, s);
            if (coggen.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            if (!jkr.fname.Equals(Path.GetFileName(GlobalVars.NedMain.ShortJKLName), StringComparison.InvariantCultureIgnoreCase))
                return;

            for (int x = 0; x < GlobalVars.Level.cogs.Count; x++)
            {
                JKCOG cg = GlobalVars.Level.cogs[x];
                if (cg.name.Equals(s, StringComparison.InvariantCultureIgnoreCase))
                    return;
            }

            {
                JKCOG cg = new JKCOG();
                cg.name = s;
                GlobalVars.Level.cogs.Add(cg);
                GlobalVars.CogForm.RefreshList();
            }
        }


    }
}
