﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Smith.NED
{
    public class Vertex
    {
        public Vec3 v;
        public int num;
        public int mark;

        public override string ToString()
        {
            return num.ToString() + ": " + v.ToString() + " [" + mark.ToString() + "]";
        }
    }

    public class TexVertex
    {
        public Vec2 v;
        public double intensity = 5.0;
        public double r = 5.0;
        public double g = 5.0;
        public double b = 5.0;
        public int num;

        public void Assign(TexVertex tv)
        {
            v = tv.v;
            intensity = tv.intensity;
            r = tv.r;
            g = tv.g;
            b = tv.b;
        }

        public TexVertex Clone()
        {
            TexVertex o = new TexVertex();

            o.v = v;
            o.intensity = intensity;
            o.r = r;
            o.g = g;
            o.b = b;
            o.num = num;

            return o;
        }
    }

    public class Vertices : List<Vertex>
    {
        public int Find(Vec3 pt)
        {
            for (int x = 0; x < Count; x++)
                if (this[x].v == pt)
                    return x;
            return -1;
        }

        public new int Add(Vertex v)
        {
            int i = Count;
            base.Add(v);
            return i;
        }
    }

    public class TexVertices : List<TexVertex>
    {
        public int Find(Vec2 pt)
        {
            for (int x = 0; x < Count; x++)
                if (this[x].v == pt)
                    return x;
            return -1;
        }

        public new int Add(TexVertex tv)
        {
            int i = Count;
            base.Add(tv);
            return i;
        }
    }

    public class Polygon
    {
        public Vec3 normal;
        public Vertices vertices = new Vertices();
        public TexVertices texvertices = new TexVertices();
        public int num;

        public Poly3 Poly3
        {
            get
            {
                Poly3 p = new Poly3();
                foreach (Vertex v in vertices)
                    p.Add(v.v);
                return p;
            }
        }

        public void CalcNormal(out Vec3 nvec)
        {
#if true
            nvec = Vec3.Zero;
            Vec3 v1 = vertices[0].v;
            Vec3 v2 = vertices[1].v;
            int cmn = 2;
            double cmax = 0.0;
            double cs = 0.0;
            for(int n=2; n<vertices.Count; n++)
            {
                Vec3 v3 = vertices[n].v;
                nvec = (v2 - v1) ^ (v3 - v1);
                cs = nvec.Magnitude;
                if (cs > 0.01)
                    break;
                if (cs > cmax)
                {
                    cmax = cs;
                    cmn = n;
                }

                // if on last one without a really good solution, do stuff
                if (n == (vertices.Count - 1))
                {
                    v3 = vertices[cmn].v;
                    nvec = (v2 - v1) ^ (v3 - v1);
                    cs = nvec.Magnitude;
                    break;
                }
            }

            if (cs != 0.0)
                nvec /= cs;

            //nvec = nvec;
#else
            // wtf is this crap.. where did i find it

            normal = Vec3.Zero;
            for (int x = 2; x < vertices.Count; x++)
            {
                Vec3 testNormal = (vertices[1].v - vertices[0].v) ^ (vertices[x].v - vertices[0].v);
                if (testNormal.Magnitude > normal.Magnitude)
                    normal = testNormal;
            }
#endif
        }

        public void CalcNormal()
        {
            CalcNormal(out normal);
        }

        public void Recalc()
        {
            CalcNormal();
        }

        public double CalcD()
        {
            return normal * vertices[0].v;
        }

        public void Planarize()
        {
            for (int x = 1; x < vertices.Count; x++)
            {
                double d = normal * (vertices[x].v - vertices[0].v);
                vertices[x].v -= d * normal;
            }
        }

        public int AddVertex(Vertex v)
        {
            int i = vertices.Count;
            vertices.Add(v);
            texvertices.Add(new TexVertex());
            return i;
        }

        public void InsertVertex(int n, Vertex v)
        {
            vertices.Insert(n, v);
            texvertices.Insert(n, new TexVertex());
        }

        public void DeleteVertex(int n)
        {
            vertices.RemoveAt(n);
            texvertices.RemoveAt(n);
        }

        public int WrapIndex(int i)
        {
            if (vertices.Count == 0)
                return 0;

            while (i < 0)
                i += vertices.Count;
            while (i >= vertices.Count)
                i -= vertices.Count;

            return i;
        }

        public Vec3 Centroid
        {
            get
            {
                if (vertices.Count == 0)
                    return Vec3.Zero;

                Vec3 c = Vec3.Zero;
                foreach (Vertex v in vertices)
                    c += v.v;

                return c / (double)vertices.Count;
            }
        }

        public double ExtrudeSize
        {
            get
            {
                double maxd = 0.0;
                Vec3 c = Centroid;
                for (int x = 0; x < vertices.Count; x++)
                {
                    Vec3 e = (normal ^ (vertices[WrapIndex(x + 1)].v - vertices[x].v)).Unit;
                    maxd = Math.Max(maxd, Math.Abs(e * (vertices[x].v - c)));                    
                }
                maxd = Math.Round(maxd * 2.0);
                return Math.Max(0.01, maxd);
            }
        }

        public double Radius
        {
            get
            {
                double maxd = 0.0;
                Vec3 c = Centroid;
                foreach (Vertex v in vertices)
                    maxd = Math.Max(maxd, (v.v - c).Magnitude);
                return maxd;
            }
        }

        // this whole routine is whack.. there's probably a much simpler way but whatever we'll plagerize from JED directly for now
        public bool IsOnPlane(Vec3 pt)
        {
            bool bx = MathLib.CompareScalar(normal.x, 0.0) == 0;
            bool by = MathLib.CompareScalar(normal.y, 0.0) == 0;
            bool bz = MathLib.CompareScalar(normal.z, 0.0) == 0;

            Vec3 n = Vec3.Zero;
            double d = CalcD();
            if (!bx && !by)
            {
                n.z = pt.z;
                n.y = (d-normal.x*pt.x-normal.z*pt.z)/normal.y;
                n.x = (d-normal.z*pt.z-normal.y*pt.y)/normal.x;
            }
            else if (!bx && !bz)
            {
                n.y = pt.y;
                n.z = (d-normal.x*pt.x-normal.y*pt.y)/normal.z;
                n.x = (d-normal.z*pt.z-normal.y*pt.y)/normal.x;
            }
            else if (!by && !bz)
            {
                n.x = pt.x;
                n.z = (d-normal.x*pt.x-normal.y*pt.y)/normal.z;
                n.y = (d-normal.z*pt.z-normal.x*pt.x)/normal.y;
            }
            else if (bx && by)
            {
                return MathLib.CompareScalar(pt.z, vertices[0].v.z) == 0;
            }
            else if (bx && bz)
            {
                return MathLib.CompareScalar(pt.y, vertices[0].v.y) == 0;
            }
            else if (by && bz)
            {
                return MathLib.CompareScalar(pt.x, vertices[0].v.x) == 0;
            }

            return (n == pt);
        }
    }

    public class Polygons : List<Polygon>
    {
        public Vertices vertices = new Vertices();

        public Poly3[] Poly3
        {
            get
            {
                Poly3[] p = new Poly3[Count];
                for (int x = 0; x < Count; x++)
                    p[x] = this[x].Poly3;
                return p;
            }
        }
    }

    public class Geometry
    {
    }
}
