﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;

namespace Smith.NED
{
    public static class GlobalVars
    {
        public static void Cleanup()
        {
            _Level = null;

            if (_NedMain != null)
            {
                _NedMain.Dispose();
                _NedMain = null;
            }

            /*if (_Preview3D != null)
            {
                _Preview3D.Dispose();
                _Preview3D = null;
            }*/

            /*if (_CogForm != null)
            {
                _CogForm.Dispose();
                _CogForm = null;
            }*/

            if (_UrqForm != null)
            {
                _UrqForm.Dispose();
                _UrqForm = null;
            }

            /*if (_ConsistencyChecker != null)
            {
                _ConsistencyChecker.Dispose();
                _ConsistencyChecker = null;
            }*/

            /*if (_MsgForm != null)
            {
                _MsgForm.Dispose();
                _MsgForm = null;
            }*/
        }

        public static JKLevel LevelOverride = null;

        private static JKLevel _Level = null;
        public static JKLevel Level
        {
            get
            {
                if (LevelOverride != null)
                    return LevelOverride;

                if (_Level == null)
                    _Level = new JKLevel();

                return _Level;
            }
        }

        // if current level is loaded from gob, this is the source gob
        public static string ProjectGob
        {
            get
            {
                if (string.IsNullOrEmpty(NedMain.LevelFile))
                    return null;

                int i = GlobalVars.NedMain.LevelFile.IndexOf('>');
                if (i == -1)
                    return null;

                return GlobalVars.NedMain.LevelFile.Substring(0, i);
            }
        }

        private static NedMain _NedMain = null;
        public static NedMain NedMain
        {
            get
            {
                return _NedMain;
            }
        }

        public static void CreateNED()
        {
            if (_NedMain == null)
            {
                _NedMain = new NedMain();
                _NedMain.Show();
            }
        }

        public static void DestroyNED()
        {
            if (_NedMain != null)
            {
                _NedMain.Dispose();
                _NedMain = null;
            }
        }

        public static Preview3D Preview3D
        {
            get
            {
                return NedMain?.preview3D1;
            }
        }

        public static MsgForm MsgForm
        {
            get
            {
                return NedMain?.msgForm1;
            }
        }

        public static CogForm CogForm
        {
            get
            {
                return NedMain?.cogForm1;
            }
        }

        private static UrqForm _UrqForm = null;
        public static UrqForm UrqForm
        {
            get
            {
                if (_UrqForm == null)
                {
                    _UrqForm = new UrqForm();
                }

                return _UrqForm;
            }
        }

        public static ConsistencyChecker ConsistencyChecker
        {
            get
            {
                if (NedMain == null)
                    return null;

                return NedMain.consistencyChecker1;
            }
        }

        public static ItemEdit ItemEdit
        {
            get
            {
                if (NedMain == null)
                    return null;

                return NedMain.itemEdit1;
            }
        }

        public static Tools ToolForm
        {
            get
            {
                if (NedMain == null)
                    return null;

                return NedMain.tools2;
            }
        }

        public static EpisodeEdit EpisodeEdit
        {
            get
            {
                if (NedMain == null)
                    return null;

                return NedMain.episodeEdit1;
            }
        }

        public static bool GetStayOnTop(Form f)
        {
            return f.TopMost;
        }

        public static void SetStayOnTop(Form f, bool ontop)
        {
            f.TopMost = ontop;
        }


        //public static Rectangle EditPos = Rectangle.Empty;
        public static Rectangle MWinPos = Rectangle.Empty;
        //public static Rectangle TBarPos = Rectangle.Empty;
        public static bool MWMaxed = false;

        //public static bool TbOnTop = false;
        //public static bool IEOnTop = false;
        public static bool CFOnTop = false;

        public static string JedVerNum = "0.951";

        public static string Recent1 = string.Empty;
        public static string Recent2 = string.Empty;
        public static string Recent3 = string.Empty;
        public static string Recent4 = string.Empty;

        public static bool ShowInfo = true;
        public static bool ShowWarnings = true;

        public static bool IsMOTS = false;

        public static bool P3DOnTop = false;
        public static bool P3DFullLit = false;
        public static bool P3DColoredLights = false;
        public static double P3DGamma = 1.0;
        public static bool P3DVisLayers = false;
        public static bool P3DThings = true;

        public static bool AutoSave = false;
        public static int SaveInterval = 0;

        public static bool SmithNeo = false;
        public static bool SmithSyncCameras = true;
        public static bool SmithMini = true;
        public static bool SmithNoPrecache = true;

        public static bool NewOnFloor = false;
        public static bool UndoEnabled = true;
        public static bool MoveFrames = true;
        public static bool GOBSmart = false;
        public static bool CheckOverlaps = false;
        private static bool _NewLightCalc = true;
        public static bool NewLightCalc
        {
            get
            {
                return _NewLightCalc;
            }

            set
            {
                if (_NewLightCalc == value)
                    return;

                _NewLightCalc = value;

                if (GlobalVars.NedMain != null)
                    GlobalVars.NedMain.lightsDirty = true;
            }
        }
        public static bool ConfirmRevert = false;
        public static bool CalculateLightsAuto = true;

        public enum MapRotation
        {
            New,
            Old
        }

        public static MapRotation MapRot = MapRotation.New;

        public static string BaseDir
        {
            get
            {
                return System.IO.Path.GetDirectoryName(Environment.GetCommandLineArgs()[0]);
            }
        }

        public static string ProjectDir = string.Empty;

        public static string GameDir = string.Empty;
        public static string JKDir = string.Empty;
        public static string MOTSDir = string.Empty;
        public static string IJIFDir = string.Empty;
        public static string DWDir = string.Empty;

        public static string res1_gob = string.Empty;
        public static string res2_gob = string.Empty;
        public static string sp_gob = string.Empty;
        public static string mp1_gob = string.Empty;
        public static string mp2_gob = string.Empty;
        public static string mp3_gob = string.Empty;


        public static NedMain.ThingView ThingView = NED.NedMain.ThingView.Wireframes;
        public static NedMain.MultiselectMode DefMSelMode = NED.NedMain.MultiselectMode.Toggle;

        public static double GridSize = 10.0;

        public static double P3DStep = 0.1;

        public static double PerpStep = 0.02;
        public static double TXStep = 0.5;
        public static double TXRotStep = 1.0;
        public static double TXScaleStep = 2.0;

        public static double DefGridStep = 0.2;
        public static double DefGridLine = 1.0;
        public static double DefGridDot = 0.2;
        public static string DefShape = "Cube";
        public static double DefGridSize = 10.0;
        public static double GridMoveStep = 0.001;
        public static double DefTXStep = 0.5;
        public static double DefTXRotStep = 1.0;
        public static double DefTXScaleStep = 2.0;
        public static double DefPerpStep = 0.02;
        public static double DefP3DStep = 0.1;

        public static double PixelsPerUnit
        {
            get
            {
                if (Level == null)
                    return 320.0;

                return Level.ppUnit;
            }
        }

        public static Color4 clMapBack = new Color4(0.0, 0.0, 0.0);
        public static Color4 clMapGeo = new Color4(1.0, 1.0, 1.0);
        public static Color4 clMapGeoBack = new Color4(0.5, 0.5, 0.5);
        public static Color4 clMapSel = new Color4(1.0, 0.0, 0.0);
        public static Color4 clMapSelBack = new Color4(0.5, 0.0, 0.0);
        public static Color4 clGrid = new Color4(1.0, 0.0, 1.0);
        public static Color4 clVertex = new Color4(0.0, 0.0, 1.0);
        public static Color4 clThing = new Color4(0.0, 1.0, 0.0);
        public static Color4 clFrame = new Color4(0.0, 0.5, 0.0);
        public static Color4 clLight = new Color4(1.0, 1.0, 0.0);
        public static Color4 clMSel = new Color4(0.0, 1.0, 1.0);
        public static Color4 clMSelBack = new Color4(0.0, 0.5, 0.5);
        public static Color4 clSelMsel = new Color4(1.0, 0.5, 0.0);
        public static Color4 clSelMselBack = new Color4(0.5, 0.25, 0.0);
        public static Color4 clGridX = new Color4(1.0, 1.0, 0.0);
        public static Color4 clGridY = new Color4(0.0, 0.0, 1.0);
        public static Color4 clExtra = new Color4(0.0, 1.0, 0.5);
    }
}
