﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public class GroupFilters
    {
        private TreeView tv;
        private TreeNode curNode = null;
        private List<TreeNode> stack = new List<TreeNode>();
        private int sp = 0;

        private string jllname = string.Empty;
        private Dictionary<string, List<string>> jll = new Dictionary<string, List<string>>(StringComparer.InvariantCultureIgnoreCase);

        private void LoadJLL(string name)
        {
            List<string> curList = null;

            Stream s = Jeddata.GetFile(name);
            if (s != null)
                using(Tokenizer t = new Tokenizer(s))
                    while (!t.EndOfStream)
                    {
                        string ln = t.ReadLine().Trim();
                        if (string.IsNullOrEmpty(ln))
                            break;

                        if (ln[0] == '[')
                        {
                            if (ln.Length < 2)
                                continue;

                            string catname = ln.Substring(1, ln.Length - 2).Trim();
                            if (jll.ContainsKey(catname))
                                curList = jll[catname];
                            else
                            {
                                curList = new List<string>();
                                jll.Add(catname, curList);
                            }

                            continue;
                        }

                        if (curList == null)
                            continue;

                        curList.Add(ln);
                    }
        }

        private void PushNode(TreeNode node)
        {
            sp = stack.Count;   // should this be pre or post increment?
            stack.Add(node);
        }

        private TreeNode PopNode()
        {
            if (sp == 0)
                return null;

            TreeNode node = stack[sp-1];
            stack.RemoveAt(sp - 1);
            sp--;
            return node;
        }

        private void AddNode(string w)
        {
            PushNode(curNode);
#if true
            curNode = curNode == null ? tv.Nodes.Add(w) : curNode.Nodes.Add(w);//curNode.Nodes.Add(w);
#else
            curNode = tv.Nodes.Add(w);//curNode.Nodes.Add(w);
#endif
        }

        private void ReadList(string w)
        {
            string[] parts = w.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
            if (parts.Length < 1)
                return;

            string id = parts[0];
            string displayname = parts.Length > 1 ? parts[1] : id;

            if (!jll.ContainsKey(id))
            {
                Utils.PanMessage(Utils.MessageType.Info, "Cannot find list " + w + " in " + jllname);
                return;
            }

            curNode.Nodes.Add(displayname).Tag = jll[id].ToArray();
        }

        private void LoadJLH(string fname)
        {
            Stream s = Jeddata.GetFile(fname);
            if (s == null)
                return;

            jllname = FileOperations.ChangeFileExt(fname, ".jll");
            LoadJLL(jllname);

            using (Tokenizer t = new Tokenizer(s))
                while (!t.EndOfStream)
                {
                    string ln = t.ReadLine();
                    if (string.IsNullOrEmpty(ln))
                        break;

                    string cmd = string.Empty;
                    string data = string.Empty;

                    int firstSpace = ln.IndexOfAny(new char[] { ' ', '\t' });
                    if (firstSpace == -1)
                        cmd = ln.Trim();
                    else
                    {
                        cmd = ln.Substring(0, firstSpace).Trim();
                        data = ln.Substring(firstSpace).Trim();
                    }

                    switch (cmd.ToLowerInvariant())
                    {
                        case "list":
                            ReadList(data);
                            break;

                        case "sub":
                            AddNode(data);
                            break;

                        case "subend":
                            curNode = PopNode();
                            break;
                    }
                }
        }

        private GroupFilters(TreeView _tv)
        {
            tv = _tv;
        }

        public static void InitializeFromFile(string fname, TreeView tv)
        {
            GroupFilters gf = new GroupFilters(tv);
            gf.LoadJLH(fname);
        }
    }
}
