﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Globalization;

namespace Smith.NED
{
    public class NEDFieldsView : DataGridView
    {
        public override DataObject GetClipboardContent()
        {
            DataObject obj = new DataObject();

            obj.SetText(CurrentRow?.Cells[1].Value as string ?? string.Empty);

            return obj;
        }
    }

    public partial class ItemEdit : UserControl
    {
        public ItemEdit()
        {
            InitializeComponent();


        }

        public new void Show()
        {
            if (GlobalVars.NedMain != null)
                GlobalVars.NedMain.tabControl1.SelectedTab = GlobalVars.NedMain.tabItemEditor;
        }

        public enum ID
        {
            Invalid = -1,

            Flags=0,
            Ambient=1,
            ExtraL = 2,
            Colormap = 3,
            Tint = 4,
            Sound = 5,
            SndVol = 6,

            AdjFlags=0,
            Material = 1,
            SurfFlags = 2,
            FaceFlags = 3,
            Geo = 4,
            LightMode = 5,
            Tex = 6,
            SFExtra = 7,
            Adjoin = 8,
            UScale = 9,
            VScale = 10,
            Layer = 15,

            Sec=0,
            Name =1,
            X=2,
            Y=3,
            Z=4,
            PCH=5,
            YAW=6,
            ROL=7,
            THLast = 20,

            Intensity=0,
            Range=1,
            RGBRange=10,
            RGBIntensity=11,
            RGB=12,
            LFlags=14,

            ThingFields=9,

            EdgeLen=2
        }

        public JKSector Cur_Sec = null;
        public JKSurface Cur_Surf = null;
        public JKVertex Cur_VX = null;
        public JKThing Cur_Thing = null;
        public JedLight Cur_Light = null;
        public TemplateValue Cur_Frame = null;
        public int cedge = -1;

        public void DoPickTexture()
        {
            if (Cur_Surf == null)
                return;

            foreach(DataGridViewRow row in SGFields.Rows)
            {
                ID id = (ID)row.Tag;
                if (id != ID.Material)
                    continue;

                string val = row.Cells[1].Value as string;
                if (SurfaceDoubleClick(id, ref val))
                {
                    row.Cells[1].Value = val;
                    SGFields.RefreshEdit();
                }

                break;
            }
        }

        public void DoPickThing()
        {
            if (Cur_Thing == null)
                return;

            foreach (DataGridViewRow row in SGFields.Rows)
            {
                ID id = (ID)row.Tag;
                if (id != ID.Name)
                    continue;

                string val = row.Cells[1].Value as string;
                if (ThingDoubleClick(id, ref val))
                {
                    row.Cells[1].Value = val;
                    SGFields.RefreshEdit();
                }

                break;
            }
        }

        public void ResetEdit()
        {
            ValueChange = null;
            ValueDoubleClick = null;
            SGFields.Rows.Clear();
            BNAdd.Visible = false;
            BNRemove.Visible = false;
            BNAsFrame.Visible = false;
            BNPaste.Visible = false;
            LBCogs.Items.Clear();
            LBText.Text = string.Empty;
        }

        public void AddFieldString(string name, ID id, string val)
        {
            DataGridViewRow row = new DataGridViewRow();
            row.Tag = id;
            row.CreateCells(SGFields, new string[] { name, val });
            SGFields.Rows.Add(row);
        }

        public void AddFieldHex(string name, ID id, uint val)
        {
            AddFieldString(name, id, val.ToString("X"));
        }

        public void AddFieldDouble(string name, ID id, double val)
        {
            AddFieldString(name, id, val.ToString());
        }

        public void AddFieldInt(string name, ID id, int val)
        {
            AddFieldString(name, id, val.ToString());
        }

        public JKLevel Level
        {
            get
            {
                return GlobalVars.Level;
            }
        }

        public void Display()
        {
            Visible = true;
        }

        public static Color? ParseColor(string s)
        {
            string[] rgb = s.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            if (rgb.Length != 3)
                return null;

            return Color.FromArgb((int)(double.Parse(rgb[0]) * 255.0),
                                (int)(double.Parse(rgb[1]) * 255.0),
                                (int)(double.Parse(rgb[2]) * 255.0));
        }

        public static string FormatColor(Color clr)
        {
            return ((double)clr.R / 255.0).ToString("0.###") + " " +
                    ((double)clr.G / 255.0).ToString("0.###") + " " +
                    ((double)clr.B / 255.0).ToString("0.###");
        }

        public bool ValSecField(string s, JKSector sec, ID id)
        {
            if (sec == null)
                return false;

            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Val);

            bool lights = false;

            switch (id)
            {
                case ID.Flags:
                    {
                        uint f;
                        if(!uint.TryParse(s, NumberStyles.HexNumber, null, out f))
                            return false;

                        sec.flags = (Sector.Flag)f;
                    }
                    break;

                case ID.ExtraL:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        sec.extra = d;

                        lights = true;
                    }
                    break;

                case ID.Colormap:
                    sec.colormap = s;
                    break;

                case ID.Tint:
                    {
                        Color? clr = ParseColor(s);
                        if (!clr.HasValue)
                            return false;

                        sec.tint = new Color4(clr.Value);
                    }
                    break;

                case ID.Sound:
                    sec.sound = s;
                    break;

                case ID.SndVol:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        sec.snd_vol = d;
                    }
                    break;

                case ID.Layer:
                    sec.layer = GlobalVars.NedMain.Toolbar.AddLayer(s);
                    break;
            }

            GlobalVars.NedMain.SectorChanged(sec, lights);
            return true;
        }

        bool SectorChange(ID id, ref string val)
        {
            Undo.StartUndoRec("Change Sector(s)");

            if (!ValSecField(val, Cur_Sec, id))
                return false;

            SCMultiSel scsl = GlobalVars.NedMain.scsel;
            for (int i = 0; i < scsl.Count; i++)
                ValSecField(val, Level.sectors[scsl.GetSC(i)], id);

            return true;
        }

        bool PickColor(ref Color clr)
        {
            ColorDialog cd = new ColorDialog();
            cd.Color = clr;
            cd.SolidColorOnly = false;
            cd.FullOpen = true;
            int cclr = 0;
            cclr |= (int)clr.R;
            cclr |= (int)clr.G << 8;
            cclr |= (int)clr.B << 16;
            cd.CustomColors = new int[] { cclr };
            if (cd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return false;

            clr = cd.Color;
            return true;
        }

        bool SectorDoubleClick(ID id, ref string val)
        {
            switch (id)
            {
                case ID.Flags:
                    {
                        uint f = uint.Parse(val, NumberStyles.HexNumber);
                        FlagEdit.EditSectorFlags(ref f);
                        val = f.ToString("X");
                    }
                    break;

                case ID.Colormap:
                    ResPicker.PickCMP(ref val);
                    break;

                case ID.Sound:
                    ResPicker.PickSecSound(ref val);
                    break;

                case ID.Tint:
                    {
                        Color? _clr = ParseColor(val);
                        if (!_clr.HasValue)
                            return false;

                        Color clr = _clr.Value;

                        if (PickColor(ref clr))
                            val = FormatColor(clr);
                    }
                    break;

                case ID.Layer:
                    ResPicker.PickLayer(ref val);
                    break;
            }

            return true;
        }

        public new string Text
        {
            set
            {
                LBTitle.Text = value;
            }
        }

        public void LoadSector(JKSector sec)
        {
            Cur_Sec = sec;
            Text = "Sector " + sec.num.ToString();
            ResetEdit();

            LBText.Text = sec.surfaces.Count.ToString() + " Surfaces " + sec.vertices.Count.ToString() + " Vertices";

            AddFieldHex("+FLAGS", ID.Flags, (uint)sec.flags);
            AddFieldDouble("EXTRA LIGHT", ID.ExtraL, sec.extra);
            AddFieldString("+COLORMAP", ID.Colormap, sec.colormap);
            AddFieldString("+TINT", ID.Tint, FormatColor(sec.tint));
            AddFieldString("+SOUND", ID.Sound, sec.sound);
            AddFieldDouble("VOLUME", ID.SndVol, sec.snd_vol);
            AddFieldString("+LAYER", ID.Layer, Level.GetLayerName(sec.layer));

            Display();
            ValueChange = SectorChange;
            ValueDoubleClick = SectorDoubleClick;
        }

        public bool ValSurfField(string s, JKSurface surf, ID id)
        {
            if (surf == null)
                return false;

            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Val);

            bool lights = false;

            switch (id)
            {
                case ID.Material:
                    surf.material = s;
                    break;

                case ID.Geo:
                    {
                        int f;
                        if (!int.TryParse(s, out f))
                            return false;

                        surf.SetGeo("ValSurfField", (Surface.GeoModeType)f);
                    }
                    break;

                case ID.Tex:
                    {
                        int f;
                        if (!int.TryParse(s, out f))
                            return false;

                        surf.tex = (Surface.TextureModeType)f;
                    }
                    break;

                case ID.LightMode:
                    {
                        int f;
                        if (!int.TryParse(s, out f))
                            return false;

                        surf.light = (Surface.LightModeType)f;

                        lights = true;
                    }
                    break;

                case ID.AdjFlags:
                    {
                        uint f;
                        if (!uint.TryParse(s, NumberStyles.HexNumber, null, out f))
                            return false;

                        surf.adjoinFlags = (Adjoin.Flag)f;

                        lights = true;// could be blocks light flag
                    }
                    break;

                case ID.SurfFlags:
                    {
                        uint f;
                        if (!uint.TryParse(s, NumberStyles.HexNumber, null, out f))
                            return false;

                        surf.surfaceFlags = (Surface.Flag)f;
                    }
                    break;

                case ID.FaceFlags:
                    {
                        uint f;
                        if (!uint.TryParse(s, NumberStyles.HexNumber, null, out f))
                            return false;

                        surf.faceFlags = (Surface.FaceFlag)f;
                    }
                    break;

                case ID.SFExtra:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        surf.extraLight = d;

                        lights = true;
                    }
                    break;

                case ID.UScale:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        surf.texScale.x = d;
                        surf.RecalcAll();
                    }
                    break;

                case ID.VScale:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        surf.texScale.y = d;
                        surf.RecalcAll();
                    }
                    break;

                case ID.Adjoin:
                    {
                        string[] parts = s.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

                        if (parts.Length < 1)
                            return false;

                        int sc;
                        if (!int.TryParse(parts[0], out sc))
                            return false;

                        if (sc == -1)
                            surf.adjoin = null;
                        else
                        {
                            if (parts.Length < 2)
                                return false;

                            int sf;
                            if (!int.TryParse(parts[1], out sf))
                                return false;

                            if (sc < 0 || sc >= Level.sectors.Count)
                                return false;

                            JKSector sec = Level.sectors[sc];
                            if (sf < 0 || sf >= sec.surfaces.Count)
                                return false;

                            surf.adjoin = sec.surfaces[sf];
                        }
                    }

                    lights = true;
                    break;
            }

            GlobalVars.NedMain.SectorChanged(surf.sector, lights);
            return true;
        }

        bool SurfaceChange(ID id, ref string val)
        {
            Undo.StartUndoRec("Change surface(s)");

            if (!ValSurfField(val, Cur_Surf, id))
                return false;

            SFMultiSel sfsel = GlobalVars.NedMain.sfsel;
            for (int i = 0; i < sfsel.Count; i++)
            {
                int sc, sf;
                sfsel.GetSCSF(i, out sc, out sf);
                ValSurfField(val, Level.sectors[sc].surfaces[sf], id);
            }
            return true;
        }

        bool SurfaceDoubleClick(ID id, ref string val)
        {
            switch (id)
            {
                case ID.Material:
                    {
                        if (Cur_Surf == null)
                            return false;

                        ResPicker.PickMAT(ref val, Level.GetMasterCMP());
                    }
                    break;

                case ID.Geo:
                    {
                        int f;
                        if (!int.TryParse(val, out f))
                            return false;

                        SCFieldPicker.PickGeo(ref f);
                        val = f.ToString();
                    }
                    break;

                case ID.Tex:
                    {
                        int f;
                        if (!int.TryParse(val, out f))
                            return false;

                        SCFieldPicker.PickTex(ref f);
                        val = f.ToString();
                    }
                    break;

                case ID.LightMode:
                    {
                        int f;
                        if (!int.TryParse(val, out f))
                            return false;

                        SCFieldPicker.PickLightMode(ref f);
                        val = f.ToString();
                    }
                    break;

                case ID.AdjFlags:
                    {
                        uint f;
                        if (!uint.TryParse(val, NumberStyles.HexNumber, null, out f))
                            return false;

                        FlagEdit.EditAdjoinFlags(ref f);
                        val = f.ToString("X");
                    }
                    break;

                case ID.SurfFlags:
                    {
                        uint f;
                        if (!uint.TryParse(val, NumberStyles.HexNumber, null, out f))
                            return false;

                        FlagEdit.EditSurfaceFlags(ref f);
                        val = f.ToString("X");
                    }
                    break;

                case ID.FaceFlags:
                    {
                        uint f;
                        if (!uint.TryParse(val, NumberStyles.HexNumber, null, out f))
                            return false;

                        FlagEdit.EditFaceFlags(ref f);
                        val = f.ToString("X");
                    }
                    break;
            }
            return true;
        }

        public void LoadSurface(JKSurface surf)
        {
            Cur_Surf = surf;
            Text = "Sector " + surf.sector.num.ToString() + " Surface " + surf.num.ToString();
            ResetEdit();
            LBText.Text = surf.vertices.Count.ToString() + " Vertices";

            AddFieldString("+MATERIAL", ID.Material, surf.material);
            AddFieldDouble("U scale", ID.UScale, surf.texScale.x);
            AddFieldDouble("V scale", ID.VScale, surf.texScale.y);
            AddFieldInt("+GEO", ID.Geo, (int)surf.geo);
            AddFieldInt("+TEX", ID.Tex, (int)surf.tex);
            AddFieldInt("+LIGHT MODE", ID.LightMode, (int)surf.light);
            AddFieldString("ADJOIN", ID.Adjoin, (surf.adjoin == null) ? "-1" : (surf.adjoin.sector.num.ToString() + " " + surf.adjoin.num.ToString()));
            AddFieldHex("+ADJOIN FLAGS", ID.AdjFlags, (uint)surf.adjoinFlags);
            AddFieldHex("+SURF FLAGS", ID.SurfFlags, (uint)surf.surfaceFlags);
            AddFieldHex("+FACE FLAGS", ID.FaceFlags, (uint)surf.faceFlags);
            AddFieldDouble("EXTRA LIGHT", ID.SFExtra, surf.extraLight);

            Display();
            AddCogs(COGType.Surface, Cur_Surf);
            ValueChange = SurfaceChange;
            ValueDoubleClick = SurfaceDoubleClick;
        }

        bool ValFRField(string s, TemplateValue frame, ID id)
        {
            if (frame == null)
                return false;

            double d;
            if (!double.TryParse(s, out d))
                return false;

            Vec3 fp;
            Orient fo;
            frame.GetFrame(out fp, out fo);
            switch (id)
            {
                case ID.X: fp.x = d; break;
                case ID.Y: fp.y = d; break;
                case ID.Z: fp.z = d; break;
                case ID.PCH: fo.pitch = d; break;
                case ID.YAW: fo.yaw = d; break;
                case ID.ROL: fo.roll = d; break;
            }
            frame.SetFrame(fp, fo);
            return true;
        }

        public bool FrameChange(ID id, ref string val)
        {
            if (!ValFRField(val, Cur_Frame, id))
                return false;

            FRMultiSel frsel = GlobalVars.NedMain.frsel;
            for (int i = 0; i < frsel.Count; i++)
            {
                int th, fr;
                frsel.GetTHFR(i, out th, out fr);
                if (fr == -1)
                    continue;

                ValFRField(val, Level.things[th].vals[fr], id);
            }
            return true;
        }

        public void LoadFrame(int th, int fr)
        {
            ResetEdit();

            if (fr < 0)
            {
                Text = "Thing " + th.ToString();
                AddFieldInt("numframes", ID.Invalid, Level.things[th].nframes);
                return;
            }

            BNAsFrame.Visible = true;
            BNPaste.Visible = true;

            JKThing thing = Level.things[th];

            int n = 0;
            for (int i = 0; i < thing.vals.Count; i++)
            {
                if (i >= fr)
                    break;

                if (thing.vals[i].adjType == AdjType.Frame)
                    n++;
            }

            Text = "Thing " + th.ToString() + " Frame " + n.ToString();

            Cur_Frame = Level.things[th].vals[fr];
            Vec3 fp;
            Orient fo;
            Cur_Frame.GetFrame(out fp, out fo);

            AddFieldDouble("X", ID.X, fp.x);
            AddFieldDouble("Y", ID.Y, fp.y);
            AddFieldDouble("Z", ID.Z, fp.z);
            AddFieldDouble("PCH", ID.PCH, fo.pitch);
            AddFieldDouble("YAW", ID.YAW, fo.yaw);
            AddFieldDouble("ROL", ID.ROL, fo.roll);

            Display();

            ValueChange = FrameChange;
        }

        bool ValTHValue(string s, JKThing thing, string fname, int n)
        {
            int cn = 0;
            TemplateValue v = null;

            for (int i = 0; i < thing.vals.Count; i++)
            {
                v = thing.vals[i];
                if (!v.name.Equals(fname, StringComparison.InvariantCultureIgnoreCase))
                {
                    v = null;
                    continue;
                }
                if (cn == n)
                    break;
                cn++;
                v = null;
            }

            if (v == null)
                return false;

            Undo.SaveThingUndo(thing, Undo.Change.Changed);

            if (!v.Val(s))
                return false;

            GlobalVars.NedMain.ThingChanged(thing);
            return true;
        }

        bool ValTHField(string s, JKThing thing, ID id)
        {
            if (thing == null)
                return false;

            Undo.SaveThingUndo(thing, Undo.Change.Changed);

            switch (id)
            {
                case ID.Name:
                    thing.name = s;
                    break;

                case ID.Sec:
                    {
                        int sc;
                        if (!int.TryParse(s, out sc))
                            return false;

                        if (sc == -1)
                            thing.sec = null;
                        else
                        {
                            if (sc < 0 || sc >= Level.sectors.Count)
                                return false;

                            thing.sec = Level.sectors[sc];
                        }
                    }
                    break;

                case ID.X:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.pos.x = d;
                    }
                    break;

                case ID.Y:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.pos.y = d;
                    }
                    break;

                case ID.Z:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.pos.z = d;
                    }
                    break;

                case ID.PCH:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.orient.pitch = d;
                    }
                    break;

                case ID.YAW:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.orient.yaw = d;
                    }
                    break;

                case ID.ROL:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        thing.orient.roll = d;
                    }
                    break;

                case ID.Layer:
                    thing.layer = GlobalVars.NedMain.Toolbar.AddLayer(s);
                    break;
            }

            GlobalVars.NedMain.ThingChanged(thing);
            return true;
        }

        bool ThingChange(ID id, ref string val)
        {
            Undo.StartUndoRec("Change thing(s)");

            if ((int)id < (int)ID.THLast)
            {
                if (!ValTHField(val, Cur_Thing, id))
                    return false;

                THMultiSel thsel = GlobalVars.NedMain.thsel;
                for (int i = 0; i < thsel.Count; i++)
                    ValTHField(val, Level.things[thsel.GetTH(i)], id);
            }
            else
            {
                string fname = Cur_Thing.vals[(int)id - (int)ID.THLast].name;
                int n = 0;
                for (int i = (int)id - (int)ID.THLast - 1; i >= 0; i--)
                    if (Cur_Thing.vals[i].name.Equals(fname, StringComparison.InvariantCultureIgnoreCase))
                        n++;

                if (!ValTHValue(val, Cur_Thing, fname, n))
                    return false;

                THMultiSel thsel = GlobalVars.NedMain.thsel;
                for (int i = 0; i < thsel.Count; i++)
                    ValTHValue(val, Level.things[thsel.GetTH(i)], fname, n);
            }
            return true;
        }

        bool ThingDoubleClick(ID id, ref string val)
        {
            if (Cur_Thing == null)
                return false;

            if (id == ID.Name)
                ResPicker.PickThing(ref val);
            else if (id == ID.Layer)
                ResPicker.PickLayer(ref val);
            else if ((int)id >= (int)ID.THLast)
            {
                int i = (int)id - (int)ID.THLast;
                if (i >= Cur_Thing.vals.Count)
                    return false;

                if (!Cur_Thing.vals[i].name.Equals("thingflags", StringComparison.InvariantCultureIgnoreCase))
                    return false;

                uint f;
                if (!uint.TryParse(val, NumberStyles.HexNumber, null, out f))
                    return false;

                FlagEdit.EditThingFlags(ref f);

                val = "0x" + f.ToString("X");
            }
            else
                // unknown
                return false;

            return true;
        }

        public void LoadThing(JKThing thing)
        {
            Text = "Thing " + thing.num.ToString();
            ResetEdit();
            BNAdd.Visible = true;
            BNRemove.Visible = true;
            BNAsFrame.Visible = true;
            BNPaste.Visible = true;
            Cur_Thing = thing;

            AddFieldString("+NAME", ID.Name, thing.name);
            AddFieldInt("SECTOR", ID.Sec, thing.sec == null ? -1 : thing.sec.num);
            AddFieldDouble("X", ID.X, thing.pos.x);
            AddFieldDouble("Y", ID.Y, thing.pos.y);
            AddFieldDouble("Z", ID.Z, thing.pos.z);
            AddFieldDouble("PCH", ID.PCH, thing.orient.pitch);
            AddFieldDouble("YAW", ID.YAW, thing.orient.yaw);
            AddFieldDouble("ROL", ID.ROL, thing.orient.roll);
            AddFieldString("+LAYER", ID.Layer, Level.GetLayerName(thing.layer));

            for (int i = 0; i < thing.vals.Count; i++)
            {
                TemplateValue tv = thing.vals[i];

                if (tv.name.Equals("thingflags", StringComparison.InvariantCultureIgnoreCase))
                    AddFieldString("+" + tv.name, (ID)(ID.THLast + i), tv.AsString);
                else
                    AddFieldString(tv.name, (ID)(ID.THLast + i), tv.AsString);
            }

            AddCogs(COGType.Thing, Cur_Thing);

            Display();

            ValueDoubleClick = ThingDoubleClick;
            ValueChange = ThingChange;
        }

        bool ValVertField(string s, JKVertex vx, ID id)
        {
            if (vx == null)
                return false;

            Undo.SaveSecUndo(vx.Sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            double d;
            if (!double.TryParse(s, out d))
                return false;
            
            switch (id)
            {
                case ID.X: vx.v.x = d; break;
                case ID.Y: vx.v.y = d; break;
                case ID.Z: vx.v.z = d; break;
            }

            foreach (JKSurface surf in vx.Sector.surfaces)
                surf.RecalcAll();
            GlobalVars.NedMain.SectorChanged(vx.Sector, true);
            return true;
        }

        bool VertexChange(ID id, ref string val)
        {
            Undo.StartUndoRec("Change vertices");

            if (!ValVertField(val, Cur_VX, id))
                return false;

            VXMultiSel vxsel = GlobalVars.NedMain.vxsel;
            for (int i = 0; i < vxsel.Count; i++)
            {
                int sc, vx;
                vxsel.GetSCVX(i, out sc, out vx);
                ValVertField(val, Level.sectors[sc].vertices[vx], id);
            }
            return true;
        }

        public void LoadVertex(JKVertex vert)
        {
            Text = "Sector " + vert.Sector.num.ToString() + " vertex " + vert.num.ToString();
            ResetEdit();

            Cur_VX = vert;
            AddFieldDouble("X", ID.X, vert.v.x);
            AddFieldDouble("Y", ID.Y, vert.v.y);
            AddFieldDouble("Z", ID.Z, vert.v.z);

            Display();
            ValueChange = VertexChange;
        }

        bool ValExtraVertex(string s, NedMain.ExtraVertex ev, ID id)
        {
            if (ev == null)
                return false;

            double d;
            if (!double.TryParse(s, out d))
                return false;

            switch (id)
            {
                // shouldnt we support name too?  whatever.. this was hopelessly useless in jed anyway
                case ID.X: ev.v.x = d; break;
                case ID.Y: ev.v.y = d; break;
                case ID.Z: ev.v.z = d; break;
            }

            return true;
        }

        bool ExtraChange(ID id, ref string val)
        {
            object o = GlobalVars.NedMain.ExtraObjs[GlobalVars.NedMain.Cur_EX];
            if (!(o is NedMain.ExtraVertex))
                return false;

            NedMain.ExtraVertex ev = o as NedMain.ExtraVertex;
            if (!ValExtraVertex(val, ev, id))
                return false;

            GlobalVars.NedMain.FireOnExtraMove(ev, false);
            return true;
        }

        public void LoadExtra(int ex)
        {
            Text = GlobalVars.NedMain.ExtraObjsName + " " + ex.ToString();
            ResetEdit();

            object o = GlobalVars.NedMain.ExtraObjs[ex];
            if (o is NedMain.ExtraVertex)
            {
                NedMain.ExtraVertex ev = o as NedMain.ExtraVertex;
                AddFieldString("Name", ID.Name, ev.Name);
                AddFieldDouble("X", ID.X, ev.v.x);
                AddFieldDouble("Y", ID.Y, ev.v.y);
                AddFieldDouble("Z", ID.Z, ev.v.z);
                ValueChange = ExtraChange;
            }

            if (o is NedMain.ExtraLine)
                AddFieldString("Name", ID.Name, (o as NedMain.ExtraLine).Name);

            if (o is NedMain.ExtraPolygon)
                AddFieldString("Name", ID.Name, (o as NedMain.ExtraPolygon).Name);

            Display();
        }

        bool ValLTField(string s, JedLight light, ID id)
        {
            if (light == null)
                return false;

            Undo.SaveLightUndo(light, Undo.Change.Changed);

            switch (id)
            {
                case ID.X:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.Pos.x = d;
                    }
                    break;

                case ID.Y:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.Pos.y = d;
                    }
                    break;

                case ID.Z:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.Pos.z = d;
                    }
                    break;

                case ID.Intensity:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.intensity = d;
                    }
                    break;

                case ID.Range:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.range = d;
                    }
                    break;

                case ID.LFlags:
                    {
                        uint f;
                        if (!uint.TryParse(s, NumberStyles.HexNumber, null, out f))
                            return false;

                        light.flags = (LightFlag)f;
                    }
                    break;

                case ID.Layer:
                    light.layer = GlobalVars.NedMain.Toolbar.AddLayer(s);
                    break;

                case ID.RGB:
                    {
                        Color? _clr = ParseColor(s);
                        if (!_clr.HasValue)
                            return false;

                        light.clr = new Color4(_clr.Value);
                    }
                    break;

                case ID.RGBIntensity:
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                            return false;

                        light.rgbintensity = d;
                    }
                    break;
            }
            GlobalVars.NedMain.LevelChanged(true);
            return true;
        }

        bool LightChange(ID id, ref string val)
        {
            Undo.StartUndoRec("Change light(s)");

            if (!ValLTField(val, Cur_Light, id))
                return false;

            LTMultiSel ltsel = GlobalVars.NedMain.ltsel;
            for(int i=0; i<ltsel.Count; i++)
                ValLTField(val, Level.lights[ltsel.GetLT(i)], id);
            return true;
        }

        bool LightDoubleClick(ID id, ref string val)
        {
            if (Cur_Light == null)
                return false;

            switch (id)
            {
                case ID.Layer:
                    ResPicker.PickLayer(ref val);
                    break;

                case ID.LFlags:
                    {
                        uint f;
                        if (!uint.TryParse(val, NumberStyles.HexNumber, null, out f))
                            return false;

                        FlagEdit.EditLightFlags(ref f);
                        val = f.ToString("X");
                    }
                    break;

                case ID.RGB:
                    {
                        Color? _clr = ParseColor(val);
                        if (!_clr.HasValue)
                            return false;

                        Color clr = _clr.Value;

                        if (!PickColor(ref clr))
                            return false;

                        val = FormatColor(clr);
                    }
                    break;
            }

            return true;
        }

        public void LoadLight(int nlight)
        {
            Text = "Light " + nlight.ToString();
            ResetEdit();
            if (nlight >= Level.lights.Count)
                return;

            Cur_Light = Level.lights[nlight];

            AddFieldDouble("Intensity", ID.Intensity, Cur_Light.intensity);
            AddFieldDouble("Range", ID.Range, Cur_Light.range);
            AddFieldDouble("X", ID.X, Cur_Light.Pos.x);
            AddFieldDouble("Y", ID.Y, Cur_Light.Pos.y);
            AddFieldDouble("Z", ID.Z, Cur_Light.Pos.z);
            AddFieldHex("+Flags", ID.LFlags, (uint)Cur_Light.flags);

            if (GlobalVars.IsMOTS)
            {
                AddFieldString("+RGB", ID.RGB, FormatColor(Cur_Light.clr));
                AddFieldDouble("RGBIntensity", ID.RGBIntensity, Cur_Light.rgbintensity);
            }

            AddFieldString("+LAYER", ID.Layer, Level.GetLayerName(Cur_Light.layer));

            Display();
            ValueDoubleClick = LightDoubleClick;
            ValueChange = LightChange;
        }

        bool EdgeChange(ID id, ref string val)
        {
            return false;
        }

        public void LoadEdge(JKSurface surf, int ed)
        {
            Text = "Sec " + surf.sector.num.ToString() + " Surf " + surf.num.ToString() + " Edge " + ed.ToString();
            ResetEdit();

            Vertex v1 = surf.vertices[ed];
            Vertex v2 = surf.vertices[surf.NextVX(ed)];

            AddFieldInt("Starting VX", ID.Invalid, v1.num);
            AddFieldInt("Ending VX", ID.Invalid, v2.num);
            AddFieldDouble("Length", ID.EdgeLen, (v2.v - v1.v).Magnitude);

            Display();
            ValueChange = EdgeChange;
        }

        class COGEntry
        {
            public string name;
            public JKCOG cog;

            public COGEntry(string _name, JKCOG _cog)
            {
                name = _name;
                cog = _cog;
            }

            public override string ToString()
            {
                return name;
            }
        }

        void AddCogs(COGType ct, object lobj)
        {
            foreach (JKCOG cog in Level.cogs)
                foreach (COGValue cv in cog.vals)
                    if (cv.cogType == ct && object.ReferenceEquals(cv.obj, lobj))
                        LBCogs.Items.Add(new COGEntry(cog.name + @"\" + cv.name, cog));
        }

        /*private void ItemEdit_FormClosed(object sender, FormClosedEventArgs e)
        {
            GlobalVars.EditPos = Bounds;
        }*/

        delegate bool ValueChangeEvent(ID id, ref string val);
        ValueChangeEvent ValueChange = null;

        delegate bool ValueDoubleClickEvent(ID id, ref string val);
        ValueDoubleClickEvent ValueDoubleClick = null;

        private void SGFields_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            if (e.RowIndex < 0 || e.RowIndex >= SGFields.Rows.Count)
                return;

            DataGridViewRow row = SGFields.Rows[e.RowIndex];
            if (ValueChange != null)
            {
                string val = row.Cells[1].Value as string;
                if (ValueChange((ID)row.Tag, ref val))
                {
                    row.Cells[1].Value = val;
                    SGFields.RefreshEdit();
                }
            }
        }

        private void SGFields_CellDoubleClick(object sender, DataGridViewCellEventArgs e)
        {
            if (e.RowIndex < 0 || e.RowIndex >= SGFields.Rows.Count)
                return;

            DataGridViewRow row = SGFields.Rows[e.RowIndex];
            if (ValueDoubleClick != null)
            {
                string val = row.Cells[1].Value as string;
                if (ValueDoubleClick((ID)row.Tag, ref val))
                {
                    row.Cells[1].Value = val;
                    SGFields.RefreshEdit();
                }
            }
        }

        private void LBCogs_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            int i = LBCogs.IndexFromPoint(e.Location);
            if (i < 0)
                return;

            COGEntry ce = LBCogs.Items[i] as COGEntry;
            i = Level.cogs.IndexOf(ce.cog);
            if (i >= 0)
                GlobalVars.CogForm.GotoCOG(i);
        }

        private void BNAdd_Click(object sender, EventArgs e)
        {
            if (Cur_Thing == null)
                return;

            string s = "THINGFLAGS";
            if (Utils.InputBox("Value name", "Enter value name", ref s) != DialogResult.OK)
                return;

            Undo.StartUndoRec("Add thing value(s)");
            THMultiSel thsel = GlobalVars.NedMain.thsel;

            int n = thsel.AddTH(GlobalVars.NedMain.Cur_TH);

            TemplateValue v = null;
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing th = GlobalVars.Level.things[thsel.GetTH(i)];
                Undo.SaveThingUndo(th, Undo.Change.Changed);

                TemplateValue cv = Templates.GetTPLField(th.name, s);
                v = new TemplateValue();
                if (cv != null)
                {
                    v.Assign(cv);
                    v.Val(cv.AsString);
                }
                v.name = s;
                v.adjType = Values.GetTPLType(v.name);
                v.vType = Values.GetTPLVType(v.name);
                th.vals.Add(v);
            }

            thsel.DeleteN(n);

            if (v != null)
            {
                if (s.Equals("thingflags", StringComparison.InvariantCultureIgnoreCase))
                    AddFieldString("+" + s, (ID)((int)ID.THLast + Cur_Thing.vals.Count - 1), v.AsString);
                else
                    AddFieldString(s, (ID)((int)ID.THLast + Cur_Thing.vals.Count - 1), v.AsString);
            }
        }

        private void BNRemove_Click(object sender, EventArgs e)
        {
            if (Cur_Thing == null)
                return;

            if (SGFields.SelectedRows.Count != 1)
                return;

            DataGridViewRow row = SGFields.SelectedRows[0];
            if (row.Index < (int)ID.ThingFields)
                return;

            int vn = row.Index - (int)ID.ThingFields;

            string fname = Cur_Thing.vals[vn].name;

            int n = 0;
            for (int i = vn - 1; i >= 0; i--)
                if (Cur_Thing.vals[i].name.Equals(fname, StringComparison.InvariantCultureIgnoreCase))
                    n++;

            THMultiSel thsel = GlobalVars.NedMain.thsel;
            int tmp = thsel.AddTH(GlobalVars.NedMain.Cur_TH);

            Undo.StartUndoRec("Remove thing value(s)");
            for (int i = 0; i < thsel.Count; i++)
                RemoveTHValue(Level.things[thsel.GetTH(i)], fname, n);

            thsel.DeleteN(tmp);

            LoadThing(Cur_Thing);
        }

        public void RemoveTHValue(JKThing thing, string fname, int n)
        {
            int cn = 0;
            int nv = -1;
            TemplateValue v = null;
            for (int i = 0; i < thing.vals.Count; i++)
            {
                nv = i;
                v = thing.vals[i];
                if (!v.name.Equals(fname, StringComparison.InvariantCultureIgnoreCase))
                {
                    v = null;
                    continue;
                }

                if (cn == n)
                    break;

                cn++;
                v = null;
            }

            if (v == null)
                return;

            Undo.SaveThingUndo(thing, Undo.Change.Changed);
            thing.vals.RemoveAt(nv);
            GlobalVars.NedMain.ThingChanged(thing);
        }

        private void BNPaste_Click(object sender, EventArgs e)
        {
            if (!Clipboard.ContainsText())
                return;

            string s = Clipboard.GetText();
            if (s.IndexOf('(') != 0)
                return;
            if (s.IndexOf(')') != s.Length - 1)
                return;

            Vec3 fp;
            Orient fo = Orient.Zero;
            JK.ParseFrame(s, out fp, ref fo);

            switch (GlobalVars.NedMain.map_mode)
            {
                case NedMain.MapMode.Thing:
                    if (Cur_Thing == null)
                        return;
                    Cur_Thing.pos = fp;
                    Cur_Thing.orient = fo;
                    GlobalVars.NedMain.ThingChanged(Cur_Thing);
                    GlobalVars.NedMain.EditObject();
                    break;

                case NedMain.MapMode.Frame:
                    if (Cur_Frame == null)
                        return;
                    /*{
                        Vec3 cfp;
                        Orient cfo;
                        Cur_Frame.GetFrame(out cfp, out cfo);
                        cfp = 
                    }*/
                    Cur_Frame.SetFrame(fp, fo);
                    GlobalVars.NedMain.LevelChanged(false);
                    GlobalVars.NedMain.EditObject();
                    break;
            }
        }

        private void BNAsFrame_Click(object sender, EventArgs e)
        {
            Vec3 fp;
            Orient fo;

            switch (GlobalVars.NedMain.map_mode)
            {
                case NedMain.MapMode.Thing:
                    if (Cur_Thing == null)
                        return;

                    fp = Cur_Thing.pos;
                    fo = Cur_Thing.orient;
                    break;

                case NedMain.MapMode.Frame:
                    if(Cur_Frame == null)
                        return;

                    Cur_Frame.GetFrame(out fp, out fo);
                    break;

                default:
                    return;
            }

            Clipboard.SetText("(" + Utils.dstr(fp.x, 5) + "/" + Utils.dstr(fp.y, 5) + "/" + Utils.dstr(fp.z, 5) + ":" +
                                          Utils.dstr(fo.pitch, 5) + "/" + Utils.dstr(fo.yaw, 5) + "/" + Utils.dstr(fo.roll, 5) + ")");
        }
    }
}
