﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Smith.NED
{
    public class JKVertex : Vertex
    {
        public JKSector Sector;

        public JKVertex(JKSector _Sector)
        {
            Sector = _Sector;
        }

        public void Assign(JKVertex o)
        {
            v = o.v;
        }

        public override string ToString()
        {
            return "Sec" + Sector.num.ToString() + ", " + base.ToString();
        }
    }

    [Flags]
    public enum LightFlag : uint
    {
        NoBlock = 0x01
    }

    public class JedLight
    {
        public int ID;
        public Vec3 Pos;
        public int mark;
        public double intensity = 10.0;
        public double range = 2.0;
        public int layer;
        public LightFlag flags = 0;
        public Color4 clr = Color4.white();
        public double rgbintensity = 10.0;
        public double rgbrange;

        public void Assign(JedLight l)
        {
            intensity = l.intensity;
            range = l.range;
            clr = l.clr;
            rgbintensity = l.rgbintensity;
            layer = l.layer;
        }

        public void SerializeTo(ZipUtil zip)
        {
            zip.WriteVec3(Pos);
            zip.WriteDouble(intensity);
            zip.WriteDouble(range);
            zip.WriteUInt((uint)flags);
            zip.WriteColor4(clr);
            zip.WriteDouble(rgbintensity);
            zip.WriteDouble(rgbrange);
        }

        public void SerializeFrom(ZipUtil zip)
        {
            Pos = zip.ReadVec3();
            intensity = zip.ReadDouble();
            range = zip.ReadDouble();
            flags = (LightFlag)zip.ReadUInt();
            clr = zip.ReadColor4();
            rgbintensity = zip.ReadDouble();
            rgbrange = zip.ReadDouble();
        }
    }

    public class JedLights : List<JedLight>
    {

    }

    public class JKVertices : List<JKVertex>
    {
        public void SerializeTo(ZipUtil zip)
        {
            foreach (JKVertex v in this)
                zip.WriteVec3(v.v);
        }

        public void SerializeFrom(ZipUtil zip)
        {
            foreach (JKVertex v in this)
                v.v = zip.ReadVec3();
        }
    }

    public class JKLine
    {
        public JKVertex v1;
        public JKVertex v2;
    }

    public class JKThing
    {
        public int ID;

        public JKLevel level;
        public int num;
        public JKSector sec;
        public string name = "walkplayer";
        public Vec3 pos;
        public Orient orient;
        public TemplateValues vals = new TemplateValues();
        public int layer;
        public JKModel model;
        public Box3 box;

        public bool NEDDynamicLight = false;

        public JKThing(JKLevel _level)
        {
            level = _level;
        }

        public void Assign(JKThing thing)
        {
            name = thing.name;
            pos = thing.pos;
            orient = thing.orient;
            foreach (TemplateValue tv in thing.vals)
            {
                TemplateValue v = new TemplateValue();
                v.Assign(tv);
                vals.Add(v);
            }
            layer = thing.layer;
        }

        public void SerializeTo(ZipUtil zip)
        {
            if (sec == null)
                zip.WriteInt(-1);
            else
                zip.WriteInt(sec.num);

            zip.WriteString(name);
            zip.WriteVec3(pos);
            zip.WriteOrient(orient);
            zip.WriteInt(layer);

            vals.SerializeTo(zip);
        }

        public void SerializeFrom(ZipUtil zip, out int secnum)
        {
            secnum = zip.ReadInt();
            name = zip.ReadString();
            pos = zip.ReadVec3();
            orient = zip.ReadOrient();
            layer = zip.ReadInt();
            vals.SerializeFrom(zip);
        }

        public int nframes
        {
            get
            {
                int numframes = 0;
                foreach (TemplateValue v in vals)
                    if (v.name.Equals("frame", StringComparison.InvariantCultureIgnoreCase))
                        numframes++;
                return numframes;
            }
        }

        public int PrevFrame(int n)
        {
            for(--n; n >= 0; n--)
                if (vals[n].name.Equals("frame", StringComparison.InvariantCultureIgnoreCase))
                    return n;

            return -1;
        }

        public int NextFrame(int n)
        {
            for (++n; n < vals.Count; n++)
                if (vals[n].name.Equals("frame", StringComparison.InvariantCultureIgnoreCase))
                    return n;

            return -1;
        }

        public TemplateValue AddValue(string name, string s)
        {
            return InsertValue(vals.Count, name, s);
        }

        public TemplateValue InsertValue(int n, string name, string s)
        {
            TemplateValue v = new TemplateValue();
            v.name = name;
            v.vType = Values.GetTPLVType(v.name);
            v.adjType = Values.GetTPLType(v.name);
            v.Val(s);
            vals.Insert(n, v);
            return v;
        }
    }

    public class JKThings : List<JKThing>
    {

    }

    public class JKSurface : Polygon
    {
        public readonly int uniqueID;
        private static int uniqueIDGen = 0;
        public JKSurface adjoin;
        public Adjoin.Flag adjoinFlags = Adjoin.Flag.Passable | Adjoin.Flag.AllowsSound | Adjoin.Flag.RenderPastAdjoin;
        public string material = "dflt.mat";
        public Surface.Flag surfaceFlags = Surface.Flag.Impassible;
        public Surface.FaceFlag faceFlags = Surface.FaceFlag.Unknown;
        public JKSector sector;
        private Surface.GeoModeType _geo = Surface.GeoModeType.Textured;
        public Surface.GeoModeType geo
        {
            get
            {
                return _geo;
            }
        }
        public void SetGeo(string who, Surface.GeoModeType __geo)
        {
            _geo = __geo;
            geoChanges.Add(new GeoChange(who, __geo));
        }

        public class GeoChange
        {
            public string who;
            public Surface.GeoModeType geo;

            public GeoChange(string _who, Surface.GeoModeType _geo)
            {
                who = _who;
                geo = _geo;
            }
        }
        public List<GeoChange> geoChanges = new List<GeoChange>();

        public Surface.LightModeType light = Surface.LightModeType.Gouraud;
        public Surface.TextureModeType tex = Surface.TextureModeType.Perspective;
        public double extraLight;
        public int mark;
        public int d3did = -1;
        public new /*should this be virtual or otherwise separate from Polygon.num?*/ int num;
        public int nadj;
        public int nmat;
        public Vec2 texScale = Vec2.One;

        public List<Vec3> vertNormals = new List<Vec3>();   // haxxed just to get vert normals from imported 3dos able to be exported

        public JKSurface(JKSector _Sector)
        {
            sector = _Sector;
            uniqueID = uniqueIDGen++;
        }

        public void Assign(JKSurface surf)
        {
            if (surf.adjoinFlags != 0)
                adjoinFlags = surf.adjoinFlags;
            if (!string.IsNullOrEmpty(surf.material))
                material = surf.material;
            if (surf.surfaceFlags != 0)
                surfaceFlags = surf.surfaceFlags;
            if (surf.faceFlags != 0)
                faceFlags = surf.faceFlags;
            if (surf.geo != 0)
                SetGeo("Assign", surf.geo);
            light = surf.light;
            extraLight = surf.extraLight;
            texScale = surf.texScale;
        }

        public void SerializeVerticesTo(ZipUtil zip)
        {
            for (int i = 0; i < vertices.Count; i++)
            {
                Vertex v = vertices[i];
                TexVertex tv = texvertices[i];

                zip.WriteInt(v.num);
                zip.WriteVec2(tv.v);
                zip.WriteDouble(tv.intensity);
                zip.WriteDouble(tv.r);
                zip.WriteDouble(tv.g);
                zip.WriteDouble(tv.b);
            }
        }

        public void SerializeVerticesFrom(ZipUtil zip)
        {
            for (int i = 0; i < vertices.Count; i++)
            {
                int nvx = zip.ReadInt();

                vertices[i] = sector.vertices[nvx];
                TexVertex tv = texvertices[i];

                tv.v = zip.ReadVec2();
                tv.intensity = zip.ReadDouble();
                tv.r = zip.ReadDouble();
                tv.g = zip.ReadDouble();
                tv.b = zip.ReadDouble();
            }
        }

        public void SerializeValTo(ZipUtil zip)
        {
            if (adjoin == null)
            {
                zip.WriteInt(-1);//adjSC
                zip.WriteInt(0);//adjSF
                zip.WriteInt(-1);//adjID
            }
            else
            {
                zip.WriteInt(adjoin.sector.num);//adjSC
                zip.WriteInt(adjoin.num);//adjSF
                zip.WriteInt(adjoin.sector.ID);//adjID
            }

            zip.WriteUInt((uint)adjoinFlags);
            zip.WriteString(material);
            zip.WriteUInt((uint)surfaceFlags);
            zip.WriteUInt((uint)faceFlags);
            zip.WriteInt((int)geo);
            zip.WriteInt((int)light);
            zip.WriteInt((int)tex);
            zip.WriteDouble(extraLight);
            zip.WriteVec2(texScale);
            zip.WriteInt(vertices.Count);
        }

        public void SerializeValFrom(ZipUtil zip)
        {
            int adjSC = zip.ReadInt();
            int adjSF = zip.ReadInt();
            int adjID = zip.ReadInt();
            mark = adjSC * 65536 + adjSF;
            nadj = adjID;
            adjoinFlags = (Adjoin.Flag)zip.ReadUInt();
            material = zip.ReadString();
            surfaceFlags = (Surface.Flag)zip.ReadUInt();
            faceFlags = (Surface.FaceFlag)zip.ReadUInt();
            SetGeo("SerializeValFrom", (Surface.GeoModeType)zip.ReadInt());
            light = (Surface.LightModeType)zip.ReadInt();
            tex = (Surface.TextureModeType)zip.ReadInt();
            extraLight = zip.ReadDouble();
            texScale = zip.ReadVec2();
            int numVerts = zip.ReadInt();
        }

        public int NextVX(int n)
        {
            return WrapIndex(n + 1);
        }

        public int PrevVX(int n)
        {
            return WrapIndex(n - 1);
        }

        public void AddVertices(JKVertex[] verts)
        {
            foreach (JKVertex j in verts)
                AddVertex(j);
        }

        public void CheckIfFloor()
        {
            double ca = normal * Vec3.PosZ;
            if(ca <= 1.0 && ca > Math.Cos(Math.PI/4.0))
                surfaceFlags |= Surface.Flag.Floor;
            else
                surfaceFlags &= ~Surface.Flag.Floor;
        }

        public Vec3? RefVector
        {
            get
            {
                if (vertices.Count < 2)
                    return null;

                Vec3 v = (vertices[1].v - vertices[0].v).Unit;
                if (MathLib.CompareScalar(v.Magnitude, 0.0) == 0)
                    return null;

                return v;
            }
        }

        public void NewRecalcAll()
        {
            Recalc();
            CalcUV(0);
            CheckIfFloor();
        }

        public void RecalcAll()
        {
            Recalc();
            Vec3 un, vn;
            CalcUVNormals(out un, out vn);
            ArrangeTexture(0, un, vn);
        }

        public void CalcUV(int org)
        {
            Vec3 un, vn;
            CalcDefaultUVNormals(org, out un, out vn);
            ArrangeTexture(org, un, vn);
        }

        public void FindUVScales()
        {
            Vec3 un, vn;
            CalcUVNormals(out un, out vn);

            int nv2, nv3;
            if (!FindTri(0, out nv2, out nv3))
                return;

            Vertex v1 = vertices[0]; TexVertex tv1 = texvertices[0];
            Vertex v2 = vertices[nv2]; TexVertex tv2 = texvertices[nv2];

            double d1 = tv2.v.x - tv1.v.x;
            if (d1 == 0.0)
            {
                v2 = vertices[nv3]; tv2 = texvertices[nv3];
                d1 = tv2.v.x - tv1.v.x;
            }

            double d2 = (un * (v2.v - v1.v)) * GlobalVars.PixelsPerUnit;
            texScale.x = d1 / d2;

            v2 = vertices[nv2]; tv2 = texvertices[nv2];

            d1 = tv2.v.y - tv1.v.y;
            if (d1 == 0.0)
            {
                v2 = vertices[nv3]; tv2 = texvertices[nv3];
                d1 = tv2.v.y - tv1.v.y;
            }

            d2 = (vn * (v2.v - v1.v)) * GlobalVars.PixelsPerUnit;
            texScale.y = d1 / d2;
        }

        public void CalcUVNormals(out Vec3 un, out Vec3 vn)
        {
            CalcUVNormalsFromVX(0, out un, out vn);
        }

        public void CalcUVNormalsFrom(int nv1, int nv2, out Vec3 un, out Vec3 vn)
        {
            SysCalcUVNormalsFrom(nv1, out un, out vn);
        }

        public void CalcUVNormalsFromVX(int nvx, out Vec3 un, out Vec3 vn)
        {
            SysCalcUVNormalsFrom(nvx, out un, out vn);
        }

        public void CalcDefaultUVNormals(int orgvx, out Vec3 un, out Vec3 vn)
        {
            Vec3 v1 = vertices[orgvx].v;
            Vec3 v2 = vertices[NextVX(orgvx)].v;

            un = (v2 - v1).Unit;

            vn = un ^ normal;
        }

        private void SysCalcUVNormalsFrom(int nv1, out Vec3 un, out Vec3 vn)
        {
            int nv2, nv3;
            if (!FindTri(nv1, out nv2, out nv3))
            {
                un = Vec3.Zero;
                vn = Vec3.Zero;
                return;
            }

            Vec3 v1 = vertices[nv1].v; Vec2 tv1 = texvertices[nv1].v;
            Vec3 v2 = vertices[nv2].v; Vec2 tv2 = texvertices[nv2].v;
            Vec3 v3 = vertices[nv3].v; Vec2 tv3 = texvertices[nv3].v;

            Vec3 a = v1 - v2;
            Vec3 b = v3 - v2;
            Vec2 d1 = tv1 - tv2;
            Vec2 d2 = tv3 - tv2;

            un = new Vec3(d2.y * a.x - d1.y * b.x,
                          d2.y * a.y - d1.y * b.y,
                          d2.y * a.z - d1.y * b.z);

            vn = new Vec3(d2.x * a.x - d1.x * b.x,
                          d2.x * a.y - d1.x * b.y,
                          d2.x * a.z - d1.x * b.z);

            if (d2.y * d1.x < d1.y * d2.x)
                un = -un;

            if (d2.x * d1.y < d1.x * d2.y)
                vn = -vn;

            un = un.Unit;
            if (MathLib.CompareScalar(un.Magnitude, 0.0) == 0)
                un = a.Unit;

            Vec3 av = normal ^ un;

            vn = av;

            UpdateSurfUVData(un, vn);
        }

        public void UpdateSurfUVData(Vec3 un, Vec3 vn)
        {
            // all this work to implement SysCalcUVNormalsFrom and this core method is commented out in JED?  WTF

        }

        public void ArrangeTexture(int orgvx, Vec3 un, Vec3 vn)
        {
            ArrangeTextureBy(un, vn, vertices[orgvx].v, texvertices[orgvx].v);
        }

        public void ArrangeTextureBy(Vec3 un, Vec3 vn, Vec3 refv, Vec2 reftv)
        {
            for (int x = 0; x < vertices.Count; x++)
            {
                Vec3 v = vertices[x].v;

                double tu = Utils.UVRound(reftv.x + un*(v-refv)*GlobalVars.PixelsPerUnit*texScale.x);

                double tv = (surfaceFlags & Surface.Flag.Flipped) != 0
                    ? Utils.UVRound(reftv.y + -vn * (v - refv) * GlobalVars.PixelsPerUnit * texScale.y)
                    : Utils.UVRound(reftv.y + vn * (v - refv) * GlobalVars.PixelsPerUnit * texScale.y);

                texvertices[x].v = new Vec2(tu, tv);
            }
        }

        public bool FindTri(int nv1, out int nv2, out int nv3)
        {
            nv2 = -1;
            nv3 = -1;
            if (vertices.Count < 3)
                return false;

            Vec3 v0 = vertices[nv1].v;
            int nvxs = vertices.Count - 1;

            int n = NextVX(nv1);
            double cdist = -1.0;
            int cn = -1;

            Vec3 v1, v2;
            bool good = false;
            for (int i = 0; i < nvxs; i++)
            {
                v1 = vertices[n].v;

                double dist = (v1 - v0).Magnitude;
                if (dist > cdist)
                {
                    cn = n;
                    cdist = dist;
                }

                if (dist > 0.0001)
                {
                    good = true;
                    nv2 = n;
                    break;
                }

                n = NextVX(n);
                nvxs--;
            }

            if (!good)
            {
                if (cn == -1)
                    return false;

                nv2 = cn;
            }

            v1 = vertices[nv2].v;
            good = false;
            n = NextVX(nv2);

            Vec3 a = ((v1 - v0) ^ normal).Unit;
            if (MathLib.CompareScalar(a.Magnitude, 0.0) == 0)
                return false;

            cdist = -1.0;
            cn = -1;

            for (int i = 0; i < nvxs; i++)
            {
                v2 = vertices[n].v;

                double dist = Math.Abs(a * (v2-v1));
                if (dist > cdist)
                {
                    cn = n;
                    cdist = dist;
                }

                if (dist > 0.0001)
                {
                    good = true;
                    nv3 = n;
                    break;
                }

                n = NextVX(n);
            }

            if (!good)
            {
                if (cn == -1)
                    return false;

                nv3 = cn;
            }

            return true;
        }
    }

    public class JKSurfaces : List<JKSurface>
    {
        public Vertices verticies = new Vertices();
    }

    public class JKSector
    {
        public int ID;
        public int num;
        public Sector.Flag flags;
        public double ambient = 5.0;
        public double extra = 0.0;
        public string colormap;
        public Color4 tint = Color4.black();
        public string sound = string.Empty;
        public double snd_vol = 0.0;

        public JKLevel level;
        public JKSurfaces surfaces = new JKSurfaces();
        public JKVertices vertices = new JKVertices();
        public int layer;
        public int mark;

        public JKSector(JKLevel _level)
        {
            level = _level;

            if (Program.DemoMode)
                colormap = "09fuel.cmp";
            else
                colormap = "dflt.cmp";
        }

        public void Assign(JKSector sec)
        {
            flags = sec.flags;
            ambient = sec.ambient;
            extra = sec.extra;
            colormap = sec.colormap;
            tint = sec.tint;
            sound = sec.sound;
            snd_vol = sec.snd_vol;
            layer = sec.layer;
        }

        public void SerializeValTo(ZipUtil zip)
        {
            zip.WriteUInt((uint)flags);
            zip.WriteDouble(ambient);
            zip.WriteDouble(extra);
            zip.WriteString(colormap);
            zip.WriteColor4(tint);
            zip.WriteString(sound);
            zip.WriteDouble(snd_vol);
            zip.WriteInt(layer);
            zip.WriteInt(vertices.Count);
            zip.WriteInt(surfaces.Count);
        }

        public void SerializeValFrom(ZipUtil zip)
        {
            flags = (Sector.Flag)zip.ReadUInt();
            ambient = zip.ReadDouble();
            extra = zip.ReadDouble();
            colormap = zip.ReadString();
            tint = zip.ReadColor4();
            sound = zip.ReadString();
            snd_vol = zip.ReadDouble();
            layer = zip.ReadInt();
            int numVerts = zip.ReadInt();
            int numSurfs = zip.ReadInt();
        }

        public JKVertex NewVertex()
        {
            JKVertex v = new JKVertex(this);
            v.num = vertices.Count;
            vertices.Add(v);
            return v;
        }

        public JKVertex NewVertex(Vec3 v)
        {
            JKVertex j = NewVertex();
            j.v = v;
            return j;
        }

        public JKVertex NewVertex(double x, double y, double z)
        {
            return NewVertex(new Vec3(x, y, z));
        }

        public JKSurface NewSurface()
        {
            JKSurface s = new JKSurface(this);
            return s;
        }

        public JKVertex[] GetVertices(IEnumerable<int> indices)
        {
            List<JKVertex> j = new List<JKVertex>();
            foreach (int i in indices)
                j.Add(vertices[i]);
            return j.ToArray();
        }

        public void Renumber()
        {
            for (int x = 0; x < vertices.Count; x++)
                vertices[x].num = x;

            for (int x = 0; x < surfaces.Count; x++)
                surfaces[x].num = x;
        }

        public int FindVX(Vec3 v)
        {
            for (int x = 0; x < vertices.Count; x++)
                if (vertices[x].v == v)
                    return x;

            return -1;
        }

        public JKVertex AddVertex(Vec3 v)
        {
            JKVertex j = null;

            int i = FindVX(v);
            if (i != -1)
                j = vertices[i];
            else
                j = NewVertex();

            j.v = v;
            return j;
        }
    }

    public class JKSectors : List<JKSector>
    {

    }

    public class JKCOG
    {
        public string name;
        public COGValues vals = new COGValues();

    }

    public class JKCOGS : List<JKCOG>
    {

    }

    public class JKLHeader
    {
        public int version;
        public double gravity;
        public double ceilingskyZ;
        public double horDistance;
        public double horPixelsPerRev;
        public Vec2 horSkyOff = Vec2.Zero;
        public Vec2 ceilingSkyOff = Vec2.Zero;
        public double[] mipMapDist = new double[4];
        public double[] LODDist = new double[4];
        public double perspDist;
        public double gouraudDist;
    }

    public class JKHierarchyNodes : List<Model.Node>
    {

    }

    public class CogGroup
    {
        public string Name;

        public CogGroup(string _Name)
        {
            Name = _Name;
        }

        public override string ToString()
        {
            return Name;
        }
    }

    public class CogGroupList : List<CogGroup>
    {
        public CogGroup GetGroup(string nm)
        {
            foreach (CogGroup cg in this)
                if (cg.Name.Equals(nm, StringComparison.InvariantCultureIgnoreCase))
                    return cg;

            return null;
        }
    }

    public class JKLevel
    {
        public JKLHeader header = new JKLHeader();
        public bool mots = false;
        public JKSectors sectors = new JKSectors();
        public JKCOGS cogs = new JKCOGS();
        public CogGroupList coggroups = new CogGroupList();
        public JKThings things = new JKThings();
        public JedLights lights = new JedLights();
        public List<string> layers = new List<string>();
        public JKModel.HNodes hnodes = new JKModel.HNodes();
        public string masterCMP = string.Empty;
        public double ppUnit = 320.0;
        public string layerVisString = string.Empty;
        public int CurID = 0;

        public int getNewID()
        {
            return CurID++;
        }

        public int GetThingByID(int id)
        {
            for (int x = 0; x < things.Count; x++)
                if (things[x].ID == id)
                    return x;

            return -1;
        }

        public int GetSectorByID(int id)
        {
            for (int x = 0; x < sectors.Count; x++)
                if (sectors[x].ID == id)
                    return x;

            return -1;
        }

        public int GetLightByID(int id)
        {
            for (int x = 0; x < lights.Count; x++)
                if (lights[x].ID == id)
                    return x;

            return -1;
        }

        public int GetNodeByID(int id)
        {
            for (int x = 0; x < hnodes.Count; x++)
                if (hnodes[x].id == id)
                    return x;

            return -1;
        }

        public JKSector NewSector()
        {
            JKSector s = new JKSector(this);
            s.ID = getNewID();
            return s;
        }

        public JKThing NewThing()
        {
            JKThing t = new JKThing(this);
            t.ID = getNewID();
            return t;
        }

        public JedLight NewLight()
        {
            JedLight j = new JedLight();
            j.ID = getNewID();
            return j;
        }

        public JKModel.HNode New3DONode()
        {
            JKModel.HNode n = new JKModel.HNode();
            n.id = getNewID();
            return n;
        }

        public void RenumSecs()
        {
            for (int x = 0; x < sectors.Count; x++)
                sectors[x].num = x;
        }

        public void RenumThings()
        {
            for (int x = 0; x < things.Count; x++)
                things[x].num = x;
        }

        public string GetLayerName(int i)
        {
            if (i < 0 || i >= layers.Count)
                return "LAYER" + i.ToString();

            return layers[i];
        }

        public int AddLayer(string name)
        {
            int i = layers.IndexOf(name);
            if (i == -1)
            {
                i = layers.Count;
                layers.Add(name);
            }
            return i;
        }

        public string GetMasterCMP()
        {
            string cmp = masterCMP;

            if (string.IsNullOrEmpty(cmp) && sectors.Count > 0)
                cmp = sectors[0].colormap;

            if (string.IsNullOrEmpty(cmp))
            {
                if (Program.DemoMode)
                    cmp = "09fuel.cmp";
                else
                    cmp = "dflt.cmp";
            }

            return cmp;
        }

        public void Clear()
        {
            GlobalVars.NedMain?.HaltLightCalc();

            sectors.Clear();
            things.Clear();
            cogs.Clear();
            coggroups.Clear();

            hnodes.Clear();

            lights.Clear();
            layers.Clear();
            mots = false;
            masterCMP = string.Empty;
            layerVisString = string.Empty;
            ppUnit = 320.0;
            CurID = 0;

            GlobalVars.NedMain?.ResumeLightCalc();
        }

        public void SetDefaultHeader()
        {
            header.version=1;
            header.gravity = 4.0;
            header.ceilingskyZ = 15.0;
            header.horDistance = 100.0;
            header.horPixelsPerRev = 768.0;
            header.horSkyOff = Vec2.Zero;
            header.ceilingSkyOff = Vec2.Zero;
            header.mipMapDist[0] = 1.0; header.mipMapDist[1] = 2.0; header.mipMapDist[2] = 3.0; header.mipMapDist[3] = 4.0;
            header.LODDist[0] = 0.3; header.LODDist[1] = 0.6; header.LODDist[2] = 0.9; header.LODDist[3] = 1.2;
            header.perspDist = 2.0;
            header.gouraudDist = 2.0;
        }

        public void AddMissingLayers()
        {
            foreach (JKSector s in sectors)
                if (s.layer < 0 || s.layer >= layers.Count)
                    s.layer = AddLayer("Layer" + s.layer.ToString());

            foreach (JKThing s in things)
                if (s.layer < 0 || s.layer >= layers.Count)
                    s.layer = AddLayer("Layer" + s.layer.ToString());

            foreach (JedLight s in lights)
                if (s.layer < 0 || s.layer >= layers.Count)
                    s.layer = AddLayer("Layer" + s.layer.ToString());
        }

        public JKSector GetSectorN(int n)
        {
            if (n < 0 || n >= sectors.Count)
                return null;

            return sectors[n];
        }

        public JKSurface GetSurfaceN(int n)
        {
            int nsf = 0;
            foreach (JKSector sec in sectors)
            {
                if (n >= nsf && n < (nsf + sec.surfaces.Count))
                    return sec.surfaces[n - nsf];

                nsf += sec.surfaces.Count;
            }

            return null;
        }

        public JKThing GetThingN(int n)
        {
            if (n < 0 || n >= things.Count)
                return null;

            return things[n];
        }

        public int GetGlobalSFN(int sec, int surf)
        {
            if (sec < 0 || sec >= sectors.Count)
                return -1;

            int nsf = 0;
            for (int s = 0; s < sec; s++)
                nsf += sectors[s].surfaces.Count;

            return nsf + surf;
        }

        public static int SFtoInt(int sc, int sf)
        {
            return (sc << 16) | sf;
        }

        public static int EDtoInt(int sc, int sf, int ed)
        {
            return (sc << 18) | (sf << 8) | ed;
        }

        public static int VXtoInt(int sc, int vx)
        {
            return (sc << 16) | vx;
        }

        public static int FRtoInt(int th, int fr)
        {
            return (th << 16) | fr;
        }

        public static int GetsfSC(int scsf)
        {
            return scsf >> 16;
        }

        public static int GetsfSF(int scsf)
        {
            return scsf & 0xFFFF;
        }

        public static int GetedSC(int scsfed)
        {
            return scsfed >> 18;
        }

        public static int GetedSF(int scsfed)
        {
            return (scsfed >> 8) & 0x3FF;
        }

        public static int GetedED(int scsfed)
        {
            return scsfed & 0xFF;
        }

        public static int GetvxSC(int scvx)
        {
            return scvx >> 16;
        }

        public static int GetvxVX(int scvx)
        {
            return scvx & 0xFFFF;
        }

        public static int GetfrTH(int thfr)
        {
            return thfr >> 16;
        }

        public static int GetfrFR(int thfr)
        {
            return thfr & 0xFFFF;
        }

        private void SkipToEOS(Tokenizer t)
        {
            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                if (ln.Trim().Equals("end", StringComparison.InvariantCultureIgnoreCase))
                    break;
            }
        }

        private void LoadHeader_JED(Tokenizer t)
        {
            t.EatToken();
            header.version = t.ReadInt();
            header.gravity = t.ReadDouble();
            header.ceilingskyZ = t.ReadDouble();
            header.perspDist = t.ReadDouble();
            header.gouraudDist = t.ReadDouble();
            masterCMP = t.ReadToken();
            ppUnit = t.ReadDouble();

            t.EatToken();
            header.horDistance = t.ReadDouble();
            header.horPixelsPerRev = t.ReadDouble();
            header.horSkyOff.x = t.ReadDouble();
            header.horSkyOff.y = t.ReadDouble();

            t.EatToken();
            header.ceilingSkyOff.x = t.ReadDouble();
            header.ceilingSkyOff.y = t.ReadDouble();

            t.EatToken();
            header.mipMapDist[0] = t.ReadDouble();
            header.mipMapDist[1] = t.ReadDouble();
            header.mipMapDist[2] = t.ReadDouble();
            header.mipMapDist[3] = t.ReadDouble();

            t.EatToken();
            header.LODDist[0] = t.ReadDouble();
            header.LODDist[1] = t.ReadDouble();
            header.LODDist[2] = t.ReadDouble();
            header.LODDist[3] = t.ReadDouble();

            if (masterCMP == "-")
                masterCMP = string.Empty;
        }

        private void LoadGeometry_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Geometry");

            t.ReadToken();
            int numsec = t.ReadInt();
            for (int s = 0; s < numsec; s++)
            {
                DO_Progress(t);

                JKSector sec = NewSector();
                sectors.Add(sec);

                sec.colormap = t.ReadToken();
                sec.sound = t.ReadToken();
                sec.snd_vol = t.ReadDouble();
                sec.flags = (Sector.Flag)t.ReadHex();
                sec.ambient = t.ReadDouble();
                sec.extra = t.ReadDouble();
                sec.tint.r = t.ReadDouble();
                sec.tint.g = t.ReadDouble();
                sec.tint.b = t.ReadDouble();
                sec.layer = t.ReadInt();

                if (sec.colormap == "-")
                    sec.colormap = string.Empty;

                if (sec.sound == "-")
                    sec.sound = string.Empty;

                t.EatToken();
                int numverts = t.ReadInt();
                for (int vx = 0; vx < numverts; vx++)
                {
                    JKVertex v = sec.NewVertex();
                    v.v.x = t.ReadDouble();
                    v.v.y = t.ReadDouble();
                    v.v.z = t.ReadDouble();
                }

                t.EatToken();
                int numsurfs = t.ReadInt();
                for (int sf = 0; sf < numsurfs; sf++)
                {
                    JKSurface surf = sec.NewSurface();
                    sec.surfaces.Add(surf);

                    surf.material = t.ReadToken();
                    surf.surfaceFlags = (Surface.Flag)t.ReadHex();
                    surf.faceFlags = (Surface.FaceFlag)t.ReadHex();
                    surf.SetGeo("LoadGeometry_JED", (Surface.GeoModeType)t.ReadInt());
                    surf.light = (Surface.LightModeType)t.ReadInt();
                    surf.tex = (Surface.TextureModeType)t.ReadInt();
                    surf.extraLight = t.ReadDouble();
                    int nsv = t.ReadInt();
                    int asec = t.ReadInt();
                    int asf = t.ReadInt();
                    surf.adjoinFlags = (Adjoin.Flag)t.ReadHex();
                    surf.texScale.x = t.ReadDouble();
                    surf.texScale.y = t.ReadDouble();

                    if (surf.material == "-")
                        surf.material = string.Empty;
                    surf.num = (asec << 16) | (asf & 0xFFFF);

                    bool endOfLine = true;
                    for (int vx = 0; vx < nsv; vx++)
                    {
                        int nv = t.ReadInt();
                        surf.AddVertex(sec.vertices[nv]);

                        surf.texvertices[vx].v.x = t.ReadDouble();
                        surf.texvertices[vx].v.y = t.ReadDouble();
                        surf.texvertices[vx].intensity = t.ReadDouble(out endOfLine);
                    }

                    // load RGB, if exist
                    if (!endOfLine)
                    {
                        for (int vx = 0; vx < nsv; vx++)
                        {
                            surf.texvertices[vx].r = t.ReadDouble();
                            surf.texvertices[vx].g = t.ReadDouble();
                            surf.texvertices[vx].b = t.ReadDouble();
                        }
                    }
                }
            }

            // link adjoin references now that we've constructed everything
            for (int s = 0; s < numsec; s++)
            {
                JKSector sec = sectors[s];

                for (int sf = 0; sf < sec.surfaces.Count; sf++)
                {
                    JKSurface surf = sec.surfaces[sf];

                    int asec = surf.num >> 16;
                    int asf = surf.num & 0xFFFF;

                    if (asec < 0 || asec >= sectors.Count ||
                       asf < 0 || asf >= sectors[asec].surfaces.Count)
                        surf.adjoin = null;
                    else
                        surf.adjoin = sectors[asec].surfaces[asf];
                }
            }
        }

        private void LoadThings_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Things");

            t.EatToken();
            bool endOfLine;
            int numthings = t.ReadInt(out endOfLine);
            if (!endOfLine)
            {
                int maxthings = t.ReadInt();
            }
            for (int x = 0; x < numthings; x++)
            {
                DO_Progress(t);

                JKThing th = NewThing();
                things.Add(th);

                th.name = t.ReadToken();
                th.pos.x = t.ReadDouble();
                th.pos.y = t.ReadDouble();
                th.pos.z = t.ReadDouble();
                th.orient.pitch = t.ReadDouble();
                th.orient.yaw = t.ReadDouble();
                th.orient.roll = t.ReadDouble();
                int ns = t.ReadInt();
                th.layer = t.ReadInt();

                if (ns < 0)
                    th.sec = null;
                else
                    th.sec = sectors[ns];

                string[] valpairs = t.ReadLine(false, true).Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string valpair in valpairs)
                {
                    string[] parts = valpair.Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.Length == 2)
                    {
                        TemplateValue tv = new TemplateValue();
                        th.vals.Add(tv);

                        tv.name = parts[0];
                        tv.vType = Values.GetTPLVType(tv.name);
                        tv.adjType = Values.GetTPLType(tv.name);
                        tv.stringval = parts[1];
                    }
                }
            }
        }

        private void LoadCogGroups_NED(Tokenizer t)
        {
            ProgressDialog.Message("Loading CogGroups");

            t.EatToken();
            int numgroups = t.ReadInt();
            for (int i = 0; i < numgroups; i++)
            {
                DO_Progress(t);

                bool endLine;
                string cgName = t.ReadToken(out endLine);
                CogGroup cg = new CogGroup(cgName);
                coggroups.Add(cg);

                if (endLine)
                    continue;


            }
        }

        private void LoadCogs_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Cogs");

            t.EatToken();
            int numcogs = t.ReadInt();
            for (int i = 0; i < numcogs; i++)
            {
                DO_Progress(t);

                JKCOG cog = new JKCOG();
                bool endLine;
                cog.name = t.ReadToken(out endLine);
                cogs.Add(cog);

                if (endLine)
                    continue;

                string[] valsets = t.ReadLine(false, true).Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                foreach (string valset in valsets)
                {
                    COGValue cv = new COGValue();
                    cog.vals.Add(cv);

                    int ps = valset.IndexOf(':');
                    int pe = valset.IndexOf('=');
                    cv.name = valset.Substring(0, ps);
                    string w = valset.Substring(ps + 1, pe - ps-1);
                    cv.vType = Values.GetCOGVType(w);
                    cv.cogType = Values.GetCOGType(w);
                    cv.Val(valset.Substring(pe + 1));
                }
            }
        }

        private void LoadLights_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Lights");

            t.EatToken();
            int numlights = t.ReadInt();
            for (int i = 0; i < numlights; i++)
            {
                JedLight l = NewLight();
                lights.Add(l);

                l.Pos.x = t.ReadDouble();
                l.Pos.y = t.ReadDouble();
                l.Pos.z = t.ReadDouble();
                l.layer = t.ReadInt();
                l.intensity = t.ReadDouble();
                l.range = t.ReadDouble();
                l.flags = (LightFlag)t.ReadHex();
                l.rgbrange = t.ReadDouble();
                l.rgbintensity = t.ReadDouble();
                l.clr.r = t.ReadDouble();
                l.clr.g = t.ReadDouble();
                l.clr.b = t.ReadDouble();

                if (!mots)
                {
                    l.rgbintensity = l.intensity;
                    l.clr = Color4.white();
                }
            }
        }

        private void LoadLayers_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Layers");

            t.EatToken();
            int numlayers = t.ReadInt();
            for (int i = 0; i < numlayers; i++)
            {
                layers.Add(t.ReadToken());
            }
        }

        private void LoadLVis_JED(Tokenizer t)
        {
            layerVisString = t.ReadToken();
        }

        private void LoadHierarchy_JED(Tokenizer t)
        {
            ProgressDialog.Message("Loading Hierarchy");

            t.EatToken();
            int numnodes = t.ReadInt();
            for (int x = 0; x < numnodes; x++)
                hnodes.Add(JKModel.HNode.NewFromString(t.ReadLine()));
        }

        private void DO_Progress(Tokenizer t)
        {
            ProgressDialog.UpdateByFileRead(t.RoughStreamPosition);
        }

        public void JKLPostLoad()
        {

        }

        private class TAdjoin
        {
            public int mirror;
            public int surf;
            public double dist;
            public Adjoin.Flag flags;
        }

        private class TVXIntensity
        {
            public double intensity;
            public double r;
            public double g;
            public double b;
        }

        private class TTempSurf
        {
            public int nmat;
            public Surface.Flag surfflags;
            public Surface.FaceFlag faceflags;
            public Surface.GeoModeType geo;
            public Surface.LightModeType light;
            public Surface.TextureModeType tex;
            public int nadj;
            public double extra;
            public List<int> VXs = new List<int>();
            public List<int> TVXs = new List<int>();
            public List<TVXIntensity> intensities = new List<TVXIntensity>();
        }

        private void LoadGeometry_JKL(Tokenizer t, ref List<string> cmps, ref List<Vec3> GVXList, ref List<Vec2> GTXVXList, ref List<TAdjoin> GAdjoins, ref List<TTempSurf> GSurfList)
        {
            t.EatToken(2);
            int numcmps = t.ReadInt();
            for (int x = 0; x < numcmps; x++)
            {
                t.EatToken();
                cmps.Add(t.ReadToken());
            }

            DO_Progress(t);

            ProgressDialog.Message("Loading Vertices");
            t.EatToken(2);
            int numverts = t.ReadInt();
            for(int x=0; x<numverts; x++)
            {
                t.EatToken();
                Vec3 v = new Vec3();
                v.x = t.ReadDouble();
                v.y = t.ReadDouble();
                v.z = t.ReadDouble();
                GVXList.Add(v);
            }

            ProgressDialog.Message("Loading Texture Vertices");
            t.EatToken(3);
            int numtexverts = t.ReadInt();
            for(int x=0; x<numtexverts; x++)
            {
                t.EatToken();
                Vec2 v= new Vec2();
                v.x = t.ReadDouble();
                v.y = t.ReadDouble();
                GTXVXList.Add(v);
            }

            DO_Progress(t);
            ProgressDialog.Message("Loading Adjoins");
            t.EatToken(2);
            int numadjoins = t.ReadInt();
            for(int x=0; x<numadjoins; x++)
            {
                TAdjoin adj = new TAdjoin();
                t.EatToken();
                adj.flags = (Adjoin.Flag)t.ReadFlags();
                adj.mirror = t.ReadInt();
                adj.dist = t.ReadDouble();
                adj.surf = -1;
                GAdjoins.Add(adj);
            }

            ProgressDialog.Message("Loading Surfaces");
            t.EatToken(2);
            int numsurfs = t.ReadInt();
            for(int x=0; x<numsurfs; x++)
            {
                DO_Progress(t);

                TTempSurf surf = new TTempSurf();
                GSurfList.Add(surf);

                t.EatToken();
                surf.nmat = t.ReadInt();
                surf.surfflags = (Surface.Flag)t.ReadFlags();
                surf.faceflags = (Surface.FaceFlag)t.ReadFlags();
                surf.geo = (Surface.GeoModeType)t.ReadInt();
                surf.light = (Surface.LightModeType)t.ReadInt();
                surf.tex = (Surface.TextureModeType)t.ReadInt();
                surf.nadj = t.ReadInt();
                surf.extra = t.ReadDouble();
                int nvx = t.ReadInt();
                for(int i=0; i<nvx; i++)
                {
                    surf.VXs.Add(t.ReadInt());
                    surf.TVXs.Add(t.ReadInt());
                }
                string[] intensities = t.ReadLine().Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                if(intensities.Length > nvx)
                {
                    mots = true;
                    for(int i=0; i<nvx; i++)
                    {
                        TVXIntensity tvi = new TVXIntensity();
                        surf.intensities.Add(tvi);
                        tvi.intensity = double.Parse(intensities[i*4+0]);
                        tvi.r = double.Parse(intensities[i*4+1]);
                        tvi.g = double.Parse(intensities[i*4+2]);
                        tvi.b = double.Parse(intensities[i*4+3]);
                    }
                } else {
                    for(int i=0; i<nvx; i++)
                    {
                        TVXIntensity tvi = new TVXIntensity();
                        surf.intensities.Add(tvi);
                        tvi.intensity = double.Parse(intensities[i]);
                        tvi.r = tvi.intensity;
                        tvi.g = tvi.intensity;
                        tvi.b = tvi.intensity;
                    }
                }
            }

            DO_Progress(t);
            ProgressDialog.Message("Loading Normals");
            t.EatLine(numsurfs);

            Options.SetMOTS(mots);
        }

        private void LoadHeader_JKL(Tokenizer t)
        {
            t.EatToken();
            header.version = t.ReadInt();

            t.EatToken(2);
            header.gravity = t.ReadDouble();

            t.EatToken(3);
            header.ceilingskyZ = t.ReadDouble();

            t.EatToken(2);
            header.horDistance = t.ReadDouble();

            t.EatToken(4);
            header.horPixelsPerRev = t.ReadDouble();

            t.EatToken(3);
            header.horSkyOff.x = t.ReadDouble();
            header.horSkyOff.y = t.ReadDouble();

            t.EatToken(3);
            header.ceilingSkyOff.x = t.ReadDouble();
            header.ceilingSkyOff.y = t.ReadDouble();

            t.EatToken(2);
            header.mipMapDist[0] = t.ReadDouble();
            header.mipMapDist[1] = t.ReadDouble();
            header.mipMapDist[2] = t.ReadDouble();
            header.mipMapDist[3] = t.ReadDouble();

            t.EatToken(2);
            header.LODDist[0] = t.ReadDouble();
            header.LODDist[1] = t.ReadDouble();
            header.LODDist[2] = t.ReadDouble();
            header.LODDist[3] = t.ReadDouble();

            t.EatToken(2);
            header.perspDist = t.ReadDouble();

            t.EatToken(2);
            header.gouraudDist = t.ReadDouble();
        }

        private void LoadSectors_JKL(Tokenizer t, List<string> cmps, List<string> mats, List<Vec3> GVXList, List<Vec2> GTXVXList, List<TTempSurf> GSurfList)
        {
            ProgressDialog.Message("Loading Sectors");

            t.EatToken(2);
            int numsec = t.ReadInt();
            for (int x = 0; x < numsec; x++)
            {
                List<int> SCVXList = new List<int>();

                DO_Progress(t);

                JKSector sec = NewSector();
                sectors.Add(sec);

                t.EatToken(2);

                t.EatToken();
                sec.flags = (Sector.Flag)t.ReadFlags();

                t.EatToken(2);
                sec.ambient = t.ReadDouble();

                t.EatToken(2);
                sec.extra = t.ReadDouble();

                t.EatToken();
                int ncmp = t.ReadInt();
                if (ncmp < 0 || ncmp >= cmps.Count)
                    sec.colormap = cmps[0];
                else
                    sec.colormap = cmps[ncmp];

                t.EatToken();
                sec.tint = Color4.white();
                sec.tint.r = t.ReadDouble();
                sec.tint.g = t.ReadDouble();
                sec.tint.b = t.ReadDouble();

                for (; ; )
                {
                    string tok = t.ReadToken();

                    if (tok.Equals("boundbox", StringComparison.InvariantCultureIgnoreCase) || tok.Equals("collidebox", StringComparison.InvariantCultureIgnoreCase))
                        t.EatToken(6);
                    else if (tok.Equals("center", StringComparison.InvariantCultureIgnoreCase))
                        t.EatToken(3);
                    else if (tok.Equals("radius", StringComparison.InvariantCultureIgnoreCase))
                        t.EatToken(1);
                    else if(tok.Equals("vertices", StringComparison.InvariantCultureIgnoreCase))
                        break;
                }

                int numverts = t.ReadInt();
                t.EatToken(numverts * 2);

                t.EatToken();
                int bsurf = t.ReadInt();
                int nsurfs = t.ReadInt();
                for (int isurf = bsurf; isurf < (bsurf + nsurfs); isurf++)
                {
                    TTempSurf tsurf = GSurfList[isurf];
                    JKSurface surf = sec.NewSurface();
                    sec.surfaces.Add(surf);

                    if (tsurf.nmat >= 0 && tsurf.nmat < mats.Count)
                        surf.material = mats[tsurf.nmat];
                    else
                        surf.material = string.Empty;

                    surf.surfaceFlags = tsurf.surfflags;
                    surf.faceFlags = tsurf.faceflags;
                    surf.SetGeo("LoadSectors_JKL", tsurf.geo);
                    surf.light = tsurf.light;
                    surf.tex = tsurf.tex;
                    surf.extraLight = tsurf.extra;    // this was missing from JED.. shouldnt it be here?
                    surf.nadj = tsurf.nadj;

                    if ((tsurf.surfflags & Surface.Flag.DoubleTextureScale) != 0)
                        surf.texScale = Vec2.One * 2.0;
                    else if ((tsurf.surfflags & Surface.Flag.HalfTextureScale) != 0)
                        surf.texScale = Vec2.One * 0.5;
                    else if ((tsurf.surfflags & Surface.Flag.EighthTextureScale) != 0)
                        surf.texScale = Vec2.One * 0.125;

                    for (int j = 0; j < tsurf.VXs.Count; j++)
                    {
                        int igvx = tsurf.VXs[j];
                        int ilvx = SCVXList.IndexOf(igvx);
                        if (ilvx == -1)
                        {
                            JKVertex jvx = sec.NewVertex();
                            jvx.v = GVXList[igvx];
                            SCVXList.Add(igvx);
                            surf.AddVertex(jvx);
                        }
                        else
                            surf.AddVertex(sec.vertices[ilvx]);

                        int itxvx = tsurf.TVXs[j];
                        if (itxvx < 0 || itxvx >= GTXVXList.Count)
                            itxvx = 0;

                        Vec2 txvx = GTXVXList[itxvx];
                        surf.texvertices[j].v = txvx;
                        TVXIntensity vxInt = tsurf.intensities[j];
                        surf.texvertices[j].intensity = vxInt.intensity;
                        surf.texvertices[j].r = vxInt.r;
                        surf.texvertices[j].g = vxInt.g;
                        surf.texvertices[j].b = vxInt.b;
                    }
                }
            }
        }

        private void LoadMats_JKL(Tokenizer t, ref List<string> mats)
        {
            t.EatToken(2);
            int nummats = t.ReadInt();
            for (int x = 0; x < nummats; x++)
            {
                string tok = t.ReadToken();
                if (tok.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                    break;

                mats.Add(t.ReadToken());

                t.EatToken(2);
            }
        }

        private void LoadThings_JKL(Tokenizer t)
        {
            ProgressDialog.Message("Loading Things");

            t.EatToken(2);
            int numthings = t.ReadInt();
            for (int x = 0; x < numthings; x++)
            {
                string tok = t.ReadToken();
                if (tok.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                    break;

                JKThing th = NewThing();
                things.Add(th);

                t.EatToken();

                th.name = t.ReadToken();

                th.pos = new Vec3();
                th.pos.x = t.ReadDouble();
                th.pos.y = t.ReadDouble();
                th.pos.z = t.ReadDouble();

                th.orient = new Orient();
                th.orient.pitch = t.ReadDouble();
                th.orient.yaw = t.ReadDouble();
                th.orient.roll = t.ReadDouble();

                bool endOfLine;
                int nsec = t.ReadInt(out endOfLine);
                if (nsec < 0 || nsec >= sectors.Count)
                    th.sec = null;
                else
                    th.sec = sectors[nsec];

                if (!endOfLine)
                {
                    string ln = t.ReadLine(false, true);

#if true
                    string[] valpairs = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string valpair in valpairs)
                    {
                        string[] vps = valpair.Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                        if (vps.Length != 2)
                            continue;

                        TemplateValue tv = new TemplateValue();
                        tv.name = vps[0];
                        tv.stringval = vps[1];
                        tv.vType = Values.GetTPLVType(tv.name);
                        tv.adjType = Values.GetTPLType(tv.name);
                        th.vals.Add(tv);
                    }
#else

                    string[] valpairs = ln.Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                    for (int i = 0; i < (valpairs.Length - 1); i += 2)
                    {
                        TemplateValue tv = new TemplateValue();
                        tv.name = valpairs[i].Trim(new char[] { ' ', '\t' });
                        tv.stringval = valpairs[i + 1].Trim(new char[] { ' ', '\t' });
                        tv.vType = Values.GetTPLVType(tv.name);
                        tv.adjType = Values.GetTPLType(tv.name);
                        th.vals.Add(tv);
                    }
#endif
                }
            }
        }

        private void LoadCogs_JKL(Tokenizer t)
        {
            CogFile cf = new CogFile();

            DO_Progress(t);
            ProgressDialog.Message("Loading Cogs");

            t.EatToken(2);
            int numcogs = t.ReadInt();
            for (int x = 0; x < numcogs; x++)
            {
                string tok = t.ReadToken();
                if (tok.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                    break;

                JKCOG cog = new JKCOG();
                cogs.Add(cog);

                bool endOfLine;
                cog.name = t.ReadToken(out endOfLine);

                cf.LoadNoLocals(cog.name);

                if (endOfLine)
                    continue;

                string[] vals = t.ReadLine().Split(new char[] { ' ', '\t' });
                for (int nval = 0; nval < vals.Length; nval++)
                {
                    COGValue cv = new COGValue();
                    cog.vals.Add(cv);

                    if (nval < cf.Count)
                        cv.Assign(cf[nval]);

                    cv.Val(vals[nval]);
                }
            }
        }

        private void LoadLights_JKL(Tokenizer t)
        {
            t.EatToken(2);
            int numlights = t.ReadInt();
            for (int x = 0; x < numlights; x++)
            {
                string tok = t.ReadToken();
                if (tok.Equals("end", StringComparison.InvariantCultureIgnoreCase))
                    break;

                JedLight jl = NewLight();
                lights.Add(jl);

                t.EatToken();

                jl.Pos = new Vec3();
                jl.Pos.x = t.ReadDouble();
                jl.Pos.y = t.ReadDouble();
                jl.Pos.z = t.ReadDouble();

                jl.intensity = t.ReadDouble();

                jl.range = t.ReadDouble();

                jl.rgbintensity = jl.intensity;
                jl.clr = Color4.white();
            }
        }

        public void LoadFromJKL(string fname)
        {
            GlobalVars.NedMain.HaltLightCalc();

            MemoryStream stream;
            int fsize;
            using (BaseFile bf = FileOperations.OpenFileRead(fname))
            {
                fsize = bf.Length;
                stream = new MemoryStream(bf.Read(fsize));
            }

            Tokenizer t = new Tokenizer(stream);

            Clear();

            List<Vec3> GVXList = new List<Vec3>();
            List<TTempSurf> GSurfList = new List<TTempSurf>();
            List<Vec2> GTXVXList = new List<Vec2>();
            List<TAdjoin> GAdjoins = new List<TAdjoin>();
            List<string> mats = new List<string>();
            List<string> cmps = new List<string>();

            const int pi_steps = 100;
            ProgressDialog.Reset(pi_steps, fsize);

            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                // eat unless section start
                if (!ln.StartsWith("section:", StringComparison.InvariantCultureIgnoreCase))
                    continue;

                string[] lnparts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                if (lnparts.Length != 2)
                    continue;

                switch (lnparts[1].ToUpperInvariant())
                {
                    case "GEORESOURCE": LoadGeometry_JKL(t, ref cmps, ref GVXList, ref GTXVXList, ref GAdjoins, ref GSurfList); break;
                    case "HEADER": LoadHeader_JKL(t); break;
                    case "SECTORS": LoadSectors_JKL(t, cmps, mats, GVXList, GTXVXList, GSurfList); break;
                    case "MATERIALS": LoadMats_JKL(t, ref mats); break;
                    case "THINGS": LoadThings_JKL(t); break;
                    case "COGS": LoadCogs_JKL(t); break;
                    case "LIGHTS": LoadLights_JKL(t); break;
                }
            }

            ProgressDialog.Cleanup();

            // post process
            int jsurf = -1;
            foreach (JKSector sec in sectors)
            {
                foreach (JKSurface surf in sec.surfaces)
                {
                    jsurf++;

                    if (surf.nadj < 0)
                    {
                        surf.adjoin = null;
                        continue;
                    }

                    TAdjoin adj = GAdjoins[surf.nadj];
                    adj.surf = jsurf;
                }
            }

            foreach (JKSector sec in sectors)
                foreach (JKSurface surf in sec.surfaces)
                {
                    if (surf.nadj < 0)
                    {
                        surf.adjoin = null;
                        continue;
                    }

                    TAdjoin adj = GAdjoins[surf.nadj];

                    surf.adjoin = GetSurfaceN(GAdjoins[adj.mirror].surf);
                    surf.adjoinFlags = adj.flags;
                }

            RenumSecs();
            RenumThings();

            foreach (JKSector sec in sectors)
            {
                sec.Renumber();
                foreach (JKSurface surf in sec.surfaces)
                {
                    surf.Recalc();

                    if (Utils.IsTXFlipped(surf))
                        surf.surfaceFlags |= Surface.Flag.Flipped;
                    else
                        surf.surfaceFlags &= ~Surface.Flag.Flipped;
                }
            }

            foreach (JKCOG cog in cogs)
                foreach (COGValue cv in cog.vals)
                    cv.Resolve();

            AddMissingLayers();

            layerVisString = string.Empty;
            for (int i = 0; i < layers.Count; i++)
                layerVisString += "1";

            GlobalVars.NedMain.ResumeLightCalc();
        }

        public void LoadFromJED(string fname)
        {
            if (!File.Exists(fname))
            {
                System.Windows.Forms.MessageBox.Show($"{fname}\nis bad");
                return;
            }

            GlobalVars.NedMain.HaltLightCalc();

            double jedver = 0.1;
            using (Stream stream = File.OpenRead(fname))
            {
                int fsize = (int)stream.Length;
                Tokenizer t = new Tokenizer(stream);

                Clear();

                const int pi_steps = 200;
                ProgressDialog.Reset(pi_steps, fsize);

                t.EatToken();
                jedver = t.ReadDouble();

                while (!t.EndOfStream)
                {
                    string tok = t.ReadToken();
                    if (string.IsNullOrEmpty(tok))
                        break;

                    if (tok.Equals("SEC:", StringComparison.InvariantCultureIgnoreCase))
                    {
                        tok = t.ReadToken().ToUpperInvariant();
                        switch (tok)
                        {
                            case "HEADER": LoadHeader_JED(t); break;
                            case "GEO": LoadGeometry_JED(t); break;
                            case "THINGS": LoadThings_JED(t); break;
                            case "LIGHTS": LoadLights_JED(t); break;
                            case "LAYERS": LoadLayers_JED(t); break;
                            case "COGS": LoadCogs_JED(t); break;
                            case "COGGROUPS": LoadCogGroups_NED(t); break;
                            case "MOTS": mots = true; break;
                            case "LVIS": LoadLVis_JED(t); break;
                            case "HIERARCHYDEF": LoadHierarchy_JED(t); break;
                        }

                        SkipToEOS(t);
                    }
                }
            }

            RenumThings();
            RenumSecs();
            foreach (JKCOG cog in cogs)
                foreach (COGValue cv in cog.vals)
                    cv.Resolve();

            foreach (JKSector sec in sectors)
            {
                sec.Renumber();

                foreach (JKSurface surf in sec.surfaces)
                {
                    surf.Recalc();
                    if (jedver < 0.81)
                    {
                        if (Utils.IsTXFlipped(surf))
                            surf.surfaceFlags |= Surface.Flag.Flipped;
                        else
                            surf.surfaceFlags &= ~(Surface.Flag.Flipped);

                        surf.texScale.y = surf.texScale.x;
                    }
                }
            }

            AddMissingLayers();
            ProgressDialog.Cleanup();

            GlobalVars.NedMain.ResumeLightCalc();
        }

        public void SaveToJED(string name)
        {
            RenumSecs();
            foreach (JKSector sec in sectors)
                sec.Renumber();

            using (StreamWriter sw = File.CreateText(name))
            {
                sw.WriteLine("NED " + GlobalVars.JedVerNum);


                if (mots)
                {
                    sw.WriteLine();
                    sw.WriteLine("SEC: MOTS");
                    sw.WriteLine("END");
                    sw.WriteLine();
                }


                sw.WriteLine("SEC: HEADER");
                string st = string.IsNullOrEmpty(masterCMP) ? "-" : masterCMP;
                sw.WriteLine("HDR " + header.version.ToString() + " " + Utils.dstr(header.gravity, 6) + " " + Utils.dstr(header.ceilingskyZ, 6) +
                    " " + Utils.dstr(header.perspDist, 6) + " " + Utils.dstr(header.gouraudDist, 6) + " " + st + " " + Utils.dstr(ppUnit, 2));

                sw.WriteLine("HOR " + Utils.dstr(header.horDistance, 6) + " " + Utils.dstr(header.horPixelsPerRev, 6) +
                    " " + Utils.dstr(header.horSkyOff.x, 6) + " " + Utils.dstr(header.horSkyOff.y, 6) );

                sw.WriteLine("CEIL " + Utils.dstr(header.ceilingSkyOff.x, 6) + " " + Utils.dstr(header.ceilingSkyOff.y, 6) );

                sw.WriteLine("MIP " + Utils.dstr(header.mipMapDist[0], 6) + " " + Utils.dstr(header.mipMapDist[1], 6) + " " + Utils.dstr(header.mipMapDist[2], 6) + " " + Utils.dstr(header.mipMapDist[3], 6) );

                sw.WriteLine("LOD " + Utils.dstr(header.LODDist[0], 6) + " " + Utils.dstr(header.LODDist[1], 6) + " " + Utils.dstr(header.LODDist[2], 6) + " " + Utils.dstr(header.LODDist[3], 6) );

                sw.WriteLine("END");
                sw.WriteLine();


                RenumSecs();
                foreach (JKSector sec in sectors)
                    sec.Renumber();

                sw.WriteLine("SEC: GEO");
                sw.WriteLine("SECS " + sectors.Count.ToString());
                foreach (JKSector sec in sectors)
                {
                    st = string.IsNullOrEmpty(sec.colormap) ? "-" : sec.colormap;
                    st += " " + (string.IsNullOrEmpty(sec.sound) ? "-" : sec.sound);

                    sw.WriteLine(st + " " + Utils.dstr(sec.snd_vol, 6) + " " + ((uint)sec.flags).ToString("X") + " " +
                        Utils.dstr(sec.ambient, 6) + " " + Utils.dstr(sec.extra, 6) + " " + Utils.dstr(sec.tint.r, 6) +
                        " " + Utils.dstr(sec.tint.g, 6) + " " + Utils.dstr(sec.tint.b, 6) + " " + sec.layer.ToString());

                    sw.WriteLine("VXS " + sec.vertices.Count.ToString());
                    foreach (JKVertex v in sec.vertices)
                        sw.WriteLine(Utils.dstr(v.v.x, 6) + " " + Utils.dstr(v.v.y, 6) + " " + Utils.dstr(v.v.z, 6));

                    sw.WriteLine("SURFS " + sec.surfaces.Count.ToString());
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        int asf, asec;
                        if (surf.adjoin != null)
                        {
                            asf = surf.adjoin.num;
                            asec = surf.adjoin.sector.num;
                        }
                        else
                        {
                            asf = -1;
                            asec = -1;
                        }

                        st = string.IsNullOrEmpty(surf.material) ? "-" : surf.material;
                        sw.WriteLine(st + " " + ((uint)surf.surfaceFlags).ToString("X") + " " + ((uint)surf.faceFlags).ToString("X") +
                            " " + ((int)surf.geo).ToString() + " " + ((int)surf.light).ToString() + " " + ((int)surf.tex).ToString() +
                            " " + Utils.dstr(surf.extraLight, 6) + " " + surf.vertices.Count.ToString() + " " + asec.ToString() + " " + asf.ToString() +
                            " " + ((uint)surf.adjoinFlags).ToString("X") + " " + Utils.dstr(surf.texScale.x, 4) + " " + Utils.dstr(surf.texScale.y, 4));

                        st = string.Empty;
                        for(int x=0; x<surf.vertices.Count; x++)
                        {
                            TexVertex tv = surf.texvertices[x];
                            st += " " + surf.vertices[x].num.ToString() + " " + Utils.dstr(tv.v.x, 2) + " " + Utils.dstr(tv.v.y, 2) + " " + Utils.dstr(tv.intensity, 4);
                        }
                        for (int x = 0; x < surf.vertices.Count; x++)
                        {
                            TexVertex tv = surf.texvertices[x];
                            st += " " + Utils.dstr(tv.r, 4) + " " + Utils.dstr(tv.g, 4) + " " + Utils.dstr(tv.b, 4);
                        }
                        sw.WriteLine(st);
                    }
                }

                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: THINGS");
                sw.WriteLine("THS " + things.Count.ToString());
                foreach (JKThing thing in things)
                {
                    int asec = thing.sec == null ? -1 : thing.sec.num;
                    sw.WriteLine(thing.name + " " + Utils.dstr(thing.pos.x, 6) + " " + Utils.dstr(thing.pos.y, 6) + " " + Utils.dstr(thing.pos.z, 6) +
                        " " + Utils.dstr(thing.orient.pitch, 6) + " " + Utils.dstr(thing.orient.yaw, 6) + " " + Utils.dstr(thing.orient.roll, 6) +
                        " " + asec.ToString() + " " + thing.layer.ToString());

                    st = string.Empty;
                    foreach (TemplateValue tv in thing.vals)
                        st += " " + tv.name + "=" + tv.AsString;
                    sw.WriteLine(st);
                }

                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: LIGHTS");
                sw.WriteLine("LTS " + lights.Count.ToString());
                foreach (JedLight light in lights)
                {
                    sw.WriteLine(Utils.dstr(light.Pos.x, 6) + " " + Utils.dstr(light.Pos.y, 6) + " " + Utils.dstr(light.Pos.z, 6) +
                        " " + light.layer.ToString() + " " + Utils.dstr(light.intensity, 6) + " " + Utils.dstr(light.range, 6) +
                        " " + ((uint)light.flags).ToString("X") + " " + Utils.dstr(light.rgbrange, 6) + " " + Utils.dstr(light.rgbintensity, 6) +
                        " " + Utils.dstr(light.clr.r, 4) + " " + Utils.dstr(light.clr.g, 4) + " " + Utils.dstr(light.clr.b, 4));
                }

                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: COGS");
                sw.WriteLine("COGS " + cogs.Count.ToString());
                foreach (JKCOG cog in cogs)
                {
                    st = cog.name;
                    foreach (COGValue cv in cog.vals)
                        st += " " + cv.name + ":" + Values.GetCogTypeName(cv.cogType, true) + "=" + cv.AsString;
                    sw.WriteLine(st);
                }

                sw.WriteLine("END");
                sw.WriteLine();


                // COG GROUPS
                sw.WriteLine("COGGROUPS");
                sw.WriteLine("GROUPS " + coggroups.Count.ToString());
                foreach (CogGroup cg in coggroups)
                {
                    st = cg.Name;

                    sw.WriteLine(st);
                }
                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: LAYERS");
                sw.WriteLine("LAYERS " + layers.Count.ToString());
                foreach (string layer in layers)
                    sw.WriteLine(layer);
                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: LVIS");
                sw.WriteLine(layerVisString);
                sw.WriteLine("END");
                sw.WriteLine();


                sw.WriteLine("SEC: HIERARCHYDEF");
                sw.WriteLine("NODES " + hnodes.Count.ToString());
                for (int x = 0; x < hnodes.Count; x++)
                    sw.WriteLine(hnodes[x].ToString(x));
                sw.WriteLine("END");
            }
        }

        public class ClassSaveToJKL
        {
            JKLevel level;
            StreamWriter sw;

            public class UsedList : List<string>
            {
                public new void Add(string s)
                {
                    if (string.IsNullOrEmpty(s) || s.Equals("none", StringComparison.InvariantCultureIgnoreCase))
                        return;

                    foreach (string t in this)
                        if (t.Equals(s, StringComparison.InvariantCultureIgnoreCase))
                            return;

                    base.Add(s);
                }

                public new void AddRange(IEnumerable<string> ls)
                {
                    foreach (string l in ls)
                        Add(l);
                }

                public new int IndexOf(string s)
                {
                    for (int x = 0; x < Count; x++)
                        if (this[x].Equals(s, StringComparison.InvariantCultureIgnoreCase))
                            return x;

                    return -1;
                }
            }

            public class UsedListMatrix
            {
                public UsedList usedWavs = new UsedList();
                public UsedList usedMats = new UsedList();
                public UsedList usedCmps = new UsedList();
                public UsedList usedCogs = new UsedList();
                public UsedList usedSnds = new UsedList();
                public UsedList usedPups = new UsedList();
                public UsedList usedKeys = new UsedList();
                public UsedList usedSprs = new UsedList();
                public UsedList usedPars = new UsedList();
                public UsedList used3DOs = new UsedList();
                public UsedList usedAIs = new UsedList();
                public UsedList usedTemplates = new UsedList();

                public readonly bool generateStaticJKL;
                public string staticJKLPath = string.Empty;
                public UsedListMatrix(bool _generateStaticJKL)
                {
                    generateStaticJKL = _generateStaticJKL;
                }

                public void SaveGOB(string filename)
                {
                    using (GOB2Creator gob = new GOB2Creator(filename))
                    {
                        List<string> files = GenerateFilesList();


                        ProgressDialog.Reset(files.Count, 0);

                        List<string> gobfiles = new List<string>();
                        foreach (string file in files)
                        {
                            BaseFile bf;
                            if (generateStaticJKL && file.EndsWith("static.jkl", StringComparison.InvariantCultureIgnoreCase))
                            {
                                gobfiles.Add(file);
                                continue;
                            }
                            else
                                bf = FileOperations.OpenGameFile(file);

                            using (bf)
                            {
                                if (bf is SubFile)
                                    gobfiles.Add(bf.FullName.Substring(bf.FullName.IndexOf('>') + 1));
                                else
                                    gobfiles.Add(bf.FullName.Substring(GlobalVars.ProjectDir.Length));

                                // need to bulletproof all this in case cant find somethin
                                //gobfiles.Add(file);
                            }
                        }
                        if (generateStaticJKL)
                            gobfiles.Add("jkl/static.jkl");
                        gob.PrepareHeader(gobfiles);


                        foreach (string file in files)
                        {
                            ProgressDialog.Step();
                            BaseFile bf;
                            if (generateStaticJKL && file.EndsWith("static.jkl", StringComparison.InvariantCultureIgnoreCase))
                                bf = FileOperations.OpenFileRead(staticJKLPath);
                            else
                                bf = FileOperations.OpenGameFile(file);

                            using(bf)
                            {
                                // need to bulletproof all this in case cant find somethin
                                gob.AddFile(bf);
                            }
                        }


                        ProgressDialog.Cleanup();

                        if (File.Exists(staticJKLPath))
                        {
                            File.Delete(staticJKLPath);
                        }
                    }
                }

                public List<string> GenerateFilesList()
                {
                    List<string> files = new List<string>();

                    files.AddRange(usedWavs);
                    files.AddRange(usedMats);
                    files.AddRange(usedCmps);
                    files.AddRange(usedCogs);
                    files.AddRange(usedSnds);
                    files.AddRange(usedPups);
                    files.AddRange(usedKeys);
                    files.AddRange(usedSprs);
                    files.AddRange(usedPars);
                    files.AddRange(used3DOs);
                    files.AddRange(usedAIs);

                    if (generateStaticJKL)
                        files.Add("jkl/static.jkl");

                    return files;
                }

                public TemplateList GenerateTemplateList()
                {
                    TemplateList wtpls = new TemplateList();
                    Dictionary<string, JKTemplate> templateMap = new Dictionary<string, JKTemplate>(StringComparer.InvariantCultureIgnoreCase);
                    for (int i = usedTemplates.Count - 1; i >= 0; i--)
                    {
                        int n = Templates.IndexOfName(usedTemplates[i]);
                        if (n == -1)
                        {
                            usedTemplates.RemoveAt(i);
                            continue;
                        }
                        if (!templateMap.ContainsKey(usedTemplates[i]))
                            templateMap.Add(usedTemplates[i], Templates.All[n]);
                    }

                    while (usedTemplates.Count > 0)
                    {
                        int n = usedTemplates.Count;
                        for (int i = usedTemplates.Count - 1; i >= 0; i--)
                        {
                            JKTemplate tp = templateMap[usedTemplates[i]];
                            if (!tp.Parent.Equals("none", StringComparison.InvariantCultureIgnoreCase) && !wtpls.Contains(tp.Parent))
                                continue;

                            bool norefs = true;
                            foreach (TemplateValue tv in tp.vals)
                                if (tv.adjType == AdjType.Template)
                                {
                                    if (!tv.stringval.Equals("none", StringComparison.InvariantCultureIgnoreCase) && !wtpls.Contains(tv.stringval))
                                    {
                                        norefs = false;
                                        break;
                                    }
                                }

                            if (!norefs)
                                continue;

                            wtpls.Add(new TemplateEntry(tp.Name, tp));
                            usedTemplates.RemoveAt(i);
                        }

                        if (n == usedTemplates.Count)
                            break;
                    }
                    return wtpls;
                }

                public int GetNCMP(int s, string name)
                {
                    int ret = usedCmps.IndexOf(name);
                    if (ret < 0)
                    {
                        ret = 0;
                        Utils.PanMessage(Utils.MessageType.Warning, "Invalid colormap in sec " + s.ToString());
                    }
                    return ret;
                }

                public int GetMatN(string mat)
                {
                    if (string.IsNullOrEmpty(mat))
                        return -1;

                    for (int x = 0; x < usedMats.Count; x++)
                        if (usedMats[x].Equals(mat, StringComparison.InvariantCultureIgnoreCase))
                            return x;

                    return -1;
                }

                UsedList newTpls = new UsedList();
                UsedList newCogs = new UsedList();
                UsedList tcogs = new UsedList();
                UsedList ttpls = new UsedList();

                public void AddTemplate(string name)
                {
                    ttpls.Add(name);
                }

                public void AddMat(string name)
                {
                    usedMats.Add(name);
                }

                public void AddColormap(string name)
                {
                    usedCmps.Add(name);
                }

                public void AddSound(string name)
                {
                    usedWavs.Add(name);
                }

                public void AddCog(string name)
                {
                    tcogs.Add(name);
                }

                public void AddCogValues(COGValues vals)
                {
                    foreach (COGValue cv in vals)
                        switch (cv.cogType)
                        {
                            case COGType.Wav: usedWavs.Add(cv.stringval); break;
                            case COGType.Template: newTpls.Add(cv.stringval); break;
                            case COGType.Model: used3DOs.Add(cv.stringval); break;
                            case COGType.KEY: usedKeys.Add(cv.stringval); break;
                            case COGType.MAT: usedMats.Add(cv.stringval); break;
                            case COGType.AI: usedAIs.Add(cv.stringval); break;
                        }
                }

                public void AddItemCog(string c)
                {
                    AddFromCog(c, newTpls, newCogs);
                }

                public void Finish(bool suppressJKLGenerate)
                {
                    if (generateStaticJKL && !suppressJKLGenerate)
                    {
                        staticJKLPath = Path.GetTempFileName();

                        JKLevel lev = new JKLevel();
                        /*JKThing t = lev.NewThing();
                        t.name = "walkplayer";
                        lev.things.Add(t);*/

                        using (StreamWriter sw = File.CreateText(staticJKLPath))
                        {
                            ClassSaveToJKL c = new ClassSaveToJKL(lev, sw);
                            c.Save(false, true, this);
                        }
                    }


                    tcogs.AddRange(newCogs);
                    ttpls.AddRange(newTpls);

                    do
                    {
                        newCogs.Clear();
                        newTpls.Clear();

                        foreach (string cog in tcogs)
                        {
                            AddFromCog(cog, newTpls, newCogs);
                            usedCogs.Add(cog);
                        }

                        foreach (string tpl in ttpls)
                        {
                            AddFromTemplate(tpl, newTpls, newCogs);
                            usedTemplates.Add(tpl);
                        }

                        tcogs.Clear();
                        ttpls.Clear();
                        tcogs.AddRange(newCogs);
                        ttpls.AddRange(newTpls);
                    } while (newCogs.Count != 0 || newTpls.Count != 0);

                    PUPFile pupf = new PUPFile();
                    foreach (string pfile in usedPups)
                    {
                        try
                        {
                            pupf.Load(pfile);
                            foreach (string kfile in pupf.keys)
                                usedKeys.Add(kfile);
                        }
                        catch
                        {

                        }
                    }

                    SNDFile sndf = new SNDFile();
                    foreach (string sfile in usedSnds)
                    {
                        try
                        {
                            sndf.Load(sfile);
                            foreach (string wfile in sndf.wavs)
                                usedWavs.Add(wfile);
                        }
                        catch
                        {

                        }
                    }

                    foreach (string sprfile in usedSprs)
                        usedMats.Add(GetMatFromSPR(sprfile));
                    foreach (string parfile in usedPars)
                        usedMats.Add(GetMatFromPAR(parfile));

                    T3DOFile f3do = new T3DOFile();
                    used3DOs.Sort();
                    foreach (string mfile in used3DOs)
                    {
                        f3do.Load(mfile);
                        foreach (string mf in f3do.mats)
                            usedMats.Add(mf);
                    }
                }

                private void AddFromCog(string name, UsedList newTpls, UsedList newCogs)
                {
                    if (usedCogs.IndexOf(name) != -1)
                        return;

                    CogFile cf = new CogFile();
                    try
                    {
                        cf.Load(name);
                        foreach (COGValue cv in cf.vals)
                            switch (cv.cogType)
                            {
                                case COGType.Wav: usedWavs.Add(cv.stringval); break;
                                case COGType.Template: newTpls.Add(cv.stringval); break;
                                case COGType.Model: used3DOs.Add(cv.stringval); break;
                                case COGType.KEY: usedKeys.Add(cv.stringval); break;
                                case COGType.COG: newCogs.Add(cv.stringval); break;
                                case COGType.MAT: usedMats.Add(cv.stringval); break;
                                case COGType.AI: usedAIs.Add(cv.stringval); break;
                            }
                    }
                    catch
                    {

                    }
                }

                void AddFromTemplate(string name, UsedList newTpls, UsedList newCogs)
                {
                    if (usedTemplates.IndexOf(name) != -1)
                        return;

                    int n = Templates.IndexOfName(name);
                    if (n == -1)
                        return;

                    JKTemplate tpl = Templates.All[n];
                    if (!string.IsNullOrEmpty(tpl.Parent) && !tpl.Parent.Equals("none", StringComparison.InvariantCultureIgnoreCase))
                        newTpls.Add(tpl.Parent);

                    foreach (TemplateValue tv in tpl.vals)
                        switch (tv.adjType)
                        {
                            case AdjType.MAT: usedMats.Add(tv.stringval); break;
                            case AdjType.COG: newCogs.Add(tv.stringval); break;
                            case AdjType.SND: usedSnds.Add(tv.stringval); break;
                            case AdjType.PUP: usedPups.Add(tv.stringval); break;
                            case AdjType.SPR: usedSprs.Add(tv.stringval); break;
                            case AdjType.PAR: usedPars.Add(tv.stringval); break;
                            case AdjType.Model: used3DOs.Add(tv.stringval); break;
                            case AdjType.AI: usedAIs.Add(tv.stringval); break;
                            case AdjType.Template: newTpls.Add(tv.stringval); break;
                        }
                }

                string GetMatFromSPR(string fname)
                {
                    MemoryStream stream;
                    try
                    {
                        using (BaseFile bf = FileOperations.OpenGameFile(fname))
                            stream = new MemoryStream(bf.Read(bf.Length));
                    }
                    catch
                    {
                        return string.Empty;
                    }

                    Tokenizer t = new Tokenizer(stream);

                    return t.ReadToken();

                }

                string GetMatFromPAR(string fname)
                {
                    MemoryStream stream;
                    try
                    {
                        using (BaseFile bf = FileOperations.OpenGameFile(fname))
                            stream = new MemoryStream(bf.Read(bf.Length));
                    }
                    catch
                    {
                        return string.Empty;
                    }

                    Tokenizer t = new Tokenizer(stream);

                    while (!t.EndOfStream)
                    {
                        string ln = t.ReadLine();
                        if (string.IsNullOrEmpty(ln))
                            break;

                        int i = ln.IndexOfAny(new char[] { ' ', '\t' });
                        if (i == -1)
                            continue;

                        string w = ln.Substring(0, i);
                        if (w.Equals("material", StringComparison.InvariantCultureIgnoreCase))
                        {
                            ln = ln.Substring(i).Trim();
                            i = ln.IndexOfAny(new char[] { ' ', '\t' });
                            if (i != -1)
                                ln = ln.Substring(0, i);

                            return ln;
                        }
                    }

                    return string.Empty;
                }
            }

            public UsedListMatrix matrix = null;

            void WriteListSection(string name, string cntname, UsedList ls)
            {
                ls.Sort();// lolz

                sw.WriteLine(name);
                sw.WriteLine(cntname + " " + ls.Count.ToString());
                for (int i = 0; i < ls.Count; i++)
                    sw.WriteLine(i.ToString() + ": " + ls[i]);
                sw.WriteLine("end");
                sw.WriteLine();
            }

            void WriteHeaderSection()
            {
                JKLHeader h = level.header;

                sw.WriteLine("SECTION: HEADER");
                sw.WriteLine("Version " + h.version.ToString());
                sw.WriteLine("World Gravity " + Utils.dstr(h.gravity, 6));
                sw.WriteLine("Ceiling Sky Z " + Utils.dstr(h.ceilingskyZ, 6));
                sw.WriteLine("Horizon Distance " + Utils.dstr(h.horDistance, 6));
                sw.WriteLine("Horizon Pixels per Rev " + Utils.dstr(h.horPixelsPerRev, 6));
                sw.WriteLine("Horizon Sky Offset " + Utils.dstr(h.horSkyOff.x, 6) + " " + Utils.dstr(h.horSkyOff.y, 6));
                sw.WriteLine("Ceiling Sky Offset " + Utils.dstr(h.ceilingSkyOff.x, 6) + " " + Utils.dstr(h.ceilingSkyOff.y, 6));
                sw.WriteLine("MipMap Distances " + Utils.dstr(h.mipMapDist[0], 6) + " " + Utils.dstr(h.mipMapDist[1], 6) + " " + Utils.dstr(h.mipMapDist[2], 6) + " " + Utils.dstr(h.mipMapDist[3], 6));
                sw.WriteLine("LOD Distances " + Utils.dstr(h.LODDist[0], 6) + " " + Utils.dstr(h.LODDist[1], 6) + " " + Utils.dstr(h.LODDist[2], 6) + " " + Utils.dstr(h.LODDist[3], 6));
                sw.WriteLine("Perspective distance " + Utils.dstr(h.perspDist, 6));
                sw.WriteLine("Gouraud distance " + Utils.dstr(h.gouraudDist, 6));
                sw.WriteLine();
            }

            Vertices vxlist;
            TexVertices txvxlist;

            Vertices vxidx;
            TexVertices txidx;

            int AddVX(Vertex v)
            {
                int n;
                if (Utils.FindVX(vxidx, v, out n))
                    return vxidx[n].num;

                int i = vxlist.Add(v);
                vxidx.Insert(n, v);
                v.num = i;
                return i;
            }

            int AddTXVX(TexVertex tv)
            {
                int n;
                if (Utils.FindTXVX(txidx, tv, out n))
                    return txidx[n].num;

                int i = txvxlist.Add(tv);
                txidx.Insert(n, tv);
                tv.num = i;
                return i;
            }

            void WriteGeometry()
            {
                ProgressDialog.Message("Pooling vertices");
                ProgressDialog.Step();

                vxlist = new Vertices();
                txvxlist = new TexVertices();

                vxidx = new Vertices();
                txidx = new TexVertices();

                foreach (JKSector sec in level.sectors)
                    foreach (JKVertex v in sec.vertices)
                        v.num = AddVX(v);

                ProgressDialog.Message("Writing vertices");

                sw.WriteLine("World vertices " + vxlist.Count.ToString());
                sw.WriteLine("#num:	vertex:");
                for (int i = 0; i < vxlist.Count; i++)
                {
                    Vertex v = vxlist[i];
                    sw.WriteLine(i.ToString() + ": " + Utils.dstr(v.v.x, 6) + " " + Utils.dstr(v.v.y, 6) + " " + Utils.dstr(v.v.z, 6));
                }
                sw.WriteLine();

                // number surfaces
                int nsurf = 0;
                foreach (JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                        surf.num = nsurf++;

                ProgressDialog.Message("Pooling texture vertices");
                ProgressDialog.Step();

                // pool texture vertices
                foreach (JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                        foreach (TexVertex tv in surf.texvertices)
                            tv.num = AddTXVX(tv);

                ProgressDialog.Message("Writing texture vertices");

                sw.WriteLine("World texture vertices " + txvxlist.Count.ToString());
                sw.WriteLine("#num:	u:	v:");
                foreach (TexVertex tv in txvxlist)
                    sw.WriteLine(tv.num.ToString() + ": " + Utils.dstr(tv.v.x, 2) + " " + Utils.dstr(tv.v.y, 2));
                sw.WriteLine();

                List<JKLevel.TAdjoin> adjoins = new List<TAdjoin>();
                foreach (JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                        if (surf.adjoin != null)
                        {
                            JKLevel.TAdjoin adj = new TAdjoin();
                            surf.nadj = adjoins.Count;
                            adjoins.Add(adj);

                            adj.surf = surf.num;
                            adj.flags = (Adjoin.Flag)((uint)surf.adjoinFlags & 0x7FFFFFFF);
                            Vec3 c = Utils.FindCenter(sec);
                            adj.dist = Utils.DistToPlane(surf, c);
                        }
                        else
                            surf.nadj = -1;

                foreach (JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                        if (surf.adjoin != null)
                            adjoins[surf.nadj].mirror = surf.adjoin.nadj;

                sw.WriteLine("World adjoins " + adjoins.Count.ToString());
                sw.WriteLine("#num:	flags:	mirror:	dist:");
                for (int i = 0; i < adjoins.Count; i++)
                {
                    JKLevel.TAdjoin adj = adjoins[i];
                    sw.WriteLine(i.ToString() + ": 0x" + ((uint)adj.flags).ToString("x") + " " + adj.mirror.ToString() + " " + Utils.dstr(adj.dist, 2));
                }
                sw.WriteLine();

                sw.WriteLine("World surfaces " + nsurf.ToString());
                sw.WriteLine("#num:   mat:     surfflags:	faceflags:	geo:	light:	tex:	adjoin:	extralight:	nverts:	vertices:			intensities:");
                foreach(JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                    {
                        int nmat = matrix.GetMatN(surf.material);
                        string st = surf.num.ToString() + ":\t" + nmat.ToString() + "\t0x" + ((uint)(surf.surfaceFlags & ~Surface.Flag.Flipped)).ToString("x") +
                            "\t0x" + ((uint)surf.faceFlags).ToString("x") + "\t" + ((int)surf.geo).ToString() + "\t" + ((int)surf.light).ToString() +
                            "\t" + ((int)surf.tex).ToString() + "\t" + surf.nadj.ToString() + "\t" + Utils.dstr(surf.extraLight, 6) + "\t" + surf.vertices.Count.ToString();

                        for (int v = 0; v < surf.vertices.Count; v++)
                            st += " " + surf.vertices[v].num.ToString() + "," + surf.texvertices[v].num.ToString();

                        for(int v=0; v<surf.vertices.Count; v++)
                        {
                            TexVertex tv = surf.texvertices[v];

                            if (level.mots)
                                st += " " + Utils.dstr(tv.intensity, 4) + " " + Utils.dstr(tv.r, 4) + " " + Utils.dstr(tv.g, 4) + " " + Utils.dstr(tv.b, 4);
                            else
                                st += " " + Utils.dstr(tv.intensity, 4);
                        }

                        sw.WriteLine(st);
                    }
                sw.WriteLine();

                // normals
                sw.WriteLine("#--- Surface normals ---");
                foreach (JKSector sec in level.sectors)
                    foreach (JKSurface surf in sec.surfaces)
                        sw.WriteLine(surf.num.ToString() + ": " + Utils.dstr(surf.normal.x, 6) + " " + Utils.dstr(surf.normal.y, 6) + " " + Utils.dstr(surf.normal.z, 6));
            }

            void WriteSectors()
            {
                sw.WriteLine();
                sw.WriteLine("World sectors " + level.sectors.Count.ToString());
                sw.WriteLine();
                int nsurf = 0;
                for (int s = 0; s < level.sectors.Count; s++)
                {
                    JKSector sec = level.sectors[s];
                    Box3 bbox = Utils.FindBBox(sec);
                    Vec3 c;
                    double rad;
                    Utils.FindBSphere(sec, out c, out rad);

                    sec.flags &= ~Sector.Flag.Preview3DO;

                    Box3? cbox = Utils.FindCollideBox(sec, bbox, c);
                    if (cbox.HasValue)
                        sec.flags |= Sector.Flag.HasCollideBox;
                    else
                        sec.flags &= ~Sector.Flag.HasCollideBox;

                    sw.WriteLine("SECTOR " + s.ToString());
                    sw.WriteLine("FLAGS 0x" + ((uint)sec.flags).ToString("x"));
                    sw.WriteLine("AMBIENT LIGHT " + Utils.dstr(sec.ambient, 6));
                    sw.WriteLine("EXTRA LIGHT " + Utils.dstr(sec.extra, 6));
                    sw.WriteLine("COLORMAP " + matrix.GetNCMP(s, sec.colormap).ToString());
                    sw.WriteLine("TINT " + Utils.dstr(sec.tint.r, 6) + " " + Utils.dstr(sec.tint.g, 6) + " " + Utils.dstr(sec.tint.b, 6));
                    sw.WriteLine("BOUNDBOX " + Utils.dstr(bbox.Left, 6) + " " + Utils.dstr(bbox.Back, 6) + " " + Utils.dstr(bbox.Bottom, 6) + " " +
                                               Utils.dstr(bbox.Right, 6) + " " + Utils.dstr(bbox.Front, 6) + " " + Utils.dstr(bbox.Top, 6));
                    if((sec.flags & Sector.Flag.HasCollideBox) != 0)
                        sw.WriteLine("COLLIDEBOX " + Utils.dstr(cbox.Value.Left, 6) + " " + Utils.dstr(cbox.Value.Back, 6) + " " + Utils.dstr(cbox.Value.Bottom, 6) + " " +
                                               Utils.dstr(cbox.Value.Right, 6) + " " + Utils.dstr(cbox.Value.Front, 6) + " " + Utils.dstr(cbox.Value.Top, 6));

                    if (!string.IsNullOrEmpty(sec.sound))
                        sw.WriteLine("SOUND " + sec.sound + " " + Utils.dstr(sec.snd_vol, 6));
                    sw.WriteLine("CENTER " + Utils.dstr(c.x, 6) + " " + Utils.dstr(c.y, 6) + " " + Utils.dstr(c.z, 6));
                    sw.WriteLine("RADIUS " + Utils.dstr(rad, 6));

                    sw.WriteLine("VERTICES " + sec.vertices.Count.ToString());
                    for (int v = 0; v < sec.vertices.Count; v++)
                        sw.WriteLine(v.ToString() + ": " + sec.vertices[v].num.ToString());
                    sw.WriteLine("SURFACES " + nsurf.ToString() + " " + sec.surfaces.Count.ToString());
                    nsurf += sec.surfaces.Count;
                    sw.WriteLine();
                }
            }

            void WriteCogs()
            {
                sw.WriteLine();
                sw.WriteLine("World cogs " + (level.cogs.Count + matrix.usedCogs.Count).ToString());
                sw.WriteLine("#Num	Script          Symbol values");
                for (int i = 0; i < level.cogs.Count; i++)
                {
                    JKCOG cog = level.cogs[i];
                    string st = i.ToString() + ": " + cog.name;
                    foreach (COGValue cv in cog.vals)
                        st += " " + cv.AsString;
                    sw.WriteLine(st);
                }
            }

            public class TemplateEntry
            {
                public string name;
                public JKTemplate template;

                public override bool Equals(object obj)
                {
                    TemplateEntry o = obj as TemplateEntry;
                    if (o == null)
                        return false;

                    if (!name.Equals(o.name, StringComparison.InvariantCultureIgnoreCase))
                        return false;

                    return true;
                }

                public override int GetHashCode()
                {
                    return name.GetHashCode();
                }

                public TemplateEntry(string _name, JKTemplate _template)
                {
                    name = _name;
                    template = _template;
                }
            }

            public class TemplateList : List<TemplateEntry>
            {
                public bool Contains(string name)
                {
                    foreach (TemplateEntry te in this)
                        if (te.name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                            return true;

                    return false;
                }
            }
            

            void WriteTemplates()
            {
                TemplateList wtpls = matrix.GenerateTemplateList();

                sw.WriteLine();
                sw.WriteLine("World templates " + wtpls.Count.ToString());
                sw.WriteLine("#Name:           Based On:        Params:");
                foreach(TemplateEntry te in wtpls)
                    sw.WriteLine(te.template.AsString);
            }

            void WriteThings()
            {
                level.RenumSecs();
                sw.WriteLine("World things " + (level.things.Count + 200).ToString());
                sw.WriteLine("#num template:       name:         	X:		Y:		Z:		Pitch:		Yaw:		Roll:		Sector:");
                for (int i = 0; i < level.things.Count; i++)
                {
                    JKThing th = level.things[i];

                    int snum;
                    if (th.sec == null)
                        snum = -1;
                    else
                        snum = th.sec.num;

                    string s = i.ToString() + ": " + th.name.ToLower() + " " + th.name.ToLower() + " " + Utils.dstr(th.pos.x, 6) + " " + Utils.dstr(th.pos.y, 6) + " " + Utils.dstr(th.pos.z, 6) + " " +
                        Utils.dstr(th.orient.pitch, 6) + " " + Utils.dstr(th.orient.yaw, 6) + " " + Utils.dstr(th.orient.roll, 6) + " " + snum.ToString();
                    if (level.mots)
                        s += " -1";
                    foreach (TemplateValue tv in th.vals)
                        s += " " + tv.name.ToLower() + "=" + tv.AsString;

                    sw.WriteLine(s);
                }
            }

            void WriteLights()
            {
                sw.WriteLine("Editor lights " + level.lights.Count.ToString());
                sw.WriteLine();
                sw.WriteLine("#num:	type:	X:		Y:		Z:		intensity:	range:");
                for (int i = 0; i < level.lights.Count; i++)
                {
                    JedLight j = level.lights[i];
                    sw.WriteLine(i.ToString() + ": 0 " + Utils.dstr(j.Pos.x, 6) + " " + Utils.dstr(j.Pos.y, 6) + " " + Utils.dstr(j.Pos.z, 6) + " " +
                        Utils.dstr(j.intensity, 6) + " " + Utils.dstr(j.range, 6));
                }
            }

            public void CompileLists(bool staticJKL)
            {
                if (!staticJKL)
                {
                    matrix.AddTemplate("+twinkle");
                    matrix.AddTemplate("+sspks_wall");
                    matrix.AddTemplate("+ssparks_wall");
                    matrix.AddTemplate("+sspks_blood");
                    matrix.AddTemplate("+ssparks_blood");
                    matrix.AddTemplate("+sspks_saber");
                    matrix.AddTemplate("+ssparks_saber");
                    matrix.AddTemplate("+backpack");
                    matrix.AddTemplate("ghostcam");

                    matrix.AddSound("00UNDERWATER.WAV");

                    ProgressDialog.Step();

                    foreach (JKSector sec in level.sectors)
                        foreach (JKSurface surf in sec.surfaces)
                            matrix.AddMat(surf.material);

                    if (!string.IsNullOrEmpty(level.masterCMP))
                        matrix.AddColormap(level.masterCMP);

                    foreach (JKSector sec in level.sectors)
                    {
                        matrix.AddSound(sec.sound);
                        matrix.AddColormap(sec.colormap);
                    }

                    ProgressDialog.Step();
                }

                foreach (JKCOG cog in level.cogs)
                    matrix.AddCog(cog.name);

                foreach (JKThing thing in level.things)
                    matrix.AddTemplate(thing.name);

                //AddFromPlacedCogs
                {
                    foreach (JKCOG cog in level.cogs)
                        matrix.AddCogValues(cog.vals);
                }

                ProgressDialog.Step();

                if (!staticJKL)
                {
                    ItemsDatFile idat = new ItemsDatFile();
                    idat.Load("items.dat");
                    foreach (string c in idat.COGs)
                        matrix.AddItemCog(c);

                    ProgressDialog.Step();
                }

                matrix.Finish(true);
            }

            public void Save(bool standardresources, bool nogeometry, UsedListMatrix _matrix)
            {
                matrix = _matrix;
                //matrix = new UsedListMatrix(staticJKL);

                ProgressDialog.Reset(10 + 9, 0);

                sw.WriteLine("#This file is created by Ned version " + GlobalVars.JedVerNum);
                sw.WriteLine("#Jed is (C) by Alexei Novikov");
                sw.WriteLine("#Ned is a BAH Production");
                sw.WriteLine();
                sw.WriteLine("SECTION: JK");
                sw.WriteLine();
                sw.WriteLine("SECTION: COPYRIGHT");
                foreach (string l in leclogo)
                    sw.WriteLine(l);
                sw.WriteLine();
                {
                    sw.WriteLine("SECTION: PENOR");
                    foreach (string l in penorlogo)
                        sw.WriteLine(l);
                    sw.WriteLine();
                }
                WriteHeaderSection();
                ProgressDialog.Message("Compiling lists");
                ProgressDialog.Step();
                CompileLists(!standardresources);

                ProgressDialog.Message("Writing sounds");
                ProgressDialog.Step();

                sw.WriteLine("SECTION: SOUNDS");
                {
                    // REMOVE ANY NON-WAV FILES SO JK DOESNT TRY TO LOAD THEM.
                    // IF SMITH IS USED THEN THESE FILES WILL BE LOOKED-UP AT RUNTIME.
                    // I GUESS I COULD ADD ANOTHER SECTION TO HINT TO SMMITH ADDITIONAL RESOURCES,
                    // OR LOADING A MOD/LEVEL COULD JUST LOAD EVERYTHING INSIDE..
                    for(int x=0; x<matrix.usedWavs.Count; x++)
                        if(!matrix.usedWavs[x].EndsWith(".wav", StringComparison.InvariantCultureIgnoreCase))
                            matrix.usedWavs.RemoveAt(x--);
                }
                matrix.usedWavs.Sort();
                sw.WriteLine("World sounds " + matrix.usedWavs.Count.ToString());
                foreach (string s in matrix.usedWavs)
                    sw.WriteLine(s);
                sw.WriteLine("end");
                sw.WriteLine();

                ProgressDialog.Message("Writing materials");
                ProgressDialog.Step();

                sw.WriteLine("SECTION: MATERIALS");
                matrix.usedMats.Sort();
                sw.WriteLine("World materials " + matrix.usedMats.Count.ToString());
                for (int i = 0; i < matrix.usedMats.Count; i++)
                    sw.WriteLine(i.ToString() + ": " + matrix.usedMats[i] + " 1 1");
                sw.WriteLine("end");
                sw.WriteLine();

                if (!nogeometry)
                {
                    sw.WriteLine("SECTION: GEORESOURCE");
                    sw.WriteLine("World Colormaps " + matrix.usedCmps.Count.ToString());
                    for (int i = 0; i < matrix.usedCmps.Count; i++)
                        sw.WriteLine(i.ToString() + ": " + matrix.usedCmps[i]);
                    sw.WriteLine();
                    WriteGeometry();
                    sw.WriteLine();

                    ProgressDialog.Message("Writing Sectors");
                    ProgressDialog.Step();

                    sw.WriteLine("Section: SECTORS");
                    WriteSectors();
                }

                ProgressDialog.Message("Writing Lists");
                ProgressDialog.Step();

                WriteListSection("Section: AICLASS", "World AIClasses", matrix.usedAIs);
                WriteListSection("Section: MODELS", "World models", matrix.used3DOs);
                WriteListSection("Section: SPRITES", "World sprites", matrix.usedSprs);
                WriteListSection("Section: KEYFRAMES", "World keyframes", matrix.usedKeys);
                WriteListSection("Section: ANIMCLASS", "World puppets", matrix.usedPups);
                WriteListSection("Section: Soundclass", "World soundclasses", matrix.usedSnds);
                WriteListSection("Section: cogscripts", "World scripts", matrix.usedCogs);

                sw.WriteLine("Section: cogs");
                WriteCogs();
                sw.WriteLine("end");
                sw.WriteLine();

                ProgressDialog.Message("Writing Templates");
                ProgressDialog.Step();

                sw.WriteLine("Section: TEMPLATES");
                WriteTemplates();
                sw.WriteLine("end");
                sw.WriteLine();

                if (!nogeometry)
                {
                    ProgressDialog.Message("Writing Things");
                    ProgressDialog.Step();

                    sw.WriteLine("Section: Things");
                    WriteThings();
                    sw.WriteLine("end");
                    sw.WriteLine();

                    if (level.mots)
                    {
                        sw.WriteLine("Section: ARCHLIGHTING");
                        sw.WriteLine("Num ArchObjects 0");
                        sw.WriteLine("end");
                        sw.WriteLine();
                    }

                    sw.WriteLine("Section: Lights");
                    WriteLights();
                    sw.WriteLine("end");
                    sw.WriteLine();
                }
            }

#region penorlogo
            private static string[] penorlogo = new string[] {
                "　　　　　　　　　　▄█▀█▀█▄",
                "　　　　　　　　▄█▀　　█　　▀█▄",
                "　　　　　　　▄█▀　　　　　　　▀█▄",
                "　　　　　　　█　　　　　　　　　　　█",
                "　　　　　　　█　　　　　　　　　　　█",
                "　　　　　　　▀█▄▄　　█　　　▄█▀",
                "　　　　　　　　　█　　▄▀▄　　█",
                "　　　　　　　　　█　▀　　　▀　█",
                "　　　　　　　　　█　　　　　　　█",
                "　　　　　　　　　█　　　　　　　█",
                "　　　　　　　　　█　　　　 　　 █",
                "　　　　　　　　　█　　　　　　　█",
                "　　　　　　　　　█　　　　　　　█",
                "　　　▄█▀▀█▄█　　　　　　　█▄█▀█▄",
                "　▄█▀▀　　　　▀　　　　　　　　　　　　▀▀█",
                "█▀　　　　　　　　　　　　　　　　　　　　　　▀█",
                "█　　　　　　　　　　　　　　　　　　　　　　　　█",
                "█　　　　　　　　　　　▄█▄　　　　　　　　　　█",
                "▀█　　　　　　　　　█▀　▀█　　　　　　　　█▀",
                "　▀█▄　　　　　　█▀　　　▀█　　　　　▄█▀",
                "　　　▀█▄▄▄█▀　　　　　　▀█▄▄▄█▀"
            };
#endregion

#region leclogo
            private static string[] leclogo = new string[] {
                "................................",
                "................@...@...@...@...",
                ".............@...@..@..@...@....",
                "................@.@.@.@.@.@.....",
                "@@@@@@@@......@...........@.....",
                "@@@@@@@@....@@......@@@....@....",
                "@@.....@.....@......@@@.....@@..",
                "@@.@@@@@......@.....@@@......@@.",
                "@@@@@@@@.......@....@@.....@@...",
                "@@@@@@@@.........@@@@@@@@@@.....",
                "@@@@@@@@..........@@@@@@........",
                "@@.....@..........@@@@@.........",
                "@@.@@@@@.........@@@@@@.........",
                "@@.....@.........@@@@@@.........",
                "@@@@@@@@.........@@@@@@.........",
                "@@@@@@@@.........@@@@@@@........",
                "@@@...@@.........@@@@@@@........",
                "@@.@@@.@.........@.....@........",
                "@@..@..@........@.......@.......",
                "@@@@@@@@........@.......@.......",
                "@@@@@@@@.......@........@.......",
                "@@..@@@@.......@........@.......",
                "@@@@..@@......@.........@.......",
                "@@@@.@.@......@.........@.......",
                "@@....@@........................",
                "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@",
                "@@@@@@@@@@@@@.@@@@@@@@@@@@@@@@@@",
                "@@.@@..@@@@@..@@@@@@@@@@.@@@@@@@",
                "@@.@.@.@@@@.@.@@@.@..@@...@@@..@",
                "@@..@@@@@@....@@@..@@@@@.@@@@.@@",
                "@@@@@@@@...@@.@@@.@@@@@..@@...@@",
                "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@",
                "@.copyright.(c).1997.lucasarts.@",
                "@@@@@@..entertainment.co..@@@@@@"
            };
#endregion

            public ClassSaveToJKL(JKLevel _level, StreamWriter _sw)
            {
                level = _level;
                sw = _sw;
            }
        }


        public void SaveToJKL(string f, bool staticJKL)
        {

            bool doNedDynamicLights = false;

            if (doNedDynamicLights)
            {
                Log.Warning("AUTOGENERATING NED DYNAMIC LIGHTS LOL");

                foreach (JedLight lt in lights)
                {
                    JKThing th = NewThing();
                    th.name = "light0.7";
                    th.NEDDynamicLight = true;
                    things.Add(th);
                    GlobalVars.NedMain.LayerThing(GetThingByID(th.ID));
                }
            }

            try
            {
                using (StreamWriter sw = File.CreateText(f))
                {
                    ClassSaveToJKL c = new ClassSaveToJKL(this, sw);
                    c.Save(!staticJKL, staticJKL, new ClassSaveToJKL.UsedListMatrix(staticJKL));
                }
            }
            finally
            {
                if (doNedDynamicLights)
                {
                    Log.Warning("CLEANING UP NED DYNAMIC LIGHTS LOL");

                    for (int x = 0; x < things.Count; x++)
                        if (things[x].NEDDynamicLight)
                            things.RemoveAt(x--);
                }


                RenumSecs();
                RenumThings();
                foreach (JKSector sec in sectors)
                    sec.Renumber();

                ProgressDialog.Cleanup();
            }
        }
    }
}
