﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class MapSettings : UserControl
    {
        public MapSettings()
        {
            InitializeComponent();

            Prefabs.LoadPrefabs("shapes.tpl");
            UpdatePrefabs();
        }

        private void SetCB(ComboBox cb, double d)
        {
            cb.Text = d.ToString();
        }

        public bool IsLayerVisible(int n)
        {
            return (n < 0 || lvis[n] || n >= nlayers);
        }

        public JKSector GetNewShape()
        {
            return GetPrefab(CBShapes.SelectedIndex);
        }

        public void LoadLayers()
        {
            RefreshLayers(true);
        }

        public JKSector GetPrefab(int n)
        {
            return Prefabs.GetPrefab(n);
        }

        public void SetDefaults()
        {
            int i = CBShapes.Items.IndexOf(GlobalVars.DefShape);
            if (i == -1)
                CBShapes.SelectedIndex = 0;
            else
                CBShapes.SelectedIndex = i;

            SetCB(CBGridStep, GlobalVars.DefGridStep);
            SetCB(CBGridLine, GlobalVars.DefGridLine);
            SetCB(CBGridDot, GlobalVars.DefGridDot);

            SetCB(CBTxStep, GlobalVars.DefTXStep);
            SetCB(CBRotStep, GlobalVars.DefTXRotStep);
            SetCB(CBScaleStep, GlobalVars.DefTXScaleStep);
            SetCB(CBPerpStep, GlobalVars.DefPerpStep);
            SetCB(CB3DPStep, GlobalVars.DefP3DStep);
            SetCB(CBGridSize, GlobalVars.DefGridSize);
        }

        public int AddLayer(string name)
        {
            // some of this seems shifty but whatever i believe its a direct replication of the original.. we'll "fix" it as necessary later


            int i = GlobalVars.Level.layers.Count;
            int n = GlobalVars.Level.AddLayer(name);
            if (GlobalVars.Level.layers.Count == i)
                return n;

            ListViewItem lvi = new ListViewItem(name);
            lvi.SubItems.Add("X");
            LVLayers.Items.Add(lvi);

            lvis[i] = true;
            nlayers = i + 1;

            return n;
        }

        public bool updating = false;

        public const int maxlayers = 2048;

        public bool[] lvis = new bool[maxlayers];

        public void SetLVis(string s)
        {
            for (int i = 0; i < GlobalVars.Level.layers.Count; i++)
            {
                if (i >= s.Length || s[i] == '1')
                    lvis[i] = true;
                else
                    lvis[i] = false;
            }

            RefreshLayers(false);
        }

        public string GetLVISString()
        {
            string s = string.Empty;
            for(int i=0; i<GlobalVars.Level.layers.Count; i++)
                s += lvis[i] ? "1" : "0";
            return s;
        }

        public int nlayers = 0;

        public void RefreshLayers(bool showAll)
        {
            updating = true;

            LVLayers.BeginUpdate();
            LVLayers.Items.Clear();
            for(int i=0; i<GlobalVars.Level.layers.Count; i++)
            {
                ListViewItem lvi = new ListViewItem(GlobalVars.Level.layers[i]);
                if (showAll)
                    lvis[i] = true;
                if (lvis[i])
                    lvi.SubItems.Add("X");
                else
                    lvi.SubItems.Add(string.Empty);

                LVLayers.Items.Add(lvi);
            }
            LVLayers.EndUpdate();

            updating = false;
            nlayers = GlobalVars.Level.layers.Count;
        }

        private void BNShow_Click(object sender, EventArgs e)
        {
            LVLayers.BeginUpdate();
            foreach (int i in LVLayers.SelectedIndices)
            {
                lvis[i] = true;
                LVLayers.Items[i].SubItems[1].Text = "X";
            }
            LVLayers.EndUpdate();
        }

        private void BNHide_Click(object sender, EventArgs e)
        {
            LVLayers.BeginUpdate();
            foreach (int i in LVLayers.SelectedIndices)
            {
                lvis[i] = false;
                LVLayers.Items[i].SubItems[1].Text = " ";
            }
            LVLayers.EndUpdate();
        }

        private void LVLayers_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            ListViewItem lvi = LVLayers.GetItemAt(e.X, e.Y);
            if (lvi == null)
                return;

            int n = LVLayers.Items.IndexOf(lvi);

            LVLayers.SelectedIndices.Clear();
            LVLayers.SelectedIndices.Add(n);

            if (lvis[n])
                BNHide_Click(BNHide, new EventArgs());
            else
                BNShow_Click(BNShow, new EventArgs());
        }

        /*private void MapSettings_FormClosed(object sender, FormClosedEventArgs e)
        {
            GlobalVars.TBarPos = Bounds;
        }*/

        public new void Show()
        {
            if (GlobalVars.NedMain != null)
                GlobalVars.NedMain.tabControl1.SelectedTab = GlobalVars.NedMain.tabMapSettings;
        }

        public void UpdatePrefabs()
        {
            CBShapes.Items.Clear();
            CBShapes.Items.AddRange(Prefabs.list.ToArray());

            int i = CBShapes.Items.IndexOf(GlobalVars.DefShape);
            if (i == -1)
                CBShapes.SelectedIndex = 0;
            else
                CBShapes.SelectedIndex = i;
        }

        private void BNDelShape_Click(object sender, EventArgs e)
        {
            int n = CBShapes.SelectedIndex;
            if (n < 1)
            {
                MessageBox.Show("You can't delete the shape \"Cube\"");
                return;
            }

            if (MessageBox.Show("Are you sure you want to delete shape \"" + CBShapes.Text + "\"", "Confirmation", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                return;

            Prefabs.DeletePrefab(n);
            Prefabs.SavePrefabs(Path.Combine(GlobalVars.BaseDir, @"JedData\shapes.tpl"));
            UpdatePrefabs();

            if (CBShapes.Items.Count > n)
                CBShapes.SelectedIndex = n;
            else
                CBShapes.SelectedIndex = CBShapes.Items.Count - 1;
        }

        void CBValDouble(ComboBox cb, ref double d, double min, double max)
        {
            if (updating)
                return;

            if (!double.TryParse(cb.Text, out d))
                return;

            d = MathLib.Clamp(d, min, max);
        }

        void CBValDouble(object sender, ref double d, double min, double max)
        {
            CBValDouble(sender as ComboBox, ref d, min, max);
        }

        private void CBGridStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.NedMain.renderer.GridStep, 0.0001, 10.0);
        }

        private void CBGridLine_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.NedMain.renderer.GridLine, 0.1, 10.0);
        }

        private void CBGridDot_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.NedMain.renderer.GridDot, 0.05, 10.0);
        }

        private void CBTxStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.TXStep, 0.0, 64.0);
        }

        private void CBRotStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.TXRotStep, 0.0, 360.0);
        }

        private void CBScaleStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.TXScaleStep, 1.0, 8.0);
        }

        private void CBPerpStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.PerpStep, 0.0, 1.0);
        }

        private void CB3DPStep_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.P3DStep, 0.001, 5.0);
        }

        private void CBGridSize_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.GridMoveStep, 1.0, 100.0);
        }

        private void CBGridMove_TextChanged(object sender, EventArgs e)
        {
            CBValDouble(sender, ref GlobalVars.GridMoveStep, 0.00001, 5.0);
        }

        private void BNSave_Click(object sender, EventArgs e)
        {
            GlobalVars.DefTXStep = GlobalVars.TXStep;
            GlobalVars.DefTXRotStep = GlobalVars.TXRotStep;
            GlobalVars.DefTXScaleStep = GlobalVars.TXScaleStep;
            GlobalVars.DefPerpStep = GlobalVars.PerpStep;
            GlobalVars.DefP3DStep = GlobalVars.P3DStep;
            //GlobalVars.DefThingView = GlobalVars.NedMain.thing_view;
            GlobalVars.DefMSelMode = GlobalVars.NedMain.msel_mode;

            GlobalVars.DefGridStep = GlobalVars.NedMain.renderer.GridStep;
            GlobalVars.DefGridLine = GlobalVars.NedMain.renderer.GridLine;
            GlobalVars.DefGridDot = GlobalVars.NedMain.renderer.GridDot;
            GlobalVars.DefGridSize = GlobalVars.GridSize;
            GlobalVars.DefShape = CBShapes.Text;

            Options.SaveSettings(true);
        }

        private void CBGrid_CheckedChanged(object sender, EventArgs e)
        {
            GlobalVars.NedMain.SnapToGrid = CBGrid.Checked;
            CBGridStep.Enabled = GlobalVars.NedMain.SnapToGrid;
        }

        private void PGStuff_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            //GlobalVars.NedMain.DoPreviewKeyDown(e);
        }

        private void PGStuff_KeyDown(object sender, KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyDown(e);
        }

        private void PGStuff_KeyUp(object sender, KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyUp(e);
        }
    }
}
