﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;

namespace Smith.NED
{
    public class JKModel
    {
        public static Dictionary<string, JKModel> Models = new Dictionary<string, JKModel>(StringComparer.InvariantCultureIgnoreCase);

        public static double OBJScale = 1.0;//10000.0;

        public Box3 GetBBox()
        {
            Box3 box = Box3.Expandable;
            foreach (Mesh mesh in meshes)
                foreach (Vertex v in mesh.vertices)
                    box.Expand(v.v);
            return box;
        }

        public double FindRadius()
        {
            double r = 0.0;
            foreach (JKModel.Mesh mesh in meshes)
                r = Math.Max(r, mesh.FindRadius());
            return r;
        }

        public class Face : Polygon
        {
            public int imat;
            public uint ftype;
            public int geo;
            public int light;
            public int tex;
            public double extra_l;

            public List<Vec3> vertNormals = new List<Vec3>();
        }

        public class Faces : Polygons
        {

        }

        public class Mesh
        {
            public string name;
            public Faces faces = new Faces();

            public Vertices vertices
            {
                get
                {
                    return faces.vertices;
                }
            }

            public double FindRadius()
            {
                double r = 0.0;
                foreach (Vertex v in vertices)
                    r = Math.Max(r, v.v.Magnitude);
                return r;
            }

            public int AddVertex(Vec3 v)
            {
                int i = vertices.Find(v);
                if (i != -1)
                    return i;

                Vertex nv = new Vertex();
                nv.v = v;
                vertices.Add(nv);
                return (vertices.Count - 1);
            }
        }

        public class Meshes : List<Mesh>
        {

        }

        public class HNode
        {
            public int id;
            public string nodename;
            public int ntype = -1;
            public int nmesh = -1;
            public int parent = -1;
            public Vec3 pos = Vec3.Zero;
            public Orient orient = Orient.Zero;
            public Vec3 pivot = Vec3.Zero;
            public Vec3 orgpos = Vec3.Zero;
            public Orient orgorient = Orient.Zero;

            public void SerializeTo(ZipUtil zip)
            {
                zip.WriteString(nodename);
                zip.WriteVec3(pos);
                zip.WriteOrient(orient);
                zip.WriteInt(ntype);
                zip.WriteInt(parent);
            }

            public void SerializeFrom(ZipUtil zip)
            {
                nodename = zip.ReadString();
                pos = zip.ReadVec3();
                orient = zip.ReadOrient();
                ntype = zip.ReadInt();
                parent = zip.ReadInt();
            }

            public void Assign(HNode o)
            {
                nodename = o.nodename;
                ntype = o.ntype;
                nmesh = o.nmesh;
                parent = o.parent;
                pos = o.pos;
                orient = o.orient;
                pivot = o.pivot;
            }

            public Mat4 RotMatrix
            {
                get
                {
                    return Utils.CreateRotMatrix(orient.pitch, orient.yaw, orient.roll);
                    //return orient.Matrix.Inverse/*think we good if we just use our existing calculation and invert it*/;
                }
            }

            public string ToString(int n)
            {
                return n.ToString() + "     0x0    0x" + ((uint)ntype).ToString("X") + "    " + nmesh.ToString() + "     " + parent.ToString() +
                    "      0     0       0     " + pos.x.ToString() + " " + pos.y.ToString() + " " + pos.z.ToString() + "  " +
                    orient.pitch.ToString() + " " + orient.yaw.ToString() + " " + orient.roll.ToString() + " " +
                    pivot.x.ToString() + " " + pivot.y.ToString() + " " + pivot.z.ToString() + " " + nodename;
            }

            public void FromString(string s)
            {
                string[] vals = s.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                ntype = (int)JK.ParseFlags(vals[2]);
                nmesh = int.Parse(vals[3]);
                parent = int.Parse(vals[4]);
                orgpos.x = double.Parse(vals[8]);
                orgpos.y = double.Parse(vals[9]);
                orgpos.z = double.Parse(vals[10]);
                orgorient.pitch = double.Parse(vals[11]);
                orgorient.yaw = double.Parse(vals[12]);
                orgorient.roll = double.Parse(vals[13]);
                pivot.x = double.Parse(vals[14]);
                pivot.y = double.Parse(vals[15]);
                pivot.z = double.Parse(vals[16]);
                nodename = vals[17];
            }

            public static HNode NewFromString(string s)
            {
                HNode n = new HNode();

                n.FromString(s);

                return n;
            }
        }

        public class HNodes : List<HNode>
        {

        }

        public List<string> mats = new List<string>();
        public Meshes meshes = new Meshes();
        public HNodes hnodes = new HNodes();

        public void UnOffsetMeshes()
        {
            for (int i = 0; i < hnodes.Count; i++)
            {
                HNode hnode = hnodes[i];
                if (hnode.nmesh == -1)
                    continue;

                List<int> branch = new List<int>();
                branch.Add(i);
                HNode hnode1 = hnode;
                while (hnode1.parent != -1)
                {
                    branch.Add(hnode1.parent);
                    hnode1 = hnodes[hnode1.parent];
                }

                // unrotate and unoffset in backwards order
                Vertices vs = meshes[hnode.nmesh].vertices;
                for (int k = branch.Count - 1; k >= 0; k--)
                {
                    hnode1 = hnodes[branch[k]];
                    Mat4 mx = Utils.CreateRotMatrix(hnode1.orient.pitch, hnode1.orient.yaw, hnode1.orient.roll);

                    Mat4 rmx = mx.Transposed;

                    foreach (Vertex v in vs)
                        v.v = (v.v - hnode1.pos).Transformed(rmx) - hnode.pivot;
                }
            }
        }

        private class OBJBuilder
        {
            public List<Vec3> Vertices = new List<Vec3>();
            public List<Vec3> Normals = new List<Vec3>();
            public List<Vec2> TexVertices = new List<Vec2>();

            public bool Is3DOSectors = true;
            public double Scale = 1.0;
            public Mat4 WorldRotation = Mat4.Identity;

            public VertRef AddVert(Vec3 v, Vec3 n, Vec2 tv, Vec2 texsize)
            {
                // massage data

                v.Transform(WorldRotation);
                n.Transform(WorldRotation);

                v *= Scale;
                if(!Is3DOSectors)
                    n *= -1.0;
                n = n.Unit;
                tv = tv.DivComps(texsize);
                tv.y = 1.0 - tv.y;


                // maybe doesnt like negative tex coords?
                //while (tv.x < 0.0)
                //    tv.x += 1.0;
                //while (tv.y < 0.0)
                //    tv.y += 1.0;


                VertRef vr = new VertRef();

                vr.VI = Vertices.IndexOf(v);
                if(vr.VI == -1)
                {
                    vr.VI = Vertices.Count;
                    Vertices.Add(v);
                }

                vr.NI = Normals.IndexOf(n);
                if(vr.NI == -1)
                {
                    vr.NI = Normals.Count;
                    Normals.Add(n);
                }

                vr.TI = TexVertices.IndexOf(tv);
                if (vr.TI == -1)
                {
                    vr.TI = TexVertices.Count;
                    TexVertices.Add(tv);
                }

                return vr;
            }

            public class VertRef
            {
                public int VI;
                public int NI;
                public int TI;
            }
        }

        public enum OBJTextureOutput
        {
            Bmp,
            Png,
            Tga
        }

        public void SaveToOBJ(string name, double scale, bool is3dosectors, string cmpname, OBJTextureOutput texoutput)
        {
            string dir = Path.GetDirectoryName(name);
            if (!string.IsNullOrEmpty(dir) && dir[1] == ':')
                Directory.CreateDirectory(dir);

            string projname = Path.GetFileNameWithoutExtension(name);

            string subdir = Path.Combine(dir, projname);
            Directory.CreateDirectory(subdir);

            string objname = projname + ".obj";
            string mtlname = projname + ".mtl";

            Dictionary<string, Material> materials = new Dictionary<string, Material>(StringComparer.InvariantCultureIgnoreCase);
            foreach (Mesh mesh in meshes)
                foreach (Face face in mesh.faces)
                {
                    string matfile = mats[face.imat];
                    if (!materials.ContainsKey(matfile))
                    {
                        Material mat = PreviewRenderer.LoadMaterial(matfile);
                        materials.Add(matfile, mat);
                    }
                }


            OBJBuilder ob = new OBJBuilder();
            ob.Scale = scale;
            ob.Is3DOSectors = is3dosectors;

            ob.WorldRotation = Mat4.RotationX(MathLib.ToRad(-90.0))/*for blender*/;

            foreach (Mesh mesh in meshes)
                foreach (Face face in mesh.faces)
                {
                    Material mat = materials[mats[face.imat]];
                    Vec2 texSize = new Vec2((double)mat.Width, (double)mat.Height);

                    for (int x = 0; x < face.vertices.Count; x++)
                    {
                        Vertex v = face.vertices[x];
                        TexVertex tv = face.texvertices[x];

                        Vec3 vnorm;
                        //if (face.vertNormals.Count == 0)
                            vnorm = face.normal;
                        //else
                        //    vnorm = face.vertNormals[x];

                        OBJBuilder.VertRef vr = ob.AddVert(v.v, vnorm, tv.v, texSize);
                    }
                }

            using (StreamWriter obj = File.CreateText(Path.Combine(subdir, objname)))
            {
                obj.WriteLine($"mtllib {mtlname}");

                obj.WriteLine();

                foreach (Vec3 v in ob.Vertices)
                    obj.WriteLine($"v {Utils.ostr(v.x, 6)} {Utils.ostr(v.y, 6)} {Utils.ostr(v.z, 6)}");

                obj.WriteLine();

                foreach (Vec3 v in ob.Normals)
                    obj.WriteLine($"vn {Utils.ostr(v.x, 6)} {Utils.ostr(v.y, 6)} {Utils.ostr(v.z, 6)}");

                obj.WriteLine();

                foreach (Vec2 v in ob.TexVertices)
                    obj.WriteLine($"vt {Utils.ostr(v.x, 6)} {Utils.ostr(v.y, 6)}");

                obj.WriteLine();

                foreach(Mesh mesh in meshes)
                {
                    //obj.WriteLine($"g {(string.IsNullOrEmpty(mesh.name)?"UNNAMED":mesh.name)}");

                    int lastimat = -1;
                    foreach(Face face in mesh.faces)
                    {
                        Material mat = materials[mats[face.imat]];
                        Vec2 texSize = new Vec2((double)mat.Width, (double)mat.Height);

                        if(face.imat != lastimat)
                        {
                            obj.WriteLine($"usemtl {Path.GetFileNameWithoutExtension(mats[face.imat])}");
                            lastimat = face.imat;
                        }

                        string ln = string.Empty;
                        for(int x=0; x< face.vertices.Count; x++)
                        //foreach(Vertex v in face.vertices)
                        {
                            int i;

                            // change winding order?
                            if (is3dosectors)
                                i = x;
                            else
                                i = face.vertices.Count - x - 1;

                            Vertex v = face.vertices[i];
                            TexVertex tv = face.texvertices[i];

                            Vec3 vnorm;
                            //if (face.vertNormals.Count == 0)
                                vnorm = face.normal;
                            //else
                            //    vnorm = face.vertNormals[i];

                            //OBJBuilder.VertRef vr = v.tag as OBJBuilder.VertRef;
                            OBJBuilder.VertRef vr = ob.AddVert(v.v, vnorm, tv.v, texSize);

                            /*if (vr.VI < 0 || vr.VI >= ob.Vertices.Count ||
                               vr.NI < 0 || vr.NI >= ob.Normals.Count ||
                               vr.TI < 0 || vr.TI >= ob.TexVertices.Count)
                            {
                                vr.VI = vr.VI;
                            }*/

                            ln += $" {vr.VI+1}/{vr.TI+1}/{vr.NI+1}";
                        }

                        obj.WriteLine($"f{ln}");
                    }
                }

                obj.WriteLine();
            }


            using (StreamWriter mtl = File.CreateText(Path.Combine(subdir, mtlname)))
            {
                Colormap cmp = PreviewRenderer.LoadColormap(/*GlobalVars.Level.GetMasterCMP()*/cmpname);

                List<string> uniqueMatNames = new List<string>();
                foreach (Mesh mesh in meshes)
                    foreach (Face face in mesh.faces)
                    {
                        string matfile = mats[face.imat];
                        string matname = Path.GetFileNameWithoutExtension(matfile);

                        if (uniqueMatNames.Contains(matname))
                            continue;

                        uniqueMatNames.Add(matname);

                        string bmpfile;
                        switch(texoutput)
                        {
                            case OBJTextureOutput.Png:
                                bmpfile = matname + ".png";
                                break;

                            case OBJTextureOutput.Tga:
                                bmpfile = matname + ".tga";
                                break;

                            default:
                                bmpfile = matname + ".bmp";
                                break;
                        }

                        mtl.WriteLine($"newmtl {matname}");

                        mtl.WriteLine($"\td 1.000000");
#if true
                        mtl.WriteLine($"\tmap_Kd {bmpfile}");
#else
                        mtl.WriteLine($"\tmap_Kd {projname}/{bmpfile}");
#endif

                        mtl.WriteLine();



                        Material mat = materials[matfile];

                        if (cmp == null || mat == null)
                        {
                            continue;
                        }

                        Bitmap bm;
                        Bitmap dummy;
                        mat.GenerateBitmap(out bm, out dummy, cmp, 0);

                        if (texoutput != OBJTextureOutput.Tga)
                            bm.Save(Path.Combine(subdir, bmpfile), texoutput == OBJTextureOutput.Png ? System.Drawing.Imaging.ImageFormat.Png : System.Drawing.Imaging.ImageFormat.Bmp);
                        else
                        {
                            ZipUtil zip = new ZipUtil_File(Path.Combine(subdir, bmpfile), ZipUtil_File.Mode.Create);

                            zip.WriteByte(0);
                            zip.WriteByte(0);
                            zip.WriteByte(2);
                            zip.WriteByte(0); zip.WriteByte(0);
                            zip.WriteByte(0); zip.WriteByte(0);
                            zip.WriteByte(0);
                            zip.WriteByte(0); zip.WriteByte(0);
                            zip.WriteByte(0); zip.WriteByte(0);
                            zip.WriteByte((byte)(bm.Width & 0xFF));
                            zip.WriteByte((byte)((bm.Width & 0xFF00)/256));
                            zip.WriteByte((byte)(bm.Height & 0xFF));
                            zip.WriteByte((byte)((bm.Height & 0xFF00) / 256));
                            zip.WriteByte(24);
                            zip.WriteByte(0);

                            for(int y=0; y<bm.Height; y++)
                                for(int x=0; x<bm.Width; x++)
                                {
                                    Color c = bm.GetPixel(x, bm.Height-y-1);

                                    zip.WriteByte((byte)c.B);
                                    zip.WriteByte((byte)c.G);
                                    zip.WriteByte((byte)c.R);
                                }

                            zip.Close();
                        }
                    }
                }
        }

        public void SaveTo3DO(string name)
        {
            UnOffsetMeshes();

            string dir = Path.GetDirectoryName(name);
            if (!string.IsNullOrEmpty(dir) && dir[1] == ':')
                Directory.CreateDirectory(dir);

            using (StreamWriter sw = File.CreateText(name))
            {
                sw.WriteLine("# MODEL " + name + " created by NED " + GlobalVars.JedVerNum);
                sw.WriteLine();
                sw.WriteLine("SECTION: HEADER");
                sw.WriteLine();
                sw.WriteLine("3DO 2.1");
                sw.WriteLine();
                sw.WriteLine("SECTION: MODELRESOURCE");
                sw.WriteLine();
                sw.WriteLine("MATERIALS " + mats.Count.ToString());
                for (int i = 0; i < mats.Count; i++)
                    sw.WriteLine(i.ToString() + ": " + mats[i]);

                double rad = FindRadius();

                sw.WriteLine();
                sw.WriteLine("SECTION: GEOMETRYDEF");
                sw.WriteLine();
                sw.WriteLine("RADIUS " + rad.ToString("0.######"));
                sw.WriteLine();
                sw.WriteLine("INSERT OFFSET 0 0 0");
                sw.WriteLine();
                sw.WriteLine("GEOSETS 1");
                sw.WriteLine();
                sw.WriteLine("GEOSET 0");
                sw.WriteLine();
                sw.WriteLine("MESHES " + meshes.Count.ToString());
                for (int i = 0; i < meshes.Count; i++)
                {
                    Mesh mesh = meshes[i];

                    sw.WriteLine();
                    sw.WriteLine("MESH " + i.ToString());
                    sw.WriteLine("NAME " + mesh.name);

                    rad = mesh.FindRadius();

                    sw.WriteLine("RADIUS " + rad.ToString("0.######"));
                    sw.WriteLine("GEOMETRYMODE  4");
                    sw.WriteLine("LIGHTINGMODE  3");
                    sw.WriteLine("TEXTUREMODE   1");
                    sw.WriteLine();

                    sw.WriteLine("VERTICES " + mesh.vertices.Count.ToString());
                    sw.WriteLine();
                    for(int j=0; j<mesh.vertices.Count; j++)
                    {
                        Vertex v = mesh.vertices[j];
                        sw.WriteLine(j.ToString() + ": " + v.v.x.ToString("0.######") + " " + v.v.y.ToString("0.######") + " " + v.v.z.ToString("0.######") + " 0");
                    }
                    sw.WriteLine();

                    // texture vertices
                    TexVertices txidx = new TexVertices();
                    TexVertices txvxlist = new TexVertices();

                    for (int j = 0; j < mesh.faces.Count; j++)
                    {
                        Face face = mesh.faces[j] as Face/*hope so*/;

                        for (int k = 0; k < face.texvertices.Count; k++)
                        {
                            TexVertex txvx = face.texvertices[k];
                            txvx.num = Utils.AddTXVX(txvx, txvxlist, txidx);
                        }
                    }

                    sw.WriteLine("TEXTURE VERTICES " + txidx.Count.ToString());
                    sw.WriteLine();
                    for (int j = 0; j < txvxlist.Count; j++)
                    {
                        TexVertex txvx = txvxlist[j];
                        sw.WriteLine($"{j}: {Utils.dstr(txvx.v.x, 6)} {Utils.dstr(txvx.v.y, 6)}");
                    }

                    // vertex normals
                    foreach (Face f in mesh.faces)
                        f.Recalc();

                    List<Vec3> vxnorms = new List<Vec3>();
                    for (int j = 0; j < mesh.vertices.Count; j++)
                    {
                        vxnorms.Add(Vec3.Zero);

                        mesh.vertices[j].num = j; // num will be number of vertex
                        mesh.vertices[j].mark = 0; // mark will be number of references to vx
                    }

                    foreach (Face face in mesh.faces)
                        foreach (Vertex facevert in face.vertices)
                        {
                            vxnorms[facevert.num] += face.normal;
                            facevert.mark++;
                        }

                    sw.WriteLine("VERTEX NORMALS");
                    sw.WriteLine();

                    for (int j = 0; j < mesh.vertices.Count; j++)
                    {
                        Vec3 pn = vxnorms[j];
                        if (mesh.vertices[j].mark != 0)
                            pn /= (double)mesh.vertices[j].mark;

                        sw.WriteLine(j.ToString() + ": " + pn.x.ToString("0.######") + " " + pn.y.ToString("0.######") + " " + pn.z.ToString("0.######"));
                    }

                    sw.WriteLine();

                    sw.WriteLine("FACES " + mesh.faces.Count.ToString());
                    sw.WriteLine();

                    for (int j = 0; j < mesh.faces.Count; j++)
                    {
                        Face face = mesh.faces[j] as Face/*hope so*/;

                        string s = j.ToString() + ": " + face.imat.ToString() + " 0x" + ((int)face.ftype).ToString("x") + " " + ((int)face.geo).ToString() + " " + ((int)face.light).ToString() + " " + ((int)face.tex).ToString() + " " + face.extra_l.ToString("0.####") + " " + face.vertices.Count.ToString();

                        for(int v=0; v<face.vertices.Count; v++)
                            s += " " + face.vertices[v].num.ToString() + ", " + face.texvertices[v].num.ToString();

                        sw.WriteLine(s);
                    }

                    sw.WriteLine();
                    sw.WriteLine("FACE NORMALS");
                    sw.WriteLine();

                    for(int j=0; j<mesh.faces.Count; j++)
                    {
                        Vec3 n = mesh.faces[j].normal;
                        sw.WriteLine(j.ToString() + ": " + n.x.ToString("0.######") + " " + n.y.ToString("0.######") + " " + n.z.ToString("0.######"));
                    }
                }

                sw.WriteLine();
                sw.WriteLine("SECTION: HIERARCHYDEF");
                sw.WriteLine();

                sw.WriteLine("HIERARCHY NODES " + hnodes.Count.ToString());
                sw.WriteLine("# num: flags: type: mesh: parent: child:  sibling:  numChildren: x:   y:  z:  pitch: yaw: roll: pivotx: pivoty: pivotz: hnodename:");
                for(int i=0; i<hnodes.Count; i++)
                {
                    // count number of children,  find values for child and sibling
                    int nc = 0;
                    int k = -1;
                    int v = -1;

                    for(int j=0; j<hnodes.Count; j++)
                    {
                        if(hnodes[j].parent != i)
                            continue;

                        nc++;
                        if(k == -1)
                            k = j;
                    }

                    for(int j=i+1; j<hnodes.Count; j++)
                        if(hnodes[j].parent == hnodes[i].parent)
                        {
                            v = j;
                            break;
                        }

                    HNode hn = hnodes[i];
                    sw.WriteLine(i.ToString() + ":     0x0    0x" + ((int)hn.ntype).ToString("x") + "    " + hn.nmesh.ToString() + "     " +
                    hn.parent.ToString() + "      " + k.ToString() + "     " + v.ToString() + "       " + nc.ToString() + "     " +
                    Utils.dstr(hn.pos.x, 6) + " " + Utils.dstr(hn.pos.y, 6) + " " + Utils.dstr(hn.pos.z, 6) + "  " +
                    Utils.dstr(hn.orient.pitch, 6) + " " + Utils.dstr(hn.orient.yaw, 6) + " " + Utils.dstr(hn.orient.roll, 6) + " " +
                    Utils.dstr(hn.pivot.x, 6) + " " + Utils.dstr(hn.pivot.y, 6) + " " + Utils.dstr(hn.pivot.z, 6) + " " +
                    (string.IsNullOrEmpty(hn.nodename) ? $"UNNAMED{i}" : hn.nodename));
                }
            }
        }

        public string GetMat(int i)
        {
            if (i < 0 || i >= mats.Count)
                return string.Empty;

            return mats[i];
        }

        private void LoadHeader(Tokenizer t)
        {
            t.EatToken(2);
        }

        private void LoadGeometry(Tokenizer t, ref int lod)
        {
            // radius
            t.EatToken(2);

            // insert offset
            t.EatToken(5);

            t.EatToken();
            int numgeosets = t.ReadInt();

            lod = Math.Min(lod, numgeosets-1);

            // jump to desired geoset
            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (!ln.StartsWith("geoset", StringComparison.InvariantCultureIgnoreCase))
                    continue;

                int geoset = int.Parse(ln.Substring(6).Trim(new char[] { ' ', '\t' }));
                if(geoset == lod)
                    break;
            }

            //for (int ig = 0; ig < numgeosets; ig++)   // only do once
            {
                //t.EatToken(2);

                t.EatToken();
                int nummeshes = t.ReadInt();

                for (int im = 0; im < nummeshes; im++)
                {
                    List<Vec2> GTXVXList = new List<Vec2>();

                    Mesh m = new Mesh();
                    meshes.Add(m);// only keep from our desired LOD

                    t.EatToken(2);

                    t.EatToken();
                    m.name = t.ReadLine(false, true);
                    if (string.IsNullOrEmpty(m.name))
                        m.name = $"UNNAMED{im}";

                    t.EatToken(8);

                    t.EatToken();
                    int numverts = t.ReadInt();

                    for (int iv = 0; iv < numverts; iv++)
                    {
                        Vertex v = new Vertex();

                        t.EatToken();
                        v.v.x = t.ReadDouble();
                        v.v.y = t.ReadDouble();
                        v.v.z = t.ReadDouble();
                        t.EatToken();// intensity

                        m.vertices.Add(v);
                    }

                    t.EatToken(2);
                    int numtex = t.ReadInt();
                    for (int itv = 0; itv < numtex; itv++)
                    {
                        Vec2 tv = new Vec2();

                        t.EatToken();
                        tv.x = t.ReadDouble();
                        tv.y = t.ReadDouble();

                        GTXVXList.Add(tv);
                    }

                    // screw all the vert normals i guess
                    List<Vec3> vertNormals = new List<Vec3>();
                    t.EatLine(1);
                    for(int x=0; x<numverts; x++)
                    {
                        Vec3 v = new Vec3();

                        t.EatToken();
                        v.x = t.ReadDouble();
                        v.y = t.ReadDouble();
                        v.z = t.ReadDouble();

                        vertNormals.Add(v);
                    }

                    t.EatToken();
                    int numfaces = t.ReadInt();
                    for (int nf = 0; nf < numfaces; nf++)
                    {
                        Face f = new Face();
                        m.faces.Add(f);

                        t.EatToken();
                        f.imat = t.ReadInt();
                        f.ftype = t.ReadFlags();
                        f.geo = t.ReadInt();
                        f.light = t.ReadInt();
                        f.tex = t.ReadInt();
                        f.extra_l = t.ReadDouble();

                        int faceverts = t.ReadInt();
                        for (int nfv = 0; nfv < faceverts; nfv++)
                        {
                            int vid = t.ReadInt();
                            f.AddVertex(m.vertices[vid]);

                            int tvid = t.ReadInt();
                            if (tvid >= 0 && tvid < GTXVXList.Count)
                            {
                                f.texvertices[nfv].v = GTXVXList[tvid];
                            }/* else {
                                tvid = tvid;
                            }*/

                            f.vertNormals.Add(vertNormals[vid]);
                        }
                    }

                    // eat face normals
                    t.EatToken(2 + numfaces * 4);
                }
            }


            // jump to start of hierarchy
            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (!ln.StartsWith("section:", StringComparison.InvariantCultureIgnoreCase))
                    continue;


                break;
            }
        }

        private void LoadMats(Tokenizer t)
        {
            t.EatToken();
            int nummats = t.ReadInt();

            for (int x = 0; x < nummats; x++)
            {
                t.EatToken();
                mats.Add(t.ReadToken());
            }

        }

        private void LoadHierarchy(Tokenizer t)
        {
            t.EatToken(2);
            int numnodes = t.ReadInt();

            for (int x = 0; x < numnodes; x++)
                hnodes.Add(HNode.NewFromString(t.ReadLine()));
        }

        public JKModel()
        {

        }

        public JKModel(string name, int lod)
        {
            Stream stream = null;
            if (FileOperations.ExtractPath(name).Length > 0)
                using(BaseFile bf = FileOperations.OpenFileRead(name))
                    stream = new MemoryStream(bf.Read(bf.Length));
            else
                using (BaseFile bf = FileOperations.OpenGameFile(name))
                    stream = new MemoryStream(bf.Read(bf.Length));

#if true
            if (Program.DumpDebug)
            {
                long oldpos = stream.Position;
                StreamReader sr = new StreamReader(stream);
                int lasti = name.LastIndexOfAny(new char[] { '\\', '/' });
                if (lasti != -1)
                    name = name.Substring(lasti + 1);
                File.WriteAllText(@"c:\smith\tmp\" + name, sr.ReadToEnd());
                stream.Position = oldpos;
            }
#endif

            Tokenizer t = new Tokenizer(stream);

            while (!t.EndOfStream)
            {
                string tok = t.ReadToken();
                if (string.IsNullOrEmpty(tok))
                    break;

                if (tok.Equals("section:", StringComparison.InvariantCultureIgnoreCase))
                {
                    string section = t.ReadToken();
                    switch (section.ToUpperInvariant())
                    {
                        case "HEADER": LoadHeader(t); break;
                        case "GEOMETRYDEF": LoadGeometry(t, ref lod); /*loading geometry eats the section: hierarchydef line so just continue directly*/LoadHierarchy(t);  break;
                        case "MODELRESOURCE": LoadMats(t); break;
                        //case "HIERARCHYDEF": LoadHierarchy(t); break;
                    }
                }
            }

            SetDefaultOffsets();
            OffsetMeshes();
        }

        private void SetDefaultOffsets()
        {
            foreach (HNode hnode in hnodes)
            {
                hnode.pos = hnode.orgpos;
                hnode.orient = hnode.orgorient;
            }
        }

        private void OffsetMeshes()
        {
            for (int i = 0; i < hnodes.Count; i++)
            {
                HNode hnode = hnodes[i];
                if (hnode.nmesh == -1)
                    continue;

                Vertices vs = meshes[hnode.nmesh].vertices;
                Mat4 mx = hnode.RotMatrix;
                foreach(Vertex v in vs)
                {
                    v.v += hnode.pivot;
                    v.v &= mx;
                    v.v += hnode.pos;
                }

                HNode hnodel = hnode;
                while (hnodel.parent != -1)
                {
                    hnodel = hnodes[hnodel.parent];

                    mx = hnodel.RotMatrix;

                    foreach (Vertex v in vs)
                        v.v = v.v.Transformed(mx) + hnodel.pos;
                }
            }
        }

        public static JKModel Load3DO(string fname)
        {
            if (!Models.ContainsKey(fname))
            {
                JKModel model = null;
                try
                {
                    model = new JKModel(fname, 3);
                }
                catch
                {
                    Utils.PanMessage(Utils.MessageType.Warning, "Cannot load " + fname);
                    return null;
                }

                Models[fname] = model;
            }

            return Models[fname];
        }
    }
}
