﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Smith.NED
{
    public partial class OpenFromContainer : GUI.BaseForm
    {
        public class FilterComboBox : ComboBox
        {
            private string _Filter = string.Empty;

            public class FilterEntry
            {
                public string FriendlyName;
                public string FilterValue;

                public FilterEntry(string _FriendlyName, string _FilterValue)
                {
                    FriendlyName = _FriendlyName;
                    FilterValue = _FilterValue;
                }

                public override string ToString()
                {
                    return FriendlyName;
                }
            }

            public string Mask
            {
                get
                {
                    FilterEntry fe = SelectedItem as FilterEntry;
                    if (fe == null)
                        return "*.*";

                    return fe.FilterValue;
                }
            }

            [Category("Data"), Browsable(true)]
            public string Filter
            {
                get
                {
                    return _Filter;
                }

                set
                {
                    _Filter = value;

                    BeginUpdate();
                    Items.Clear();
                    string[] filts = value.Split('|');
                    for (int x = 0; x < (filts.Length - 1); x += 2)
                        Items.Add(new FilterEntry(filts[x], filts[x + 1]));
                    EndUpdate();

                    if (Items.Count > 0)
                        SelectedIndex = 0;
                }
            }
        }

        public FileOperations.MaskedDirectoryControl DirControl;
        public ContainerFile Dir = null;
        public string SubDir = string.Empty;

        public OpenFromContainer()
        {
            InitializeComponent();

            DirControl = new FileOperations.MaskedDirectoryControl(DirList);
        }

        public string FileName
        {
            get
            {
                if(Dir == null)
                    return string.Empty;

                if (string.IsNullOrEmpty(SubDir))
                    return Dir.FullName + ">" + EBFname.Text;
                else
                    return Dir.FullName + ">" + SubDir + @"\" + EBFname.Text;
            }
        }

        public void SetFilter(string filter)
        {
            CBFilter.Filter = filter;
        }

        public void SetContainer(string container)
        {
            Dir = FileOperations.OpenContainer(container);
            DirControl.SetDir(Dir);
            SubDir = string.Empty;
            LBContainer.Text = container;
        }

        private void CBFilter_SelectedIndexChanged(object sender, EventArgs e)
        {
            DirControl.SetMask(CBFilter.Mask);
        }

        private void EBFname_TextChanged(object sender, EventArgs e)
        {
            for (int i = 0; i < DirList.Items.Count; i++)
            {
                FileOperations.MaskedDirectoryControl.ListEntry le = DirList.Items[i] as FileOperations.MaskedDirectoryControl.ListEntry;
                if (le.Name.Equals(EBFname.Text, StringComparison.InvariantCultureIgnoreCase))
                {
                    DirList.SelectedIndex = i;
                    return;
                }
            }
        }

        private void DirList_SelectedIndexChanged(object sender, EventArgs e)
        {
            FileOperations.MaskedDirectoryControl.ListEntry le = DirList.SelectedItem as FileOperations.MaskedDirectoryControl.ListEntry;
            if (le == null)
                return;

            if(le.FInfo != null)
                EBFname.Text = le.Name;

            if (le.FInfo == null)
                LBFileSize.Text = "Directory";
            else
                LBFileSize.Text = le.FInfo.size.ToString();
        }

        private void DirList_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            BNOpen_Click(BNOpen, new EventArgs());
        }

        private void BNOpen_Click(object sender, EventArgs e)
        {
            FileOperations.MaskedDirectoryControl.ListEntry le = DirList.SelectedItem as FileOperations.MaskedDirectoryControl.ListEntry;
            if (le == null)
                return;

            if (le.FInfo == null)
            {
                if (string.IsNullOrEmpty(SubDir))
                    SubDir = le.Name;
                else
                    SubDir += @"\" + le.Name;

                Dir.ChDir(SubDir);
                DirControl.SetMask(CBFilter.Mask);
                return;
            }

            if (!Dir.ListFiles.ContainsKey(EBFname.Text))
                MessageBox.Show("The file " + EBFname.Text + " is not in the container", "Error");
            else
            {
                DialogResult = System.Windows.Forms.DialogResult.OK;
                Close();
            }
        }

        private void BNDirUp_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(SubDir))
                return;

            int ps = SubDir.LastIndexOf('\\');
            if (ps == -1)
                SubDir = string.Empty;
            else
                SubDir = SubDir.Substring(0, ps);

            Dir.ChDir(SubDir);
            DirControl.SetMask(CBFilter.Mask);
        }
    }
}
