﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Microsoft.Win32;
using System.Net;
using System.Threading;

namespace Smith.NED
{
    public partial class Options : GUI.BaseForm
    {
        public void ReceiveSettings()
        {
            jkDirectory.Text = GlobalVars.JKDir;
            motsDirectory.Text = GlobalVars.MOTSDir;
            ijifDirectory.Text = GlobalVars.IJIFDir;
            dwDirectory.Text = GlobalVars.DWDir;

            RGNewStyle.Checked = GlobalVars.MapRot == GlobalVars.MapRotation.New;
            RGOldStyle.Checked = GlobalVars.MapRot == GlobalVars.MapRotation.Old;

            CB3DLayers.Checked = GlobalVars.P3DVisLayers;
            CBShowThings.Checked = GlobalVars.P3DThings;
            CBFullLight.Checked = GlobalVars.P3DFullLit;
            CBColored.Checked = GlobalVars.P3DColoredLights;
            CBP3DOnTop.Checked = GlobalVars.P3DOnTop;
            EBGamma.Text = GlobalVars.P3DGamma.ToString();

            CBAutoSave.Checked = GlobalVars.AutoSave;
            EBSaveInt.Value = (decimal)GlobalVars.SaveInterval;

            CBThingsOnFloor.Checked = GlobalVars.NewOnFloor;
            CBUndo.Checked = GlobalVars.UndoEnabled;
            CBMoveFrames.Checked = GlobalVars.MoveFrames;
            CBGobSmart.Checked = GlobalVars.GOBSmart;
            CBCheckOverlaps.Checked = GlobalVars.CheckOverlaps;
            CBNewLightCalc.Checked = GlobalVars.NewLightCalc;
            CBConfRevert.Checked = GlobalVars.ConfirmRevert;
            CBCalculateLightsAuto.Checked = GlobalVars.CalculateLightsAuto;

            smithNeo.Checked = GlobalVars.SmithNeo;
            smithMini.Checked = GlobalVars.SmithMini;
            smithSync.Checked = GlobalVars.SmithSyncCameras;
            smithNoPrecache.Checked = GlobalVars.SmithNoPrecache;
            UpdateSmithCmdLine();

            LBColors.BeginUpdate();
            LBColors.Items.Clear();
            AddColor("Map background");
            AddColor("Geometry");
            AddColor("Geometry - backside");
            AddColor("Map selection");
            AddColor("Map selection - backside");
            AddColor("Grid");
            AddColor("Vertices");
            AddColor("Things");
            AddColor("Frames");
            AddColor("Lights");
            AddColor("Multiselection");
            AddColor("Multiselection - backside");
            AddColor("Multi&Selected");
            AddColor("Multi&Selected - backside");
            AddColor("Grid X");
            AddColor("Grid Y");
            AddColor("Plug-in objects");   
            LBColors.EndUpdate();
        }

        private void AddColor(string desc)
        {
            ColorEntry ce = new ColorEntry(desc, LBColors.Items.Count);
            LBColors.Items.Add(ce);
        }

        const int numcolors = 17;

        public void ApplySettings()
        {
            double dval;

            GlobalVars.JKDir = jkDirectory.Text;
            GlobalVars.MOTSDir = motsDirectory.Text;
            GlobalVars.IJIFDir = ijifDirectory.Text;
            GlobalVars.DWDir = dwDirectory.Text;

            GlobalVars.MapRot = RGNewStyle.Checked ? GlobalVars.MapRotation.New : GlobalVars.MapRotation.Old;

            GlobalVars.P3DVisLayers = CB3DLayers.Checked;
            GlobalVars.P3DThings = CBShowThings.Checked;
            GlobalVars.P3DFullLit = CBFullLight.Checked;
            GlobalVars.P3DColoredLights = CBColored.Checked;
            GlobalVars.P3DOnTop = CBP3DOnTop.Checked;
            if (double.TryParse(EBGamma.Text, out dval))
                GlobalVars.P3DGamma = dval;

            GlobalVars.AutoSave = CBAutoSave.Checked;
            GlobalVars.SaveInterval = (int)EBSaveInt.Value;

            GlobalVars.NewOnFloor = CBThingsOnFloor.Checked;
            GlobalVars.UndoEnabled = CBUndo.Checked;
            GlobalVars.MoveFrames = CBMoveFrames.Checked;
            GlobalVars.GOBSmart = CBGobSmart.Checked;
            GlobalVars.CheckOverlaps = CBCheckOverlaps.Checked;
            GlobalVars.NewLightCalc = CBNewLightCalc.Checked;
            GlobalVars.ConfirmRevert = CBConfRevert.Checked;
            GlobalVars.CalculateLightsAuto = CBCalculateLightsAuto.Checked;

            GlobalVars.SmithNeo = smithNeo.Checked;
            GlobalVars.SmithMini = smithMini.Checked;
            GlobalVars.SmithSyncCameras = smithSync.Checked;
            GlobalVars.SmithNoPrecache = smithNoPrecache.Checked;

            foreach (ColorEntry ce in LBColors.Items)
                SetColor(ce.ClrIndex, ce.Color4);
        }

        void UpdateSmithCmdLine()
        {
            if (GlobalVars.NedMain == null)
            {
                smithCmdline.Text = string.Empty;
                return;
            }

            string s = string.Empty;
            string[] c = GlobalVars.NedMain.GenerateSmithCmdLine(smithNeo.Checked, smithMini.Checked, smithSync.Checked, smithNoPrecache.Checked);
            if (c.Length > 0)
            {
                s = c[0];
                for (int x = 1; x < c.Length; x++)
                    s += " " + c[x];
            }
            smithCmdline.Text = s;
        }

        //private static INIFile _ConfigINI = null;
        public static INIFile ConfigINI
        {
            get
            {
                return Program.ConfigINI;
            }
        }

        public const string JEDRegBase = @"Software\Code Alliance\Jed";

        public static bool? ImportedJEDRegistry = null;

        public unsafe static string ReadString(string name)
        {
            const string nonexist = "NONEXISTANT";
            string s = ConfigINI.GetKeyString("NED", name, nonexist);
            if (s != nonexist)
            {
                if (!ImportedJEDRegistry.HasValue)
                    ImportedJEDRegistry = false;

                return s;
            }

            s = string.Empty;

            // try to get from JED registry?
            RegistryKey k = Registry.CurrentUser.OpenSubKey(JEDRegBase);
            if (k == null)
                return null;

            object o = k.GetValue(name);
            if (o != null)
            {
                if (!ImportedJEDRegistry.HasValue)
                    ImportedJEDRegistry = true;

                bool alreadyGot = false;

                RegistryValueKind t = k.GetValueKind(name);
                if (t == RegistryValueKind.Binary)
                {
                    byte[] buf = o as byte[];
                    if (buf.Length == 8)
                    {
                        fixed (byte* pbuf = buf)
                        {
                            s = (*((double*)pbuf)).ToString();
                            alreadyGot = true;
                        }
                    }
                }

                if(!alreadyGot)
                    s = o.ToString();
            }

            k.Close();

            return s;
        }

        public static int? ReadInt(string name)
        {
            string s = ReadString(name);
            if (string.IsNullOrEmpty(s))
                return null;

            return int.Parse(s);
        }

        public static uint? ReadUInt(string name)
        {
            string s = ReadString(name);
            if (string.IsNullOrEmpty(s))
                return null;

            return uint.Parse(s);
        }

        public static bool? ReadBool(string name)
        {
            int? i = ReadInt(name);
            if (!i.HasValue)
                return null;

            return i.Value != 0;
        }

        public static Color4? ReadColor(string name)
        {
            uint? i = ReadUInt(name);
            if (!i.HasValue)
                return null;

            Color4 clr = new Color4();
            clr.a = 1.0;// (double)((i.Value & 0xFF000000) >> 24) / 255.0;
            clr.b = (double)((i.Value & 0x00FF0000) >> 16) / 255.0;
            clr.g = (double)((i.Value & 0x0000FF00) >> 8) / 255.0;
            clr.r = (double)(i.Value & 0x000000FF) / 255.0;

            return clr;
        }

        public static double? ReadDouble(string name)
        {
            string s = ReadString(name);
            if (string.IsNullOrEmpty(s))
                return null;

            return double.Parse(s);
        }

        public static void WriteString(string name, string val)
        {
            ConfigINI.WriteKey("NED", name, val);
        }

        public static void WriteInt(string name, int i)
        {
            WriteString(name, i.ToString());
        }

        public static void WriteUInt(string name, uint i)
        {
            WriteString(name, i.ToString());
        }

        public static void WriteDouble(string name, double d)
        {
            WriteString(name, d.ToString());
        }

        public static void WriteBool(string name, bool b)
        {
            WriteInt(name, b ? 1 : 0);
        }

        public static void WriteColor(string name, Color4 clr)
        {
            uint i = 0;

            //i |= (uint)(clr.a * 255.0) << 24;
            i |= (uint)(clr.b * 255.0) << 16;
            i |= (uint)(clr.g * 255.0) << 8;
            i |= (uint)(clr.r * 255.0);

            WriteUInt(name, i);
        }

        public static bool LoadSettings()
        {
            //if (!Program.DemoMode)
            //{
                GlobalVars.JKDir = ReadString("JKDir") ?? string.Empty;
                GlobalVars.MOTSDir = ReadString("MOTSDir") ?? string.Empty;
            GlobalVars.IJIFDir = ReadString("IJIFDir") ?? string.Empty;
            GlobalVars.DWDir = ReadString("DWDir") ?? string.Empty;
            //}
            //else
            //{
            //    GlobalVars.JKDir = Program.ExecutableDirectory;
            //    GlobalVars.MOTSDir = Program.ExecutableDirectory;
            //}

            // refresh demomode status to begin with;  this helps our logic determine if going from demo to full version even if done at startup
            if (!File.Exists(Path.Combine(Path.Combine(GlobalVars.JKDir, "episode"), "jk1.gob")))
                Program.DemoMode = true;

            /*GlobalVars.EditPos.X = ReadInt("IEdit.X") ?? GlobalVars.EditPos.X;
            GlobalVars.EditPos.Y = ReadInt("IEdit.Y") ?? GlobalVars.EditPos.Y;
            GlobalVars.EditPos.Width = ReadInt("IEdit.W") ?? GlobalVars.EditPos.Width;
            GlobalVars.EditPos.Height = ReadInt("IEdit.H") ?? GlobalVars.EditPos.Height;*/
            GlobalVars.MWinPos.X = ReadInt("MWin.X") ?? GlobalVars.MWinPos.X;
            GlobalVars.MWinPos.Y = ReadInt("MWin.Y") ?? GlobalVars.MWinPos.Y;
            GlobalVars.MWinPos.Width = ReadInt("MWin.W") ?? GlobalVars.MWinPos.Width;
            GlobalVars.MWinPos.Height = ReadInt("MWin.H") ?? GlobalVars.MWinPos.Height;
            GlobalVars.MWMaxed = ReadBool("MWMaxed") ?? GlobalVars.MWMaxed;
            /*GlobalVars.TBarPos.X = ReadInt("TBar.X") ?? GlobalVars.TBarPos.X;
            GlobalVars.TBarPos.Y = ReadInt("TBar.Y") ?? GlobalVars.TBarPos.Y;
            GlobalVars.TBarPos.Width = ReadInt("TBar.W") ?? GlobalVars.TBarPos.Width;
            GlobalVars.TBarPos.Height = ReadInt("TBar.H") ?? GlobalVars.TBarPos.Height;
            GlobalVars.TbOnTop = ReadBool("TbOnTop") ?? GlobalVars.TbOnTop;
            GlobalVars.IEOnTop = ReadBool("IEOnTop") ?? GlobalVars.IEOnTop;*/

            GlobalVars.P3DOnTop = ReadBool("P3DOnTop") ?? GlobalVars.P3DOnTop;
            //GlobalVars.P3DFullLit = ReadBool("P3DFullLit") ?? GlobalVars.P3DFullLit;  // doesnt look like JED had this (probably should though)
            GlobalVars.P3DColoredLights = ReadBool("P3DColored") ?? GlobalVars.P3DColoredLights;
            GlobalVars.P3DGamma = ReadDouble("P3Gamma") ?? GlobalVars.P3DGamma;
            GlobalVars.P3DVisLayers = ReadBool("P3DVisLayers") ?? GlobalVars.P3DVisLayers;
            GlobalVars.P3DThings = ReadBool("P3DThings") ?? GlobalVars.P3DThings;

            GlobalVars.Recent1 = ReadString("Recent1") ?? GlobalVars.Recent1;
            GlobalVars.Recent2 = ReadString("Recent2") ?? GlobalVars.Recent2;
            GlobalVars.Recent3 = ReadString("Recent3") ?? GlobalVars.Recent3;
            GlobalVars.Recent4 = ReadString("Recent4") ?? GlobalVars.Recent4;

            GlobalVars.clMapBack = ReadColor("clMapBack") ?? GlobalVars.clMapBack;
            GlobalVars.clMapGeo = ReadColor("clMapGeo") ?? GlobalVars.clMapGeo;
            GlobalVars.clMapGeoBack = ReadColor("clMapGeoBack") ?? GlobalVars.clMapGeoBack;
            GlobalVars.clMapSel = ReadColor("clMapSel") ?? GlobalVars.clMapSel;
            GlobalVars.clGrid = ReadColor("clGrid") ?? GlobalVars.clGrid;
            GlobalVars.clVertex = ReadColor("clVertex") ?? GlobalVars.clVertex;
            GlobalVars.clMapSelBack = ReadColor("clMapSelBack") ?? GlobalVars.clMapSelBack;
            GlobalVars.clThing = ReadColor("clThing") ?? GlobalVars.clThing;
            GlobalVars.clFrame = ReadColor("clFrame") ?? GlobalVars.clFrame;
            GlobalVars.clLight = ReadColor("clLight") ?? GlobalVars.clLight;
            GlobalVars.clMSel = ReadColor("clMsel") ?? GlobalVars.clMSel;
            GlobalVars.clMSelBack = ReadColor("clMselBack") ?? GlobalVars.clMSelBack;
            GlobalVars.clSelMsel = ReadColor("clSelMsel") ?? GlobalVars.clSelMsel;
            GlobalVars.clSelMselBack = ReadColor("clSelMselBack") ?? GlobalVars.clSelMselBack;
            GlobalVars.clGridX = ReadColor("clGridX") ?? GlobalVars.clGridX;
            GlobalVars.clGridY = ReadColor("clGridY") ?? GlobalVars.clGridY;
            GlobalVars.clExtra = ReadColor("clExtra") ?? GlobalVars.clExtra;
            GlobalVars.ThingView = (NedMain.ThingView)(ReadInt("ThingView") ?? (int)GlobalVars.ThingView);

            GlobalVars.MapRot = (GlobalVars.MapRotation)(ReadInt("MapRot") ?? (int)GlobalVars.MapRot);

            GlobalVars.DefTXStep = ReadDouble("DefTxStep") ?? GlobalVars.DefTXStep;
            GlobalVars.DefTXRotStep = ReadDouble("DefTXRotStep") ?? GlobalVars.DefTXRotStep;
            GlobalVars.DefTXScaleStep = ReadDouble("DefTXScaleStep") ?? GlobalVars.DefTXScaleStep;
            GlobalVars.DefPerpStep = ReadDouble("DefPerpStep") ?? GlobalVars.DefPerpStep;
            GlobalVars.DefP3DStep = ReadDouble("DefP3DStep") ?? GlobalVars.DefP3DStep;

            GlobalVars.DefGridStep = ReadDouble("DefGridStep") ?? GlobalVars.DefGridStep;
            GlobalVars.DefGridLine = ReadDouble("DefGridLine") ?? GlobalVars.DefGridLine;
            GlobalVars.DefGridDot = ReadDouble("DefGridDot") ?? GlobalVars.DefGridDot;
            GlobalVars.DefGridSize = ReadDouble("DefGridSize") ?? GlobalVars.DefGridSize;

            GlobalVars.SmithMini = ReadBool("SmithMini") ?? GlobalVars.SmithMini;
            GlobalVars.SmithNeo = ReadBool("SmithNeo") ?? GlobalVars.SmithNeo;
            GlobalVars.SmithSyncCameras = ReadBool("SmithSyncCameras") ?? GlobalVars.SmithSyncCameras;
            GlobalVars.SmithNoPrecache = ReadBool("SmithNoPrecache") ?? GlobalVars.SmithNoPrecache;

            GlobalVars.SaveInterval = ReadInt("SaveInterval") ?? GlobalVars.SaveInterval;
            GlobalVars.AutoSave = ReadBool("AutoSave") ?? GlobalVars.AutoSave;
            GlobalVars.NewOnFloor = ReadBool("NewOnFloor") ?? GlobalVars.NewOnFloor;

            GlobalVars.UndoEnabled = ReadBool("UndoEnabled") ?? GlobalVars.UndoEnabled;
            GlobalVars.MoveFrames = ReadBool("MoveFrames") ?? GlobalVars.MoveFrames;
            GlobalVars.GOBSmart = ReadBool("GOBSmart") ?? GlobalVars.GOBSmart;
            GlobalVars.CheckOverlaps = ReadBool("CheckOverlaps") ?? GlobalVars.CheckOverlaps;
            GlobalVars.NewLightCalc = ReadBool("NewLightCalc") ?? GlobalVars.NewLightCalc;
            GlobalVars.ConfirmRevert = ReadBool("ConfirmRevert") ?? GlobalVars.ConfirmRevert;
            GlobalVars.CalculateLightsAuto = ReadBool("CalculateLightsAuto") ?? GlobalVars.CalculateLightsAuto;

            //if (Program.DemoMode)
            //    goto goforit;

            if (ImportedJEDRegistry.HasValue && ImportedJEDRegistry.Value)
            {

                MessageBox.Show("Successfully imported existing JED settings from registry!\n\nFrom this point on, Smith/NED's settings will be stored in config.ini.\nYour existing JED registry will not be changed.", "JED settings import", MessageBoxButtons.OK, MessageBoxIcon.Information);
                SaveSettings(false);
                ImportedJEDRegistry = null;
            }



        doituphomie:


            // if only one dir is good, try it for the other
            if (CheckJKDir(GlobalVars.JKDir) != CheckMOTSDir(GlobalVars.MOTSDir))
            {
                if (CheckJKDir(GlobalVars.JKDir))
                    GlobalVars.MOTSDir = GlobalVars.JKDir;
                else
                    GlobalVars.JKDir = GlobalVars.MOTSDir;
            }

            // now to try to resolve any other possible sources
            if (!CheckJKDir(GlobalVars.JKDir) || !CheckMOTSDir(GlobalVars.MOTSDir))
            {
                // try to resolve JK first and foremost
                List<string> possibleJKDirs = new List<string>();
                possibleJKDirs.Add(Program.ExecutableDirectory);
                //possibleJKDirs.Add(Program.GetLocalMachineReg(@"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 32380", "InstallLocation", string.Empty) as string);
                foreach (string possibleDir in possibleJKDirs)
                {
                    if (CheckJKDir(possibleDir))
                    {
                        GlobalVars.JKDir = possibleDir;
                        break;
                    }
                }

                List<string> possibleMOTSDirs = new List<string>();
                possibleMOTSDirs.Add(Program.ExecutableDirectory);
                //possibleMOTSDirs.Add(Program.GetLocalMachineReg(@"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 32380", "InstallLocation", string.Empty) as string);
                possibleMOTSDirs.Add(GlobalVars.JKDir);// maybe we picked it up somehow before
                foreach (string possibleDir in possibleMOTSDirs)
                {
                    if (CheckMOTSDir(possibleDir))
                    {
                        GlobalVars.MOTSDir = possibleDir;
                        break;
                    }
                }
            }

            // if we couldnt even find JK then we're losers
            if (!CheckJKDir(GlobalVars.JKDir))
            {
                /*byte[] buf;
                using (Stream s = System.Reflection.Assembly.GetExecutingAssembly().GetManifestResourceStream("Smith.toiletscene.mp3"))
                {
                    buf = new byte[s.Length];
                    s.Read(buf, 0, buf.Length);
                }

                Audio.Sound snd = Audio.GetSound("toiletscene.mp3", buf, Audio.DimensionMode._2D, false);
                if (snd != null)
                {
                    Audio.Channel dump = Audio.PlaySound(snd, true);

                    FMOD.DSP reverb;
                    Audio.fmod.createDSPByType(FMOD.DSP_TYPE.SFXREVERB, out reverb);

                    FMOD.DSP fart;
                    Audio.fmod.createDSPByType(FMOD.DSP_TYPE.CHORUS, out fart);

                    dump.channel.addDSP(0, fart);
                    dump.channel.addDSP(1, reverb);

                    dump.Volume = StartupSplash.musicvol;
                    dump.Play();
                }*/

                FolderBrowserDialog fbd = new FolderBrowserDialog();
                fbd.Description = "Please locate a JK installation path";
                fbd.ShowNewFolderButton = false;
                if (fbd.ShowDialog() != DialogResult.OK)
                {
                    goto goforit;
                    //return false;
                }

                GlobalVars.JKDir = fbd.SelectedPath;
                goto doituphomie;
            }


        goforit:

            bool wasDemoMode = Program.DemoMode;

            // if cant find full jk then force demo mode
            if (!File.Exists(Path.Combine(Path.Combine(GlobalVars.JKDir, "episode"), "jk1.gob")))
            {
                Program.DemoMode = true;

                GlobalVars.JKDir = Program.ExecutableDirectory;
                GlobalVars.MOTSDir = string.Empty;

                if (!File.Exists(Path.Combine(Program.ExecutableDirectory, "res1demo.gob")))
                {
                    if (!Program.InAppDownloadSupported)
                    {
                        DialogResult dr = MessageBox.Show("Normally you'd be able to download the demo files, but...\n\nUnable to set TLS 1.2 network security protocol.\nIn-app downloading may not work.", "Can't download demo", MessageBoxButtons.AbortRetryIgnore);

                        if (dr == DialogResult.Retry)
                            goto doituphomie;

                        if (dr != DialogResult.Ignore)
                            return false;
                    }

                    bool allowDownloadDemo = false;
                    if (!allowDownloadDemo)
                    {
                        DialogResult dr = MessageBox.Show("Normally you'd be able to download the demo files, but...\n\nDemo file download is disabled.\nMight be blocked so its just disabled.", "Can't download demo", MessageBoxButtons.OK);
                        return false;
                    }

                    string[] demofiles = new string[] { "jk1demo.gob", /*"jk1mpdem.gob",*/ "res1demo.gob" };

                    string s = "You didn't pick a valid JK installation path.\nWould you like to retrieve the official LEC demo gob files? (34mb)\n\n" + demofiles.Length.ToString() + " files will be downloaded";
                    foreach (string df in demofiles)
                        s += "\n" + Path.Combine(Program.ExecutableDirectory, df);
                    if (MessageBox.Show(s, "Download Demo", MessageBoxButtons.OKCancel) != DialogResult.OK)
                        return false;

                    if (StartupSplash.inst != null && StartupSplash.inst.IsHandleCreated)
                    {
                        StartupSplash.inst.Text = "Smith - Initiating Demo Mode";
                    }

                    Download("https://smith.bah.wtf/", demofiles, Program.ExecutableDirectory);
                }
                else
                {
                    //MessageBox.Show("DEMO MODE");
                }
            }
            else
                Program.DemoMode = false;



        if (!Program.DemoMode)
        {
            string[] demofiles = new string[] { "jk1demo.gob", /*"jk1mpdem.gob",*/ "res1demo.gob" };
            bool anyExist = false;
            foreach(string demoFile in demofiles)
                if (File.Exists(Path.Combine(Program.ExecutableDirectory, demoFile)))
                {
                    anyExist = true;
                    break;
                }

            if (anyExist)
            {
                Resource.Cleanup(true, true, false, "NED Options LoadSettings");

                if (wasDemoMode)
                {
                    if (MessageBox.Show("It looks like you're running full JK but still have demo files.\n\nDo you want to clean them up?", "Demo Cleanup", MessageBoxButtons.OKCancel) == DialogResult.OK)
                        foreach (string demoFile in demofiles)
                        {
                            string fullfile = Path.Combine(Program.ExecutableDirectory, demoFile);
                            if (File.Exists(fullfile))
                                File.Delete(fullfile);
                        }
                }
            }
        }


            // see if we can find the music and cutscenes

            Log.Error("Need to figure out searchingf or music/video source");
            //throw new Exception("Need to handle these cd resources");
            //Resource.AddDirectory(@"D:\Steam\steamapps\common\Jedi Knight Mysteries of the Sith\MUSIC");
            //Resource.AddDirectory(@"D:\Steam\steamapps\common\Jedi Knight Mysteries of the Sith\Resource\VIDEO");



            // whatever go for it
            return true;
        }

        public static void Download(string baseurl, string[] files, string outdir, bool forcetmpextension=false)
        {
            Form ProgressForm = new Form();
            ProgressForm.FormBorderStyle = FormBorderStyle.FixedToolWindow;
            ProgressForm.ControlBox = false;
            ProgressForm.TopMost = false;// true;
            ProgressForm.StartPosition = FormStartPosition.CenterScreen;
            ProgressForm.Size = new System.Drawing.Size(375, 35);

            //f.Text = "Precaching...";

            ProgressBar ProgressBar = new ProgressBar();
            ProgressBar.Dock = DockStyle.Fill;
            ProgressBar.Style = ProgressBarStyle.Blocks;
            ProgressForm.Controls.Add(ProgressBar);

            ProgressForm.Show(Program.MainForm);
            ProgressForm.Update();

            for(int x=0; x<files.Length; x++)
            {
                string file = files[x];

                ProgressForm.Text = "Downloading " + file + " (" + (x + 1).ToString() + "/" + files.Length.ToString() + ")";
                ProgressForm.Update();

                string tmpfile = Path.GetTempFileName();
                string dstfile = Path.Combine(outdir, file);

                if (forcetmpextension)
                {
                    int i = dstfile.LastIndexOf('.');
                    if (i != -1)
                        dstfile = dstfile.Substring(0, i) + ".tmp";
                }

                if (File.Exists(dstfile))
                    File.Delete(dstfile);

                WebClient wc = new WebClient();
                wc.DownloadProgressChanged += new DownloadProgressChangedEventHandler(wc_DownloadProgressChanged);
                wc.DownloadFileAsync(new Uri(new Uri(baseurl), file + "?t=" + Program.random.Next().ToString()), tmpfile);

                while (wc.IsBusy)
                {
                    ProgressBar.Value = (x * 100 + downloadprogress) / files.Length;
                    ProgressBar.Update();

                    Application.DoEvents();
                    Thread.Sleep(100);
                }

                File.Move(tmpfile, dstfile);

                wc.Dispose();
            }

            ProgressForm.Close();
            ProgressForm.Dispose();
            ProgressForm = null;
            ProgressBar = null;
        }

        static volatile int downloadprogress = 0;

        static void wc_DownloadProgressChanged(object sender, DownloadProgressChangedEventArgs e)
        {
            downloadprogress = e.ProgressPercentage;
        }

        public static void SaveSettings(bool isExit)
        {
            //if (!isExit)
            {
                if (!Program.DemoMode)
                {
                    WriteString("JKDir", GlobalVars.JKDir);
                    WriteString("MOTSDir", GlobalVars.MOTSDir);
                    WriteString("IJIFDir", GlobalVars.IJIFDir);
                    WriteString("DWDir", GlobalVars.DWDir);
                }

                WriteColor("clMapBack", GlobalVars.clMapBack);
                WriteColor("clMapGeo", GlobalVars.clMapGeo);
                WriteColor("clMapGeoBack", GlobalVars.clMapGeoBack);
                WriteColor("clMapSel", GlobalVars.clMapSel);
                WriteColor("clGrid", GlobalVars.clGrid);
                WriteColor("clVertex", GlobalVars.clVertex);
                WriteColor("clMapSelBack", GlobalVars.clMapSelBack);
                WriteColor("clThing", GlobalVars.clThing);
                WriteColor("clFrame", GlobalVars.clFrame);
                WriteColor("clLight", GlobalVars.clLight);
                WriteColor("clMsel", GlobalVars.clMSel);
                WriteColor("clMselBack", GlobalVars.clMSelBack);
                WriteColor("clSelMsel", GlobalVars.clSelMsel);
                WriteColor("clSelMselBack", GlobalVars.clSelMselBack);
                WriteColor("clGridX", GlobalVars.clGridX);
                WriteColor("clGridY", GlobalVars.clGridY);
                WriteColor("clExtra", GlobalVars.clExtra);
                WriteInt("DefThingView", (int)GlobalVars.ThingView);

                WriteBool("P3DVisLayers", GlobalVars.P3DVisLayers);
                WriteBool("P3DThings", GlobalVars.P3DThings);

                WriteInt("MapRot", (int)GlobalVars.MapRot);

                WriteBool("SmithNeo", GlobalVars.SmithNeo);
                WriteBool("SmithSyncCameras", GlobalVars.SmithSyncCameras);
                WriteBool("SmithMini", GlobalVars.SmithMini);
                WriteBool("SmithNoPrecache", GlobalVars.SmithNoPrecache);

                WriteInt("SaveInterval", GlobalVars.SaveInterval);
                WriteBool("AutoSave", GlobalVars.AutoSave);
                WriteBool("NewOnFloor", GlobalVars.NewOnFloor);

                WriteBool("UndoEnabled", GlobalVars.UndoEnabled);
                WriteBool("MoveFrames", GlobalVars.MoveFrames);
                WriteBool("GOBSmart", GlobalVars.GOBSmart);
                WriteBool("CheckOverlaps", GlobalVars.CheckOverlaps);
                WriteBool("NewLightCalc", GlobalVars.NewLightCalc);
                WriteBool("ConfirmRevert", GlobalVars.ConfirmRevert);
                WriteBool("CalculateLightsAuto", GlobalVars.CalculateLightsAuto);
            }
            //else
            {
                /*WriteInt("IEdit.X", GlobalVars.EditPos.X);
                WriteInt("IEdit.Y", GlobalVars.EditPos.Y);
                WriteInt("IEdit.W", GlobalVars.EditPos.Width);
                WriteInt("IEdit.H", GlobalVars.EditPos.Height);*/
                WriteInt("MWin.X", GlobalVars.MWinPos.X);
                WriteInt("MWin.Y", GlobalVars.MWinPos.Y);
                WriteInt("MWin.W", GlobalVars.MWinPos.Width);
                WriteInt("MWin.H", GlobalVars.MWinPos.Height);
                WriteBool("MWMaxed", GlobalVars.MWMaxed);
                /*WriteInt("TBar.X", GlobalVars.TBarPos.X);
                WriteInt("TBar.Y", GlobalVars.TBarPos.Y);
                WriteInt("TBar.W", GlobalVars.TBarPos.Width);
                WriteInt("TBar.H", GlobalVars.TBarPos.Height);
                WriteBool("TbOnTop", GlobalVars.TbOnTop);
                WriteBool("IEOnTop", GlobalVars.IEOnTop);*/

                WriteBool("P3DOnTop", GlobalVars.P3DOnTop);
                //WriteBool("P3DFullLit", GlobalVars.P3DFullLit);  // doesnt look like JED had this (probably should though)
                WriteBool("P3DColored", GlobalVars.P3DColoredLights);
                WriteDouble("P3Gamma", GlobalVars.P3DGamma);

                WriteString("Recent1", GlobalVars.Recent1);
                WriteString("Recent2", GlobalVars.Recent2);
                WriteString("Recent3", GlobalVars.Recent3);
                WriteString("Recent4", GlobalVars.Recent4);

                WriteDouble("DefTxStep", GlobalVars.DefTXStep);
                WriteDouble("DefTXRotStep", GlobalVars.DefTXRotStep);
                WriteDouble("DefTXScaleStep", GlobalVars.DefTXScaleStep);
                WriteDouble("DefPerpStep", GlobalVars.DefPerpStep);
                WriteDouble("DefP3DStep", GlobalVars.DefP3DStep);

                WriteDouble("DefGridStep", GlobalVars.DefGridStep);
                WriteDouble("DefGridLine", GlobalVars.DefGridLine);
                WriteDouble("DefGridDot", GlobalVars.DefGridDot);
                WriteDouble("DefGridSize", GlobalVars.DefGridSize);
            }
        }

        public static string FindGob(string name)
        {
            string s;

            s = Path.Combine(Path.Combine(GlobalVars.JKDir, "Episode"), name);
            if (File.Exists(s))
                return s;

            s = Path.Combine(Path.Combine(GlobalVars.JKDir, "Resource"), name);
            if (File.Exists(s))
                return s;

            return string.Empty;
        }

        public static string FindGoo(string name)
        {
            string s;

            s = Path.Combine(Path.Combine(GlobalVars.MOTSDir, "Episode"), name);
            if (File.Exists(s))
                return s;

            s = Path.Combine(Path.Combine(GlobalVars.MOTSDir, "Resource"), name);
            if (File.Exists(s))
                return s;

            return string.Empty;
        }

        public static bool SetMOTS(bool isit)
        {
            bool result;

            GlobalVars.IsMOTS = isit;

            if (isit)
            {
                GlobalVars.GameDir = GlobalVars.MOTSDir;
                GlobalVars.res1_gob = FindGoo("JKMsndLO.goo");
                GlobalVars.res2_gob = FindGoo("Jkmres.goo");
                GlobalVars.sp_gob = FindGoo("Jkm.goo");
                GlobalVars.mp1_gob = FindGoo("Jkm_mp.goo");
                GlobalVars.mp2_gob = FindGoo("Jkm_kfy.goo");
                GlobalVars.mp3_gob = FindGoo("JKM_SABER.GOO");

                result = !string.IsNullOrEmpty(GlobalVars.GameDir) &&
                         !string.IsNullOrEmpty(GlobalVars.res1_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.res2_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.sp_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.mp1_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.mp2_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.mp3_gob);

                GlobalVars.Level.mots = GlobalVars.IsMOTS;

                if (!result)
                    Utils.PanMessage(Utils.MessageType.Warning, "Some resources weren't found! Check MOTS settings in Options");
            }
            else
            {
                GlobalVars.GameDir = GlobalVars.JKDir;
                GlobalVars.res1_gob = FindGob("Res1hi.gob");
                if (string.IsNullOrEmpty(GlobalVars.res1_gob))
                    GlobalVars.res1_gob = FindGob("Res1low.gob");
                GlobalVars.res2_gob = FindGob("Res2.gob");
                GlobalVars.sp_gob = FindGob("Jk1.gob");
                GlobalVars.mp1_gob = FindGob("Jk1mp.gob");
                GlobalVars.mp2_gob = FindGob("Jk1ctf.gob");
                GlobalVars.mp3_gob = string.Empty;

                if (string.IsNullOrEmpty(GlobalVars.res1_gob) || !File.Exists(GlobalVars.res1_gob))
                    GlobalVars.res1_gob = Path.Combine(Program.ExecutableDirectory, "res1demo.gob");

                if (string.IsNullOrEmpty(GlobalVars.res2_gob) || !File.Exists(GlobalVars.res2_gob))
                    GlobalVars.res2_gob = Path.Combine(Program.ExecutableDirectory, "res1demo.gob");

                result = !string.IsNullOrEmpty(GlobalVars.GameDir) &&
                         !string.IsNullOrEmpty(GlobalVars.res1_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.res2_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.sp_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.mp1_gob) &&
                         !string.IsNullOrEmpty(GlobalVars.mp2_gob);

                GlobalVars.Level.mots = GlobalVars.IsMOTS;

                if (!result)
                    Utils.PanMessage(Utils.MessageType.Warning, "Some resources weren't found! Check JK settings in Options");
            }

            GlobalVars.NedMain.LoadGOBResources();

            return result;
        }

        public Options()
        {
            InitializeComponent();
        }

        private void jkDirectory_TextChanged(object sender, EventArgs e)
        {
            if (Program.DemoMode)
                jkDirectoryValid.BackColor = Color.Orange;
            else
                jkDirectoryValid.BackColor = CheckJKDir(jkDirectory.Text) ? Color.SpringGreen : Color.Red;
        }

        private void motsDirectory_TextChanged(object sender, EventArgs e)
        {
            motsDirectoryValid.BackColor = CheckMOTSDir(motsDirectory.Text) ? Color.SpringGreen : Color.Red;
        }

        public static bool CheckJKDir(string dir)
        {
            if (string.IsNullOrEmpty(dir))
                return false;

            return File.Exists(Path.Combine(Path.Combine(dir, "Resource"), "res2.gob"));
        }

        public static bool CheckMOTSDir(string dir)
        {
            if (string.IsNullOrEmpty(dir))
                return false;

            return File.Exists(Path.Combine(Path.Combine(dir, "Resource"), "Jkmres.goo"));
        }

        public static bool CheckIJIFDir(string dir)
        {
            if (string.IsNullOrEmpty(dir))
                return false;

            return File.Exists(Path.Combine(Path.Combine(dir, "Resource"), "Jones3D.GOB"));
        }

        public static bool CheckDWDir(string dir)
        {
            if (string.IsNullOrEmpty(dir))
                return false;

            return File.Exists(Path.Combine(dir, "dwMin.gob"));
        }

        private void jkDirectoryBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.ShowNewFolderButton = false;
            fbd.SelectedPath = jkDirectory.Text;
            if (fbd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            jkDirectory.Text = fbd.SelectedPath;

            // see if MOTS is there too if we currently have an invalid mots path
            if (!CheckMOTSDir(motsDirectory.Text) && CheckMOTSDir(jkDirectory.Text))
                motsDirectory.Text = jkDirectory.Text;
        }

        private void motsDirectoryBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.ShowNewFolderButton = false;
            fbd.SelectedPath = motsDirectory.Text;
            if (fbd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            motsDirectory.Text = fbd.SelectedPath;

            // see if JK is there too if we currently have an invalid jk path
            if (!CheckJKDir(jkDirectory.Text) && CheckJKDir(motsDirectory.Text))
                jkDirectory.Text = motsDirectory.Text;
        }

        private void Options_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (DialogResult == System.Windows.Forms.DialogResult.OK)
            {
                ApplySettings();
                SaveSettings(false);
            }
        }

        static Color4 GetColor(int i)
        {
            switch (i)
            {
                case 0: return GlobalVars.clMapBack;
                case 1: return GlobalVars.clMapGeo;
                case 2: return GlobalVars.clMapGeoBack;
                case 3: return GlobalVars.clMapSel;
                case 4: return GlobalVars.clMapSelBack;
                case 5: return GlobalVars.clGrid;
                case 6: return GlobalVars.clVertex;
                case 7: return GlobalVars.clThing;
                case 8: return GlobalVars.clFrame;
                case 9: return GlobalVars.clLight;
                case 10: return GlobalVars.clMSel;
                case 11: return GlobalVars.clMSelBack;
                case 12: return GlobalVars.clSelMsel;
                case 13: return GlobalVars.clSelMselBack;
                case 14: return GlobalVars.clGridX;
                case 15: return GlobalVars.clGridY;
                case 16: return GlobalVars.clExtra;
                default: return Color4.black();
            }
        }

        static void SetColor(int i, Color4 clr)
        {
            switch (i)
            {
                case 0: GlobalVars.clMapBack = clr; break;
                case 1: GlobalVars.clMapGeo = clr; break;
                case 2: GlobalVars.clMapGeoBack = clr; break;
                case 3: GlobalVars.clMapSel = clr; break;
                case 4: GlobalVars.clMapSelBack = clr; break;
                case 5: GlobalVars.clGrid = clr; break;
                case 6: GlobalVars.clVertex = clr; break;
                case 7: GlobalVars.clThing = clr; break;
                case 8: GlobalVars.clFrame = clr; break;
                case 9: GlobalVars.clLight = clr; break;
                case 10: GlobalVars.clMSel = clr; break;
                case 11: GlobalVars.clMSelBack = clr; break;
                case 12: GlobalVars.clSelMsel = clr; break;
                case 13: GlobalVars.clSelMselBack = clr; break;
                case 14: GlobalVars.clGridX = clr; break;
                case 15: GlobalVars.clGridY = clr; break;
                case 16: GlobalVars.clExtra = clr; break;
            }
        }

        public class ColorEntry
        {
            public string Desc;
            public int ClrIndex;
            public Color4 Color4;

            public ColorEntry(string _Desc, int _ClrIndex)
            {
                Desc = _Desc;
                ClrIndex = _ClrIndex;
                Color4 = GetColor(ClrIndex);
            }

            public Color Color
            {
                get
                {
                    return Color.FromArgb((int)(Color4.r * 255.0), (int)(Color4.g * 255.0), (int)(Color4.b * 255.0));
                }

                set
                {
                    Color4.r = (double)value.R / 255.0;
                    Color4.g = (double)value.G / 255.0;
                    Color4.b = (double)value.B / 255.0;
                }
            }

            public override string ToString()
            {
                return Desc;
            }
        }

        private void Options_Load(object sender, EventArgs e)
        {
            ReceiveSettings();
        }

        private void LBColors_DrawItem(object sender, DrawItemEventArgs e)
        {
            e.DrawBackground();

            if (e.Index >= 0 || e.Index < LBColors.Items.Count)
            {
                ColorEntry ce = LBColors.Items[e.Index] as ColorEntry;

                StringFormat sf = new StringFormat();
                sf.LineAlignment = StringAlignment.Center;
                sf.Alignment = StringAlignment.Near;
                e.Graphics.DrawString(ce.Desc, e.Font, new SolidBrush(e.ForeColor), new Box2(e.Bounds), sf);

                Box2 clrrc = new Box2(e.Bounds);
                clrrc.Left = clrrc.Right - clrrc.Height * 1.5;
                e.Graphics.FillRectangle(new SolidBrush(ce.Color), clrrc);
            }

            e.DrawFocusRectangle();
        }

        private void LBColors_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            int i = LBColors.IndexFromPoint(e.Location);
            if (i < 0 || i >= LBColors.Items.Count)
                return;

            ColorEntry ce = LBColors.Items[i] as ColorEntry;

            ColorDialog cd = new ColorDialog();
            cd.Color = ce.Color;
            cd.SolidColorOnly = false;
            cd.FullOpen = true;
            int cclr = 0;
            cclr |= (int)ce.Color.R;
            cclr |= (int)ce.Color.G << 8;
            cclr |= (int)ce.Color.B << 16;
            cd.CustomColors = new int[] { cclr };
            if (cd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            ce.Color = cd.Color;

            LBColors.Refresh();
        }

        private void extractJEDData_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(GlobalVars.BaseDir) || !Directory.Exists(GlobalVars.BaseDir))
            {
                MessageBox.Show("Invalid base dir");
                return;
            }

            string jdpath = Path.Combine(GlobalVars.BaseDir, "Jeddata");
            if (!Directory.Exists(jdpath))
                Directory.CreateDirectory(jdpath);

            string[] files = new string[] {
                "master.tpl",
                "mats.jlh",
                "mats.jll",
                "mots.tpl",
                "mtpls.jlh",
                "mtpls.jll",
                "shapes.tpl",
                "tpls.jlh",
                "tpls.jll",
                "materials.db"
            };

            bool hasany = false;
            foreach(string file in files)
                if (File.Exists(Path.Combine(jdpath, file)))
                {
                    hasany = true;
                    break;
                }

            if (hasany)
            {
                if (MessageBox.Show("Jeddata folder already exists and contains items.\nAre you sure you want to replace with defaults?", "Extract Jeddata", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                    return;
            }

            foreach (string file in files)
            {
                using (Stream src = Jeddata.GetEmbeddedFile(file))
                using (FileStream dst = File.Create(Path.Combine(jdpath, file)))
                {
                    byte[] data = new byte[src.Length];
                    src.Read(data, 0, (int)src.Length);
                    dst.Write(data, 0, (int)src.Length);
                }
            }

            System.Diagnostics.Process.Start("explorer.exe", "\"" + jdpath + "\"");
        }

        private void SmithOptChange(object sender, EventArgs e)
        {
            UpdateSmithCmdLine();
        }

        private void ijifDirectoryBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.ShowNewFolderButton = false;
            fbd.SelectedPath = ijifDirectory.Text;
            if (fbd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            ijifDirectory.Text = fbd.SelectedPath;

            // see if JK is there too if we currently have an invalid jk path
            //if (!CheckJKDir(jkDirectory.Text) && CheckJKDir(motsDirectory.Text))
              //  jkDirectory.Text = motsDirectory.Text;
        }

        private void ijifDirectory_TextChanged(object sender, EventArgs e)
        {
            ijifDirectoryValid.BackColor = CheckIJIFDir(ijifDirectory.Text) ? Color.SpringGreen : Color.Red;
        }

        private void dwDirectory_TextChanged(object sender, EventArgs e)
        {
            dwDirectoryValid.BackColor = CheckDWDir(dwDirectory.Text) ? Color.SpringGreen : Color.Red;
        }

        private void dwDirectoryBrowse_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog fbd = new FolderBrowserDialog();
            fbd.ShowNewFolderButton = false;
            fbd.SelectedPath = dwDirectory.Text;
            if (fbd.ShowDialog(this) != System.Windows.Forms.DialogResult.OK)
                return;

            dwDirectory.Text = fbd.SelectedPath;

            // see if JK is there too if we currently have an invalid jk path
            //if (!CheckJKDir(jkDirectory.Text) && CheckJKDir(motsDirectory.Text))
            //  jkDirectory.Text = motsDirectory.Text;
        }
    }
}
