﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Smith.NED
{
    public class Prefab
    {
        public string Name;
        public JKSector Sector;

        public override string ToString()
        {
            return Name;
        }

        public override int GetHashCode()
        {
            return Name.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            return Name.Equals(obj.ToString()/*should work regardless if an actual string or a prefab*/, StringComparison.InvariantCultureIgnoreCase);
        }
    }

    public static class Prefabs
    {
        private static List<Prefab> _list = null;
        public static List<Prefab> list
        {
            get
            {
                if (_list == null)
                {
                    _list = new List<Prefab>();

                    JKSector sec = GlobalVars.Level.NewSector();
                    Utils.CreateCubeSec(sec, Vec3.Zero, Vec3.PosZ, Vec3.PosX);

                    Prefab cube = new Prefab();
                    cube.Name = "Cube";
                    cube.Sector = sec;
                    _list.Add(cube);
                }

                return _list;
            }
        }

        public static void AddPrefab(string pname, JKSector s)
        {
            Vec3 c = Vec3.Infinite;

            foreach (JKVertex v in s.vertices)
            {
                c.x = Math.Min(c.x, v.v.x);
                c.y = Math.Min(c.y, v.v.y);
                c.z = Math.Min(c.z, v.v.z);
            }

            if (c.x == MathLib.Infinity || c.y == MathLib.Infinity || c.z == MathLib.Infinity)
                return;

            JKSector newsec = GlobalVars.Level.NewSector();
            Utils.DuplicateSector(s, newsec, c, Vec3.PosX, Vec3.PosY, Vec3.PosZ, -c);

            Prefab p = new Prefab();
            p.Name = pname;
            p.Sector = newsec;
            list.Add(p);
        }

        public static void DeletePrefab(int n)
        {
            if (n == 0)
                return;

            list.RemoveAt(n);
        }

        public static void SavePrefabs(string fname)
        {
            FileOperations.BackUpFile(fname);
            using (StreamWriter sw = File.CreateText(fname))
                for (int i = 1; i < list.Count; i++)
                {
                    JKSector sec = list[i].Sector;

                    sw.WriteLine("SEC " + list[i].Name);

                    string st;
                    if (string.IsNullOrEmpty(sec.colormap))
                        st = "-";
                    else
                        st = sec.colormap;

                    if(string.IsNullOrEmpty(sec.sound))
                        st += " -";
                    else
                        st += " " + sec.sound;

                    sw.WriteLine(st + " " + sec.snd_vol.ToString("0.######") + " " + ((uint)sec.flags).ToString("x") + " " + sec.ambient.ToString("0.######") + " " +
                        sec.extra.ToString("0.######") + " " + sec.tint.r.ToString("0.######") + " " + sec.tint.g.ToString("0.######") + " " + sec.tint.b.ToString("0.######") + " "+
                        sec.layer.ToString());

                    sw.WriteLine("VXS " + sec.vertices.Count.ToString());
                    foreach(JKVertex v in sec.vertices)
                        sw.WriteLine(v.v.x.ToString("0.######") + " " + v.v.y.ToString("0.######") + " " + v.v.z.ToString("0.######"));

                    sw.WriteLine("SURFS " + sec.surfaces.Count.ToString());
                    foreach(JKSurface surf in sec.surfaces)
                    {
                        int asf;
                        int asec;
                        if(surf.adjoin != null)
                        {
                            asf = surf.adjoin.num;
                            asec = surf.adjoin.sector.num;
                        } else
                        {
                            asf = -1;
                            asec = -1;
                        }

                        st = surf.material;
                        if(string.IsNullOrEmpty(st))
                            st = "-";

                        sw.WriteLine(st + " " + ((uint)surf.surfaceFlags).ToString("x") + " " + ((uint)surf.faceFlags).ToString("x") + " " + ((int)surf.geo).ToString() + " " +
                            ((int)surf.light).ToString() + " " + ((int)surf.tex).ToString() + " " + surf.extraLight.ToString("0.######") + " " +
                            surf.vertices.Count.ToString() + " " + asec.ToString() + " " + asf.ToString() + " " + ((uint)surf.adjoinFlags).ToString("x") + " " +
                            surf.texScale.x.ToString("0.####") + " " + surf.texScale.y.ToString("0.####"));

                        st = string.Empty;
                        foreach(TexVertex tv in surf.texvertices)
                            st += " " + tv.num.ToString() + " " + tv.v.x.ToString("0.##") + " " + tv.v.y.ToString("0.##") + " " + tv.intensity.ToString("0.####");
                        foreach(TexVertex tv in surf.texvertices)
                            st += " " + tv.r.ToString("0.####") + " " + tv.g.ToString("0.####") + " " + tv.b.ToString("0.####");
                        sw.WriteLine(st);
                    }

                    sw.WriteLine("END");
                    sw.WriteLine();
                }
        }

        public static JKSector GetPrefab(int n)
        {
            return list[n].Sector;
        }

        private static JKSector LoadSector(Tokenizer t)
        {
            JKSector sec = GlobalVars.Level.NewSector();

            sec.colormap = t.ReadToken();
            sec.sound = t.ReadToken();
            sec.snd_vol = t.ReadDouble();
            sec.flags = (Sector.Flag)t.ReadFlags();
            sec.ambient = t.ReadDouble();
            sec.extra = t.ReadDouble();
            sec.tint.r = t.ReadDouble();
            sec.tint.g = t.ReadDouble();
            sec.tint.b = t.ReadDouble();
            sec.layer = t.ReadInt();

            if (sec.colormap == "-")
                sec.colormap = string.Empty;
            if (sec.sound == "-")
                sec.sound = string.Empty;

            t.EatToken();
            int numverts = t.ReadInt();
            for (int x = 0; x < numverts; x++)
            {
                JKVertex v = sec.NewVertex();
                v.v.x = t.ReadDouble();
                v.v.y = t.ReadDouble();
                v.v.z = t.ReadDouble();
            }

            t.EatToken();
            int numsurfs = t.ReadInt();
            for (int x = 0; x < numsurfs; x++)
            {
                JKSurface surf = sec.NewSurface();
                sec.surfaces.Add(surf);

                surf.material = t.ReadToken();
                surf.surfaceFlags = (Surface.Flag)t.ReadFlags();
                surf.faceFlags = (Surface.FaceFlag)t.ReadFlags();
                surf.SetGeo("LoadSector", (Surface.GeoModeType)t.ReadInt());
                surf.light = (Surface.LightModeType)t.ReadInt();
                surf.tex = (Surface.TextureModeType)t.ReadInt();
                surf.extraLight = t.ReadDouble();
                int nsv = t.ReadInt();
                int asec = t.ReadInt();
                int asf = t.ReadInt();
                bool endOfLine;
                surf.adjoinFlags = (Adjoin.Flag)t.ReadFlags(out endOfLine);
                if (!endOfLine)
                {
                    surf.texScale.x = t.ReadDouble();
                    surf.texScale.y = t.ReadDouble();
                }

                if (surf.material == "-")
                    surf.material = string.Empty;

                endOfLine = true;
                for (int vx = 0; vx < nsv; vx++)
                {
                    surf.AddVertex(sec.vertices[t.ReadInt()]);

                    TexVertex tv = surf.texvertices[vx];
                    tv.v.x = t.ReadDouble();
                    tv.v.y = t.ReadDouble();
                    tv.intensity = t.ReadDouble(out endOfLine);
                }

                if (!endOfLine)
                    for (int vx = 0; vx < nsv; vx++)
                    {
                        TexVertex tv = surf.texvertices[vx];
                        tv.r = t.ReadDouble();
                        tv.g = t.ReadDouble();
                        tv.b = t.ReadDouble();
                    }
            }

            sec.Renumber();

            return sec;
        }

        public static void LoadPrefabs(string fname)
        {
            // remove everything except the cube
            while (list.Count > 1)
                list.RemoveAt(1);

            Stream stream = Jeddata.GetFile(fname);
            if (stream == null)
                return;

            Tokenizer t = new Tokenizer(stream);

            while (!t.EndOfStream)
            {
                string ln = t.ReadLine();
                if (string.IsNullOrEmpty(ln))
                    break;

                if (!ln.StartsWith("SEC", StringComparison.InvariantCultureIgnoreCase))
                    continue;

                string name = ln.Substring(3).Trim();

                JKSector sec = LoadSector(t);
                if(sec == null)
                    continue;

                Prefab prefab = new Prefab();
                prefab.Name = name;
                prefab.Sector = sec;
                list.Add(prefab);
            }
        }
    }
}
