﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Smith.NED
{
    public partial class Preview3D : UserControl
    {
        public Preview3D()
        {
            InitializeComponent();
        }

        public PreviewRenderer renderer
        {
            get
            {
                return previewRenderControl1.renderer;
            }
        }

        public JKLevel Level
        {
            get
            {
                return GlobalVars.Level;
            }
        }

        public MapSettings Toolbar
        {
            get
            {
                return GlobalVars.NedMain.Toolbar;
            }
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            DoKeyDown(e);
        }

        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            DoMouseDoubleClick(e);
        }

        protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            DoPreviewKeyDown(e);
        }

        public void DoPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            e.IsInputKey = true;
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            DoMouseUp(e);
        }

        public void DoMouseDoubleClick(MouseEventArgs e)
        {
            switch (GlobalVars.NedMain.map_mode)
            {
                case NedMain.MapMode.Surface:
                    GlobalVars.ItemEdit.DoPickTexture();
                    break;

                case NedMain.MapMode.Thing:
                    GlobalVars.ItemEdit.DoPickThing();
                    break;
            }
        }

        public void DoMouseUp(MouseEventArgs e)
        {
            // dont use the location passed in since we have this whole mess between NedMain form client rectangle VS rendercontrol client rectangle
            Vec2 pt = new Vec2(previewRenderControl1.PointToClient(Cursor.Position));

            if (renderer == null)
                return;

            switch (renderer.PickAt(pt))
            {
                case PreviewRenderer.PickType.Nothing:
                    return;

                case PreviewRenderer.PickType.Surface:
                    {
                        int i = Level.sectors.IndexOf(renderer.SelSC);
                        if (i != -1)
                        {
                            GlobalVars.NedMain.SetMapMode(NedMain.MapMode.Surface);
                            GlobalVars.NedMain.SetCurSF(i, renderer.SelSF);

                            if ((ModifierKeys & Keys.Shift) != 0)
                                GlobalVars.NedMain.Do_MultiSelect();
                            else
                                GlobalVars.NedMain.ClearMultiSelection();
                        }
                        else
                        {
                            Utils.PanMessage(Utils.MessageType.Warning, "3D preview is out of sync with the level! Reload it");
                        }
                    }
                    break;

                case PreviewRenderer.PickType.Thing:
                    {
                        int i = Level.things.IndexOf(renderer.SelTH);
                        if (i != -1)
                        {
                            GlobalVars.NedMain.SetMapMode(NedMain.MapMode.Thing);
                            GlobalVars.NedMain.SetCurTH(i);

                            if ((ModifierKeys & Keys.Shift) != 0)
                                GlobalVars.NedMain.Do_MultiSelect();
                            // should we be clearing multiselect just like surface here or no...???
                        }
                        else
                        {
                            Utils.PanMessage(Utils.MessageType.Warning, "3D preview is out of sync with the level! Reload it");
                        }
                    }
                    break;
            }
        }

        public void DoKeyDown(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Left:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Left);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RotateTexture(NedMain.ShiftTextureType.Left);
                    else if ((e.Modifiers & Keys.Alt) != 0)
                    {
                        renderer.CamPos.x -= Math.Cos(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                        renderer.CamPos.y += Math.Sin(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                    }
                    else
                    {
                        renderer.CamOrient.yaw -= 10.0;
                        if (renderer.CamOrient.yaw < 0.0)
                            renderer.CamOrient.yaw += 360.0;
                    }
                    break;

                case Keys.Right:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Right);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RotateTexture(NedMain.ShiftTextureType.Right);
                    else if ((e.Modifiers & Keys.Alt) != 0)
                    {
                        renderer.CamPos.x += Math.Cos(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                        renderer.CamPos.y -= Math.Sin(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                    }
                    else
                    {
                        renderer.CamOrient.yaw += 10.0;
                        if (renderer.CamOrient.yaw > 360.0)
                            renderer.CamOrient.yaw -= 360.0;
                    }
                    break;

                case Keys.Up:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Up);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RaiseObject(NedMain.RaiseObjectType.Up);
                    else
                    {
                        renderer.CamPos.x += Math.Sin(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                        renderer.CamPos.y += Math.Cos(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                    }
                    break;

                case Keys.Down:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Down);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RaiseObject(NedMain.RaiseObjectType.Down);
                    else
                    {
                        renderer.CamPos.x -= Math.Sin(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                        renderer.CamPos.y -= Math.Cos(MathLib.ToRad(renderer.CamOrient.yaw)) * GlobalVars.P3DStep;
                    }
                    break;

                case Keys.Next:
                    renderer.CamOrient.pitch = Math.Max(-80.0, renderer.CamOrient.pitch - 10.0);
                    break;

                case Keys.Prior:
                    renderer.CamOrient.pitch = Math.Min(80.0, renderer.CamOrient.pitch + 10.0);
                    break;

                case Keys.Home:
                    if ((e.Modifiers & (Keys.Control | Keys.Alt | Keys.Shift)) != 0)
                        GlobalVars.NedMain.StraightenTexture((e.Modifiers & Keys.Control) != 0, (e.Modifiers & Keys.Shift) != 0);
                    else
                        renderer.CamOrient.pitch = 0.0;
                    break;

                case Keys.A:
                    if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.DoKeyDown(new KeyEventArgs(Keys.A));
                    else if ((e.Modifiers & (Keys.Alt | Keys.Shift)) != 0)
                        GlobalVars.NedMain.DoKeyDown(e);
                    else
                        renderer.CamPos.z += GlobalVars.P3DStep;
                    break;

                case Keys.Z:
                    renderer.CamPos.z -= GlobalVars.P3DStep;
                    break;

                case Keys.N:
                    GotoWPlayer(cwPlayer + 1, 1);
                    break;

                case Keys.P:
                    if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.DoKeyDown(e);
                    else
                        GotoWPlayer(cwPlayer - 1, -1);
                    break;

                case Keys.Oemcomma:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Up);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RotateTexture(NedMain.ShiftTextureType.Left);
                    else if ((e.Modifiers & Keys.Alt) != 0)
                        GlobalVars.NedMain.ScaleTexture(NedMain.ShiftTextureType.Down);
                    else
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Left);
                    break;

                case Keys.OemPeriod:
                    if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Down);
                    else if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.RotateTexture(NedMain.ShiftTextureType.Right);
                    else if ((e.Modifiers & Keys.Alt) != 0)
                        GlobalVars.NedMain.ScaleTexture(NedMain.ShiftTextureType.Up);
                    else
                        GlobalVars.NedMain.ShiftTexture(NedMain.ShiftTextureType.Right);
                    break;

                case Keys.I:
                    if ((e.Modifiers & (Keys.Control | Keys.Alt | Keys.Shift)) == 0)
                        GlobalVars.NedMain.AddThingAtSurf();
                    else if ((e.Modifiers & Keys.Shift) != 0)
                        GlobalVars.NedMain.AddThingAtXYZPYR(renderer.CamPos, new Orient(-renderer.CamOrient.pitch, 360.0 - renderer.CamOrient.yaw, 0.0));
                    else
                        GlobalVars.NedMain.DoKeyDown(e);
                    break;

                case Keys.OemSemicolon:
                    GlobalVars.NedMain.StartStitch();
                    break;

                case Keys.Insert:
                    if ((e.Modifiers & Keys.Control) != 0)
                        GlobalVars.NedMain.StartStitch();
                    else
                        GlobalVars.NedMain.DoStitch();
                    break;

                case Keys.OemQuotes:
                    GlobalVars.NedMain.DoStitch();
                    break;

                case Keys.OemQuestion:
                    GlobalVars.NedMain.StraightenTexture((e.Modifiers & Keys.Control) != 0, (e.Modifiers & Keys.Shift) != 0);
                    break;

                case Keys.S:
                    if ((e.Modifiers & (Keys.Alt | Keys.Control)) == 0)
                        GlobalVars.NedMain.DoKeyDown(e);
                    break;

                case Keys.Delete:
                    if ((e.Modifiers & (Keys.Shift | Keys.Control)) == 0)
                        GlobalVars.NedMain.DoKeyDown(e);
                    break;

                case Keys.X:
                case Keys.Return:
                case Keys.OemCloseBrackets:
                case Keys.OemOpenBrackets:
                    if ((e.Modifiers & (Keys.Shift | Keys.Control | Keys.Alt)) == 0)
                        GlobalVars.NedMain.DoKeyDown(e);
                    break;

                case Keys.F:
                    if ((e.Modifiers & (Keys.Shift | Keys.Control | Keys.Alt)) == 0)
                        Clipboard.SetText("(" + renderer.CamPos.x.ToString("0.#####") + "/" + renderer.CamPos.y.ToString("0.#####") + "/" + renderer.CamPos.z.ToString("0.#####") + ":" +
                            (-renderer.CamOrient.pitch).ToString("0.#####") + "/" + (360.0 - renderer.CamOrient.yaw).ToString("0.#####") + "/0.0)");
                    break;
            }

            /*procedure TPreview3D.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if Render3D=nil then exit;
 
 With Render3D do
 case key of
  VK_ADD: If P3DGamma<254.9 then
          begin
           P3DGamma:=P3DGamma+0.1;
           if Render3D.SetGamma(P3DGamma)=1 then ReloadLevel;
           Invalidate;
          end;
  VK_SUBTRACT: If P3DGamma>0.1 then
          begin
           P3DGamma:=P3DGamma-0.1;
           if Render3D.SetGamma(P3DGamma)=1 then ReloadLevel;
           Invalidate;
          end;

  else exit;
 end;
  Invalidate;

end;     */
        }

        public void UpdateSector(JKSector sec)
        {
            AddSecChange(sec, 'C');
        }

        public void AddSector(JKSector sec)
        {
            AddSecChange(sec, 'A');
        }

        public void DeleteSector(JKSector sec)
        {
            AddSecChange(sec, 'D');
        }

        public void AddSecChange(JKSector sec, char change)
        {
            switch(change)
            {
                case 'A':
                    scs.Add(new SectorChange(change, sec));
                    break;

                case 'C':
                    {
                        int i = scs.IndexOf(sec);
                        if(i != -1 && scs[i].code == 'C')
                            return;
                        scs.Add(new SectorChange(change, sec));
                    }
                    break;

                case 'D':
                    {
                        int i = scs.IndexOf(sec);
                        if(i != -1)
                            scs.RemoveAt(i);
                        scs.Add(new SectorChange(change, sec));
                    }
                    break;
            }
        }

        public void UpdateThing(JKThing thing)
        {
            AddThingChange(thing, 'C');
        }

        public void AddThing(JKThing thing)
        {
            AddThingChange(thing, 'A');
        }

        public void DeleteThing(JKThing thing)
        {
            AddThingChange(thing, 'D');
        }

        public void AddThingChange(JKThing thing, char change)
        {
            switch (change)
            {
                case 'A':
                    ths.Add(new ThingChange(change, thing));
                    break;

                case 'C':
                    {
                        int i = ths.IndexOf(thing);
                        if (i != -1 && ths[i].code == 'C')
                            return;
                        ths.Add(new ThingChange(change, thing));
                    }
                    break;

                case 'D':
                    {
                        int i = ths.IndexOf(thing);
                        if (i != -1)
                            ths.RemoveAt(i);
                        ths.Add(new ThingChange(change, thing));
                    }
                    break;
            }
        }

        public void ReloadLevel()
        {
            if (renderer == null)
                return;
            
            scs.Clear();
            ths.Clear();
            renderer.ClearThings();
            renderer.ClearSectors();
            renderer.SetGamma(GlobalVars.P3DGamma);

            ProgressDialog.Reset(Level.sectors.Count + Level.things.Count, 0);

            ProgressDialog.Message("Loading sectors...");
            foreach (JKSector sec in Level.sectors)
            {
                ProgressDialog.Step();
                if (!GlobalVars.P3DVisLayers || Toolbar.IsLayerVisible(sec.layer))
                    SysAddSector(sec);
            }

            ProgressDialog.Message("Loading things...");
            foreach (JKThing thing in Level.things)
            {
                ProgressDialog.Step();
                if (!GlobalVars.P3DVisLayers || Toolbar.IsLayerVisible(thing.layer))
                    SysAddThing(thing);
            }

            ProgressDialog.Cleanup();
        }

        long lastTimestamp = 0;
        public void Process()
        {
            long curtimestamp = PerfTimer.Timestamp;
            double truedt = PerfTimer.TimeBetween(lastTimestamp, curtimestamp);//curtime.Subtract(lastProcess).TotalSeconds;
            if (truedt <= 0.0)
                return;

            double dt;
            if (truedt > (1.0 / 20.0))      // frame min..
                dt = 1.0 / 20.0;
            else
                dt = truedt;
            lastTimestamp = curtimestamp;


            /*VerifySelection();
            VerifyMultiSelection();

            if (updatestuff)
            {
                updatestuff = false;
                EditObject();
            }*/


            if (renderer != null)
                renderer.Process(dt);

            Redraw();
        }

        public void Redraw()
        {
            if (renderer == null)
                return;

            if (!Visible)
                return;

            using (renderer.Context)
            {
                bool smooth = true;// (ModifierKeys & Keys.Shift) == 0;

                if (smooth)
                {
                    gl.Enable(gl.LINE_SMOOTH);
                    gl.Enable(gl.POINT_SMOOTH);
                    gl.Enable(gl.POLYGON_SMOOTH);
                }
                else
                {
                    gl.Disable(gl.LINE_SMOOTH);
                    gl.Disable(gl.POINT_SMOOTH);
                    gl.Disable(gl.POLYGON_SMOOTH);
                }

                renderer.PushFBO(renderer.fbo, renderer, Width, Height, smooth ? 4 : 1);

                gl.Viewport(0, 0, previewRenderControl1.ClientRectangle.Width, previewRenderControl1.ClientRectangle.Height);

                gl.ClearColor(GlobalVars.clMapBack);
                gl.Clear(gl.DEPTH_BUFFER_BIT | gl.COLOR_BUFFER_BIT);
                renderer.PushStencil(Renderer.Stencil.Off);
                renderer.PushTextureEnable(true);
                renderer.PushBlend(Renderer.Blend.Normal);


                //RenderWorld(visCam, Mat4.Identity, dt, truedt, camsec, false);
                //RenderGUI(visCam, dt, truedt, lastFramePerfReport);


                //gl.Disable(gl.DEPTH_TEST);
                //gl.BlendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);


                ApplyChanges();
                renderer.Redraw();

                renderer.DrawNEDShit();

                /*renderer.BeginScene();
                DrawInternal();
                renderer.EndScene();*/

                /*if (false)
                {

                    gl.MatrixMode(gl.PROJECTION);
                    gl.LoadMatrixd(Mat4.Ortho(new Box2(renderControl1.ClientRectangle)));
                    gl.MatrixMode(gl.MODELVIEW);
                    gl.LoadIdentity();

                    Box2 rc = new Box2(renderControl1.ClientRectangle);
                    renderer.DrawLine(rc.FromUL(0.25, 0.25), rc.FromUL(0.7, 0.6), Color4.white());


                    gl.Begin(gl.POINTS);
                    gl.Vertex(rc.FromUL(0.1, 0.4));
                    gl.End();
                }*/


                RendererBase.FinishFBOStruct ffs = new RendererBase.FinishFBOStruct();
                ffs.resolveOverride = 0;
                renderer.PopFBO(ffs);

                renderer.PopBlend();
                renderer.PopTextureEnable();
                renderer.PopStencil();

                gl.SwapBuffers(renderer.hDC);
            }
        }

        public class SectorChange
        {
            public char code;
            public JKSector sector;

            public SectorChange(char _code, JKSector _sector)
            {
                code = _code;
                sector = _sector;
            }
        }

        public class ThingChange
        {
            public char code;
            public JKThing thing;

            public ThingChange(char _code, JKThing _thing)
            {
                code = _code;
                thing = _thing;
            }
        }

        public class SectorChangeList : List<SectorChange>
        {
            public int IndexOf(JKSector sec)
            {
                for(int x=0; x<Count; x++)
                    if(this[x].sector == sec)
                        return x;

                return -1;
            }
        }

        public class ThingChangeList : List<ThingChange>
        {
            public int IndexOf(JKThing thing)
            {
                for(int x=0; x<Count; x++)
                    if(this[x].thing == thing)
                        return x;

                return -1;
            }
        }

        SectorChangeList scs = new SectorChangeList();
        ThingChangeList ths = new ThingChangeList();

        void ApplyChanges()
        {
            if (renderer == null)
            {
                scs.Clear();
                ths.Clear();
                return;
            }

            foreach (SectorChange sc in scs)
                switch (sc.code)
                {
                    case 'A': SysAddSector(sc.sector); break;
                    case 'D': renderer.DeleteSector(sc.sector); break;
                    case 'C': renderer.UpdateSector(sc.sector); break;
                }

            scs.Clear();

            foreach(ThingChange tc in ths)
                switch (tc.code)
                {
                    case 'A': SysAddThing(tc.thing); break;
                    case 'D': renderer.DeleteThing(tc.thing); break;
                    case 'C': renderer.UpdateThing(tc.thing); break;
                }

            ths.Clear();
        }

        public void SetCam(Vec3 p, Orient o)
        {
            if (renderer == null)
                return;

            renderer.CamPos = p;
            renderer.CamOrient = new Orient(-o.pitch, -o.yaw, 0.0);// uh sure
        }

        public void SysAddSector(JKSector sec)
        {
            renderer.AddSector(sec);
        }

        public void SysAddThing(JKThing thing)
        {
            if (!GlobalVars.P3DThings)
                return;

            GlobalVars.NedMain.LoadThing3DO(thing, false);
            renderer.AddThing(thing);
        }

        protected override void OnHandleDestroyed(EventArgs e)
        {
            previewRenderControl1.Shutdown();

            base.OnHandleDestroyed(e);
        }

        private void Preview3D_Load(object sender, EventArgs e)
        {
            previewRenderControl1.Init();

            //ShowPreview();
        }

        public void ShowPreview()
        {
            if (IsHandleCreated)//Visible)
            {
                //BringToFront();
                if (GlobalVars.NedMain != null)
                    GlobalVars.NedMain.tabControl2.SelectedTab = GlobalVars.NedMain.tab3dPreview;
            }
            else
            {

                //GlobalVars.SetP3DPos(this, GlobalVars.P3DX, GlobalVars.P3DY, GlobalVars.P3DWinSize);

                Show();

                if (renderer == null)
                {
                    Hide();
                    return;
                }
            }

            ReloadLevel();
            GotoWPlayer(0, 1);
        }

        public void GotoWPlayer(int wPlayer, int step)
        {
            if (wPlayer >= Level.things.Count)
                wPlayer = 0;
            if (wPlayer < 0)
                wPlayer = Level.things.Count - 1;

            for (int i = 0; i < Level.things.Count; i++)
            {
                JKThing th = Level.things[wPlayer];

                if (th.name.Equals("walkplayer", StringComparison.InvariantCultureIgnoreCase))
                {
                    renderer.CamPos = th.pos;
                    renderer.CamOrient.pitch = -th.orient.pitch;
                    renderer.CamOrient.yaw = -th.orient.yaw;
                    cwPlayer = wPlayer;

                    return;
                }

                wPlayer += step;
                if (wPlayer >= Level.things.Count)
                    wPlayer = 0;
                if (wPlayer < 0)
                    wPlayer = Level.things.Count - 1;
            }
        }

        int cwPlayer = 0;
    }
}
