﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Smith.NED
{
    public partial class PreviewRenderControl : Control
    {
        public PreviewRenderControl()
        {
            InitializeComponent();
        }

        protected override void OnMouseEnter(EventArgs e)
        {
            base.OnMouseEnter(e);

            Focus();
        }

        public void Init()
        {
            Shutdown();

            renderer = new PreviewRenderer(this);
            using (renderer.Context)
            {
                renderer.SetupViewport(ClientRectangle, new Box2(ClientRectangle));

                gl.CullFace(gl.BACK);
                gl.FrontFace(gl.CCW);
                gl.Enable(gl.CULL_FACE);
                gl.Disable(gl.ALPHA_TEST);
            }
        }

        public void Shutdown()
        {
            if (renderer != null)
            {
                renderer.Dispose();
                renderer = null;
            }

            Invalidate();
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            //GlobalVars.NedMain.DoMouseMove(e);
        }

        protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
        {
            //GlobalVars.NedMain.DoPreviewKeyDown(e);
            GlobalVars.Preview3D.DoPreviewKeyDown(e);
        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyDown(e);
            GlobalVars.Preview3D.DoKeyDown(e);
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyUp(e);
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            GlobalVars.Preview3D.DoMouseUp(e);
        }

        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            GlobalVars.Preview3D.DoMouseDoubleClick(e);
        }

        public PreviewRenderer renderer = null;

        protected override void OnResize(EventArgs e)
        {
            if (renderer != null)
                renderer.SetupViewport(ClientRectangle, new Box2(ClientRectangle));
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            if (renderer != null)
                return;

            if (DesignMode)
            {
                e.Graphics.FillRectangle(new SolidBrush(Color.DarkBlue), ClientRectangle);
                e.Graphics.DrawString("designer mode", new Font("Verdana", 16.0f), new SolidBrush(Color.White), new Point(0, 0));
                return;
            }

            Graphics gfx = e.Graphics;

            gfx.FillRectangle(new HatchBrush(HatchStyle.WideUpwardDiagonal, Color.FromArgb(30, 5, 0), Color.Black), ClientRectangle);
        }

        protected override void OnPaintBackground(PaintEventArgs pevent)
        {

        }
    }
}
