﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class ResPicker : GUI.BaseForm
    {
        public static void Pick3DO(ref string model)
        {
            ResPicker r = new ResPicker(PickType.Model, model);
            r.Inita3DOPreview();
            if (r.ShowDialog() == DialogResult.OK)
                model = r.Choice;
            r.DoneA3DOPreview();
        }

        public static void PickTemplate(ref string tpl)
        {
            ResPicker r = new ResPicker(PickType.Template, tpl);
            r.InitTplPreview();
            if (r.ShowDialog() == DialogResult.OK)
                tpl = r.Choice;
            r.DoneTplPreview();
        }

        public static void PickCMP(ref string cmp)
        {
            ResPicker r = new ResPicker(PickType.Colormap, cmp);
            if (r.ShowDialog() == DialogResult.OK)
                cmp = r.Choice;
        }

        public static void PickSecSound(ref string wav)
        {
            ResPicker r = new ResPicker(PickType.SecSound, wav);
            if (r.ShowDialog() == DialogResult.OK)
                wav = r.Choice;
        }

        public static void PickMAT(ref string mat, string cmp)
        {
            ResPicker r = new ResPicker(PickType.Material, mat);
            r.pickCmpName = cmp;
            if (r.ShowDialog() == DialogResult.OK)
                mat = r.Choice;
        }

        public static void PickCOG(ref string cog)
        {
            ResPicker r = new ResPicker(PickType.COG, cog);
            if (r.ShowDialog() == DialogResult.OK)
                cog = r.Choice;
        }

        public static void PickAI(ref string ai)
        {
            ResPicker r = new ResPicker(PickType.AI, ai);
            if (r.ShowDialog() == DialogResult.OK)
                ai = r.Choice;
        }

        public static void PickSND(ref string snd)
        {
            ResPicker r = new ResPicker(PickType.SND, snd);
            if (r.ShowDialog() == DialogResult.OK)
                snd = r.Choice;
        }

        public static void PickPUP(ref string pup)
        {
            ResPicker r = new ResPicker(PickType.PUP, pup);
            if (r.ShowDialog() == DialogResult.OK)
                pup = r.Choice;
        }

        public static void PickPAR(ref string par)
        {
            ResPicker r = new ResPicker(PickType.PAR, par);
            if (r.ShowDialog() == DialogResult.OK)
                par = r.Choice;
        }

        public static void PickSPR(ref string spr)
        {
            ResPicker r = new ResPicker(PickType.SPR, spr);
            if (r.ShowDialog() == DialogResult.OK)
                spr = r.Choice;
        }

        public static void PickKEY(ref string key)
        {
            ResPicker r = new ResPicker(PickType.KEY, key);
            if (r.ShowDialog() == DialogResult.OK)
                key = r.Choice;
        }

        public static void PickLayer(ref string layer)
        {
            ResPicker r = new ResPicker(PickType.Layer, layer);
            if (r.ShowDialog() == DialogResult.OK)
                layer = r.Choice;
        }

        public static void PickThing(ref string thing)
        {
            ResPicker r = new ResPicker(PickType.Thing, thing);
            r.InitTplPreview();

            if(r.ShowDialog() == DialogResult.OK)
                thing = r.Choice;

            r.DoneTplPreview();
        }

        private string Cur_CMP = string.Empty;
        private bool loadcmp = false;

        public void SetCMP(string cmp)
        {
            Cur_CMP = cmp;
            loadcmp = true;
        }

        public string Choice
        {
            get
            {
                return EBResName.Text;
            }

            /*set
            {
                EBResName.Text = value;
            }*/
        }

        private void LoadJLH(string fname)
        {
            GroupFilters.InitializeFromFile(fname, TVGroups);
        }

        public void AddGobGroup(string gob, string dir, bool perm)
        {
            if (string.IsNullOrEmpty(gob))
                return;

            try
            {
                ContainerFile cf;
                if (perm)
                    cf = FileOperations.OpenGameContainer(gob);
                else
                    cf = FileOperations.OpenContainer(gob);

                cf.ChDir(dir);

                TVGroups.Nodes.Add(FileOperations.ExtractName(gob) + ">" + dir).Tag = cf.ListFiles.Keys.ToArray();
            }
            catch (Exception e)
            {
                Utils.PanMessage(Utils.MessageType.Error, e.Message);
            }
        }

        public void AddProjSubDirGroup(string subdir, string mask)
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
                return;

            List<string> files = new List<string>();
            TreeNode node = null;
            if (TVGroups.Nodes.ContainsKey(subdir))
            {
                node = TVGroups.Nodes[subdir];
                files.AddRange(node.Tag as string[]);
            }
            else
                node = TVGroups.Nodes.Add(subdir, subdir);

            files.AddRange(FileOperations.ListDirMask(System.IO.Path.Combine(GlobalVars.ProjectDir, subdir), mask));
            node.Tag = files.ToArray();
        }

        public void AddProjDirGroup(string mask)
        {
            AddDirGroup(GlobalVars.ProjectDir, "Project Directory", mask);
        }

        public void AddDirGroup(string dir, string name, string mask)
        {
            if (string.IsNullOrEmpty(dir))
                return;

            List<string> files = new List<string>();
            TreeNode node = null;
            if (TVGroups.Nodes.ContainsKey(name))
            {
                node = TVGroups.Nodes[name];
                files.AddRange(node.Tag as string[]);
            }
            else
                node = TVGroups.Nodes.Add(name, name);

            files.AddRange(FileOperations.ListDirMask(dir, mask));
            node.Tag = files.ToArray();
        }

        public void ResetGroups(string curnode="")
        {
            TreeNode firstNode = (TVGroups.Nodes.Count == 0 ? null : TVGroups.Nodes[0]);
            TreeNode node = firstNode;
            foreach(TreeNode tn in TVGroups.Nodes)
            {
                if (tn.Text.Equals(curnode))
                {
                    node = tn;
                    break;
                }
            }

            updatingmask = true;
            EBMask.Text = string.Empty;
            updatingmask = false;

            TVGroups.SelectedNode = node ?? TVGroups.TopNode;
        }

        public void SetItemList(string[] ls)
        {
            ResList.BeginUpdate();
            ResList.Items.Clear();
            try
            {
                if (ls == null)
                    return;

                string m = EBMask.Text;
                if (string.IsNullOrEmpty(m))
                {
                    ResList.Items.AddRange(ls);
                    return;
                }

                // if we dont end with a wildcard character, add one
                if (m.LastIndexOfAny(new char[] { '*', '?' }) != m.Length - 1)
                    m += "*";

                FileOperations.WildCardMask wcm = new FileOperations.WildCardMask(m);
                foreach (string s in ls)
                    if (wcm.Match(s))
                        ResList.Items.Add(s);
            }
            finally
            {
                ResList.EndUpdate();
            }
        }

        public enum PickType
        {
            Thing,
            Material,
            Colormap,
            COG,
            SecSound,
            Layer,
            Template,
            Model,
            AI,
            SND,
            PUP,
            SPR,
            KEY,
            PAR
        }

        // used so we can do our logic within form load;  since events are not getting fired prior to the dialog getting created
        PickType pickType;
        string pickName;
        string pickCmpName;

        public ResPicker(PickType _pickType, string _pickName)
        {
            ResListClick = _ResListClick;
            pickType = _pickType;
            pickName = _pickName;

            InitializeComponent();
        }

        // not really a thread.. whatever
        class PreviewThread : IDisposable
        {
            ResPicker rp;
            public PictureBox IM;
            public TextBox MM;
            public Label LB;
            public ScrollBar SBMat;

            ~PreviewThread()
            {
                Dispose(false);
            }

            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            private bool disposed = false;
            protected virtual void Dispose(bool disposing)
            {
                if (disposed)
                    return;

                if (disposing)
                {

                }

                if (soundchannel != null && soundchannel.IsPlaying)
                {
                    soundchannel.Stop();
                    soundchannel = null;
                }

                disposed = true;
            }

            public PreviewThread(ResPicker _rp, PictureBox aIM, TextBox aMM, Label aLB, ScrollBar aSB)
            {
                rp = _rp;
                IM = aIM;
                MM = aMM;
                LB = aLB;
                SBMat = aSB;
            }

            public void SBChange(int val)
            {
                nMatCell = val;
                NewFile = true;
                Execute();
            }

            string fname = string.Empty;
            bool NewFile = false;
            int nMatCell = -1;
            //string cmp_name = string.Empty;

            /*public void SetCmp(string name)
            {
                cmp_name = name;
            }*/

            public void StartPreview(string name)
            {
                fname = name;
                NewFile = true;
                nMatCell = SBMat.Value;

                Execute();
            }

            string bmComment = string.Empty;
            Bitmap bm = null;
            int nCells = 0;
            string ptext = string.Empty;
            Audio.Channel soundchannel = null;

            public unsafe void Execute()
            {
                if (!NewFile)
                    return;

                NewFile = false;

                bmComment = string.Empty;

                string ext = FileOperations.ExtractExt(fname).ToUpperInvariant();
                if (ext == ".3DO")
                {
                    //rp.Set3DOCMP(cmp_name);
                    rp.View3DO(fname);
                    return;
                }

                if (ext == ".MAT")
                {
                    Colormap cmp = PreviewRenderer.LoadColormap(rp.Cur_CMP);
                    Material mat = PreviewRenderer.LoadMaterial(fname);

                    if (cmp == null || mat == null)
                    {
                        ShowNothing();
                        if (cmp == null)
                        {
                            Utils.PanMessage(Utils.MessageType.Warning, rp.Cur_CMP + " not found");
                            return;
                        }

                        Utils.PanMessage(Utils.MessageType.Warning, fname + " not found");
                        return;
                    }

                    Bitmap dummy;
                    mat.GenerateBitmap(out bm, out dummy, cmp, nMatCell);

                    nCells = mat.NumCels;
                    if (nCells > 1)
                        bmComment = ", Multiple";

                    ShowMAT();
                }

                if (new List<string>(new string[] { ".COG", ".PAR", ".SPR", ".AI", ".AI0", ".AI2", ".SND", ".PUP" }).IndexOf(ext) != -1)
                {
                    using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    {
                        byte[] buf = bf.Read(Math.Min(4000, bf.Length)+1);
                        fixed (byte* pbuf = buf)
                        {
                            pbuf[buf.Length - 1] = 0;
                            ptext = new string((sbyte*)pbuf);
                        }
                    }

                    ShowText();
                }

                if (Audio.IsAcceptableExtension(ext))
                {
                    using (BaseFile bf = FileOperations.OpenGameFile(fname))
                    {
                        if (bf == null)
                            throw new Exception(fname + " not found");

                        if (soundchannel != null && soundchannel.IsPlaying)
                            soundchannel.Stop();

                        Audio.Sound snd = Audio.GetSound("NED_SOUND_PREVIEW", bf.Read(bf.Length), Audio.DimensionMode._2D, false);
                        soundchannel = Audio.PlaySound(snd);
                    }

                    ShowNothing();
                }
            }

            void ShowMAT()
            {
                LB.Visible = true;
                IM.Visible = true;
                MM.Visible = false;
                rp.updatingmatscroll = true;
                SBMat.Maximum = nCells - 1;
                SBMat.Visible = !string.IsNullOrEmpty(bmComment);
                rp.updatingmatscroll = false;
                LB.Text = "Width " + bm.Width.ToString() + " height " + bm.Height.ToString() + bmComment;
                if (bm.Width > IM.Width || bm.Height > IM.Height)
                    IM.SizeMode = PictureBoxSizeMode.Zoom;
                else
                    IM.SizeMode = PictureBoxSizeMode.Normal;
                IM.Image = bm;
                bm = null;
            }

            void ShowText()
            {
                IM.Visible = false;
                LB.Visible = false;
                MM.Visible = true;
                MM.Text = ptext;
                ptext = string.Empty;
            }

            void ShowNothing()
            {
                IM.Visible = false;
                MM.Visible = false;
                LB.Text = string.Empty;
            }
        }

        PreviewThread pv = null;

        void FreshLoad()
        {
            BNImport.Visible = false;
            BNNew.Visible = false;
            BNEdit.Visible = false;
            BNExport.Visible = false;
            SBMatCell.Visible = false;
            SBYAW.Value = 0;
            SBPCH.Value = 0;
            pv = new PreviewThread(this, BMPreview, memo, LBDetails, SBMatCell);

            TVGroups.Nodes.Clear();

            switch (pickType)
            {
                case PickType.Colormap:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\cmp", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\cmp", true);
                    AddProjDirGroup("*.cmp");
                    AddProjSubDirGroup(@"misc\cmp\", "*.cmp");
                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.COG:
                    AddGobGroup(GlobalVars.ProjectGob, "cog", false);
                    AddGobGroup(GlobalVars.sp_gob, "cog", true);
                    AddGobGroup(GlobalVars.mp1_gob, "cog", true);
                    AddGobGroup(GlobalVars.mp2_gob, "cog", true);
                    AddGobGroup(GlobalVars.mp3_gob, "cog", true);
                    AddGobGroup(GlobalVars.res2_gob, "cog", true);
                    AddProjDirGroup("*.cog");
                    AddProjSubDirGroup(@"cog\", "*.cog");

                    if (GlobalVars.IsMOTS)
                        LoadJLH("mcogs.jlh");
                    else
                        LoadJLH("cogs.jlh");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.Layer:
                    TVGroups.Nodes.Add("Layers").Tag = GlobalVars.Level.layers.ToArray();
                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.Material:
                    BNImport.Visible = true;
                    BNExport.Visible = true;
                    SetCMP(pickCmpName);
                    AddGobGroup(GlobalVars.ProjectGob, "mat", false);
                    AddGobGroup(GlobalVars.ProjectGob, @"3do\mat", false);
                    AddGobGroup(GlobalVars.res2_gob, "mat", true);
                    AddGobGroup(GlobalVars.res2_gob, @"3do\mat", true);
                    AddProjDirGroup("*.mat");
                    AddProjSubDirGroup(@"mat\", "*.mat");
                    AddProjSubDirGroup(@"3do\mat\", "*.mat");

                    if (GlobalVars.IsMOTS)
                        LoadJLH("mmats.jlh");
                    else
                        LoadJLH("mats.jlh");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.SecSound:
                    AddGobGroup(GlobalVars.ProjectGob, "sound", false);
                    AddGobGroup(GlobalVars.ProjectGob, "voice", false);
                    AddGobGroup(GlobalVars.res1_gob, "sound", true);
                    AddGobGroup(GlobalVars.res1_gob, "voice", true);
                    if (GlobalVars.IsMOTS)
                    {
                        AddGobGroup(GlobalVars.ProjectGob, "voiceuu", false);
                        AddGobGroup(GlobalVars.res2_gob, "voiceuu", true);
                    }
                    foreach (string ext in Audio.AcceptableAudioExtensions)
                    {
                        string mask = "*" + ext.ToLowerInvariant();
                        AddProjDirGroup(mask);
                        AddProjSubDirGroup(@"sound\", mask);
                        AddProjSubDirGroup(@"voice\", mask);
                    }

                    if (GlobalVars.IsMOTS)
                        LoadJLH("mwavs.jlh");
                    else
                        LoadJLH("wavs.jlh");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.Thing:
                    {
                        BNImport.Visible = !string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNNew.Visible = !string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNEdit.Visible = true;//!string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNExport.Visible = true;
                        SetCMP(GlobalVars.Level.GetMasterCMP());
                        List<string> tpls = new List<string>();
                        foreach (JKTemplate tpl in Templates.All)
                        {
                            if (tpl.Name.StartsWith("_") ||
                                tpl.Name.StartsWith("+"))
                                continue;

                            tpls.Add(tpl.Name);
                        }
                        TVGroups.Nodes.Add("Templates").Tag = tpls.ToArray();

                        if (GlobalVars.IsMOTS)
                            LoadJLH("mtpls.jlh");
                        else
                            LoadJLH("tpls.jlh");

                        ResetGroups();
                        ChangeResName(pickName);
                    }
                    break;

                case PickType.Template:
                    {
                        BNImport.Visible = !string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNNew.Visible = false/*i dont think we want to recursively let template creator spawn itself*/;// !string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNEdit.Visible = false/*i dont think we want to recursively let template creator spawn itself*/;// !string.IsNullOrEmpty(GlobalVars.ProjectDir);
                        BNExport.Visible = true;
                        SetCMP(GlobalVars.Level.GetMasterCMP());
                        List<string> tpls = new List<string>();
                        foreach (JKTemplate tpl in Templates.All)
                            tpls.Add(tpl.Name);
                        TVGroups.Nodes.Add("Templates").Tag = tpls.ToArray();

                        if (GlobalVars.IsMOTS)
                            LoadJLH("mtpls.jlh");
                        else
                            LoadJLH("tpls.jlh");

                        ResetGroups();
                        ChangeResName(pickName);
                    }
                    break;

                case PickType.Model:
                    {
                        BNExport.Visible = true;
                        SetCMP(GlobalVars.Level.GetMasterCMP());
                        AddGobGroup(GlobalVars.ProjectGob, "3do", false);
                        AddGobGroup(GlobalVars.res2_gob, "3do", true);
                        AddProjDirGroup("*.3do");
                        AddProjSubDirGroup(@"3do\", "*.3do");

                        if (GlobalVars.IsMOTS)
                            LoadJLH("m3dos.jlh");
                        else
                            LoadJLH("3dos.jlh");

                        ResetGroups();
                        ChangeResName(pickName);
                    }
                    break;

                case PickType.AI:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\ai", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\ai", true);
                    AddProjDirGroup("*.ai;*.ai0;*.ai2");
                    AddProjSubDirGroup(@"misc\ai\", "*.ai;*.ai0;*.ai2");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.KEY:
                    AddGobGroup(GlobalVars.ProjectGob, @"3do\key", false);
                    AddGobGroup(GlobalVars.res2_gob, @"3do\key", true);
                    AddProjDirGroup("*.key");
                    AddProjSubDirGroup(@"3do\key\", "*.key");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.SND:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\snd", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\snd", true);
                    AddProjDirGroup("*.snd");
                    AddProjSubDirGroup(@"misc\snd\", "*.snd");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.PUP:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\pup", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\pup", true);
                    AddProjDirGroup("*.pup");
                    AddProjSubDirGroup(@"misc\pup\", "*.pup");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.SPR:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\spr", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\spr", true);
                    AddProjDirGroup("*.spr");
                    AddProjSubDirGroup(@"misc\spr\", "*.spr");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;

                case PickType.PAR:
                    AddGobGroup(GlobalVars.ProjectGob, @"misc\par", false);
                    AddGobGroup(GlobalVars.res2_gob, @"misc\par", true);
                    AddProjDirGroup("*.par");
                    AddProjSubDirGroup(@"misc\par\", "*.par");

                    ResetGroups();
                    ChangeResName(pickName);
                    break;
            }



            if (loadcmp)
            {
                CBCMPs.Visible = true;
                ContainerFile cf = FileOperations.OpenGameContainer(GlobalVars.res2_gob);
                cf.ChDir(@"misc\cmp");
                CBCMPs.Items.AddRange(cf.ListFiles.Keys.ToArray());
                int matchingIndex = -1;
                for (int x = 0; x < CBCMPs.Items.Count; x++)
                    if (Cur_CMP.Equals(CBCMPs.Items[x] as string, StringComparison.InvariantCultureIgnoreCase))
                    {
                        matchingIndex = x;
                        break;
                    }

                if (matchingIndex == -1)
                    matchingIndex = CBCMPs.Items.Add(Cur_CMP);

                CBCMPs.SelectedIndex = matchingIndex;
                CBCMPs_SelectedIndexChanged(CBCMPs, new EventArgs());
            }
        }

        private void ResPicker_Load(object sender, EventArgs e)
        {
            FreshLoad();
        }

        bool updatingmask = false;
        private void EBMask_TextChanged(object sender, EventArgs e)
        {
            if (updatingmask)
                return;

            if (TVGroups.SelectedNode == null)
                SetItemList(null);

            SetItemList(TVGroups.SelectedNode.Tag as string[]);
        }

        private void TVGroups_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node == null)
            {
                BNImport.Enabled = false;
                BNNew.Enabled = false;
                BNEdit.Enabled = false;
                BNExport.Enabled = false;
                SetItemList(null);
            }
            else
            {
                SetItemList(e.Node.Tag as string[]);

                switch(pickType)
                {
                    case PickType.Material:
                        BNImport.Enabled = e.Node.Text.Equals(@"mat\", StringComparison.InvariantCultureIgnoreCase)||
                                                e.Node.Text.Equals(@"3do\mat\", StringComparison.InvariantCultureIgnoreCase);

                        BNExport.Enabled = false;
                        break;

                    case PickType.Thing:
                    case PickType.Template:
                        BNImport.Enabled = true;
                        BNExport.Enabled = true;
                        BNNew.Enabled = true;
                        BNEdit.Enabled = true;
                        break;

                    case PickType.Model:
                        BNExport.Enabled = true;
                        break;

                    default:
                        BNImport.Enabled = false;
                        BNExport.Enabled = false;
                        BNNew.Enabled = false;
                        BNEdit.Enabled = true;
                        break;
                }
            }
        }

        delegate void ResListClickEvent();
        ResListClickEvent ResListClick = null;
        ResListClickEvent OldResListClick = null;

        PreviewRenderer Render3D
        {
            get
            {
                return Panel3D.renderer;
            }
        }

        void InitRenderer()
        {
            if (Render3D != null)
                return;

            try
            {
                Panel3D.Init();
                //Render3D = new PreviewRenderer(Panel3D);
                //Render3D.Initialize();
            }
            catch
            {
                //Render3D = null;
            }
        }

        /*void Set3DOCMP(string cmp)
        {
            Cur_CMP = cmp;
        }*/

        void View3DO(string name)
        {           
            InitRenderer();
            if (Render3D == null)
                return;

            JKThing tmpthing = new JKThing(null/*aight? lol*/);

            JKModel a3do = tmpthing.model;
            tmpthing.model = null;

            tmpthing.model = new JKModel(name, 0);
            tmpthing.orient.yaw = 180.0;
            Box3 box = tmpthing.model.GetBBox();

            Render3D.thing_cmp = Cur_CMP;

            Render3D.SetGamma(GlobalVars.P3DGamma);
            Render3D.ClearThings();
            Render3D.AddThing(tmpthing);
            Render3D.SetViewToThing(tmpthing);

            P3DO_SetPCHYAW(cPCH, cYAW);
            Redraw();
        }

        double cPCH = 0.0;
        double cYAW = 0.0;

        void P3DO_SetPCHYAW(double pch, double yaw)
        {
            cPCH = pch;
            cYAW = yaw;
            if (Render3D == null)
                return;

            Render3D.SetPCHYAW(pch, yaw);
            Redraw();
        }

        public void Redraw()
        {
            if (Render3D == null)
                return;

            //if (!Visible)
            //    return;

            using (Render3D.Context)
            {
                bool smooth = true;// (ModifierKeys & Keys.Shift) == 0;

                if (smooth)
                {
                    gl.Enable(gl.LINE_SMOOTH);
                    gl.Enable(gl.POINT_SMOOTH);
                    gl.Enable(gl.POLYGON_SMOOTH);
                }
                else
                {
                    gl.Disable(gl.LINE_SMOOTH);
                    gl.Disable(gl.POINT_SMOOTH);
                    gl.Disable(gl.POLYGON_SMOOTH);
                }

                Render3D.PushFBO(Render3D.fbo, Render3D, Width, Height, smooth ? 4 : 1);

                gl.Viewport(0, 0, Panel3D.ClientRectangle.Width, Panel3D.ClientRectangle.Height);

                gl.ClearColor(GlobalVars.clMapBack);
                gl.Clear(gl.DEPTH_BUFFER_BIT | gl.COLOR_BUFFER_BIT);


                //RenderWorld(visCam, Mat4.Identity, dt, truedt, camsec, false);
                //RenderGUI(visCam, dt, truedt, lastFramePerfReport);


                //gl.Disable(gl.DEPTH_TEST);
                //gl.BlendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

                //gl.MatrixMode(gl.MODELVIEW);
                //gl.PushMatrix();
                //gl.LoadIdentity();
                //gl.MultMatrix(Render3D.WorldOrient.Matrix.Inverse);

                //ApplyChanges();
                //renderer.Redraw();
                Render3D.Redraw();


                //gl.PopMatrix();

                /*renderer.BeginScene();
                DrawInternal();
                renderer.EndScene();*/

                /*if (false)
                {

                    gl.MatrixMode(gl.PROJECTION);
                    gl.LoadMatrixd(Mat4.Ortho(new Box2(renderControl1.ClientRectangle)));
                    gl.MatrixMode(gl.MODELVIEW);
                    gl.LoadIdentity();

                    Box2 rc = new Box2(renderControl1.ClientRectangle);
                    renderer.DrawLine(rc.FromUL(0.25, 0.25), rc.FromUL(0.7, 0.6), Color4.white());


                    gl.Begin(gl.POINTS);
                    gl.Vertex(rc.FromUL(0.1, 0.4));
                    gl.End();
                }*/


                RendererBase.FinishFBOStruct ffs = new RendererBase.FinishFBOStruct();
                ffs.resolveOverride = 0;
                Render3D.PopFBO(ffs);

                gl.SwapBuffers(Render3D.hDC);
            }
        }

        void InitTplPreview()
        {
            LBDetails.Text = string.Empty;
            LBDetails.Visible = true;
            OldResListClick = ResListClick;
            ResListClick = TplListClick;
            Panel3D.Visible = true;
            SBPCH.Visible = true;
            SBYAW.Visible = true;
        }

        void DoneTplPreview()
        {
            LBDetails.Visible = false;
            ResListClick = OldResListClick;
            Panel3D.Visible = false;

            SBPCH.Visible = false;
            SBYAW.Visible = false;
        }

        void Inita3DOPreview()
        {
            LBDetails.Text = string.Empty;
            LBDetails.Visible = true;
            OldResListClick = ResListClick;
            ResListClick = A3DOListClick;
            Panel3D.Visible = true;
            SBPCH.Visible = true;
            SBYAW.Visible = true;
        }

        void DoneA3DOPreview()
        {
            LBDetails.Visible = false;
            ResListClick = OldResListClick;
            Panel3D.Visible = false;
            SBPCH.Visible = false;
            SBYAW.Visible = false;
        }

        void A3DOListClick()
        {
            int n = ResList.SelectedIndex;
            if (n < 0)
                return;

            string fname = ResList.Items[n] as string;
            SetResName(fname);
            pv.StartPreview(fname);
        }

        void _ResListClick()
        {
            if (ResList.SelectedIndex < 0)
                return;

            SetResName(ResList.Items[ResList.SelectedIndex] as string);
            pv.StartPreview(EBResName.Text);
        }

        public void SetResName(string s)
        {
            updatingresname = true;
            EBResName.Text = s;
            updatingresname = false;
        }

        public void ChangeResName(string s)
        {
            SetResName(s);

            EBResName_TextChanged(EBResName, new EventArgs());

            if (ResList.SelectedIndex < 0)
            {
                for (int i = 0; i < TVGroups.Nodes.Count; i++)
                {
                    string[] sl = TVGroups.Nodes[i].Tag as string[];
                    if (sl == null)
                        continue;

                    int matchingIndex = -1;
                    for(int x=0; x<sl.Length; x++)
                        if (sl[x].Equals(s, StringComparison.InvariantCultureIgnoreCase))
                        {
                            matchingIndex = x;
                            break;
                        }

                    if (matchingIndex != -1)
                    {
                        TVGroups.SelectedNode = TVGroups.Nodes[i];
                        EBResName_TextChanged(EBResName, new EventArgs());
                        break;
                    }
                }
            }
        }

        void TplListClick()
        {
            int n = ResList.SelectedIndex;
            if (n < 0)
                return;

            string tpname = ResList.Items[n] as string;
            SetResName(tpname);
            n = Templates.IndexOfName(tpname);
            if (n == -1)
                LBDetails.Text = string.Empty;
            else
            {
                TemplateValue tplval = Templates.GetTPLField(tpname, "model3d");
                if (tplval != null)
                    pv.StartPreview(tplval.AsString);
                else
                    Panel3D.Invalidate();

                LBDetails.Text = Templates.All[n].desc + "\n" + Templates.All[n].resourceSummary;
            }
        }

        private void ResList_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (ResListClick != null)
                ResListClick();
        }

        bool updatingresname = false;
        private void EBResName_TextChanged(object sender, EventArgs e)
        {
            if (updatingresname)
                return;

            int matchingIndex = -1;
            for(int x=0; x<ResList.Items.Count; x++)
                if (EBResName.Text.Equals(ResList.Items[x] as string, StringComparison.InvariantCultureIgnoreCase))
                {
                    matchingIndex = x;
                    break;
                }

            if (matchingIndex != -1)
                ResList.SelectedIndex = matchingIndex;
        }

        private void ResList_DoubleClick(object sender, EventArgs e)
        {
            // maybe this should be mousedoubleclick-  will have to test
            DialogResult = System.Windows.Forms.DialogResult.OK;
            Close();
        }

        private void CBCMPs_SelectedIndexChanged(object sender, EventArgs e)
        {
            SetCMP(CBCMPs.SelectedItem as string); //pv.SetCmp(CBCMPs.SelectedItem as string);
            if (ResListClick != null)
                ResListClick();
        }

        private void ResPicker_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (pv != null)
            {
                pv.Dispose();
                pv = null;
            }

            Panel3D.Shutdown();

            if (loadcmp)
            {
                CBCMPs.Visible = false;
                loadcmp = false;
            }
        }

        private void SBYAW_ValueChanged(object sender, EventArgs e)
        {
            P3DO_SetPCHYAW(SBPCH.Value, SBYAW.Value);
        }

        private void SBPCH_ValueChanged(object sender, EventArgs e)
        {
            P3DO_SetPCHYAW(SBPCH.Value, SBYAW.Value);
        }

        bool updatingmatscroll = false;
        private void SBMatCell_ValueChanged(object sender, EventArgs e)
        {
            if (updatingmatscroll)
                return;

            pv.SBChange(SBMatCell.Value);
        }

        private void ResPicker_Shown(object sender, EventArgs e)
        {
            ResList.Focus();
        }

        private void ResList_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Escape)
            {
                DialogResult = System.Windows.Forms.DialogResult.Cancel;
                Close();
                return;
            }

            if (e.KeyCode == Keys.Enter)
            {
                DialogResult = System.Windows.Forms.DialogResult.OK;
                Close();
                return;
            }
        }

        private unsafe void ImportImage()
        {
            TreeNode node = TVGroups.SelectedNode;
            if (node == null)
                return;

            if (!node.Text.Equals(@"mat\", StringComparison.InvariantCultureIgnoreCase) &&
                !node.Text.Equals(@"3do\mat\", StringComparison.InvariantCultureIgnoreCase))
                return;

            //BMP, GIF, JPEG, PNG, TIFF, Exif, WMF, and EMF
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Image Files (*.bmp;*.gif;*.jpg;*.png)|*.bmp;*.gif;*.jpg;*.png";
            ofd.Multiselect = true;
            if (ofd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            if (ofd.FileNames.Length == 0)
                return;

            Bitmap[] bmps;
            try
            {
                List<string> files = new List<string>(ofd.FileNames);
                files.Sort();
                bmps = new Bitmap[files.Count];
                for(int x=0; x<files.Count; x++)
                    bmps[x] = (Bitmap)Bitmap.FromFile(files[x]);
            }
            catch
            {
                MessageBox.Show("Failed to load image");
                return;
            }

            // doesnt really do anything.. just a hook for when we do stuff later
            ImportToMAT dlg = new ImportToMAT();
            dlg.ShowDialog();

            string outfile = Path.Combine(node.Text, Path.GetFileNameWithoutExtension(ofd.FileName) + ".mat");
            bool alreadyExist = File.Exists(outfile);
            if (alreadyExist)
            {
                if (MessageBox.Show("File " + Path.Combine(node.Text, Path.GetFileName(outfile)) + " already exists. Overwrite?", "Warning", MessageBoxButtons.OKCancel) != System.Windows.Forms.DialogResult.OK)
                    return;
            }

            string outdir = Path.GetDirectoryName(outfile);

            // ensure directory exists
            if (!Directory.Exists(outdir))
                Directory.CreateDirectory(outdir);

            

            DOMAGICIMPORTBMPSHIT(bmps, outfile, true);



            // add file entry and refresh
            if (!alreadyExist)
            {
                List<string> files = new List<string>(node.Tag as string[]);
                files.Add(Path.GetFileName(outfile));
                node.Tag = files.ToArray();
            }
            SetItemList(node.Tag as string[]);
        }

        public static unsafe void DOMAGICIMPORTBMPSHIT(Bitmap[] bmps, string outfile, bool flipX)
        {
            int width = bmps[0].Width;
            int height = bmps[0].Height;
            bool hasTransparentPixels = false;

            for (int i = 0; i < bmps.Length; i++)
            {
                if (flipX)
                {
                    bmps[i].RotateFlip(RotateFlipType.Rotate180FlipX);
                }
            }

            BitmapData[] bmdats = new BitmapData[bmps.Length];
            for (int i = 0; i < bmdats.Length; i++)
            {
                Bitmap bmp = bmps[i];

                BitmapData bmdat = bmp.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                bmdats[i] = bmdat;

                // lets first scan for any transparent pixels so we know what type of MAT we should be making
                for (int y = 0; y < height; y++)
                {
                    byte* pBMPRow = (byte*)((ulong)bmdat.Scan0 + (ulong)(y * bmdat.Stride));
                    for (int x = 0; x < width; x++)
                    {
                        /*  pBMPRow[x * 4 + 0] = rgb.B;
                            pBMPRow[x * 4 + 1] = rgb.G;
                            pBMPRow[x * 4 + 2] = rgb.R;
                            pBMPRow[x * 4 + 3] = (byte)((th.Transparent && b == TransparentColor) ? 0 : 255);*/

                        if (pBMPRow[3] == 0)    // what should threshold be?
                        {
                            hasTransparentPixels = true;
                            break;
                        }
                        pBMPRow += 4;
                    }
                }
            }


            using (BinaryWriter bw = new BinaryWriter(File.Create(outfile)))
            {
                bw.Write((byte)'M');
                bw.Write((byte)'A');
                bw.Write((byte)'T');
                bw.Write((byte)' ');
                bw.Write(0x32); // ver
                bw.Write(2);    // type
                bw.Write(bmps.Length);    // nummaterials
                bw.Write(bmps.Length);    // numtextures
                bw.Write(1);    // transparency value
                bw.Write(16);    // colorbits
                bw.Write(5);    // bluebits
                bw.Write(hasTransparentPixels ? 5 : 6);    // greenbits
                bw.Write(5);    // redbits

                bw.Write(0x0B);// unknown
                bw.Write(0x05);//unknown
                bw.Write(0);//unknown
                bw.Write(0x03);//unknown
                bw.Write(0x02);//unknown
                bw.Write(0x03);//unknown
                for (int x = 0; x < 3*4; x++)
                    bw.Write((byte)0);

                for (int bmpIndex = 0; bmpIndex < bmps.Length; bmpIndex++)
                {
                    Bitmap bmp = bmps[bmpIndex];

                    bw.Write(8);    // texture
                    bw.Write(0);    // unknown
                    for (int x = 0; x < 4; x++)
                        bw.Write((int)0x3F800000);
                    bw.Write(0);    // dunno
                    bw.Write(0);    // dunno
                    bw.Write((uint)0xBFF78482);  // dunno
                    bw.Write(bmpIndex);    // texture index
                }

                for (int bmpIndex = 0; bmpIndex < bmps.Length; bmpIndex++)
                {
                    bw.Write(width);
                    bw.Write(height);
                    bw.Write(0);    // transparent
                    bw.Write(0);    // pad
                    bw.Write(0);    // pad
                    bw.Write(1);    // mipmaps

                    BitmapData bmdat = bmdats[bmpIndex];

                    for (int y = 0; y < height; y++)
                    {
                        byte* pBMPRow = (byte*)((ulong)bmdat.Scan0 + (ulong)bmdat.Stride * (ulong)(height - y - 1));

                        for (int x = 0; x < width; x++)
                        {
                            int b = pBMPRow[0];
                            int g = pBMPRow[1];
                            int r = pBMPRow[2];
                            int a = pBMPRow[3];

                            pBMPRow += 4;

                            ushort pixelword = 0;
                            if (hasTransparentPixels)
                            {
                                if (a > 0)
                                {
                                    pixelword |= 0x8000;

                                    pixelword |= (ushort)((double)b / 255.0 * (double)0x1F);
                                    pixelword |= (ushort)((int)((double)g / 255.0 * (double)0x1F) << 5);
                                    pixelword |= (ushort)((int)((double)r / 255.0 * (double)0x1F) << 10);
                                }
                                else
                                {
                                    // leave all zero?
                                }
                            }
                            else
                            {
                                pixelword |= (ushort)((double)b / 255.0 * (double)0x1F);
                                pixelword |= (ushort)((int)((double)g / 255.0 * (double)0x3F) << 5);
                                pixelword |= (ushort)((int)((double)r / 255.0 * (double)0x1F) << 11);
                            }


                            bw.Write(pixelword);
                        }
                    }
                }
            }

            for (int i = 0; i < bmps.Length; i++)
            {
                bmps[i].UnlockBits(bmdats[i]);
            }
        }

        void ImportTemplate()
        {
            if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                MessageBox.Show("Level not saved");
                return;
            }

            TreeNode node = TVGroups.SelectedNode;
            if (node == null)
                return;

            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Resources (*.gob)|*.gob";
            ofd.Multiselect = true;
            if (ofd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            if (ofd.FileNames.Length == 0)
                return;

            foreach (string gobfile in ofd.FileNames)
            {
                //public void ExtractProject(string gobpath, string dstpath)

                GlobalVars.NedMain.ExtractProject(gobfile, GlobalVars.ProjectDir, false);
            }

            MessageBox.Show("Please reopen resource picker");

            DialogResult = System.Windows.Forms.DialogResult.Cancel;
            Close();

            //SetItemList(node.Tag as string[]);
        }

        private void BNImportImage_Click(object sender, EventArgs e)
        {
            switch(pickType)
            {
                case PickType.Material:
                    ImportImage();
                    return;

                case PickType.Template:
                case PickType.Thing:
                    ImportTemplate();
                    return;

                default:
                    MessageBox.Show("Import not supported for " + pickType.ToString());
                    return;
            }
        }

        void ExportOBJ()
        {
            if (string.IsNullOrEmpty(Choice))
            {
                MessageBox.Show("No choice");
                return;
            }

            JKModel model = new JKModel(Choice, 0);
            if(model == null)
            {
                MessageBox.Show("Couldn't load model " + Choice);
                return;
            }

            foreach (JKModel.Mesh mesh in model.meshes)
                foreach (JKModel.Face face in mesh.faces)
                    face.Recalc();

            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "OBJ Files (*.OBJ)|*.OBJ";
            sfd.Title = "Save OBJ";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            model.SaveToOBJ(sfd.FileName, JKModel.OBJScale, true, Cur_CMP, JKModel.OBJTextureOutput.Tga);
        }

        void ExportTemplate()
        {
            if (string.IsNullOrEmpty(Choice))
            {
                MessageBox.Show("No choice");
                return;
            }

            int i = Templates.IndexOfName(Choice);
            if (i == -1)
            {
                MessageBox.Show("Couldn't find template " + Choice);
                return;
            }

            JKLevel.ClassSaveToJKL.UsedListMatrix matrix = Templates.All[i].BuildUsedListMatrix(true);

            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Resources (*.gob)|*.gob";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            matrix.SaveGOB(sfd.FileName);
        }

        private void BNExport_Click(object sender, EventArgs e)
        {
            switch (pickType)
            {
                case PickType.Template:
                case PickType.Thing:
                    ExportTemplate();
                    return;

                case PickType.Model:
                    ExportOBJ();
                    return;

                default:
                    MessageBox.Show("Export not supported for " + pickType.ToString());
                    return;
            }
        }

        void CreateTemplate(bool editTemplate)
        {
            string templateName = editTemplate ? ResList.SelectedItem as string : null;

            templateName = TemplateCreator.CreateTemplate(templateName);

            FreshLoad();

            if (!string.IsNullOrEmpty(templateName))
            {
                //EBMask.Text = templateName;
                for (int x = 0; x < ResList.Items.Count; x++)
                    if (templateName.Equals(ResList.Items[x] as string))
                    {
                        ResList.SelectedIndex = x;
                        break;
                    }
            }

                /*MessageBox.Show("Please reopen resource picker");

                DialogResult = System.Windows.Forms.DialogResult.Cancel;
                Close();*/
        }

        private void BNNew_Click(object sender, EventArgs e)
        {
            switch (pickType)
            {
                //case PickType.Template:
                case PickType.Thing:
                    CreateTemplate(false);
                    return;

                default:
                    MessageBox.Show("New not supported for " + pickType.ToString());
                    return;
            }
        }

        private void BNEdit_Click(object sender, EventArgs e)
        {
            switch (pickType)
            {
                //case PickType.Template:
                case PickType.Thing:
                    CreateTemplate(true);
                    return;

                default:
                    MessageBox.Show("New not supported for " + pickType.ToString());
                    return;
            }
        }
    }
}
