﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Smith.NED
{
    public partial class TemplateCreator : GUI.BaseForm
    {
        // returns name of edited/created template, if any
        public static string CreateTemplate(string prepopulateTemplateName=null)
        {
            /*if (string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                MessageBox.Show("Save project first");
                return null;
            }*/

            TemplateCreator tc = new TemplateCreator(prepopulateTemplateName);
            if (tc.ShowDialog() != DialogResult.OK)
                return null;

            return tc.EBTPL.Text;
        }

        string prepopulateTemplateName = null;

        public TemplateCreator(string _prepopulateTemplateName)
        {
            prepopulateTemplateName = _prepopulateTemplateName;

            InitializeComponent();
        }

        bool changed = false;

        private void TemplateCreator_Load(object sender, EventArgs e)
        {
            EBTPL.Text = string.Empty;
            EBParent.Text = "none";

            if (!string.IsNullOrEmpty(prepopulateTemplateName))
            {
                LoadTemplate(prepopulateTemplateName);
            }


            // if no saved project, still allow editor.. just no saving
            if(string.IsNullOrEmpty(GlobalVars.ProjectDir))
            {
                BNSave.Enabled = false;
            }
        }

        void LoadTemplate(string val)
        {
            if (string.IsNullOrEmpty(val))
                return;

            EBTPL.Text = val;
            int i = Templates.IndexOfName(val);
            if (i == -1)
                return;

            EBParent.Text = Templates.All[i].Parent;

            BNUpdate_Click(BNUpdate, new EventArgs());
        }

        bool CheckChanged()
        {
            if (!changed)
                return true;

            if (MessageBox.Show("You're about to lose changes you've made. Proceed?", "Warning", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                return false;

            return true;
        }

        private void BNTPL_Click(object sender, EventArgs e)
        {
            string val = EBTPL.Text;
            ResPicker.PickTemplate(ref val);
            LoadTemplate(val);
        }

        void SetField(string fname, string fval)
        {
            List<string> lines = new List<string>(MMFields.Lines);

            bool found = false;
            for (int i = 0; i < lines.Count; i++)
            {
                string s = lines[i];
                string[] sparts = s.Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                if (sparts.Length != 2)
                    continue;

                if (sparts[0].Equals(fname, StringComparison.InvariantCultureIgnoreCase))
                {
                    lines[i] = fname + "=" + fval;
                    found = true;
                    break;
                }
            }

            if (!found)
                lines.Add(fname + "=" + fval);

            MMFields.Lines = lines.ToArray();
        }

        private void BNParent_Click(object sender, EventArgs e)
        {
            string val = EBParent.Text;
            ResPicker.PickTemplate(ref val);
            EBParent.Text = val;
        }

        private void MMFields_TextChanged(object sender, EventArgs e)
        {
            changed = true;
        }

        private void TemplateCreator_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (!CheckChanged())
            {
                e.Cancel = true;
                return;
            }
        }

        private void BNSave_Click(object sender, EventArgs e)
        {
            int i = Templates.IndexOfName(EBTPL.Text);
            if (i != -1)
            {
                if (MessageBox.Show("Template " + EBTPL.Text + " already exists! Replace?", "Warning", MessageBoxButtons.YesNo) != System.Windows.Forms.DialogResult.Yes)
                {
                    DialogResult = DialogResult.None;
                    return;
                }

                Templates.All.RemoveAt(i);
            }

            string s = EBTPL.Text.PadRight(18) + " " + EBParent.Text.PadRight(18);
            foreach (string _t in MMFields.Lines)
            {
                string t = _t;

                for (int j = 0; j < t.Length; j++)
                {
                    if (t[j] == ' ' || t[j] == '\t')
                    {
                        t = t.Remove(j, 1);
                        j--;
                    }
                }

                s += " " + t;
            }

            i = Templates.AddFromString(s);
            if (i != -1)
            {
                JKTemplate tpl = Templates.All[i];
                tpl.desc = EBDesc.Text;

                string[] boxvals = EBBBox.Text.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                if (boxvals.Length != 6)
                    tpl.box = Box3.Empty;
                else
                    tpl.box = new Box3(double.Parse(boxvals[0]), double.Parse(boxvals[1]), double.Parse(boxvals[2]), double.Parse(boxvals[3]), double.Parse(boxvals[4]), double.Parse(boxvals[5]));

                Templates.SaveToFile(System.IO.Path.Combine(GlobalVars.ProjectDir, GlobalVars.IsMOTS ? "mots.tpl" : "master.tpl"));

                foreach (JKThing th in GlobalVars.Level.things)
                    if (th.name.Equals(tpl.Name, StringComparison.InvariantCultureIgnoreCase))
                        GlobalVars.NedMain.ThingChanged(th);
            }

            changed = false;
        }

        private void BNUpdate_Click(object sender, EventArgs e)
        {
            if (!CheckChanged())
                return;

            int i = Templates.IndexOfName(EBTPL.Text);
            List<string> lines = new List<string>();
            if (i != -1)
            {
                JKTemplate tpl = Templates.All[i];

                EBDesc.Text = tpl.desc;
                EBBBox.Text = tpl.box.Left.ToString() + " " + tpl.box.Back.ToString() + " " + tpl.box.Bottom.ToString() + " " + tpl.box.Right.ToString() + " " + tpl.box.Front.ToString() + " " + tpl.box.Top.ToString();
                foreach (TemplateValue tv in tpl.vals)
                    lines.Add(tv.name + "=" + tv.AsString);
            }
            MMFields.Lines = lines.ToArray();

            changed = false;
        }

        private void BN3DO_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            ResPicker.Pick3DO(ref s);
            if (string.IsNullOrEmpty(s))
                return;

            JKModel model = JKModel.Load3DO(s);
            if (model != null)
            {
                Box3 bbox = model.GetBBox();
                EBBBox.Text = bbox.Left.ToString() + " " + bbox.Back.ToString() + " " + bbox.Bottom.ToString() + " " + bbox.Right.ToString() + " " + bbox.Front.ToString() + " " + bbox.Top.ToString();
                SetField("size", model.FindRadius().ToString());
                SetField("movesize", model.FindRadius().ToString());
            }
            SetField("model3d", s);
        }

        private void BNAI_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            ResPicker.PickAI(ref s);
            if (string.IsNullOrEmpty(s))
                return;

            SetField("aiclass", s);
        }

        private void BNSND_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            ResPicker.PickSND(ref s);
            if (string.IsNullOrEmpty(s))
                return;

            SetField("soundclass", s);
        }

        private void BNPup_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            ResPicker.PickPUP(ref s);
            if (string.IsNullOrEmpty(s))
                return;

            SetField("puppet", s);
        }

        private void BNCog_Click(object sender, EventArgs e)
        {
            string s = string.Empty;
            ResPicker.PickCOG(ref s);
            if (string.IsNullOrEmpty(s))
                return;

            SetField("cog", s);
        }
    }
}
