﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Smith.NED
{
    public class JKTemplate
    {
        public string Name;
        public string Parent;
        public TemplateValues vals = new TemplateValues();
        public string desc;
        public Box3 box;

        public string AsString
        {
            get
            {
                string s = Name.PadRight(17) + " " + Parent.PadRight(17) + " ";
                foreach (TemplateValue tv in vals)
                    s += " " + tv.name + "=" + tv.AsString;
                return s;
            }
        }

        public string resourceSummary
        {
            get
            {
                JKTemplate[] tpls = TemplateChain;


                return "import/export disabled";
            }
        }

        public JKLevel.ClassSaveToJKL.UsedListMatrix BuildUsedListMatrix(bool generateStaticJKL)
        {
            JKLevel.ClassSaveToJKL.UsedListMatrix matrix = new JKLevel.ClassSaveToJKL.UsedListMatrix(generateStaticJKL);

            foreach (JKTemplate t in TemplateChain)
                matrix.AddTemplate(t.Name);

            matrix.Finish(false);

            return matrix;
        }

        public JKTemplate[] TemplateChain
        {
            get
            {
                List<JKTemplate> templates = new List<JKTemplate>();
                for (JKTemplate curTemp = this; curTemp != null; )
                {
                    templates.Insert(templates.Count, curTemp);

                    int parentIndex = Templates.IndexOfName(curTemp.Parent);
                    if (parentIndex == -1)
                        break;

                    curTemp = Templates.All[parentIndex];
                }
                return templates.ToArray();
            }
        }

        public static JKTemplate FromSmithTemplate(Template s)
        {
            JKTemplate t = new JKTemplate();

            t.Name = s.Name;
            t.Parent = s.ParentName;
            foreach(string key in s.Properties.Keys)
            {
                foreach (Flex val in s.Properties[key])
                {
                    TemplateValue tv = new TemplateValue();
                    t.vals.Add(tv);

                    tv.name = key;
                    tv.vType = Values.GetTPLVType(tv.name);
                    tv.adjType = Values.GetTPLType(tv.name);
                    tv.Val(val.Text);
                }
            }

            return t;
        }
    }

    public static class Templates
    {
        public static List<JKTemplate> All = new List<JKTemplate>();

        public static int IndexOfName(string name)
        {
            if (string.IsNullOrEmpty(name) || name.Equals("none", StringComparison.InvariantCultureIgnoreCase))
                return -1;

            for (int x = 0; x < All.Count; x++)
                if (All[x].Name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                    return x;
            return -1;
        }

        public static JKTemplate Get(int i)
        {
            if (i < 0 || i >= All.Count)
                return null;

            return All[i];
        }

        public static TemplateValue GetTPLField(string tpl, string field)
        {
            for (int n = 0; n < 100; n++)
            {
                if (tpl.Equals("none", StringComparison.InvariantCultureIgnoreCase))
                    return null;

                int i = IndexOfName(tpl);
                if (i == -1)
                    return null;

                JKTemplate tp = All[i];
                foreach (TemplateValue tv in tp.vals)
                    if (tv.name.Equals(field, StringComparison.InvariantCultureIgnoreCase))
                        return tv;

                tpl = tp.Parent;
            }
            return null;
        }

        public static void Clear()
        {
            All.Clear();
        }

        public static void AddFromJKL(string jklpath)
        {
            // not very efficient in performance but efficient in that we can reuse some simth code
            using (BaseFile bf = FileOperations.OpenGameFile(jklpath))
            {
                MemoryStream ms = new MemoryStream(bf.Read(bf.Length));

                Tokenizer t = new Tokenizer(ms);

                if (t.SeekToLine("section: templates"))
                //Log.Warning("Couldn't find section: templates");
                //else
                {
                    //Log.Info("Parsing section: templates");

                    t.EatToken(2);
                    int numTemplates = t.ReadInt();
                    //JK.AddTemplates(Templates.ToArray());
                    //Templates.AddRange(JK.ParseTemplates(t, numTemplates));

                    foreach (Template src in JK.ParseTemplates(t, numTemplates))
                    {
                        int i = IndexOfName(src.Name);
                        if (i != -1)
                        {
                            // skip or keep override?
                            Utils.PanMessage(Utils.MessageType.Warning, "Template already specified " + src.Name + ", removing previous");
                            All.RemoveAt(i);
                            //continue;
                        }
                        All.Add(JKTemplate.FromSmithTemplate(src));
                    }
                }
            }
        }

        public static int AddFromString(string ln, string desc = null, Box3? box = null)
        {
            string[] parts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
            if (parts.Length < 2)
                return -1;

            if (IndexOfName(parts[0]) != -1)
                return -1;

            JKTemplate tpl = new JKTemplate();
            tpl.Name = parts[0];
            int i = All.Count;
            All.Add(tpl);

            tpl.Parent = parts[1];

            for (int x = 2; x < parts.Length; x++)
            {
                string[] valpair = parts[x].Split(new char[] { '=' }, StringSplitOptions.RemoveEmptyEntries);
                if (valpair.Length != 2)
                    continue;

                TemplateValue tv = new TemplateValue();
                tpl.vals.Add(tv);

                tv.name = valpair[0].Trim();
                tv.vType = Values.GetTPLVType(tv.name);
                tv.adjType = Values.GetTPLType(tv.name);
                tv.stringval = valpair[1].Trim();
            }

            tpl.desc = desc ?? string.Empty;
            tpl.box = box ?? Box3.Empty;

            return i;
        }

        public static void SaveToFile(string filepath)
        {
            using (StreamWriter sw = File.CreateText(filepath))
                foreach (JKTemplate t in All)
                {
                    sw.WriteLine("# DESC " + t.desc);
                    sw.WriteLine("# BBOX: " + t.box.Left.ToString("0.######") + " " + t.box.Back.ToString("0.######") + " " + t.box.Bottom.ToString("0.######") + " " +
                                t.box.Right.ToString("0.######") + " " + t.box.Front.ToString("0.######") + " " + t.box.Top.ToString("0.######"));
                    sw.WriteLine(t.AsString);
                }
        }

        public static void LoadFromFile(string name)
        {
            Clear();
            string cdesc = string.Empty;
            Box3 cbox = Box3.Empty;
            using (StreamReader sr = new StreamReader(Jeddata.GetFile(name)))
                while (!sr.EndOfStream)
                {
                    string ln = sr.ReadLine();
                    string[] parts = ln.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.Length == 0)
                        continue;

                    if (parts[0] == "#")
                    {
                        if (parts[1].Equals("DESC:", StringComparison.InvariantCultureIgnoreCase))
                        {
                            int i = ln.IndexOf("DESC:", StringComparison.InvariantCultureIgnoreCase);
                            cdesc = ln.Substring(i + 5).Trim();
                        }
                        else if (parts[1].Equals("DESC", StringComparison.InvariantCultureIgnoreCase))
                        {
                            int i = ln.IndexOf("DESC", StringComparison.InvariantCultureIgnoreCase);
                            cdesc = ln.Substring(i + 4).Trim();
                        }
                        else
                        {
                            if (parts.Length >= 8)
                            {
                                cbox.Left = double.Parse(parts[2]);
                                cbox.Back = double.Parse(parts[3]);
                                cbox.Bottom = double.Parse(parts[4]);
                                cbox.Right = double.Parse(parts[5]);
                                cbox.Front = double.Parse(parts[6]);
                                cbox.Top = double.Parse(parts[7]);
                            }
                        }

                        continue;
                    }

                    AddFromString(ln, cdesc, cbox);
                    cdesc = string.Empty;
                    cbox = Box3.Empty;
                }
        }
    }
}
