﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Smith.NED
{
    public partial class Tools : UserControl
    {
        public Tools()
        {
            InitializeComponent();
        }

        public new void Show()
        {
            if (GlobalVars.NedMain != null)
                GlobalVars.NedMain.tabControl1.SelectedTab = GlobalVars.NedMain.tabTools;
        }

        NedMain.ToolAxis RGIndex
        {
            get
            {
                if (RBXAxis.Checked)
                    return NedMain.ToolAxis.X;

                if (RBYAxis.Checked)
                    return NedMain.ToolAxis.Y;

                if (RBZAxis.Checked)
                    return NedMain.ToolAxis.Z;

                if (RBGrid.Checked)
                    return NedMain.ToolAxis.Grid;

                return 0;
            }
        }

        private void BNRotate_Click(object sender, EventArgs e)
        {
            double angle;
            if(!double.TryParse(EBAngle.Text, out angle))
                return;

            GlobalVars.NedMain.RotateObject(angle, RGIndex);
        }

        private void BNFlip_Click(object sender, EventArgs e)
        {
            GlobalVars.NedMain.FlipObject(RGIndex);
        }

        private void BNScale_Click(object sender, EventArgs e)
        {
            double sfactor;
            if (!double.TryParse(EBSfactor.Text, out sfactor))
                return;

            NedMain.ToolAxis how = 0;
            if (CBScaleTX.Checked)
                how |= NedMain.ToolAxis.TX;

            if (!CBScaleXYZ.Checked)
                how |= RGIndex;// way cleaner.. love it

            GlobalVars.NedMain.ScaleObject(sfactor, how);
        }

        private void BNTranslate_Click(object sender, EventArgs e)
        {
            Vec3 d = Vec3.Zero;
            double.TryParse(EBDX.Text, out d.x);
            double.TryParse(EBDY.Text, out d.y);
            double.TryParse(EBDZ.Text, out d.z);

            GlobalVars.NedMain.TranslateObject(d);
        }

        public JKLevel level
        {
            get
            {
                return GlobalVars.Level;
            }
        }

        private void BNStraighten_Click(object sender, EventArgs e)
        {
            foreach (JKSector sec in level.sectors)
                foreach (JKSurface surf in sec.surfaces)
                    surf.RecalcAll();

            GlobalVars.Preview3D.ReloadLevel();
            GlobalVars.NedMain.LevelChanged(false);
        }

        private void BNDoor_Click(object sender, EventArgs e)
        {
            GlobalVars.NedMain.MakeDoor();
        }

        private void BNCalcAngle_Click(object sender, EventArgs e)
        {
            double a = 0.0;
            double.TryParse(EBAngle.Text, out a);
            GlobalVars.NedMain.CalcAngle(ref a);
            EBAngle.Text = a.ToString();

            RBGrid.Checked = true;
            GlobalVars.NedMain.SetMapMode(NedMain.MapMode.Sector);
        }

        private void BNSetGrid_Click(object sender, EventArgs e)
        {
            Renderer r = GlobalVars.NedMain.renderer;

            double.TryParse(EBGridX.Text, out r.GridPos.x);
            double.TryParse(EBGridY.Text, out r.GridPos.y);
            double.TryParse(EBGridZ.Text, out r.GridPos.z);

            Orient o = Orient.Zero;
            double.TryParse(EBGridPCH.Text, out o.pitch);
            double.TryParse(EBGridYAW.Text, out o.yaw);
            double.TryParse(EBGridROL.Text, out o.roll);

            Vec3 x = Vec3.PosX;
            Vec3 y = Vec3.PosY;
            Vec3 z = Vec3.PosZ;
            Utils.RotateVector(ref x, o.pitch, 0.0, 0.0); Utils.RotateVector(ref x, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref x, 0.0, 0.0, o.yaw/*WTF*/);
            Utils.RotateVector(ref y, o.pitch, 0.0, 0.0); Utils.RotateVector(ref y, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref y, 0.0, 0.0, o.yaw/*WTF*/);
            Utils.RotateVector(ref z, o.pitch, 0.0, 0.0); Utils.RotateVector(ref z, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref z, 0.0, 0.0, o.yaw/*WTF*/);

            r.SetGridNormal(z);
            r.SetGridXNormal(x);
        }

        private void Button1_Click(object sender, EventArgs e)
        {
            Vec3 c;
            double.TryParse(EBCamX.Text, out c.x);
            double.TryParse(EBCamY.Text, out c.y);
            double.TryParse(EBCamZ.Text, out c.z);

            Orient o = Orient.Zero;
            double.TryParse(EBCamPCH.Text, out o.pitch);
            double.TryParse(EBCamYAW.Text, out o.yaw);
            double.TryParse(EBCamROL.Text, out o.roll);

            GlobalVars.NedMain.SetCam(c, o);
        }

        private void Button2_Click(object sender, EventArgs e)
        {
            Vec3 c;
            double.TryParse(EBP3DX.Text, out c.x);
            double.TryParse(EBP3DY.Text, out c.y);
            double.TryParse(EBP3DZ.Text, out c.z);

            Orient o = Orient.Zero;
            double.TryParse(EBP3DPCH.Text, out o.pitch);
            double.TryParse(EBP3DYAW.Text, out o.yaw);

            GlobalVars.Preview3D.SetCam(c, o);
        }

        private void Pages_SelectedIndexChanged(object sender, EventArgs e)
        {
            timer1.Enabled = (Pages.SelectedTab == GridNCamera);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (!CBLiveUpdate.Checked)
                return;

            // dont update if any control has focus
            if (GridNCamera.ContainsFocus)
                return;

            if (GlobalVars.NedMain == null)
                return;

            Renderer r = GlobalVars.NedMain.renderer;
            if (r == null)
                return;

            EBGridX.Text = r.GridPos.x.ToString();
            EBGridY.Text = r.GridPos.y.ToString();
            EBGridZ.Text = r.GridPos.z.ToString();

            // dunno how we're supposed to deal with this.  might wait until the framework is improved first
#if false
            Orient o = Orient.Zero;
            double.TryParse(EBGridPCH.Text, out o.pitch);
            double.TryParse(EBGridYAW.Text, out o.yaw);
            double.TryParse(EBGridROL.Text, out o.roll);

            Vec3 x = Vec3.PosX;
            Vec3 y = Vec3.PosY;
            Vec3 z = Vec3.PosZ;
            Utils.RotateVector(ref x, o.pitch, 0.0, 0.0); Utils.RotateVector(ref x, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref x, 0.0, 0.0, o.yaw/*WTF*/);
            Utils.RotateVector(ref y, o.pitch, 0.0, 0.0); Utils.RotateVector(ref y, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref y, 0.0, 0.0, o.yaw/*WTF*/);
            Utils.RotateVector(ref z, o.pitch, 0.0, 0.0); Utils.RotateVector(ref z, 0.0, o.roll/*WTF*/, 0.0); Utils.RotateVector(ref z, 0.0, 0.0, o.yaw/*WTF*/);

            r.SetGridNormal(z);
            r.SetGridXNormal(x);
#endif


            EBCamX.Text = r.CamPos.x.ToString();
            EBCamY.Text = r.CamPos.y.ToString();
            EBCamZ.Text = r.CamPos.z.ToString();

            Orient co = GlobalVars.NedMain.camOrient;
            EBCamPCH.Text = co.pitch.ToString();
            EBCamYAW.Text = co.yaw.ToString();
            EBCamROL.Text = co.roll.ToString();


            Vec3 p3dpos = Vec3.Zero;
            Orient p3dorient = Orient.Zero;
            if(GlobalVars.Preview3D != null && GlobalVars.Preview3D.renderer != null)
            {
                p3dpos = GlobalVars.Preview3D.renderer.CamPos;
                p3dorient = GlobalVars.Preview3D.renderer.CamOrient;
            }

            EBP3DX.Text = p3dpos.x.ToString();
            EBP3DY.Text = p3dpos.y.ToString();
            EBP3DZ.Text = p3dpos.z.ToString();

            EBP3DPCH.Text = p3dorient.pitch.ToString();
            EBP3DYAW.Text = p3dorient.yaw.ToString();
        }

        private void Pages_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            //GlobalVars.NedMain.DoPreviewKeyDown(e);
        }

        private void Pages_KeyDown(object sender, KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyDown(e);
        }

        private void Pages_KeyUp(object sender, KeyEventArgs e)
        {
            //GlobalVars.NedMain.DoKeyUp(e);
        }
    }
}
