﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace Smith.NED
{
    public partial class UrqForm : SubForm
    {
        public UrqForm()
        {
            InitializeComponent();
        }

        private TreeNode AddNode(JKModel.HNodes hnodes, int nnode, TreeNode[] tnodes, ref int recurdepth)
        {
            recurdepth++;
            if (recurdepth > 50)
            {
                Utils.PanMessage(Utils.MessageType.Warning, "Possible circular hierarchy! Node " + hnodes[nnode].nodename);
                recurdepth--;
                return null;
            }

            if (tnodes[nnode] != null)
            {
                recurdepth--;
                return tnodes[nnode];
            }

            JKModel.HNode node = hnodes[nnode];
            TreeNode pnode;
            if (node.parent < 0 || node.parent >= hnodes.Count)
                pnode = null;
            else
                pnode = AddNode(hnodes, node.parent, tnodes, ref recurdepth);

            TreeNode tnode = (pnode == null) ? TVNodes.Nodes.Add(node.nodename) : pnode.Nodes.Add(node.nodename);
            tnode.Tag = node;
            tnodes[nnode] = tnode;
            recurdepth--;

            return tnode;
        }

        public void Reload()
        {
            int recurdepth = 0;
            JKModel.HNodes hnodes = GlobalVars.Level.hnodes;
            JKModel.HNode lastselected = null;
            if (TVNodes.SelectedNode != null)
                lastselected = TVNodes.SelectedNode.Tag as JKModel.HNode;

            CBParent.Items.Clear();
            CBParent.Items.Add("None");
            foreach (JKModel.HNode n in hnodes)
                CBParent.Items.Add(n.nodename);

            TVNodes.Nodes.Clear();
            TreeNode[] tnodes = new TreeNode[256];
            for (int x = 0; x < tnodes.Length; x++)
                tnodes[x] = null;
            for(int i=0; i<hnodes.Count; i++)
                AddNode(hnodes, i, tnodes, ref recurdepth);
            TVNodes.ExpandAll();
            SelectNode(lastselected);
        }

        bool updating = false;

        public void SelectNode(JKModel.HNode node)
        {
            foreach(TreeNode tn in TVNodes.Nodes)
                if (object.ReferenceEquals(tn.Tag, node))
                {
                    TVNodes.SelectedNode = tn;
                    return;
                }
        }

        public void GotoNode(int n)
        {
            SelectNode(GlobalVars.Level.hnodes[n]);
        }

        void SetCBItem(ComboBox cb, int n)
        {
            updating = true;
            cb.SelectedIndex = n;
            updating = false;
        }

        void SetEditText(TextBox ed, string s)
        {
            updating = true;
            ed.Text = s;
            updating = false;
        }

        JKModel.HNode lasthnode = null;

        public void LoadNode(TreeNode node)
        {
            if (node == null)
                return;

            JKModel.HNode hnode = node.Tag as JKModel.HNode;
            lasthnode = hnode;
            SetCBItem(CBParent, hnode.parent + 1);
            SetEditText(EBXYZ, Utils.dstr(hnode.pos.x, 6) + " " + Utils.dstr(hnode.pos.y, 6) + " " + Utils.dstr(hnode.pos.z, 6));
            SetEditText(EBPYR, Utils.dstr(hnode.orient.pitch, 6) + " " + Utils.dstr(hnode.orient.yaw, 6) + " " + Utils.dstr(hnode.orient.roll, 6));
            SetEditText(EBType, hnode.ntype.ToString());
        }

        public void UpdateMeshOffsets()
        {
            if (GlobalVars.NedMain.ExtraObjsName.Equals("Mesh Offset", StringComparison.InvariantCultureIgnoreCase))
                LoadMeshOffsets(false);
        }

        public void LoadMeshOffsets(bool reload)
        {
            if (reload)
                GlobalVars.NedMain.ClearExtraObjs();
            GlobalVars.NedMain.ExtraObjsName = "Mesh Offset";
            for(int i=0; i<GlobalVars.Level.hnodes.Count; i++)
            {
                JKModel.HNode hnode = GlobalVars.Level.hnodes[i];

                Vec3 a = hnode.pos;
                if (hnode.parent >= 0)
                    Utils.OffsetByNode(GlobalVars.Level, hnode.parent, ref a);
                if (reload)
                    GlobalVars.NedMain.AddExtraVertex(a, hnode.nodename);
                else
                {
                    NedMain.ExtraVertex v = GlobalVars.NedMain.ExtraObjs[i] as NedMain.ExtraVertex;
                    if (v != null)
                    {
                        v.v = a;
                    }
                }
            }
        }

        private void BNSave_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Title = "Save Hierarchy";
            sfd.Filter = "3DH Files|*.3dh";
            sfd.DefaultExt = ".3dh";
            if (sfd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            using (StreamWriter sw = File.CreateText(sfd.FileName))
            {
                JKModel.HNodes hnodes = GlobalVars.Level.hnodes;

                sw.WriteLine("SECTION: HIERARCHYDEF");
                sw.WriteLine("HIERARCHY NODES " + hnodes.Count.ToString());
                for (int i = 0; i < hnodes.Count; i++)
                    sw.WriteLine(hnodes[i].ToString(i));
            }
        }

        private void BNLoad_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Title = "Load Hierarchy";
            ofd.Filter = "3DH Files|*.3dh;*.3do";
            ofd.DefaultExt = ".3dh";
            if (ofd.ShowDialog() != System.Windows.Forms.DialogResult.OK)
                return;

            Undo.ClearNodesUndo();

            using (Tokenizer t = new Tokenizer(File.OpenRead(ofd.FileName)))
            {
                JKModel.HNodes hnodes = GlobalVars.Level.hnodes;
                hnodes.Clear();

                if (t.SeekToLine("SECTION: HIERARCHYDEF"))
                {
                    t.EatToken(2);
                    int n = t.ReadInt();
                    for (int i = 0; i < n; i++)
                    {
                        JKModel.HNode hnode = GlobalVars.Level.New3DONode();
                        hnode.FromString(t.ReadLine());
                        hnodes.Add(hnode);
                        hnode.pivot = Vec3.Zero;
                    }
                }
            }

            Reload();
        }

        public void MoveOffset(object ex, bool continued)
        {
            if (!(ex is NedMain.ExtraVertex))
                return;
            if (!GlobalVars.NedMain.IsSelValid)
                return;

            if (!continued)
                Undo.StartUndoRec("Node XYZ change");

            int cur_ex = GlobalVars.NedMain.Cur_EX;
            Undo.SaveNodeUndo(GlobalVars.Level.hnodes[cur_ex], Undo.Change.Changed);

            NedMain.ExtraVertex evx = ex as NedMain.ExtraVertex;
            JKModel.HNode node = GlobalVars.Level.hnodes[cur_ex];

            Vec3 a = evx.v;
            if (node.parent >= 0)
                Utils.UnOffsetByNode(GlobalVars.Level, node.parent, ref a);
            node.pos = a;

            UpdateMeshOffsets();
            GlobalVars.NedMain.Cur_EX = cur_ex;
        }

        private void BNEditOffsets_Click(object sender, EventArgs e)
        {
            LoadMeshOffsets(true);
            GlobalVars.NedMain.OnExtraMove = MoveOffset;
            GlobalVars.NedMain.SetMapMode(NedMain.MapMode.Extra);
            GlobalVars.NedMain.Focus();
        }

        private void EBXYZ_TextChanged(object sender, EventArgs e)
        {
            if (updating)
                return;

            if (lasthnode == null)
                return;

            Undo.StartUndoRec("Node XYZ change");
            Undo.SaveNodeUndo(lasthnode, Undo.Change.Changed);
            lasthnode.pos = Utils.ValVector(EBXYZ.Text) ?? lasthnode.pos;
            UpdateMeshOffsets();
        }

        private void EBXYZ_Leave(object sender, EventArgs e)
        {
            LoadNode(TVNodes.SelectedNode);
        }

        private void EBPYR_TextChanged(object sender, EventArgs e)
        {
            if (updating)
                return;

            if (lasthnode == null)
                return;

            Undo.StartUndoRec("Node PYR change");
            Undo.SaveNodeUndo(lasthnode, Undo.Change.Changed);
            lasthnode.orient = Utils.ValOrient(EBPYR.Text) ?? lasthnode.orient;
            UpdateMeshOffsets();
        }

        private void EBType_TextChanged(object sender, EventArgs e)
        {
            if (updating)
                return;

            if (lasthnode == null)
                return;

            Undo.StartUndoRec("Node change");
            Undo.SaveNodeUndo(lasthnode, Undo.Change.Changed);
            int.TryParse(EBType.Text, out lasthnode.ntype);
        }

        private void TVNodes_AfterLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
            if (e.Node == null)
                return;

            JKModel.HNode node = e.Node.Tag as JKModel.HNode;

            Undo.StartUndoRec("Node name change");
            Undo.SaveNodeUndo(/*lasthnode*/node, Undo.Change.Changed);
            node.nodename = e.Label;
        }

        private void CBParent_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (updating)
                return;

            if (lasthnode == null)
                return;

            Undo.StartUndoRec("Node change");
            Undo.SaveNodeUndo(lasthnode, Undo.Change.Changed);
            SetNodeParent(lasthnode, CBParent.SelectedIndex - 1);
        }

        public void SetNodeParent(JKModel.HNode node, int parent)
        {
            node.parent = parent;
            Reload();
        }

        private void AddMissing_Click(object sender, EventArgs e)
        {
            Undo.StartUndoRec("Add Nodes");

            if (GlobalVars.Level.hnodes.Count == 0)
            {
                JKModel.HNode hnode = GlobalVars.Level.New3DONode();
                hnode.nodename = "$$DUMMY";
                GlobalVars.Level.hnodes.Add(hnode);
                Undo.SaveNodeUndo(hnode, Undo.Change.Added);
            }

            for (int i = 0; i < GlobalVars.Level.layers.Count; i++)
            {
                string s = GlobalVars.Level.layers[i];
                bool found = false;
                for(int j=0; j<GlobalVars.Level.hnodes.Count; j++)
                    if (s.Equals(GlobalVars.Level.hnodes[j].nodename, StringComparison.InvariantCultureIgnoreCase))
                    {
                        found = true;
                        break;
                    }

                if (!found)
                {
                    JKModel.HNode hnode = GlobalVars.Level.New3DONode();
                    hnode.nodename = s;
                    hnode.parent = 0;
                    GlobalVars.Level.hnodes.Add(hnode);
                    Undo.SaveNodeUndo(hnode, Undo.Change.Added);
                }
            }

            Reload();
        }

        private void BNAddNode_Click(object sender, EventArgs e)
        {
            Undo.StartUndoRec("Add Node");
            JKModel.HNode hnode = GlobalVars.Level.New3DONode();
            hnode.nodename = "NewNode";
            GlobalVars.Level.hnodes.Add(hnode);
            Undo.SaveNodeUndo(hnode, Undo.Change.Added);
            Reload();
            SelectNode(hnode);
            TVNodes.Focus();
            TVNodes.SelectedNode.BeginEdit();
        }

        private void BNDelNode_Click(object sender, EventArgs e)
        {
            Undo.ClearNodesUndo();
            int n = -1;
            for(int i=0; i<GlobalVars.Level.hnodes.Count; i++)
                if (object.ReferenceEquals(GlobalVars.Level.hnodes[i], lasthnode))
                {
                    n = i;
                    break;
                }

            if (n == -1)
                return;

            for (int i = 0; i < GlobalVars.Level.hnodes.Count; i++)
            {
                JKModel.HNode hnode = GlobalVars.Level.hnodes[i];
                if (hnode.parent >= n)
                {
                    if (hnode.parent == n)
                        hnode.parent = -1;
                    else
                        hnode.parent--;
                }
            }

            GlobalVars.Level.hnodes.RemoveAt(n);
            lasthnode = null;//?
            Reload();
        }


    }
}
