﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;

namespace Smith.NED
{
    public static class Utils
    {
        public const bool rs_surfs = true;

        public const bool newLight = true;

        public enum ScaleDataMode
        {
            XYZ,
            X,
            Y,
            Z,
            Vector
        }

        public class ScaleData
        {
            public Vec3 c;
            public double sfactor;
            public Vec3 vec;
            public ScaleDataMode how;
        }

        public static void FlipSectorsOverPlane(JKLevel lev, SCMultiSel scsel, Vec3 pnorm, Vec3 p)
        {
            ClearVXMarks(lev);
            for (int i = 0; i < scsel.Count; i++)
            {
                JKSector sec = lev.sectors[i];
                Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);

                foreach(JKSurface surf in sec.surfaces)
                    if (surf.adjoin != null)
                    {
                        if (scsel.FindSC(surf.adjoin.sector.num) == -1)
                            Unadjoin(surf);
                    }
            }

            for (int i = 0; i < scsel.Count; i++)
            {
                JKSector sec = lev.sectors[scsel.GetSC(i)];

                foreach (JKVertex v in sec.vertices)
                    FlipVertexOverPlane(v, pnorm, p);

                foreach (JKSurface surf in sec.surfaces)
                    FlipSurface(surf);

                GlobalVars.NedMain.SectorChanged(sec, true);
            }
        }

        public static void FlipSurface(JKSurface surf)
        {
            for (int i = 0; i < (surf.vertices.Count / 2); i++)
            {
                int n = surf.vertices.Count - i - 1;
                Vertex vx = surf.vertices[i];
                surf.vertices[i] = surf.vertices[n];
                surf.vertices[n] = vx;

                TexVertex tv = surf.texvertices[i];
                surf.texvertices[i] = surf.texvertices[n];
                surf.texvertices[n] = tv;
            }

            surf.Recalc();

            if (IsTXFlipped(surf))
                surf.surfaceFlags |= Surface.Flag.Flipped;
            else
                surf.surfaceFlags &= ~Surface.Flag.Flipped;
        }

        public static void FlipSurfaces(JKLevel lev, SFMultiSel sfsel)
        {
            for (int i = 0; i < sfsel.Count; i++)
            {
                int sc, sf;
                sfsel.GetSCSF(i, out sc, out sf);
                JKSurface surf = lev.sectors[sc].surfaces[sf];
                FlipSurface(surf);
                GlobalVars.NedMain.SectorChanged(surf.sector, false/*prolly not*/);
            }
        }

        public static void FlipVertexOverPlane(JKVertex v, Vec3 pnorm, Vec3 c)
        {
            if (v.mark != 0)
                return;

            double d = pnorm * (v.v - c);
            v.v -= pnorm * 2.0 * d;
            v.mark = 1;
        }

        public static Vec3 CalcSecCenter(JKSector sec)
        {
            Vec3 s = Vec3.Zero;
            foreach (JKVertex v in sec.vertices)
                s += v.v;
            return s / (double)sec.vertices.Count;
        }

        public static void FlipSectors(JKLevel lev, SCMultiSel scsel, Vec3 c, NedMain.ToolAxis how)
        {
            ClearVXMarks(lev);

            for (int i = 0; i < scsel.Count; i++)
            {
                JKSector sec = lev.sectors[scsel.GetSC(i)];
                Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);

                foreach (JKSurface surf in sec.surfaces)
                    if (surf.adjoin != null)
                    {
                        if (scsel.FindSC(surf.adjoin.sector.num) == -1)
                            Unadjoin(surf);
                    }
            }

            for (int i = 0; i < scsel.Count; i++)
            {
                JKSector sec = lev.sectors[scsel.GetSC(i)];

                foreach (JKVertex v in sec.vertices)
                    FlipVertex(v, c, how);

                foreach (JKSurface surf in sec.surfaces)
                    FlipSurface(surf);

                GlobalVars.NedMain.SectorChanged(sec, true);
            }
        }

        public static void FlipVertex(JKVertex v, Vec3 c, NedMain.ToolAxis how)
        {
            if (v.mark != 0)
                return;

            switch (how)
            {
                case NedMain.ToolAxis.X: v.v.x = c.x - (v.v.x - c.x); /* what the fuck lol */ break;
                case NedMain.ToolAxis.Y: v.v.y = c.y - (v.v.y - c.y); /* what the fuck lol */ break;
                case NedMain.ToolAxis.Z: v.v.z = c.z - (v.v.z - c.z); /* what the fuck lol */ break;
            }
            v.mark = 1;
        }

        public static void FlipThings(JKLevel lev, THMultiSel thsel, Vec3 c, NedMain.ToolAxis how)
        {
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing th = lev.things[thsel.GetTH(i)];
                Undo.SaveThingUndo(th, Undo.Change.Changed);
                switch (how)
                {
                    case NedMain.ToolAxis.X: th.pos.x = c.x - (th.pos.x - c.x); /*what the fuck lol*/ break;
                    case NedMain.ToolAxis.Y: th.pos.y = c.y - (th.pos.y - c.y); /*what the fuck lol*/ break;
                    case NedMain.ToolAxis.Z: th.pos.z = c.z - (th.pos.z - c.z); /*what the fuck lol*/ break;
                }
                GlobalVars.NedMain.ThingChanged(th);
            }
        }

        public static void ScaleVertex(JKVertex v, ScaleData sd)
        {
            if (v.mark != 0)
                return;
            switch (sd.how)
            {
                case ScaleDataMode.XYZ:
                    v.v = sd.c + sd.sfactor * (v.v - sd.c);
                    break;

                case ScaleDataMode.X: v.v.x = sd.c.x + sd.sfactor * (v.v.x - sd.c.x); break;
                case ScaleDataMode.Y: v.v.y = sd.c.y + sd.sfactor * (v.v.y - sd.c.y); break;
                case ScaleDataMode.Z: v.v.z = sd.c.z + sd.sfactor * (v.v.z - sd.c.z); break;

                case ScaleDataMode.Vector:
                    {
                        double d = (v.v - sd.c) * sd.vec;
                        v.v += d * sd.vec;
                    }
                    break;
            }
            v.mark = 1;
        }

        public static void ScaleThings(JKLevel lev, THMultiSel thsel, ScaleData sd)
        {
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing th = lev.things[thsel.GetTH(i)];
                Undo.SaveThingUndo(th, Undo.Change.Changed);
                ScaleXYZ(ref th.pos, sd);
                GlobalVars.NedMain.ThingChanged(th);
            }
        }

        public static void ScaleLights(JKLevel lev, LTMultiSel ltsel, ScaleData sd)
        {
            for (int i = 0; i < ltsel.Count; i++)
            {
                JedLight lt = lev.lights[ltsel.GetLT(i)];
                Undo.SaveLightUndo(lt, Undo.Change.Changed);
                ScaleXYZ(ref lt.Pos, sd);
                GlobalVars.NedMain.LevelChanged(true);
            }
        }

        public static void ScaleXYZ(ref Vec3 p, ScaleData sd)
        {
            switch (sd.how)
            {
                case ScaleDataMode.XYZ:
                    p = sd.c + sd.sfactor * (p - sd.c);
                    break;

                case ScaleDataMode.X: p.x = sd.c.x + sd.sfactor * (p.x - sd.c.x); break;
                case ScaleDataMode.Y: p.y = sd.c.y + sd.sfactor * (p.y - sd.c.y); break;
                case ScaleDataMode.Z: p.z = sd.c.z + sd.sfactor * (p.z - sd.c.z); break;

                case ScaleDataMode.Vector:
                    double d = sd.sfactor - (p - sd.c) * sd.vec;
                    p += d * sd.vec;
                    break;
            }
        }

        public static void ScaleFrames(JKLevel lev, FRMultiSel frsel, ScaleData sd)
        {
            for (int i = 0; i < frsel.Count; i++)
            {
                int th, fr;
                frsel.GetTHFR(i, out th, out fr);
                TemplateValue vl = lev.things[th].vals[fr];
                Undo.SaveThingUndo(lev.things[th], Undo.Change.Changed);
                Vec3 fp;
                Orient fo;
                vl.GetFrame(out fp, out fo);
                ScaleXYZ(ref fp, sd);
                vl.SetFrame(fp, fo);
                GlobalVars.NedMain.ThingChanged(lev.things[th]);
            }
        }

        public static void FlipXYZ(ref Vec3 p, Vec3 pnorm, Vec3 c)
        {
            double d = pnorm * (p - c);
            p -= pnorm * 2.0 * d;
        }

        public static void FlipThingsOverPlane(JKLevel lev, THMultiSel thsel, Vec3 pnorm, Vec3 c)
        {
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing th = lev.things[thsel.GetTH(i)];
                Undo.SaveThingUndo(th, Undo.Change.Changed);
                FlipXYZ(ref th.pos, pnorm, c);
                GlobalVars.NedMain.ThingChanged(th);
            }
        }

        public static void FlipLightsOverPlane(JKLevel lev, LTMultiSel ltsel, Vec3 pnorm, Vec3 c)
        {
            for (int i = 0; i < ltsel.Count; i++)
            {
                JedLight lt = lev.lights[ltsel.GetLT(i)];
                Undo.SaveLightUndo(lt, Undo.Change.Changed);
                FlipXYZ(ref lt.Pos, pnorm, c);
                GlobalVars.NedMain.LevelChanged(true);
            }
        }

        public static void FlipFramesOverPlane(JKLevel lev, FRMultiSel frsel, Vec3 pnorm, Vec3 c)
        {
            for (int i = 0; i < frsel.Count; i++)
            {
                int th, fr;
                frsel.GetTHFR(i, out th, out fr);
                TemplateValue tv = lev.things[th].vals[fr];
                Undo.SaveThingUndo(lev.things[th], Undo.Change.Changed);
                Vec3 fp;
                Orient fo;
                tv.GetFrame(out fp, out fo);
                FlipXYZ(ref fp, pnorm, c);
                tv.SetFrame(fp, fo);
                GlobalVars.NedMain.ThingChanged(lev.things[th]);
            }
        }

        public static void RotatePoint(Vec3 a1, Vec3 a2, double angle, ref Vec3 o)
        {
            Vec3 nz = (a2 - a1).Unit;
            double d = nz * o;
            Vec3 p;
            PlaneLineXn(nz, d, a1, a2, out p);
            Vec3 nx = o - p;
            double l = nx.Magnitude;
            if(MathLib.CompareScalar(l, 0.0) == 0)
                return;
            nx = nx.Unit;
            Vec3 ny = nz ^ nx;// should we be normalizing this?
            double u = Math.Cos(MathLib.ToRad(angle)) * l;
            double v = Math.Sin(MathLib.ToRad(angle)) * l;
            o = p + nx * u + ny * v;
        }

        public struct RotateData
        {
            public Vec3 c;
            public Mat4 m;
        }


        static RotateData PrepareRotData(Vec3 axis, Vec3 c, double angle)
        {
            RotateData rd = new RotateData();
            rd.c = c;

            Vec3 vx = Vec3.PosX;
            Vec3 vy = Vec3.PosY;
            Vec3 vz = Vec3.PosZ;

            RotatePoint(Vec3.Zero, axis, angle, ref vx);
            RotatePoint(Vec3.Zero, axis, angle, ref vy);
            RotatePoint(Vec3.Zero, axis, angle, ref vz);

            // check this in case   public Mat4(Vec3 right, Vec3 up, Vec3 forward, Vec3 pos)
            // ever changes.
            rd.m = new Mat4(vx, vy, vz, Vec3.Zero);

            return rd;
        }

        public static void RotateVertex(JKVertex v, RotateData rd)
        {
            if (v.mark != 0)
                return;

            Vec3 p = v.v - rd.c;
            p.Transform(rd.m);
            v.v = p + rd.c;
            v.mark = 1;
        }

        public static void RotSec(JKSector sec, RotateData rd)
        {
            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);
            foreach (JKVertex v in sec.vertices)
                RotateVertex(v, rd);

            foreach (JKSurface surf in sec.surfaces)
                if (surf.adjoin != null)
                {
                    Undo.SaveSecUndo(surf.adjoin.sector, Undo.Change.Changed, Undo.SectorChange.Geo);
                    foreach (JKVertex/*hope so*/ v in surf.adjoin.vertices)
                        RotateVertex(v, rd);
                }

            RecalcSector(sec);
            GlobalVars.NedMain.SectorChanged(sec, true);

            foreach (JKSurface surf in sec.surfaces)
                if (surf.adjoin != null)
                    GlobalVars.NedMain.SectorChanged(surf.adjoin.sector, true);
        }

        public static void RotateSectors(JKLevel lev, SCMultiSel scsel, Vec3 axis, Vec3 c, double angle)
        {
            RotateData rd = PrepareRotData(axis, c, angle);

            ClearVXMarks(lev);
            for (int i = 0; i < scsel.Count; i++)
                RotSec(lev.sectors[scsel.GetSC(i)], rd);
        }

        public static void RotateXYZ(ref Vec3 a, RotateData rd)
        {
            Vec3 p = a - rd.c;
            p.Transform(rd.m);
            a = p + rd.c;
        }

        public static void RotateThings(JKLevel lev, THMultiSel thsel, Vec3 axis, Vec3 c, double angle)
        {
            RotateData rd = PrepareRotData(axis, c, angle);
            for (int i = 0; i < thsel.Count; i++)
            {
                JKThing th = lev.things[thsel.GetTH(i)];
                Undo.SaveThingUndo(th, Undo.Change.Changed);
                RotateXYZ(ref th.pos, rd);
                GlobalVars.NedMain.ThingChanged(th);
            }
        }

        public static void RotateLights(JKLevel lev, LTMultiSel ltsel, Vec3 axis, Vec3 c, double angle)
        {
            RotateData rd = PrepareRotData(axis, c, angle);
            for (int i = 0; i < ltsel.Count; i++)
            {
                JedLight lt = lev.lights[ltsel.GetLT(i)];
                Undo.SaveLightUndo(lt, Undo.Change.Changed);
                RotateXYZ(ref lt.Pos, rd);
                GlobalVars.NedMain.LevelChanged(true);
            }
        }

        public static void RotateFrames(JKLevel lev, FRMultiSel frsel, Vec3 axis, Vec3 c, double angle)
        {
            RotateData rd = PrepareRotData(axis, c, angle);
            for (int i = 0; i < frsel.Count; i++)
            {
                int th, fr;
                frsel.GetTHFR(i, out th, out fr);
                TemplateValue vl = lev.things[th].vals[fr];
                Undo.SaveThingUndo(lev.things[th], Undo.Change.Changed);
                Vec3 fp;
                Orient fo;
                vl.GetFrame(out fp, out fo);
                RotateXYZ(ref fp, rd);
                vl.SetFrame(fp, fo);
                GlobalVars.NedMain.ThingChanged(lev.things[th]);
            }
        }

        public static void ScaleSectors(JKLevel lev, SCMultiSel scsel, ScaleData sd, bool scaletx)
        {
            ClearVXMarks(lev);
            for (int i = 0; i < scsel.Count; i++)
            {
                JKSector sec = lev.sectors[scsel.GetSC(i)];

                Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);
                foreach (JKVertex v in sec.vertices)
                    ScaleVertex(v, sd);

                foreach (JKSurface surf in sec.surfaces)
                {
                    if (scaletx)
                        surf.texScale /= sd.sfactor;

                    if (surf.adjoin == null)
                        continue;

                    if (scaletx)
                        surf.adjoin.texScale /= sd.sfactor;

                    foreach (JKVertex/*hope so*/ v in surf.adjoin.vertices)
                        ScaleVertex(v, sd);
                }

                RecalcSector(sec);
                GlobalVars.NedMain.SectorChanged(sec, true);

                foreach (JKSurface surf in sec.surfaces)
                    if (surf.adjoin != null)
                        GlobalVars.NedMain.SectorChanged(surf.adjoin.sector, true);
            }
        }

        public static bool IsPointWithInLine(Vec3 v, Vec3 v1, Vec3 v2)
        {
            // is this even correct?
            return (v.x >= Math.Min(v1.x, v2.x)) && (v.x <= Math.Max(v1.x, v2.x)) &&
                   (v.y >= Math.Min(v1.y, v2.y)) && (v.y <= Math.Max(v1.y, v2.y)) &&
                   (v.z >= Math.Min(v1.z, v2.z)) && (v.z <= Math.Max(v1.z, v2.z));
        }

        public static bool IsPointOnSurface(JKSurface surf, Vec3 v)
        {
            int vc = surf.vertices.Count;

            for (int i = 0; i < vc; i++)
            {
                Vec3 v1 = surf.vertices[i].v;
                Vec3 v2 = surf.vertices[NextIdx(i, vc)].v;

                Vec3 vct = surf.normal ^ (v2 - v1);
                double dist = vct * (v - v1);
                if (MathLib.CompareScalar(dist, 0.0) < 0)
                    return false;
            }

            return true;
        }

        public static bool IsSurfCrossed(JKSurface surf, Vec3 v1, Vec3 v2)
        {
            if (surf.vertices.Count == 0)
                return false;

            Vec3 v0 = surf.vertices[0].v;

            double dist1 = surf.normal * (v1 - v0);
            double dist2 = surf.normal * (v2 - v0);

            if (dist1 > 0.0001 && dist2 > 0.0001)
                return false;

            if (dist1 < -0.0001 && dist2 < -0.0001)
                return false;

            if (Math.Abs(dist1) <= 0.0001 && Math.Abs(dist2) <= 0.0001)
                return true;

            // find intersection of a line with plane
            double k = dist1 / (dist1 - dist2);
            Vec3 p = v1 + k * (v2 - v1);

            // find if intersection is on surface
            return IsPointOnSurface(surf, p);
        }

        public static Box3 CalcSurfBBox(JKSurface surf)
        {
            Box3 box = Box3.Expandable;
            foreach (Vertex v in surf.vertices)
                box.Expand(v.v);
            return box;
        }

        public static int FindSectorForXYZ(JKLevel lev, Vec3 p)
        {
            for (int i = 0; i < lev.sectors.Count; i++)
                if (IsInSector(lev.sectors[i], p))
                    return i;

            return -1;
        }

        public unsafe class CalcLightingImpl
        {
            JKLevel lev;
            SCMultiSel scs;

            public CalcLightingImpl(JKLevel _lev, SCMultiSel _scs)
            {
                lev = _lev;
                scs = _scs;
            }

            bool SurfBlocks(Vec3 v1, Vec3 v2, JKSurface surf, JKSurface asurf, Box3 lbox)
            {
                if (surf == asurf)
                    return false;

                if (asurf.adjoin != null)
                {
                    if ((asurf.adjoinFlags & Adjoin.Flag.BlocksLight) == 0)
                        return false;
                }

                LCSurfData* csd = sfd + asurf.mark;
                double d = csd->d;

                if (newLight)
                {
                    if (!DoBoxesIntersect(csd->bbox, lbox))
                        return false;
                }

                Vec3 v;
                if (!PlaneLineXn(asurf.normal, d, v1, v2, out v))
                    return false;

                if (!IsPointWithInLine(v, v1, v2))
                    return false;

                double dist = Math.Pow(v.x - v2.x, 2.0) + Math.Pow(v.y - v2.y, 2.0) + Math.Pow(v.z - v2.z, 2.0);
                if (dist < 0.0001)
                    return false;

                dist = Math.Pow(v.x - v1.x, 2.0) + Math.Pow(v.y - v1.y, 2.0) + Math.Pow(v.z - v1.z, 2.0);
                if (dist < 0.0001)
                    return false;

                return IsPointOnSurface(asurf, v);
            }

            bool IsBlocked(int s, JedLight lt, JKSurface surf, Vec3 v1, Vec3 v2)
            {
                Box3 lbox = Box3.Expandable;
                lbox.Expand(v1);
                lbox.Expand(v2);

                if (lt.mark != -1)
                    foreach (JKSurface asurf in lev.sectors[lt.mark].surfaces)
                        if (SurfBlocks(v1, v2, surf, asurf, lbox))
                            return true;

                foreach (JKSurface asurf in lev.sectors[s].surfaces)
                    if (SurfBlocks(v1, v2, surf, asurf, lbox))
                        return true;

                for (int ls = 0; ls < lev.sectors.Count; ls++)
                {
                    if (ls == s || ls == lt.mark)
                        continue;

                    foreach (JKSurface asurf in lev.sectors[ls].surfaces)
                        if (SurfBlocks(v1, v2, surf, asurf, lbox))
                            return true;
                }

                return false;
            }

            struct LCSurfData
            {
                public Box3 bbox;
                public double d;
            }

            LCSurfData* sfd;

            public void Do(bool background)
            {
                int gsfs = 0;

                Utils.ZeroVertexLights(lev, scs);

                foreach (JKSector sec in lev.sectors)
                {
                    for (int sf = 0; sf < sec.surfaces.Count; sf++)
                    {
                        JKSurface surf = sec.surfaces[sf];
                        surf.mark = gsfs + sf;
                    }
                    gsfs += sec.surfaces.Count;
                }

                LCSurfData[] __sfd = new LCSurfData[gsfs];
                fixed (LCSurfData* _sfd = __sfd)
                {
                    sfd = _sfd;

                    for (int s = 0; s < lev.sectors.Count; s++)
                    {
                        JKSector sec = lev.sectors[s];
                        for (int sf = 0; sf < sec.surfaces.Count; sf++)
                        {
                            JKSurface surf = sec.surfaces[sf];
                            LCSurfData* csd = sfd + surf.mark;
                            csd->d = surf.CalcD();
                            csd->bbox = Utils.CalcSurfBBox(surf);
                        }
                    }


                    // find sectors for lights
                    for (int l = 0; l < lev.lights.Count; l++)
                    {
                        JedLight jl = lev.lights[l];
                        jl.mark = Utils.FindSectorForXYZ(lev, jl.Pos);
                    }


                    // calc lights

                    if (!background)
                    {
                        ProgressDialog.Reset(lev.sectors.Count, 0);
                        ProgressDialog.Message("Calculating lights");
                    }

                    for (int s = 0; s < scs.Count; s++)
                    {
                        GlobalVars.NedMain.calcLightProgress = (int)Math.Round((double)s / (double)(scs.Count - 1) * 100.0);
                        if (background && GlobalVars.NedMain.WantLightCalcCancel)
                            return;

                        JKSector sec = lev.sectors[scs.GetSC(s)];

                        for (int sf = 0; sf < sec.surfaces.Count; sf++)
                        {
                            JKSurface surf = sec.surfaces[sf];

                            Vertex v0 = surf.vertices[0];

                            for (int l = 0; l < lev.lights.Count; l++)
                            {
                                JedLight lt = lev.lights[l];

                                // check if the light is in front of the surface
                                if (surf.normal * (lt.Pos - v0.v) < 0.001)
                                    continue;

                                for (int v = 0; v < surf.vertices.Count; v++)
                                {
                                    Vertex cv = surf.vertices[v];

                                    double dist = (lt.Pos - cv.v).Magnitude;
                                    if (dist >= lt.range)
                                        continue;

                                    // if light and vertex arent in teh same sector
                                    // check if the light is blocked
                                    if (s != lt.mark && (lt.flags & LightFlag.NoBlock) == 0)
                                    {
                                        if (IsBlocked(s, lt, surf, lt.Pos, cv.v))
                                            continue;
                                    }

                                    // light vertex
                                    TexVertex tv = surf.texvertices[v];
                                    double pow = Math.Pow((lt.range - dist) / lt.range, 2.0);
                                    tv.intensity += lt.intensity * pow;
                                    tv.r += lt.rgbintensity * lt.clr.r * pow;
                                    tv.g += lt.rgbintensity * lt.clr.g * pow;
                                    tv.b += lt.rgbintensity * lt.clr.b * pow;
                                }
                            }
                        }

                        if(!background)
                            ProgressDialog.Step();
                    }
                }

                CalcSectorAmbients(lev, scs);

                if(!background)
                    ProgressDialog.Cleanup();
            }
        }

        public static void CalcLighting(JKLevel lev, SCMultiSel scs, bool background)
        {
            new CalcLightingImpl(lev, scs).Do(background);
        }

        public class CalcLightingNewImpl
        {
            JKLevel lev;
            SCMultiSel scs;

            public CalcLightingNewImpl(JKLevel _lev, SCMultiSel _scs)
            {
                lev = _lev;
                scs = _scs;
            }

            bool CanReach(JKSector viasec, JedLight lt, Vec3 lvec, JKVertex vx)
            {
                if (viasec.mark != 0)
                    return false;

                if (object.ReferenceEquals(viasec, vx.Sector))
                    return true;

                viasec.mark = 1;

                for (int i = 0; i < viasec.surfaces.Count; i++)
                {
                    JKSurface surf = viasec.surfaces[i];
                    if (surf.adjoin == null || surf.vertices.Count == 0 || (surf.adjoinFlags & Adjoin.Flag.BlocksLight) != 0)
                        continue;

                    // if a surface doesn't face the ray, skip
                    if (lvec * surf.normal <= -0.0001)
                        continue;

                    // check if ray intersects the surface
                    if (IsSurfCrossed(surf, lt.Pos, vx.v))
                    {
                        bool ret = CanReach(surf.adjoin.sector, lt, lvec, vx);
                        if (ret)
                            return true;
                    }
                }

                return false;
            }

            bool IsBlocked(JKSector lsec, JedLight lt, Vec3 lvec, JKVertex vx)
            {
                for (int i = 0; i < lev.sectors.Count; i++)
                    lev.sectors[i].mark = 0;

                return !CanReach(lsec, lt, lvec, vx);
            }

            public void Do(bool background)
            {
                int nl = 0;
                for (int l = 0; l < lev.lights.Count; l++)
                {
                    JedLight lt = lev.lights[l];
                    lt.mark = FindSectorForXYZ(lev, lt.Pos);
                    if (lt.mark == -1)
                    {
                        if(!background)
                            Utils.PanMessage(MessageType.Warning, "Light " + l.ToString() + " is not in any sector");
                    }
                    else
                        nl++;
                }

                if (nl == 0)
                {
                    if(!background)
                        Utils.PanMessage(MessageType.Error, "No valid lights to calculate!");
                    return;
                }

                ZeroVertexLights(lev, scs);

                if (!background)
                {
                    ProgressDialog.Reset(lev.sectors.Count, 0);
                    ProgressDialog.Message("Calculating lights");
                }

                for (int s = 0; s < scs.Count; s++)
                {
                    GlobalVars.NedMain.calcLightProgress = (int)Math.Round((double)s / (double)(scs.Count - 1) * 100.0);
                    if (background && GlobalVars.NedMain.WantLightCalcCancel)
                        return;

                    JKSector sec = lev.sectors[scs.GetSC(s)];
                    for (int sf = 0; sf < sec.surfaces.Count; sf++)
                    {
                        JKSurface surf = sec.surfaces[sf];

                        for (int l = 0; l < lev.lights.Count; l++)
                        {
                            JedLight lt = lev.lights[l];
                            if (lt.mark == -1)
                                continue;

                            JKSector lsec = lev.sectors[lt.mark];
                            double range2 = Math.Pow(lt.range, 2.0);
                            JKVertex vx = surf.vertices[0] as JKVertex/*hope so*/;

                            // if a surface doesnt face the light, skip it
                            double d = (lt.Pos - vx.v) * surf.normal;
                            if (d <= 0.0)
                                continue;

                            for (int v = 0; v < surf.vertices.Count; v++)
                            {
                                vx = surf.vertices[v] as JKVertex/*hope so*/;

                                Vec3 lvec = (lt.Pos - vx.v);

                                double dist2 = Math.Pow(lvec.x, 2.0) + Math.Pow(lvec.y, 2.0) + Math.Pow(lvec.z, 2.0);
                                if (dist2 >= range2)
                                    continue;

                                if ((lt.flags & LightFlag.NoBlock) == 0)
                                {
                                    if (IsBlocked(lsec, lt, lvec, vx))
                                        continue;
                                }

                                double dist = Math.Sqrt(dist2);
                                dist2 = Math.Pow((lt.range - dist) / lt.range, 2.0);

                                TexVertex tv = surf.texvertices[v];
                                double pow = dist2;// same calc
                                tv.intensity += lt.intensity * pow;
                                tv.r += lt.rgbintensity * lt.clr.r * pow;
                                tv.g += lt.rgbintensity * lt.clr.g * pow;
                                tv.b += lt.rgbintensity * lt.clr.b * pow;
                            }
                        }
                    }

                    if(!background)
                        ProgressDialog.Step();
                }

                CalcSectorAmbients(lev, scs);

                if(!background)
                    ProgressDialog.Cleanup();
            }

        }

        public static void CalcLightingNew(JKLevel lev, SCMultiSel scs, bool background)
        {
            new CalcLightingNewImpl(lev, scs).Do(background);
        }

        public static void ZeroVertexLights(JKLevel lev, SCMultiSel scs)
        {
            for(int s=0; s<scs.Count; s++)
                foreach(JKSurface surf in lev.sectors[scs.GetSC(s)].surfaces)
                    foreach (TexVertex tv in surf.texvertices)
                    {
                        tv.intensity = 0.0;
                        tv.r = 0.0;
                        tv.g = 0.0;
                        tv.b = 0.0;
                    }
        }

        public static void CalcSectorAmbients(JKLevel lev, SCMultiSel scs)
        {
            for(int s=0; s<scs.Count; s++)
            {
                JKSector sec = lev.sectors[scs.GetSC(s)];

                double vlight = 0.0;
                double sflight = 0.0;
                int nv = 0;
                for (int sf = 0; sf < sec.surfaces.Count; sf++)
                {
                    JKSurface surf = sec.surfaces[sf];
                    sflight += surf.extraLight;
                    nv += surf.vertices.Count;
                    foreach(TexVertex tv in surf.texvertices)
                    {
                        if (tv.intensity < 0.0)
                            tv.intensity = 0.0;
                        vlight += tv.intensity;
                    }
                }

                vlight /= (double)nv;
                sflight /= (double)sec.surfaces.Count;
                sec.ambient = Math.Max(vlight, sflight);
            }
        }

        // same as dstr but includes 0 at the beginning
        public static string ostr(double d, int precision)
        {
            // probably should be optimized. whatever for now these files are ancient

            string s = "0.";
            for (int x = 0; x < precision; x++)
                s += "0";

            // using all #'s as digit placeholders means that a 0.0 value will become a null string.
            // normally this is 0.###  but predescessor's format specifies leave out the prefixed zero such as  0.5   vs   .5
            // therefore we'll use the most obviously compatible approach and if it produces an empty string we'll shove in a zero

            s = d.ToString(s);
            if (string.IsNullOrEmpty(s))
                s = "0";

            return s;
        }

        public static string dstr(double d, int precision)
        {
            // probably should be optimized. whatever for now these files are ancient

            string s = "#.";
            for (int x = 0; x < precision; x++)
                s += "#";

            // using all #'s as digit placeholders means that a 0.0 value will become a null string.
            // normally this is 0.###  but predescessor's format specifies leave out the prefixed zero such as  0.5   vs   .5
            // therefore we'll use the most obviously compatible approach and if it produces an empty string we'll shove in a zero

            s = d.ToString(s);
            if (string.IsNullOrEmpty(s))
                s = "0";

            return s;
        }

        public static bool DoBoxesIntersect(Box3 box1, Box3 box2)
        {
            return box1.Intersect(box2);
        }

        public static Vec3? ValVector(string s)
        {
            string[] parts = s.Split(new char[] { '(', ')', ' ', '\t', '/', '\\' }, StringSplitOptions.RemoveEmptyEntries);
            if (parts.Length != 3)
                return null;

            Vec3 v = new Vec3();
            if (!double.TryParse(parts[0], out v.x) ||
                !double.TryParse(parts[1], out v.y) ||
                !double.TryParse(parts[2], out v.z))
                return null;

            return v;
        }

        public static Orient? ValOrient(string s)
        {
            Vec3? v = ValVector(s);
            if (!v.HasValue)
                return null;

            return new Orient(v.Value.x, v.Value.y, v.Value.z);
        }

        public static void UnOffsetByNode(JKLevel lev, int nnode, ref Vec3 a)
        {
            List<int> branch = new List<int>();

            // trace back all parent nodes
            JKModel.HNode hnode = lev.hnodes[nnode];
            branch.Add(nnode);
            JKModel.HNode hnode1 = hnode;
            while (hnode1.parent != -1)
            {
                branch.Add(hnode1.parent);
                hnode1 = lev.hnodes[hnode1.parent];
            }

            // unrotate and unoffset in backwards order
            for (int k = branch.Count - 1; k >= 0; k--)
            {
                hnode1 = lev.hnodes[branch[k]];

                Mat4 rmx = CreateRotMatrix(hnode1.orient.pitch, hnode1.orient.yaw, hnode1.orient.roll).Transposed;

                a -= hnode1.pos;
                a.Transform(rmx);
            }
        }

        public static void OffsetByNode(JKLevel lev, int nnode, ref Vec3 a)
        {
            JKModel.HNode hnode = lev.hnodes[nnode];

            a.Transform(CreateRotMatrix(hnode.orient.pitch, hnode.orient.yaw, hnode.orient.roll));
            a += hnode.pos;

            JKModel.HNode hnode1 = hnode;

            while (hnode1.parent != -1)
            {
                hnode1 = lev.hnodes[hnode1.parent];

                a.Transform(CreateRotMatrix(hnode1.orient.pitch, hnode1.orient.yaw, hnode1.orient.roll));
                a += hnode1.pos;
            }
        }

        public static void StitchSurfaces(JKSurface surf, JKSurface surf1)
        {
            Vec3 un, vn;
            int nv1, nv2;
            if (FindCommonVXs(surf, surf1, out nv1, out nv2))
                surf.CalcUVNormalsFromVX(nv1, out un, out vn);
            else
            {
                surf.CalcUVNormals(out un, out vn);
                nv1 = -1;
                nv2 = 0;
            }

            Vec3 lv = (surf.normal ^ surf1.normal).Unit;
            if (lv.Magnitude > 0.0)
            {
                Vec3 p1v = surf.normal ^ lv;
                Vec3 p2v = surf1.normal ^ lv;

                double a = p1v * un;
                double b = lv * un;
                un = new Vec3(a * p2v.x + b * lv.x,
                              a * p2v.y + b * lv.y,
                              a * p2v.z + b * lv.z);

                a = p1v * vn;
                b = lv * vn;
                vn = new Vec3(a * p2v.x + b * lv.x,
                              a * p2v.y + b * lv.y,
                              a * p2v.z + b * lv.z);
            }

            Undo.SaveSecUndo(surf1.sector, Undo.Change.Changed, Undo.SectorChange.Both);

            if (nv1 != -1)
                surf1.texvertices[nv2].v = surf.texvertices[nv1].v;

            surf1.texScale = surf.texScale;

            if ((surf.surfaceFlags & Surface.Flag.Flipped) != 0)
                surf1.surfaceFlags |= Surface.Flag.Flipped;
            else
                surf1.surfaceFlags &= ~Surface.Flag.Flipped;

            surf1.ArrangeTexture(nv2, un, vn);

            // copy stuff
            surf1.surfaceFlags = surf.surfaceFlags;
            surf1.SetGeo("StitchSurfaces", surf.geo);
            surf1.faceFlags = surf.faceFlags;
            surf1.adjoinFlags = surf.adjoinFlags;
        }

        public static bool FindCommonVXs(JKSurface surf, JKSurface withsurf, out int v1, out int v2)
        {
            for (int i = 0; i < surf.vertices.Count; i++)
            {
                Vertex tv = surf.vertices[i];
                for (int j = 0; j < withsurf.vertices.Count; j++)
                {
                    Vertex tv2 = withsurf.vertices[j];
                    if (tv.v == tv2.v)
                    {
                        v1 = i;
                        v2 = j;
                        return true;
                    }
                }
            }

            v1 = -1;
            v2 = -1;
            return false;
        }

        public static Orient GetJKPYR(Vec3 x, Vec3 y, Vec3 z)
        {
            Orient o = Orient.Zero;

            double l = y.XY.Magnitude;
            if (l != 0.0)
                o.yaw = MathLib.ToDeg(Math.Acos(y.y / l));

            if (y.x > 0.0)
                o.yaw = 360.0 - o.yaw;

            Vec3 ny = y;
            RotateVector(ref ny, 0.0, -o.yaw, 0.0);

            o.pitch = MathLib.ToDeg(Math.Acos(ny.y));
            if (ny.z < 0.0)
                o.pitch = 360.0 - o.pitch;

            Vec3 nz = z;
            RotateVector(ref nz, 0.0, -o.yaw, 0.0);
            RotateVector(ref nz, -o.pitch, 0.0, 0.0);

            o.roll = MathLib.ToDeg(Math.Acos(nz.z));
            if (nz.x < 0.0)
                o.roll = 360.0 - o.roll;

            return o;
        }

        public static void SysCalcSurfData(JKSurface surf, out Vec3 xv, out Vec3 yv, out Vec2 min, out Vec2 max)
        {
            Vertex v0 = surf.vertices[0];
            Vertex v1 = surf.vertices[1];
            xv = (v1.v - v0.v).Unit;
            yv = surf.normal ^ xv;
            min = Vec2.Infinity;
            max = -Vec2.Infinity;
            for (int i = 1; i < surf.vertices.Count; i++)
            {
                v1 = surf.vertices[i];
                double x = (v1.v - v0.v) * xv;
                double y = (v1.v - v0.v) * yv;
                min.x = Math.Min(min.x, x);
                min.y = Math.Min(min.y, y);
                max.x = Math.Max(max.x, x);
                max.y = Math.Max(max.y, y);
            }
        }

        public static Vec3 CalcSurfCenter(JKSurface surf)
        {
            Vec3 xv, yv;
            Vec2 min, max;
            SysCalcSurfData(surf, out xv, out yv, out min, out max);
            Vec2 v = (max - min) / 2.0;
            Vertex v0 = surf.vertices[0];
            return new Vec3(v0.v.x + v.x * xv.x + v.y * yv.x,
                            v0.v.y + v.x * xv.y + v.y * yv.y,
                            v0.v.z + v.x * xv.z + v.y * yv.z);
        }

        public static void DeleteCOG(JKLevel lev, int n)
        {
            lev.cogs.RemoveAt(n);

            foreach(JKCOG cog in lev.cogs)
                foreach(COGValue cv in cog.vals)
                    if (cv.cogType == COGType.COG)
                    {
                        if (cv.intval == n)
                            cv.intval = 1;
                        else if (cv.intval > n)
                            cv.intval--;
                    }

            GlobalVars.CogForm.RefreshList();
        }

        public static DialogResult InputBox(string title, string promptText, ref string value)
        {
            Form form = new Form();
            Label label = new Label();
            TextBox textBox = new TextBox();
            Button buttonOk = new Button();
            Button buttonCancel = new Button();

            form.Text = title;
            label.Text = promptText;
            textBox.Text = value;

            buttonOk.Text = "OK";
            buttonCancel.Text = "Cancel";
            buttonOk.DialogResult = DialogResult.OK;
            buttonCancel.DialogResult = DialogResult.Cancel;

            label.SetBounds(9, 20, 372, 13);
            textBox.SetBounds(12, 36, 372, 20);
            buttonOk.SetBounds(228, 72, 75, 23);
            buttonCancel.SetBounds(309, 72, 75, 23);

            label.AutoSize = true;
            textBox.Anchor = textBox.Anchor | AnchorStyles.Right;
            buttonOk.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;
            buttonCancel.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;

            form.ClientSize = new Size(396, 107);
            form.Controls.AddRange(new Control[] { label, textBox, buttonOk, buttonCancel });
            form.ClientSize = new Size(Math.Max(300, label.Right + 10), form.ClientSize.Height);
            form.FormBorderStyle = FormBorderStyle.FixedDialog;
            form.StartPosition = FormStartPosition.CenterScreen;
            form.MinimizeBox = false;
            form.MaximizeBox = false;
            form.AcceptButton = buttonOk;
            form.CancelButton = buttonCancel;

            DialogResult dialogResult = form.ShowDialog();
            value = textBox.Text;
            return dialogResult;
        }

        public static int CompareVX(Vertex v1, Vertex v2)
        {
            double d = v1.v.x - v2.v.x;
            if (MathLib.CompareScalar(d, 0.0) != 0)
            {
                if (d > 0.0)
                    return 1;
                else
                    return -1;
            }

            d = v1.v.y - v2.v.y;
            if (MathLib.CompareScalar(d, 0.0) != 0)
            {
                if (d > 0.0)
                    return 1;
                else
                    return -1;
            }

            d = v1.v.z - v2.v.z;
            if (MathLib.CompareScalar(d, 0.0) != 0)
            {
                if (d > 0.0)
                    return 1;
                else
                    return -1;
            }

            return 0;
        }

        public static int CompareTXVX(TexVertex v1, TexVertex v2)
        {
            double d = v1.v.x - v2.v.x;
            if (MathLib.CompareScalar(d, 0.0) != 0)
            {
                if (d > 0.0)
                    return 1;
                else
                    return -1;
            }

            d = v1.v.y - v2.v.y;
            if (MathLib.CompareScalar(d, 0.0) != 0)
            {
                if (d > 0.0)
                    return 1;
                else
                    return -1;
            }

            return 0;
        }

        public static double DistToPlane(Polygon surf, Vec3 p)
        {
            Vertex vx = surf.vertices[0];
            return Math.Abs(surf.normal * (p - vx.v));
        }

        public static Vec3 FindCenter(JKSector sec)
        {
            if (sec.vertices.Count == 0)
                return Vec3.Zero;

            Vec3 sum = Vec3.Zero;
            foreach (JKVertex vert in sec.vertices)
                sum += vert.v;

            return sum / (double)sec.vertices.Count;
        }

        public static void FindBSphere(JKSector sec, out Vec3 c, out double rad)
        {
            c = FindCenter(sec);
            rad = 0.0;

            foreach (Vertex v in sec.vertices)
                rad = Math.Max(rad, (v.v - c).Magnitude);
        }

        public static bool IsInSecNS(JKSector sec, double x, double y, double z)
        {
            return SysIsInSector(sec, new Vec3(x, y, z), false);
        }

        public static Box3? FindCollideBox(JKSector sec, Box3 bbox, Vec3 c)
        {
            // DUDE wtf is this horseradish

            const int cbSteps = 25;

            bool bx1 = true, bx2 = true;
            bool by1 = true, by2 = true;
            bool bz1 = true, bz2 = true;

            Vec3 minus = new Vec3((bbox.Left - c.x) / (double)cbSteps,
                                  (bbox.Back - c.y) / (double)cbSteps,
                                  (bbox.Bottom - c.z) / (double)cbSteps);

            Vec3 plus = new Vec3((bbox.Right - c.x) / (double)cbSteps,
                                 (bbox.Front - c.y) / (double)cbSteps,
                                 (bbox.Top - c.z) / (double)cbSteps);

            Vec3 v1 = c + minus;
            Vec3 v2 = c + plus;

            if (!IsInSecNS(sec, v1.x, v1.y, v1.z) ||
                !IsInSecNS(sec, v1.x, v1.y, v2.z) ||
                !IsInSecNS(sec, v1.x, v2.y, v2.z) ||
                !IsInSecNS(sec, v1.x, v2.y, v1.z))
                return null;

            if (!IsInSecNS(sec, v2.x, v1.y, v1.z) ||
                !IsInSecNS(sec, v2.x, v1.y, v2.z) ||
                !IsInSecNS(sec, v2.x, v2.y, v2.z) ||
                !IsInSecNS(sec, v2.x, v2.y, v1.z))
                return null;

            for (int i = 0; i < cbSteps; i++)
            {
                if(bx1)
                {
                    v1.x += minus.x;
                    if (!IsInSecNS(sec, v1.x, v1.y, v1.z) ||
                        !IsInSecNS(sec, v1.x, v1.y, v2.z) ||
                        !IsInSecNS(sec, v1.x, v2.y, v2.z) ||
                        !IsInSecNS(sec, v1.x, v2.y, v1.z))
                    {
                        v1.x -= minus.x;
                        bx1 = false;
                    }
                }

                if (bx2)
                {
                    v2.x += plus.x;
                    if (!IsInSecNS(sec, v2.x, v1.y, v1.z) ||
                        !IsInSecNS(sec, v2.x, v1.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v2.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v2.y, v1.z))
                    {
                        v2.x -= plus.x;
                        bx2 = false;
                    }
                }

                if (by1)
                {
                    v1.y += minus.y;
                    if (!IsInSecNS(sec, v1.x, v1.y, v1.z) ||
                        !IsInSecNS(sec, v1.x, v1.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v1.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v1.y, v1.z))
                    {
                        v1.y -= minus.y;
                        by1 = false;
                    }
                }

                if (by2)
                {
                    v2.y += plus.y;
                    if (!IsInSecNS(sec, v1.x, v2.y, v1.z)||
                        !IsInSecNS(sec, v1.x, v2.y, v2.z)||
                        !IsInSecNS(sec, v2.x, v2.y, v2.z)||
                        !IsInSecNS(sec, v2.x, v2.y, v1.z))
                    {
                        v2.y -= plus.y;
                        by2 = false;
                    }
                }

                if (bz1)
                {
                    v1.z += minus.z;
                    if (!IsInSecNS(sec, v1.x, v1.y, v1.z) ||
                        !IsInSecNS(sec, v1.x, v2.y, v1.z) ||
                        !IsInSecNS(sec, v2.x, v2.y, v1.z) ||
                        !IsInSecNS(sec, v2.x, v1.y, v1.z))
                    {
                        v1.z -= minus.z;
                        bz1 = false;
                    }
                }

                if (bz2)
                {
                    v2.z += plus.z;
                    if (!IsInSecNS(sec, v1.x, v1.y, v2.z) ||
                        !IsInSecNS(sec, v1.x, v2.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v2.y, v2.z) ||
                        !IsInSecNS(sec, v2.x, v1.y, v2.z))
                    {
                        v2.z -= plus.z;
                        bz2 = false;
                    }
                }

                if (!bx1 && !bx2 && !by1 && !by2 && !bz1 && !bz2)
                    break;
            }

            return new Box3(v1, v2);
        }

        public static bool FindVX(Vertices vs, Vertex v, out int index)
        {
            //  LOLLL  WHAT THE HELL IS GOING ON HERE

            bool ret = false;
            int l = 0;
            int h = vs.Count - 1;
            while (l <= h)
            {
                int i = (l + h) >> 1;
                int c = CompareVX(vs[i], v);
                if (c < 0)
                    l = i + 1;
                else
                {
                    h = i - 1;
                    if (c == 0)
                    {
                        ret = true;
                        l = i;
                    }
                }
            }
            index = l;
            return ret;
        }

        public static bool FindTXVX(TexVertices vs, TexVertex v, out int index)
        {
            //  LOLLL  WHAT THE HELL IS GOING ON HERE

            bool ret = false;
            int l = 0;
            int h = vs.Count - 1;
            while (l <= h)
            {
                int i = (l + h) >> 1;
                int c = CompareTXVX(vs[i], v);
                if (c < 0)
                    l = i + 1;
                else
                {
                    h = i - 1;
                    if (c == 0)
                    {
                        ret = true;
                        l = i;
                    }
                }
            }
            index = l;
            return ret;
        }

        public static int AddVX(Vertex tv, Vertices vxlist, Vertices vxidx)
        {
            int n;
            if (FindVX(vxidx, tv, out n))
                return vxidx[n].num;

            int i = vxlist.Count;
            vxlist.Add(tv);
            vxidx.Insert(n, tv);
            tv.num = i;
            return i;
        }

        public static int AddTXVX(TexVertex tv, TexVertices txvxlist, TexVertices txidx)
        {
            int n;
            if (FindTXVX(txidx, tv, out n))
                return txidx[n].num;

            int i = txvxlist.Count;
            txvxlist.Add(tv);
            txidx.Insert(n, tv);
            tv.num = i;
            return i;
        }

        public static bool UnadjoinSectors(JKSector sec, JKSector sec1)
        {
            bool ret = false;
            foreach (JKSurface surf in sec.surfaces)
            {
                if (surf.adjoin == null)
                    continue;

                if (surf.adjoin.sector == sec1)
                    ret |= Unadjoin(surf);
            }
            return ret;
        }

        public static bool Unadjoin(JKSurface surf)
        {
            if (surf.adjoin == null)
                return false;

            JKSector sec1 = surf.adjoin.sector;
            if (surf.adjoin.adjoin != surf)
            {
                Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Val);
                PanMessage(MessageType.Warning, "Incorrect reverse adjoin SC " + surf.sector.num.ToString() + " SF " + surf.num.ToString());
                ClearAdjoinParams(surf);
                surf.adjoin = null;
            }
            else
            {
                Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Val);
                Undo.SaveSecUndo(surf.adjoin.sector, Undo.Change.Changed, Undo.SectorChange.Val);

                ClearAdjoinParams(surf);
                ClearAdjoinParams(surf.adjoin);

                surf.CheckIfFloor();
                surf.adjoin.CheckIfFloor();

                surf.adjoin.adjoin = null;
                surf.adjoin = null;
                GlobalVars.NedMain.SectorChanged(sec1, true);
            }
            GlobalVars.NedMain.SectorChanged(surf.sector, true);
            return true;
        }

        public static void ExtrudeSurface(JKSurface surf, double by)
        {
            if (surf.adjoin != null)
                return;

            JKLevel lev = surf.sector.level;
            JKSector nsec = lev.NewSector();
            nsec.Assign(surf.sector);
            nsec.num = lev.sectors.Count;

            Undo.SaveSecUndo(nsec, Undo.Change.Added, Undo.SectorChange.Both);
            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Val);

            // construct adjoin
            JKSurface asurf = nsec.NewSurface();
            asurf.Assign(surf);
            for (int v = surf.vertices.Count - 1; v >= 0; v--)
            {
                JKVertex nvx = nsec.NewVertex();
                nvx.Assign(surf.vertices[v] as JKVertex/*hope so*/);
                asurf.AddVertex(nvx);
            }
            nsec.surfaces.Add(asurf);

            SysAdjoinSurfaces(surf, asurf);

            // construct opposite surface
            JKSurface osurf = nsec.NewSurface();
            osurf.Assign(surf);
            for (int v = 0; v < surf.vertices.Count; v++)
            {
                JKVertex nvx = nsec.NewVertex();
                nvx.v = surf.vertices[v].v + (-surf.normal) * by;
                osurf.AddVertex(nvx);
            }
            nsec.surfaces.Add(osurf);
            osurf.NewRecalcAll();
            osurf.Planarize();

            // construct other surfaces
            int vc = surf.vertices.Count;
            for (int i = 0; i < vc; i++)
            {
                JKSurface nsurf = nsec.NewSurface();
                nsurf.Assign(surf);

                nsurf.AddVertex(asurf.vertices[i]);

                if (i == 0)
                    nsurf.AddVertex(asurf.vertices[vc - 1]);
                else
                    nsurf.AddVertex(asurf.vertices[i - 1]);

                if (i == 0)
                    nsurf.AddVertex(osurf.vertices[0]);
                else
                    nsurf.AddVertex(osurf.vertices[vc - i]);

                nsurf.AddVertex(osurf.vertices[vc - i - 1]);
                nsec.surfaces.Add(nsurf);
            }
            lev.sectors.Add(nsec);
            lev.RenumSecs();
            foreach (JKSurface nsurf in nsec.surfaces)
                nsurf.NewRecalcAll();
            nsec.Renumber();
            GlobalVars.NedMain.SectorAdded(nsec);
            GlobalVars.NedMain.SectorChanged(surf.sector, true);
        }

        public static bool ConnectSurfaces(JKSurface surf1, JKSurface surf2)
        {
            if (surf1.adjoin != null || surf2.adjoin != null)
            {
                PanMessage(MessageType.Warning, "One (or both) surfaces are adjoined!");
                return false;
            }

            if (surf1.sector == surf2.sector)
            {
                PanMessage(MessageType.Warning, "Surfaces belong to the same sector!");
                return false;
            }

            // check that surfaces are back to back
            Vec3 v1 = surf1.vertices[0].v;
            Vec3 v2 = surf2.vertices[0].v;

            double d = surf1.normal * surf2.normal;
            if (d > -0.9999)
            {
                PanMessage(MessageType.Warning, "The surfaces aren't coplanar!");
                return false;
            }

            if (Math.Abs((v2 - v1) * surf1.normal) > 0.0001)
            {
                PanMessage(MessageType.Warning, "The surfaces are not placed back to back");
                return false;
            }

            CleaveSurfBySurf(surf2, surf1);
            CleaveSurfBySurf(surf1, surf2);

            if (Do_Surf_Overlap(surf1, surf2))
            {
                SysAdjoinSurfaces(surf1, surf2);
                return true;
            }

            PanMessage(MessageType.Warning, "The resulting surfaces don't overlap!");
            return false;
        }

        private static void CleaveSurfBySurf(JKSurface surf, JKSurface bysurf)
        {
            for (int i = 0; i < bysurf.vertices.Count; i++)
            {
                Vec3 v1 = bysurf.vertices[i].v;
                Vec3 v2 = bysurf.vertices[bysurf.NextVX(i)].v;
                Vec3 vec = (v2 - v1).Unit;
                if (vec.Magnitude == 0.0)
                    continue;

                vec = vec ^ bysurf.normal;
                CleaveSurface(surf, vec, v1);
            }
        }

        public static bool ConnectSectors(JKSector sec1, JKSector sec2)
        {
            if (!DoSectorsOverlap(sec1, sec2))
            {
                PanMessage(MessageType.Warning, "The sectors don't overlap");
                return false;
            }

            int lsec = sec1.level.sectors.Count;
            CleaveSecBySec(sec2, sec1);
            CleaveSecBySec(sec1, sec2);
            DeleteSector(sec2.level, sec2.num);
            foreach (JKSurface surf in sec1.surfaces)
                MakeAdjoinSCUP(surf, lsec);

            return true;
        }

        private static bool DoSectorsOverlap_DoThey(JKSector sec1, JKSector sec2)
        {
            foreach (JKSurface surf in sec1.surfaces)
            {
                if (surf.vertices.Count < 1)
                    continue;

                Vec3 v0 = surf.vertices[0].v;

                bool onright = false;
                foreach (Vertex v in sec2.vertices)
                {
                    double d = (v.v - v0) * surf.normal;
                    if (d > 0.01)
                    {
                        onright = true;
                        break;
                    }
                }
                if (!onright)
                    return false;
            }

            return true;
        }

        public static bool DoSectorsOverlap(JKSector sec1, JKSector sec2)
        {
            return DoSectorsOverlap_DoThey(sec1, sec2) && DoSectorsOverlap_DoThey(sec2, sec1);
        }

        private static void CleaveSecBySec(JKSector sec, JKSector bysec)
        {
            foreach (JKSurface surf in bysec.surfaces)
            {
                if (surf.normal.Magnitude == 0.0)
                    continue;

                Vec3 v = surf.vertices[0].v;
                CleaveSector(sec, surf.normal, v);
            }
        }

        public static JKSector MergeSectors(JKSector sec, JKSector sec1)
        {
            JKSurfaces sfs = new JKSurfaces();
            JKSurfaces sfs1 = new JKSurfaces();

            foreach (JKSurface surf in sec.surfaces)
                if (surf.adjoin != null && surf.adjoin.sector == sec1 && Do_Surf_Overlap(surf, surf.adjoin))
                {
                    sfs.Add(surf);
                    sfs1.Add(surf.adjoin);
                }

            if (sfs.Count == 0)
                return null;

            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Both);
            Undo.SaveSecUndo(sec1, Undo.Change.Deleted, Undo.SectorChange.Both);

            for (int i = sec1.surfaces.Count - 1; i >= 0; i--)
            {
                JKSurface surf = sec1.surfaces[i];

                // if surface is to be deleted, skip it
                if (sfs1.IndexOf(surf) != -1)
                    continue;

                // add all vertices used in the surface to sec
                for (int j = 0; j < surf.vertices.Count; j++)
                    surf.vertices[j] = sec.AddVertex(surf.vertices[j].v);

                // add surface to sec and delete from sec1
                surf.sector = sec;
                sec.surfaces.Add(surf);
                sec1.surfaces.RemoveAt(i);
            }

            RemoveSecRefs(sec1.level, sec1, false);

            foreach (JKSurface surf in sfs)
                RemoveSurfRefs(sec1.level, surf);

            foreach (JKSurface surf in sfs1)
                RemoveSurfRefs(sec1.level, surf);

            foreach (JKSurface surf in sfs)
                sec.surfaces.Remove(surf);

            sec.Renumber();

            GlobalVars.NedMain.SectorChanged(sec, true);

            sec1.level.sectors.Remove(sec1);
            GlobalVars.NedMain.SectorDeleted(sec1);

            sec.level.RenumSecs();

            if (!IsSectorConvex(sec))
                PanMessage(MessageType.Info, "The sector " + sec.num.ToString() + " is not convex");

            return sec;
        }

        public static bool IsSectorConvex(JKSector sec)
        {
            foreach (JKSurface surf in sec.surfaces)
            {
                Vec3 v0 = surf.vertices[0].v;

                foreach (JKVertex v in sec.vertices)
                    if ((v.v - v0) * surf.normal < -0.001)
                        return false;
            }
            return true;
        }

        private static bool SysFindCommonEdge(JKSurface surf1, JKSurface surf2, out int ed1, out int ed2)
        {
            ed1 = -1;
            ed2 = -1;

            for (int i = 0; i < surf1.vertices.Count; i++)
            {
                Vertex v1 = surf1.vertices[i];
                int j = surf2.vertices.IndexOf(v1);
                if (j == -1)
                    continue;

                if (surf1.vertices[surf1.NextVX(i)] == surf2.vertices[surf2.PrevVX(j)])
                {
                    ed1 = i;
                    ed2 = surf2.PrevVX(j);
                    return true;
                }
            }

            return false;
        }

        private static JKSurface SysMergeSurfs(JKSurface surf1, JKSurface surf2, int edge1, int edge2)
        {
            int nv = 0;

            if ((surf1.adjoin == null && surf2.adjoin != null) ||
                (surf2.adjoin == null && surf1.adjoin != null))
            {
                PanMessage(MessageType.Error, "Can't merge - one surface is adjoined, the other isn't");
                return null;
            }

            foreach (Vertex v in surf1.vertices)
                if (surf2.vertices.IndexOf(v) == -1)
                    nv++;

            foreach (Vertex v in surf2.vertices)
                if (surf1.vertices.IndexOf(v) == -1)
                    nv++;

            if (nv > 22)
            {
                PanMessage(MessageType.Error, "Can't merge surfaces - resulting surface is >24 vertices");
                return null;
            }

            // merge surfaces

            Undo.SaveSecUndo(surf1.sector, Undo.Change.Changed, Undo.SectorChange.Both);

            {
                int v = surf2.PrevVX(edge2);
                int lastv = surf2.NextVX(edge2);
                int ipos = surf1.NextVX(edge1);
                while (v != lastv)
                {
                    surf1.InsertVertex(ipos, surf2.vertices[v]);
                    v = surf2.PrevVX(v);
                }

                do
                {
                    nv = 0;
                    for (int i = surf1.vertices.Count - 1; i >= 0; i--)
                    {
                        if (surf1.vertices[surf1.PrevVX(i)] == surf1.vertices[surf1.NextVX(i)])
                        {
                            Vertex vx = surf1.vertices[i];
                            surf1.DeleteVertex(i);
                            DeleteVXIfUnused(surf1.sector, vx, surf2);
                            nv++;
                        }
                    }

                    for (int i = surf1.vertices.Count - 1; i >= 0; i--)
                    {
                        if (surf1.vertices[i] == surf1.vertices[surf1.NextVX(i)])
                        {
                            surf1.DeleteVertex(i);
                            nv++;
                        }
                    }
                } while (nv != 0);

                surf1.RecalcAll();

                // if the surfaces were adjoined, try to merge the adjoins and readjoin

                if (surf1.adjoin != null && surf2.adjoin != null)
                {
                    JKSurface sf1 = surf1.adjoin;
                    JKSurface sf2 = surf2.adjoin;

                    ClearAdjoinParams(sf1);
                    ClearAdjoinParams(sf2);

                    int i, j;
                    if (SysFindCommonEdge(sf1, sf2, out i, out j))
                    {
                        ClearAdjoinParams(sf1);
                        ClearAdjoinParams(sf2);
                        SysMergeSurfs(sf1, sf2, i, j);
                        if(Do_Surf_Overlap(surf1, sf1))
                        {
                            SysAdjoinSurfaces(surf1, sf1);
                            GlobalVars.NedMain.SectorChanged(sf1.sector, true);
                        }
                    }
                }

                RemoveSurfRefs(surf1.sector.level, surf2);
                surf2.sector.surfaces.Remove(surf2);

                surf1.sector.Renumber();
                GlobalVars.NedMain.SectorChanged(surf1.sector, true);

                return surf1;
            }
        }

        private static void DeleteVXIfUnused(JKSector sec, Vertex v, JKSurface surf2)
        {
            foreach (JKSurface surf in sec.surfaces)
                if (surf != surf2 && surf.vertices.IndexOf(v) != -1)
                    return;

            int i = sec.vertices.IndexOf(v as JKVertex/*hope so*/);
            if (i == -1)
                return;

            sec.vertices.RemoveAt(i);
        }

        public static JKSurface MergeSurfaces(JKSurface surf, JKSurface surf1)
        {
            int ed1, ed2;
            if (!SysFindCommonEdge(surf, surf1, out ed1, out ed2))
                return null;

            JKSurface ret = SysMergeSurfs(surf, surf1, ed1, ed2);
            if (ret == null)
                return null;

            if (!IsSurfPlanar(ret))
                PanMessage(MessageType.Info, "The surface " + surf.sector.num.ToString() + "," + surf.num.ToString() + " is not planar");

            if (!IsSurfConvex(ret))
                PanMessage(MessageType.Info, "The surface " + surf.sector.num.ToString() + "," + surf.num.ToString() + " is not convex");

            return ret;
        }

        public static bool IsSurfPlanar(JKSurface surf)
        {
            double d = surf.CalcD();
            foreach (Vertex vx in surf.vertices)
                if (Math.Abs(surf.normal * vx.v - d) > 0.001)
                    return false;

            return true;
        }

        public static bool IsSurfConvex(JKSurface surf)
        {
            for (int i = 0; i < surf.vertices.Count; i++)
            {
                Vertex v1 = surf.vertices[i];
                int n = surf.NextVX(i);
                Vertex v2 = surf.vertices[n];
                n = surf.NextVX(n);
                Vertex v3 = surf.vertices[n];
                Vec3 v = (v3.v - v2.v) ^ (v1.v - v2.v);
                if (v * surf.normal < -0.0001)
                    return false;
            }

            return true;
        }

        public static bool MergeSurfacesAt(JKSurface surf, int edge)
        {
            Vertex v1 = surf.vertices[edge];
            Vertex v2 = surf.vertices[surf.NextVX(edge)];

            JKSurface surf1 = null;
            int edge1 = -1;
            int nsf1 = -1;
            JKSector sec = surf.sector;
            for(int sf=0; sf<sec.surfaces.Count; sf++)
            {
                JKSurface asf = sec.surfaces[sf];

                if (asf == surf)
                    continue;

                for (int v = 0; v < asf.vertices.Count; v++)
                {
                    Vertex av1 = asf.vertices[v];
                    Vertex av2 = asf.vertices[asf.NextVX(v)];

                    if (v1 != av2 || v2 != av1)
                        continue;

                    if (surf1 != null)
                    {
                        PanMessage(MessageType.Error, "The edge is common to more than one surface");
                        return false;
                    }

                    surf1 = asf;
                    edge1 = v;
                    nsf1 = sf;
                }
            }

            if (surf1 == null)
            {
                PanMessage(MessageType.Error, "The edge belongs to just one surface");
                return false;
            }

            return (SysMergeSurfs(surf, surf1, edge, edge1) != null);
        }

        public static int MergeVertices(JKLevel lev, VXMultiSel vxsel)
        {
            Undo.StartUndoRec("Merge vertices");
            vxsel.Sort();

            int result = 0;
            for (int i = 0; i < vxsel.Count; i++)
            {
                for (int j = i + 1; j < vxsel.Count; j++)
                {
                    int sc, vx;
                    vxsel.GetSCVX(i, out sc, out vx);
                    JKVertex v1 = lev.sectors[sc].vertices[vx];
                    vxsel.GetSCVX(j, out sc, out vx);
                    JKVertex v2 = lev.sectors[sc].vertices[vx];
                    int nv2 = vx;
                    if (v1 == v2)
                        continue;
                    if (v1.Sector != v2.Sector)
                        continue;
                    if (v1.v != v2.v)
                        continue;

                    JKSector sec = v1.Sector;

                    Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);

                    foreach (JKSurface surf in sec.surfaces)
                    {
                        for (vx = surf.vertices.Count - 1; vx >= 0; vx--)
                        {
                            bool sfchanged = false;
                            if (object.ReferenceEquals(surf.vertices[vx], v2))
                            {
                                surf.vertices[vx] = v1;
                                if (object.ReferenceEquals(surf.vertices[surf.NextVX(vx)], v1))
                                    surf.vertices.RemoveAt(vx);
                                if (object.ReferenceEquals(surf.vertices[surf.PrevVX(vx)], v1))
                                    surf.vertices.RemoveAt(vx);
                                sfchanged = true;
                            }
                            if (sfchanged)
                                surf.Recalc();
                        }
                    }

                    sec.vertices.RemoveAt(nv2);
                    result++;
                    GlobalVars.NedMain.SectorChanged(sec, true);
                }
            }

            return result;
        }

        private static bool AdjoinToSecl(JKSurface surf, JKSector secl)
        {
            if (surf.adjoin != null)
                return false;

            bool result = false;            
            foreach(JKSurface surfl in secl.surfaces)
                if (surfl.adjoin == null && Do_Surf_Overlap(surf, surfl))
                {
                    result = true;
                    SysAdjoinSurfaces(surf, surfl);
                }

            return result;
        }

        public static bool AdjoinSectors(JKSector sec, JKSector secl)
        {
            bool result = false;
            foreach (JKSurface surf in sec.surfaces)
                result |= AdjoinToSecl(surf, secl);

            if (result)
            {
                GlobalVars.NedMain.SectorChanged(sec, true);
                GlobalVars.NedMain.SectorChanged(secl, true);
            }

            return result;
        }

        public static bool MakeAdjoin(JKSurface surf)
        {
            if (surf.adjoin != null)
                return false;

            JKLevel lev = surf.sector.level;

            foreach(JKSector secl in lev.sectors)
                foreach (JKSurface surfl in secl.surfaces)
                    if (Do_Surf_Overlap(surf, surfl))
                    {
                        Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Val);
                        Undo.SaveSecUndo(surfl.sector, Undo.Change.Changed, Undo.SectorChange.Val);

                        SysAdjoinSurfaces(surf, surfl);
                        GlobalVars.NedMain.SectorChanged(surf.sector, true);
                        GlobalVars.NedMain.SectorChanged(surfl.sector, true);

                        return true;
                    }

            return false;
        }

        public static bool MakeAdjoinSCUP(JKSurface surf, int sc)
        {
            if (surf.adjoin != null)
                return false;

            JKLevel lev = surf.sector.level;

            for (int s = sc; s < lev.sectors.Count; s++)
            {
                foreach (JKSurface surf1 in lev.sectors[s].surfaces)
                    if (Do_Surf_Overlap(surf, surf1))
                    {
                        SysAdjoinSurfaces(surf, surf1);
                        return true;
                    }
            }

            return false;
        }

        public static void ClearVXMarks(JKLevel lev)
        {
            foreach (JKSector sec in lev.sectors)
                foreach (JKVertex vert in sec.vertices)
                    vert.mark = 0;
        }

        public static void ClearSCMarks(JKLevel lev)
        {
            foreach (JKSector sec in lev.sectors)
                sec.mark = 0;
        }

        public static void TranslateSectors(JKLevel lev, SCMultiSel scsel, int cursc, Vec3 d)
        {
            ClearVXMarks(lev);

            if (cursc >= 0)
                MoveSec(lev.sectors[cursc], d);

            for (int i = 0; i < scsel.Count; i++)
                MoveSec(lev.sectors[scsel.GetSC(i)], d);
        }

        public static void MoveSec(JKSector sec, Vec3 d)
        {
            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);

            foreach (JKVertex vert in sec.vertices)
                TranslateVX(vert, d);

            foreach (JKSurface surf in sec.surfaces)
            {
                if (surf.adjoin == null)
                    continue;

                Undo.SaveSecUndo(surf.adjoin.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

                foreach (Vertex vert in surf.adjoin.vertices)
                    TranslateVX(vert, d);
            }

            RecalcSector(sec);

            GlobalVars.NedMain.SectorChanged(sec, true);

            foreach (JKSurface surf in sec.surfaces)
            {
                if (surf.adjoin == null)
                    continue;

                GlobalVars.NedMain.SectorChanged(surf.adjoin.sector, true);
            }
        }

        public static void RecalcSector(JKSector sec)
        {
            ClearSCMarks(sec.level);
            foreach (JKSurface surf in sec.surfaces)
            {
                surf.RecalcAll();

                if (surf.adjoin == null)
                    continue;

                if (surf.adjoin.sector.mark != 0)
                    continue;

                foreach (JKSurface asurf in surf.adjoin.sector.surfaces)
                    asurf.RecalcAll();

                surf.adjoin.sector.mark = 0;
            }
        }

        public static void TranslateSurfaces(JKLevel lev, SFMultiSel sfsel, int cursc, int cursf, Vec3 d)
        {
            ClearVXMarks(lev);

            if (cursc >= 0)
                MoveSurf(lev.sectors[cursc].surfaces[cursf], d);

            for (int i = 0; i < sfsel.Count; i++)
            {
                int sc, sf;
                sfsel.GetSCSF(i, out sc, out sf);
                MoveSurf(lev.sectors[sc].surfaces[sf], d);
            }
        }

        public static void MoveSurf(JKSurface surf, Vec3 d)
        {
            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            foreach (Vertex vert in surf.vertices)
                TranslateVX(vert, d);

            if (surf.adjoin != null)
            {
                Undo.SaveSecUndo(surf.adjoin.sector, Undo.Change.Changed, Undo.SectorChange.Geo);
                foreach (Vertex vert in surf.adjoin.vertices)
                    TranslateVX(vert, d);
            }

            foreach (JKSurface jsurf in surf.sector.surfaces)
                jsurf.RecalcAll();

            if (surf.adjoin != null)
                foreach (JKSurface jsurf in surf.adjoin.sector.surfaces)
                    jsurf.RecalcAll();

            GlobalVars.NedMain.SectorChanged(surf.sector, true);
            if (surf.adjoin != null)
                GlobalVars.NedMain.SectorChanged(surf.adjoin.sector, true);
        }

        public static void TranslateVertices(JKLevel lev, VXMultiSel vxsel, int cursc, int curvx, Vec3 d)
        {
            ClearVXMarks(lev);
            MoveVX(lev.sectors[cursc].vertices[curvx], d);
            for (int i = 0; i < vxsel.Count; i++)
            {
                int sc, vx;
                vxsel.GetSCVX(i, out sc, out vx);
                MoveVX(lev.sectors[sc].vertices[vx], d);
            }
        }

        public static void MoveVX(JKVertex vx, Vec3 d)
        {
            Undo.SaveSecUndo(vx.Sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            TranslateVX(vx, d);
            foreach (JKSurface surf in vx.Sector.surfaces)
                surf.RecalcAll();
            GlobalVars.NedMain.SectorChanged(vx.Sector, true);
        }

        public static void TranslateVX(Vertex v, Vec3 d)
        {
            if (v.mark != 0)
                return;
            v.v += d;
            v.mark = 1;
        }

        public static void TranslateEdges(JKLevel lev, EDMultiSel edsel, int cursc, int cursf, int cured, Vec3 d)
        {
            ClearVXMarks(lev);
            MoveED(lev.sectors[cursc].surfaces[cursf], cured, d);
            for (int i = 0; i < edsel.Count; i++)
            {
                int sc, sf, ed;
                edsel.GetSCSFED(i, out sc, out sf, out ed);
                MoveED(lev.sectors[sc].surfaces[sf], ed, d);
            }
        }

        public static void MoveED(JKSurface surf, int ed, Vec3 d)
        {
            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            TranslateVX(surf.vertices[ed], d);
            TranslateVX(surf.vertices[surf.NextVX(ed)], d);

            foreach (JKSurface jsurf in surf.sector.surfaces)
                jsurf.RecalcAll();

            GlobalVars.NedMain.SectorChanged(surf.sector, true);
        }

        public static void TranslateThings(JKLevel lev, THMultiSel thsel, int curth, Vec3 d, bool moveframes)
        {
            if (thsel.FindTH(curth) == -1)
                MoveTH(lev.things[curth], d, moveframes);

            for (int i = 0; i < thsel.Count; i++)
                MoveTH(lev.things[thsel.GetTH(i)], d, moveframes);
        }

        public static void MoveTH(JKThing thing, Vec3 d, bool moveframes)
        {
            Undo.SaveThingUndo(thing, Undo.Change.Changed);
            
            thing.pos += d;

            if (moveframes)
            {
                foreach (TemplateValue tv in thing.vals)
                {
                    if (tv.adjType != AdjType.Frame)
                        continue;

                    Vec3 fp;
                    Orient fo;
                    tv.GetFrame(out fp, out fo);
                    fp += d;
                    tv.SetFrame(fp, fo);
                }
            }

            GlobalVars.NedMain.ThingChanged(thing);
            GlobalVars.NedMain.LayerThing(thing.num);
        }

        public static void TranslateThingKeepFrame(JKLevel lev, int curth, Vec3 d)
        {
            JKThing thing = lev.things[curth];
            Undo.SaveThingUndo(thing, Undo.Change.Changed);
            thing.pos += d;
            GlobalVars.NedMain.ThingChanged(thing);
            GlobalVars.NedMain.LayerThing(curth);
        }

        public static void TranslateFrames(JKLevel lev, FRMultiSel frsel, int th, int fr, Vec3 d)
        {
            int n = frsel.AddFR(th, fr);
            for (int i = 0; i < frsel.Count; i++)
            {
                frsel.GetTHFR(i, out th, out fr);
                Undo.SaveThingUndo(lev.things[th], Undo.Change.Changed);

                TemplateValue tv = lev.things[th].vals[fr];

                Vec3 fp;
                Orient fo;
                tv.GetFrame(out fp, out fo);
                fp += d;
                tv.SetFrame(fp, fo);
            }

            frsel.DeleteN(n);
        }

        public static void TranslateLights(JKLevel lev, LTMultiSel ltsel, int curlt, Vec3 d)
        {
            if (ltsel.FindLT(curlt) == -1)
                MoveLT(lev.lights[curlt], d);

            for (int i = 0; i < ltsel.Count; i++)
                MoveLT(lev.lights[ltsel.GetLT(i)], d);
        }

        public static void MoveLT(JedLight light, Vec3 d)
        {
            Undo.SaveLightUndo(light, Undo.Change.Changed);
            light.Pos += d;
            GlobalVars.NedMain.LevelChanged(true);
        }

        public static void DeleteSector(JKLevel level, int n)
        {
            JKSector sec = level.sectors[n];
            Undo.SaveSecUndo(sec, Undo.Change.Deleted, Undo.SectorChange.Both);

            level.sectors.RemoveAt(n);
            RemoveSecRefs(level, sec, rs_surfs);
            GlobalVars.NedMain.SectorDeleted(sec);
            level.RenumSecs();
        }

        public static void DeleteThing(JKLevel lev, int n)
        {
            JKThing th = lev.things[n];

            Undo.SaveThingUndo(th, Undo.Change.Deleted);

            foreach (JKCOG cog in lev.cogs)
                foreach (COGValue cv in cog.vals)
                    if (cv.cogType == COGType.Thing && object.ReferenceEquals(cv.obj, th))
                        cv.obj = null;

            GlobalVars.NedMain.ThingDeleted(th);

            lev.things.RemoveAt(n);
            lev.RenumThings();
        }

        public static void DeleteLight(JKLevel lev, int n)
        {
            Undo.SaveLightUndo(lev.lights[n], Undo.Change.Deleted);
            lev.lights.RemoveAt(n);
            GlobalVars.NedMain.LevelChanged(true);
        }

        public static bool DeleteVertex(JKSector sec, int vx, bool silent)
        {
            JKSurface sf1 = null;
            JKSurface sf2 = null;
            JKVertex v1 = sec.vertices[vx];
            int ns = 0;
            int ed1 = -1;
            int ed2 = -1;
            foreach (JKSurface asf in sec.surfaces)
            {
                int iv = asf.vertices.IndexOf(v1);
                if (iv == -1)
                    continue;

                switch (ns)
                {
                    case 0:
                        sf1 = asf;
                        ed1 = iv;
                        break;

                    case 1:
                        sf2 = asf;
                        ed2 = iv;
                        break;

                    default:
                        if (!silent)
                            PanMessage(MessageType.Error, "More than two surfaces use this vertex");

                        return false;
                }
                ns++;
            }

            if (sf2 != null)
            {
                if (sf1.vertices.Count < 4 || sf2.vertices.Count < 4)
                {
                    if (!silent)
                        PanMessage(MessageType.Error, "One of the surfaces has less than 4 vertices!");

                    return false;
                }

                if (sf1.vertices[sf1.PrevVX(ed1)] != sf2.vertices[sf2.NextVX(ed2)] ||
                    sf1.vertices[sf1.NextVX(ed1)] != sf2.vertices[sf2.PrevVX(ed2)])
                    return false;
            }

            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Geo);

            if (sf1 != null)
                sf1.DeleteVertex(ed1);

            if (sf2 != null)
                sf2.DeleteVertex(ed2);

            sec.vertices.RemoveAt(vx);
            sec.Renumber();
            GlobalVars.NedMain.SectorChanged(sec, true);
            return true;
        }

        public static void RemoveSecRefs(JKLevel level, JKSector sec, bool opt)
        {
            if (opt && rs_surfs)
            {
                foreach(JKSector lsec in level.sectors)
                    foreach (JKSurface lsurf in lsec.surfaces)
                    {
                        if (lsurf.adjoin == null)
                            continue;

                        if (lsurf.adjoin.sector == sec)
                        {
                            Undo.SaveSecUndo(lsec, Undo.Change.Changed, Undo.SectorChange.Val);
                            lsurf.adjoin = null;
                            ClearAdjoinParams(lsurf);
                            GlobalVars.NedMain.SectorChanged(lsec, true/*probably not but whatever*/);
                        }
                    }
            }

            foreach(JKCOG cog in level.cogs)
                foreach (COGValue cv in cog.vals)
                {
                    if (cv.cogType == COGType.Sector && object.ReferenceEquals(cv.obj, sec))
                        cv.obj = null;

                    if (opt && rs_surfs)
                    {
                        if (cv.cogType == COGType.Surface && cv.obj != null && (cv.obj as JKSurface).sector == sec)
                            cv.obj = null;
                    }
                }

            foreach (JKThing thing in level.things)
                if (thing.sec == sec)
                    thing.sec = null;
        }

        public static void RemoveSurfRefs(JKLevel level, JKSurface surf)
        {
            foreach(JKSector lsec in level.sectors)
                foreach (JKSurface lsurf in lsec.surfaces)
                {
                    if (lsurf.adjoin == null)
                        continue;

                    if (lsurf.adjoin == surf)
                    {
                        Undo.SaveSecUndo(lsurf.adjoin.sector, Undo.Change.Changed, Undo.SectorChange.Val);
                        lsurf.adjoin = null;
                        ClearAdjoinParams(lsurf);
                        GlobalVars.NedMain.SectorChanged(lsec, true /*probably not but whatever*/);
                    }
                }

            foreach (JKCOG cog in level.cogs)
                foreach (COGValue cv in cog.vals)
                    if (cv.cogType == COGType.Surface && object.ReferenceEquals(cv.obj, surf))
                        cv.obj = null;
        }

        static void SplitVertices(Vertices vs, Vec3 cnormal, Vec3 c)
        {
            foreach (Vertex v in vs)
            {
                double dist = cnormal * (v.v - c);
                if (MathLib.CompareScalar(dist, 0.0) == 0)
                    v.mark = 0;
                else if (dist < 0.0)
                    v.mark = -1;
                else
                    v.mark = 1;
            }
        }

        static void SplitVertices(JKVertices vs, Vec3 cnormal, Vec3 c)
        {
            foreach (JKVertex v in vs)
            {
                double dist = cnormal * (v.v - c);
                if (MathLib.CompareScalar(dist, 0.0) == 0)
                    v.mark = 0;
                else if (dist < 0.0)
                    v.mark = -1;
                else
                    v.mark = 1;
            }
        }

        public static int NextIdx(int n, int maxn)
        {
            if (n == (maxn - 1))
                return 0;
            else
                return n + 1;
        }

        public static int PrevIdx(int n, int maxn)
        {
            if (n == 0)
                return maxn - 1;
            else
                return n - 1;
        }

        public static void SysAdjoinSurfaces(JKSurface surf1, JKSurface surf2, AdjoinVals av1=null, AdjoinVals av2=null)
        {
            surf1.adjoin = surf2;
            surf2.adjoin = surf1;
            SetAdjoinParams(surf1, av1);
            SetAdjoinParams(surf2, av2);
        }

        public static bool CleaveSector(JKSector sec, Vec3 cnormal, Vec3 c)
        {
            JKLevel lev = sec.level;

            SplitVertices(sec.vertices, cnormal, c);

            int nleft = 0;
            int nright = 0;
            foreach (JKVertex v in sec.vertices)
                if (v.mark == -1)
                    nleft++;
                else if (v.mark == 1)
                    nright++;

            if (nleft == 0 || nright == 0)
                return false;

            Undo.SaveSecUndo(sec, Undo.Change.Changed, Undo.SectorChange.Both);

            for (int sf = 0; sf < sec.surfaces.Count; sf++)
            {
                JKSurface surf = sec.surfaces[sf];
                CleaveAdjoin(surf, cnormal, c, SysCleaveSurface(surf, cnormal, c));
            }

            JKVertices vxs = new JKVertices();
            foreach (JKSurface surf in sec.surfaces)
            {
                nleft = 0;
                for (int i = 0; i < surf.vertices.Count; i++)
                    if (surf.vertices[i].mark != 0)
                    {
                        nleft = 1;
                        break;
                    }

                if (nleft == 0)
                    continue;

                for (int i = 0; i < surf.vertices.Count; i++)
                {
                    JKVertex v1 = surf.vertices[i] as JKVertex;//presumably these are indeed JKVertex?
                    JKVertex v2 = surf.vertices[Utils.NextIdx(i, surf.vertices.Count)] as JKVertex;//presumably these are indeed JKVertex?
                    if (v1.mark == 0 && v2.mark == 0)
                    {
                        bool skip = false;
                        for (int k = 0; k < vxs.Count / 2; k++)
                        {
                            JKVertex cv1 = vxs[k * 2];
                            JKVertex cv2 = vxs[k * 2 + 1];
                            if (((cv1 == v1) && (cv2 == v2)) || ((cv1 == v2) && (cv2 == v1)))
                            {
                                skip = true;
                                break;
                            }
                        }

                        if (!skip)
                        {
                            vxs.Add(v1);
                            vxs.Add(v2);
                        }
                    }
                }
            }

            if (vxs.Count == 0)
                return false;

            JKVertices nsf = new JKVertices();
            {
                nsf.Add(vxs[0]);
                nsf.Add(vxs[1]);
                JKVertex v1 = vxs[1];
                vxs.RemoveAt(0);
                vxs.RemoveAt(0);

                int nv;
                do
                {
                    nv = vxs.Count;
                    for (int sf = vxs.Count / 2 - 1; sf >= 0; sf--)
                    {
                        if (vxs[sf * 2] == v1)
                        {
                            v1 = vxs[sf * 2 + 1];
                            nsf.Add(v1);
                            vxs.RemoveAt(sf * 2);
                            vxs.RemoveAt(sf * 2);
                            break;
                        }

                        if (vxs[sf * 2 + 1] == v1)
                        {
                            v1 = vxs[sf * 2];
                            nsf.Add(v1);
                            vxs.RemoveAt(sf * 2);
                            vxs.RemoveAt(sf * 2);
                            break;
                        }
                    }
                } while (nv != vxs.Count);

                vxs.Clear();
            }

            if (nsf[nsf.Count - 1] == nsf[0])
                nsf.RemoveAt(nsf.Count - 1);

            JKSector sec1 = lev.NewSector();
            sec1.num = lev.sectors.Count;
            sec1.Assign(sec);
            lev.sectors.Add(sec1);

            Undo.SaveSecUndo(sec1, Undo.Change.Added, Undo.SectorChange.Both);

            for (int i = 0; i < nsf.Count; i++)
            {
                JKVertex v1 = sec1.NewVertex();
                v1.Assign(nsf[i]);
                vxs.Add(v1);
            }

            for (int sf = sec.surfaces.Count - 1; sf >= 0; sf--)
            {
                int sfside = 0;
                JKSurface surf = sec.surfaces[sf];
                for (int i = 0; i < surf.vertices.Count; i++)
                {
                    sfside = surf.vertices[i].mark;
                    if (sfside != 0)
                        break;
                }

                switch (sfside)
                {
                    case -1:
                        surf.sector = sec1;
                        sec1.surfaces.Add(surf);
                        sec.surfaces.RemoveAt(sf);
                        break;

                    case 0:
                        if ((cnormal * surf.normal) < 0.0)
                        {
                            surf.sector = sec1;
                            sec1.surfaces.Add(surf);
                            sec.surfaces.RemoveAt(sf);
                        }
                        break;
                }
            }

            for (int sf = sec.vertices.Count - 1; sf >= 0; sf--)
            {
                JKVertex v1 = sec.vertices[sf];
                if (v1.mark == -1)
                {
                    v1.Sector = sec1;
                    sec1.vertices.Add(v1);
                    sec.vertices.RemoveAt(sf);
                }
            }

            {
                JKSurface surf = sec.NewSurface();
                foreach (JKVertex v in nsf)
                    surf.AddVertex(v);

                JKSurface surf1 = sec.NewSurface();
                for (int i = nsf.Count - 1; i >= 0; i--)
                    surf1.AddVertex(nsf[i]);

                if (sec.surfaces.Count > 0)
                {
                    surf.Assign(sec.surfaces[0]);
                    surf1.Assign(sec.surfaces[0]);
                }

                SysAdjoinSurfaces(surf, surf1);

                surf.NewRecalcAll();
                surf1.NewRecalcAll();

                if ((cnormal * surf.normal) > 0.0)
                {
                    sec.surfaces.Add(surf);
                    surf1.sector = sec1;
                    sec1.surfaces.Add(surf1);
                }
                else
                {
                    sec.surfaces.Add(surf1);
                    surf.sector = sec1;
                    sec1.surfaces.Add(surf);
                }
            }

            foreach(JKSurface surf in sec1.surfaces)
                for (int i = 0; i < surf.vertices.Count; i++)
                {
                    JKVertex v1 = surf.vertices[i] as JKVertex;//presumably these are indeed JKVertex?
                    if (v1.mark != 0)
                        continue;

                    surf.vertices[i] = sec1.AddVertex(v1.v);
                }

            sec.Renumber();
            GlobalVars.NedMain.SectorChanged(sec, true);
            sec1.Renumber();
            GlobalVars.NedMain.SectorAdded(sec1);
            RelayerSCThings(sec);

            return true;
        }

        public static bool CleaveSurface(JKSurface surf, Vec3 cnormal, Vec3 c)
        {
            SplitVertices(surf.vertices, cnormal, c);

            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Both);

            JKSurface nsf = SysCleaveSurface(surf, cnormal, c);
            CleaveAdjoin(surf, cnormal, c, nsf);
            GlobalVars.NedMain.SectorChanged(surf.sector, true);
            return (nsf != null);
        }

        public static bool CleaveEdge(JKSurface surf, int edge, Vec3 cnormal, Vec3 c)
        {
            SplitVertices(surf.vertices, cnormal, c);

            JKVertex v1 = surf.vertices[edge] as JKVertex/*hope so*/;
            JKVertex v2 = surf.vertices[surf.NextVX(edge)] as JKVertex/*hope so*/;
            if (v1.mark == 0 || v1.mark != -v2.mark)
                return false;

            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Both);

            double sd = cnormal * c;

            Vec3 a;
            PlaneLineXn(cnormal, sd, v1.v, v2.v, out a);

            // again?
            Undo.SaveSecUndo(surf.sector, Undo.Change.Changed, Undo.SectorChange.Geo);

            JKVertex nvx = surf.sector.AddVertex(a);

            surf.InsertVertex(surf.NextVX(edge), nvx);

            foreach(JKSurface jsurf in surf.sector.surfaces)
                for (int j = 0; j < jsurf.vertices.Count; j++)
                {
                    JKVertex nv1 = jsurf.vertices[j] as JKVertex/*hope so*/;
                    JKVertex nv2 = jsurf.vertices[jsurf.NextVX(j)] as JKVertex/*hope so*/;

                    if (((v1 == nv1) && (v2 == nv2)) ||
                        ((v1 == nv2) && (v2 == nv1)))
                        jsurf.InsertVertex(jsurf.NextVX(j), nvx);
                }

            surf.sector.Renumber();

            foreach (JKSurface jsurf in surf.sector.surfaces)
                jsurf.RecalcAll();

            GlobalVars.NedMain.SectorChanged(surf.sector, true);

            return true;
        }

        public static void CleaveAdjoin(JKSurface surf, Vec3 cnormal, Vec3 c, JKSurface nsf)
        {
            if (nsf == null)
                return;
            if (surf.adjoin == null)
                return;

            JKSurface sfl = surf.adjoin;

            AdjoinVals av1 = GetAdjoinParams(surf);
            AdjoinVals av2 = GetAdjoinParams(sfl);

            Undo.SaveSecUndo(sfl.sector, Undo.Change.Changed, Undo.SectorChange.Both);

            sfl.adjoin = null;
            surf.adjoin = null;
            SplitVertices(sfl.vertices, cnormal, c);

            JKSurface nsfl = SysCleaveSurface(sfl, cnormal, c);

            if (Do_Surf_Overlap(surf, sfl))
                SysAdjoinSurfaces(surf, sfl, av1, av2);

            if (Do_Surf_Overlap(surf, nsfl))
                SysAdjoinSurfaces(nsfl, surf, av2, av1);

            if (Do_Surf_Overlap(nsf, sfl))
                SysAdjoinSurfaces(nsf, sfl, av1, av2);

            if (Do_Surf_Overlap(nsf, nsfl))
                SysAdjoinSurfaces(nsf, nsfl, av1, av2);

            GlobalVars.NedMain.SectorChanged(sfl.sector, true);
        }

        public static bool Do_Surf_Overlap(JKSurface surf, JKSurface surfl)
        {
            if (surf == null || surfl == null)
                return false;

            Vertices vxs0 = new Vertices();
            Vertices vxs1 = new Vertices();

            for (int i = 0; i < surf.vertices.Count; i++)
            {
                Vertex v0 = surf.vertices[surf.PrevVX(i)];
                Vertex v1 = surf.vertices[i];
                Vertex v2 = surf.vertices[surf.NextVX(i)];
                Vec3 e1 = (v1.v - v0.v).Unit;
                Vec3 e2 = (v2.v - v1.v).Unit;
                if (e1.Magnitude == 0.0)
                    continue;

                if (e1 == e2)
                    continue;

                vxs0.Add(v1);
            }

            for (int i = 0; i < surfl.vertices.Count; i++)
            {
                Vertex v0 = surfl.vertices[surfl.PrevVX(i)];
                Vertex v1 = surfl.vertices[i];
                Vertex v2 = surfl.vertices[surfl.NextVX(i)];
                Vec3 e1 = (v1.v - v0.v).Unit;
                Vec3 e2 = (v2.v - v1.v).Unit;
                if (e1.Magnitude == 0.0)
                    continue;

                if (e1 == e2)
                    continue;

                vxs1.Add(v1);
            }

            {
                if (vxs0.Count != vxs1.Count)
                    return false;

                Vertex v0 = vxs0[0];

                int fv = -1;
                for (int v = 0; v < vxs1.Count; v++)
                {
                    Vertex v1 = vxs1[v];
                    if (v0.v == v1.v)
                    {
                        fv = v;
                        break;
                    }
                }

                if (fv == -1)
                    return false;

                fv = PrevIdx(fv, vxs1.Count);
                for (int v = 1; v < vxs0.Count; v++)
                {
                    v0 = vxs0[v];
                    Vertex v1 = vxs1[fv];
                    if (v0.v != v1.v)
                        return false;
                    fv = PrevIdx(fv, vxs1.Count);
                }
            }

            return true;
        }

        public class TXData
        {
            public Vec3 v;
            public Vec2 tv;
            public Vec3 un;
            public Vec3 vn;
        }

        static TXData SaveTXData(JKSurface surf, int vx)
        {
            TXData txd = new TXData();
            txd.v = surf.vertices[vx].v;
            txd.tv = surf.texvertices[vx].v;
            surf.CalcUVNormals(out txd.un, out txd.vn);
            return txd;
        }

        public static void CreateCubeSec(JKSector s, Vec3 p, Vec3 pnormal, Vec3 edge)
        {
            JKSector nsec = s;
            JKSurface asurf = nsec.NewSurface();
            nsec.surfaces.Add(asurf);

            Vec3 enormal = (pnormal ^ edge).Unit;

            JKVertex nvx = nsec.NewVertex();
            nvx.v = p;
            asurf.AddVertex(nvx);
            nvx = nsec.NewVertex();
            nvx.v = p + edge;
            asurf.AddVertex(nvx);

            double a = edge.Magnitude;
            nvx = nsec.NewVertex();
            nvx.v = p + edge + a * enormal;
            asurf.AddVertex(nvx);

            nvx = nsec.NewVertex();
            nvx.v = p + a * enormal;
            asurf.AddVertex(nvx);

            asurf.NewRecalcAll();

            // construct opposite surface
            JKSurface osurf = nsec.NewSurface();
            osurf.Assign(asurf);
            for (int v = asurf.vertices.Count - 1; v >= 0; v--)
            {
                nvx = nsec.NewVertex();
                nvx.v = asurf.vertices[v].v + asurf.normal * a;
                osurf.AddVertex(nvx);
            }
            nsec.surfaces.Add(osurf);

            // construct other surfaces
            int vc = asurf.vertices.Count;
            JKSurface nsurf = nsec.NewSurface();
            nsurf.Assign(asurf);
            for (int i = 0; i < vc; i++)
            {
                nsurf = nsec.NewSurface();// oh ok, we'll just leak the one we just made before the loop? lol
                nsurf.Assign(asurf);

                nsurf.AddVertex(asurf.vertices[i] as JKVertex/*hope so*/);

                if (i == 0)
                    nsurf.AddVertex(asurf.vertices[vc - 1] as JKVertex/*hope so*/);
                else
                    nsurf.AddVertex(asurf.vertices[i - 1] as JKVertex/*hope so*/);

                if (i == 0)
                    nsurf.AddVertex(osurf.vertices[0] as JKVertex/*hope so*/);
                else
                    nsurf.AddVertex(osurf.vertices[vc - i] as JKVertex/*hope so*/);

                nsurf.AddVertex(osurf.vertices[vc - i - 1] as JKVertex/*hope so*/);
                nsec.surfaces.Add(nsurf);
            }

            foreach (JKSurface surf in nsec.surfaces)
                surf.NewRecalcAll();
            nsec.Renumber();
        }

        public class CleaveData
        {
            public JKVertex vx = null;
        }

        // my word this whole thing is friggin disgusting
        public static JKSurface SysCleaveSurface(JKSurface surf, Vec3 cnormal, Vec3 c)
        {
            TXData txd = SaveTXData(surf, 0);

            double d = cnormal * c;

            JKLevel lev = surf.sector.level;

            int vc = surf.vertices.Count;

            CleaveData[] cd = new CleaveData[1000];
            for (int x = 0; x < cd.Length; x++)
                cd[x] = new CleaveData();

            for (int sf = 0; sf < vc; sf++)
            {
                CleaveData cd1 = cd[sf];
                CleaveData cd2 = cd[NextIdx(sf, vc)];

                Vertex v1 = surf.vertices[sf];
                Vertex v2 = surf.vertices[NextIdx(sf, vc)];

                if (v1.mark == 0 || v1.mark != -v2.mark)
                    continue;

                Vec3 a;
                PlaneLineXn(cnormal, d, v1.v, v2.v, out a);
                cd1.vx = surf.sector.AddVertex(a);
            }

            int nleft = 0;
            int nright = 0;
            foreach (Vertex v in surf.vertices)
                if (v.mark == 1)
                    nright++;
                else if (v.mark == -1)
                    nleft++;

            if (nleft == 0 || nright == 0)
                return null;

            JKSurface nsf = surf.sector.NewSurface();
            nsf.Assign(surf);

            for (int sf = 0; sf < surf.vertices.Count; sf++)
            {
                Vertex v1 = surf.vertices[sf];
                if (v1.mark == 1 || v1.mark == 0)
                    nsf.AddVertex(v1 as JKVertex/*right?*/);

                if (cd[sf].vx != null)
                    nsf.AddVertex(cd[sf].vx);
            }
            surf.sector.surfaces.Add(nsf);
            nsf.NewRecalcAll();

            Vertices vxs = surf.vertices;

            surf.vertices = new Vertices();

            for (int sf = 0; sf < vxs.Count; sf++)
            {
                Vertex v1 = vxs[sf];
                if (v1.mark == -1 || v1.mark == 0)
                    surf.vertices.Add(v1);
                if (cd[sf].vx != null)
                    surf.InsertVertex(surf.vertices.Count, cd[sf].vx);
            }

            Vec3 rvec = surf.normal ^ cnormal;

            for (int i = surf.vertices.Count - 1; i >= 0; i--)
            {
                Vertex v0 = surf.vertices[surf.PrevVX(i)];
                Vertex v1 = surf.vertices[i];
                Vertex v2 = surf.vertices[surf.NextVX(i)];
                if (v0.mark != 0 || v1.mark != 0 || v2.mark != 0)
                    continue;
                if ((v2.v - v1.v) * rvec < 0.0 ||
                    (v1.v - v0.v) * rvec < 0.0)
                    surf.vertices.RemoveAt(i);// what about texture coords DERPP
            }

            for (int i = nsf.vertices.Count - 1; i >= 0; i--)
            {
                Vertex v0 = nsf.vertices[nsf.PrevVX(i)];
                Vertex v1 = nsf.vertices[i];
                Vertex v2 = nsf.vertices[nsf.NextVX(i)];
                if (v0.mark != 0 || v1.mark != 0 || v2.mark != 0)
                    continue;

                if ((v2.v - v1.v) * rvec > 0.0 ||
                    (v1.v - v0.v) * rvec > 0.0)
                    nsf.vertices.RemoveAt(i);// what about texture coords DERPP
            }

            for (int sf = 0; sf < vc; sf++)
            {
                CleaveData pcd = cd[sf];
                if (pcd.vx == null)
                    continue;

                for (int i = 0; i < surf.sector.surfaces.Count; i++)
                {
                    JKSurface asurf = surf.sector.surfaces[i];
                    if (asurf == surf)
                        continue;

                    TXData txd1 = SaveTXData(asurf, 0);
                    bool sfchange = false;
                    for (int j = asurf.vertices.Count - 1; j >= 0; j--)
                    {
                        Vertex v1 = asurf.vertices[j];
                        Vertex v2 = asurf.vertices[NextIdx(j, asurf.vertices.Count)];
                        if (object.ReferenceEquals(v1, vxs[sf]) && object.ReferenceEquals(v2, vxs[NextIdx(sf, vxs.Count)]))
                        {
                            sfchange = true;
                            asurf.InsertVertex(asurf.NextVX(j), pcd.vx);
                            InsertVXTIntoSurf(asurf.adjoin, vxs[sf] as JKVertex/*hope so*/, vxs[NextIdx(sf, vxs.Count)] as JKVertex/*hope so*/, pcd.vx);
                        }

                        if (object.ReferenceEquals(v2, vxs[sf]) && object.ReferenceEquals(v1, vxs[NextIdx(sf, vxs.Count)]))
                        {
                            sfchange = true;
                            asurf.InsertVertex(NextIdx(j, asurf.vertices.Count), pcd.vx);
                            InsertVXTIntoSurf(asurf.adjoin, vxs[sf] as JKVertex/*hope so*/, vxs[NextIdx(sf, vxs.Count)] as JKVertex/*hope so*/, pcd.vx);
                        }

                        if (sfchange)
                            ApplyTXData(asurf, txd1);
                    }
                }
            }

            nsf.sector.Renumber();
            ApplyTXData(surf, txd);
            ApplyTXData(nsf, txd);

            return nsf;
        }

        public static void InsertVXTIntoSurf(JKSurface surf, JKVertex v1, JKVertex v2, JKVertex vx)
        {
            if (surf == null)
                return;

            for (int j = 0; j < surf.sector.surfaces.Count; j++)
            {
                JKSurface asurf = surf.sector.surfaces[j];

                bool changed = false;
                TXData txd = SaveTXData(asurf, 0);

                for (int i = asurf.vertices.Count - 1; i >= 0; i--)
                {
                    Vertex v3 = asurf.vertices[i];
                    Vertex v4 = asurf.vertices[asurf.NextVX(i)];

                    if (VXEq(v3, v1) && VXEq(v4, v2))
                    {
                        asurf.InsertVertex(asurf.NextVX(i), asurf.sector.AddVertex(vx.v));
                        changed = true;
                    }

                    if (VXEq(v3, v2) && VXEq(v4, v1))
                    {
                        asurf.InsertVertex(asurf.NextVX(i), asurf.sector.AddVertex(vx.v));
                        changed = true;
                    }
                }

                if (changed)
                    ApplyTXData(asurf, txd);
            }

            surf.sector.Renumber();
            GlobalVars.NedMain.SectorChanged(surf.sector, true);
        }

        public static bool VXEq(Vertex v1, Vertex v2)
        {
            return (v1.v == v2.v);
        }

        static void ApplyTXData(JKSurface surf, TXData txd)
        {
            surf.ArrangeTextureBy(txd.un, txd.vn, txd.v, txd.tv);
        }

        public static void RelayerSCThings(JKSector sec)
        {
            JKLevel lev = sec.level;
            foreach (JKThing thing in lev.things)
            {
                if (thing.sec != sec)
                    continue;

                int n = FindSectorForThing(thing);
                if (n == -1)
                    thing.sec = null;
                else
                    thing.sec = lev.sectors[n];
            }
        }

        public static bool SysIsInSector(JKSector sec, Vec3 pos, bool onsides)
        {
            foreach (JKSurface surf in sec.surfaces)
            {
                double dist = surf.normal * (pos - surf.vertices[0].v);

                if (onsides)
                {
                    if (dist < -0.00001)
                        return false;
                }
                else
                {
                    if (dist <= 0.001)
                        return false;
                }
            }

            return true;
        }

        public static bool IsInSector(JKSector sec, Vec3 pos)
        {
            return SysIsInSector(sec, pos, true);
        }

        public static int FindSectorForThing(JKThing thing)
        {
            for (int i = 0; i < thing.level.sectors.Count; i++)
                if (IsInSector(thing.level.sectors[i], thing.pos))
                    return i;

            return -1;
        }

        public class AdjoinVals
        {
            public Surface.Flag surfaceFlags;
            public Surface.FaceFlag faceFlags;
            public Surface.GeoModeType geo;
            public Adjoin.Flag adjoinFlags;
        }

        public static AdjoinVals GetAdjoinParams(JKSurface sf)
        {
            AdjoinVals v = new AdjoinVals();

            v.surfaceFlags = sf.surfaceFlags;
            v.faceFlags = sf.faceFlags;
            v.geo = sf.geo;
            v.adjoinFlags = sf.adjoinFlags;

            return v;
        }

        public static void SetAdjoinParams(JKSurface sf, AdjoinVals av)
        {
            if (av == null)
            {
                SetAdjoinParams(sf);
                return;
            }

            sf.surfaceFlags = av.surfaceFlags;
            sf.faceFlags = av.faceFlags;
            sf.SetGeo("SetAdjoinParams AdjoinVals", av.geo);
            sf.adjoinFlags = av.adjoinFlags;
        }

        public static void SetAdjoinParams(JKSurface sf)
        {
            sf.surfaceFlags = 0;
            sf.faceFlags = 0;
            sf.SetGeo("SetAdjoinParams", Surface.GeoModeType.NotRendered);
            sf.adjoinFlags = Adjoin.Flag.Passable | Adjoin.Flag.AllowsSound | Adjoin.Flag.RenderPastAdjoin;
        }

        public static void ClearAdjoinParams(JKSurface sf)
        {
            sf.surfaceFlags = Surface.Flag.Impassible;
            sf.faceFlags = Surface.FaceFlag.Unknown;
            sf.SetGeo("ClearAdjoinParams", Surface.GeoModeType.Textured);
            if (string.IsNullOrEmpty(sf.material))
                sf.material = "dflt.mat";
        }

        public static bool IsTXFlipped(JKSurface surf)
        {
            int nv1, nv2, nv3;
            nv1 = 0;
            if (!FindTri(surf, nv1, out nv2, out nv3))
                return false;

            Vec3 v1 = surf.vertices[nv1].v;
            Vec3 v2 = surf.vertices[nv2].v;
            Vec3 v3 = surf.vertices[nv3].v;
            Vec2 tv1 = surf.texvertices[nv1].v;
            Vec2 tv2 = surf.texvertices[nv2].v;
            Vec2 tv3 = surf.texvertices[nv3].v;

            return (tv1.x - tv2.x) * (tv3.y - tv2.y) > (tv3.x - tv2.x) * (tv1.y - tv2.y);
        }

        public static bool FindTri(JKSurface surf, int nv1, out int nv2, out int nv3)
        {
            nv2 = -1;
            nv3 = -1;
            if (surf.vertices.Count < 3)
                return false;

            Vec3 v0 = surf.vertices[nv1].v;
            int nvxs = surf.vertices.Count - 1;

            int n = surf.NextVX(nv1);

            double cdist = -1.0;
            int cn = -1;

            bool res = false;
            for (int i = 0; i < nvxs; i++)
            {
                Vec3 v1 = surf.vertices[n].v;

                double dist = Math.Sqrt(v1.x-v0.x)+Math.Sqrt(v1.y-v0.y)+Math.Sqrt(v1.z-v0.z);
                if (dist > cdist)
                {
                    cn = n;
                    cdist = dist;
                }

                if (dist > 0.0001)
                {
                    res = true;
                    nv2 = n;
                    break;
                }

                n = surf.NextVX(n);
                nvxs--;
            }

            if (!res)
            {
                if (cn == -1)
                    return false;
                else
                    nv2 = cn;
            }

            {
                Vec3 v1 = surf.vertices[nv2].v;
                res = false;
                n = surf.NextVX(nv2);

                Vec3 a = (v1 - v0) ^ surf.normal;
                if (MathLib.CompareScalar(a.Magnitude, 0.0) == 0)
                    return false;

                cdist = -1.0;
                cn = -1;
                for (int i = 0; i < nvxs; i++)
                {
                    Vec3 v2 = surf.vertices[n].v;

                    double dist = Math.Abs(a * (v2-v1));

                    if (dist > cdist)
                    {
                        cn = n;
                        cdist = dist;
                    }

                    if (dist > 0.0001)
                    {
                        res = true;
                        nv3 = n;
                        break;
                    }

                    n = surf.NextVX(n);
                }
            }

            if (!res)
            {
                if (cn != -1)
                    nv3 = cn;
                else
                    return false;
            }

            return true;
        }

        public static Mat4 CreateRotMatrix(double pitch, double yaw, double roll)
        {
            // i have no idea why we are negating these values..
            // in fact, if i remove the negation then the old map rotate mode seems to work properly..
            // so maybe we'll experiment a bit later

#if true
            // yeah i dunno after implementing the R to rotate  this works perfectly.. so i guess we'll keep it

            Mat4 mx = Mat4.RotationZ(MathLib.ToRad(yaw));
            Mat4 tmx = Mat4.RotationX(MathLib.ToRad(pitch));
            mx *= tmx;
            tmx = Mat4.RotationY(MathLib.ToRad(roll));
            mx *= tmx;
            return mx;
#else
            Mat4 mx = Mat4.RotationZ(MathLib.ToRad(-yaw));
            Mat4 tmx = Mat4.RotationX(MathLib.ToRad(-pitch));
            mx *= tmx;
            tmx = Mat4.RotationY(MathLib.ToRad(-roll));
            mx *= tmx;
            return mx;
#endif
            /* CreateMatrix(mx,-YAW,rt_z);

 CreateMatrix(tmx,-PCH,rt_x);
 MultM3(mx,tmx);

 CreateMatrix(tmx,-ROL,rt_y);
 MultM3(mx,tmx);*/
        }

        public static void RotateVector(ref Vec3 v, double pitch, double yaw, double roll)
        {
            v.Transform(CreateRotMatrix(pitch, yaw, roll));
        }

        // is this seriously just plane/line-segment intersection?   this ought to be repealed and replaced
        public static bool PlaneLineXn(Vec3 normal, double sd, Vec3 v1, Vec3 v2, out Vec3 v)
        {
            Vec3 d = v2 - v1;
            v = new Vec3();

            double sm = normal * d;

            if (MathLib.CompareScalar(d.x, 0.0) != 0)
            {
                v.x = (v1.x * (normal.y * d.y + normal.z * d.z) - d.x * (normal.y * v1.y + normal.z * v1.z - sd)) / sm;
                v.y = v1.y + d.y / d.x * (v.x - v1.x);
                v.z = v1.z + d.z / d.x * (v.x - v1.x);
                return true;
            }
            else if (MathLib.CompareScalar(d.y, 0.0) != 0)
            {
                v.y = (v1.y * (normal.x * d.x + normal.z * d.z) - d.y * (normal.x * v1.x + normal.z * v1.z - sd)) / sm;
                v.x = v1.x + d.x / d.y * (v.y - v1.y);
                v.z = v1.z + d.z / d.y * (v.y - v1.y);
                return true;
            }
            else if (MathLib.CompareScalar(d.z, 0.0) != 0)
            {
                v.z = (v1.z * (normal.x * d.x + normal.y * d.y) - d.z * (normal.x * v1.x + normal.y * v1.y - sd)) / sm;
                v.y = v1.y + d.y / d.z * (v.z - v1.z);
                v.x = v1.x + d.x / d.z * (v.z - v1.z);
                return true;
            }
            else
                return false;
        }

        public static bool PlaneLineXnNew(Vec3 normal, Vec3 p, Vec3 a1, Vec3 a2, out Vec3 r)
        {
            double dist1 = (a1 - p) * normal;
            double dist2 = (a2 - p) * normal;
            if (MathLib.CompareScalar(dist1, dist2) == 0)
            {
                r = Vec3.Zero;
                return false;
            }

            double k = dist1 / (dist1 - dist2);

            r = a1 + k * (a2 - a1);
            return true;
        }

        public static double Sqr(double d)
        {
            return d * d;
        }

        public static bool GetNearestVX(JKLevel level, ref Vec3 v, double mdist)
        {
            Vec3 o = v;
            bool ret = false;
            double cmdist = mdist;
            foreach(JKSector sec in level.sectors)
                foreach (JKVertex vert in sec.vertices)
                {
                    double cdist = (vert.v - o).Magnitude;
                    if (cdist <= cmdist)
                    {
                        cmdist = cdist;
                        v = vert.v;
                        ret = true;
                    }
                }

            return ret;
        }

        public static Box3 FindBBox(JKSector sec)
        {
            Box3 box = Box3.Expandable;
            foreach (JKVertex vert in sec.vertices)
                box.Expand(vert.v);
            return box;
        }

        public static void DuplicateSector(JKSector sec, JKSector newsc, Vec3 c, Vec3 nx, Vec3 ny, Vec3 nz, Vec3 d)
        {
            newsc.Assign(sec);

            foreach (JKVertex v in sec.vertices)
            {
                JKVertex nv = newsc.NewVertex();
                nv.v.x = (v.v.x - c.x) * nx.x + (v.v.y - c.y) * ny.x + (v.v.z - c.z) * nz.x + d.x + c.x;
                nv.v.y = (v.v.x - c.x) * nx.y + (v.v.y - c.y) * ny.y + (v.v.z - c.z) * nz.y + d.y + c.y;
                nv.v.z = (v.v.x - c.x) * nx.z + (v.v.y - c.y) * ny.z + (v.v.z - c.z) * nz.z + d.z + c.z;
            }

            sec.Renumber();
            foreach (JKSurface sf in sec.surfaces)
            {
                JKSurface sfl = newsc.NewSurface();
                sfl.Assign(sf);
                if (sf.adjoin != null)
                    sfl.nadj = 1;
                newsc.surfaces.Add(sfl);
                for (int j = 0; j < sf.vertices.Count; j++)
                {
                    Vertex v = sf.vertices[j];
                    sfl.AddVertex(newsc.vertices[v.num]);
                    sfl.texvertices[j].Assign(sf.texvertices[j]);
                }
            }

            foreach (JKSurface surf in newsc.surfaces)
                surf.RecalcAll();
            newsc.Renumber();
        }

        public static double JKRound(double d)
        {
            return Math.Round(d * 10000.0) / 10000.0;
        }

        public static void JKRound(ref Vec3 v)
        {
            v.x = JKRound(v.x);
            v.y = JKRound(v.y);
            v.z = JKRound(v.z);
        }

        public static double UVRound(double d)
        {
            return Math.Round(d*1000.0) / 1000.0;
        }

        public static bool IsInContainer(string fname)
        {
            return (fname.IndexOf('>') != -1);
        }

        public enum MessageType
        {
            Info,
            Warning,
            Error
        }

        public static void PanMessage(MessageType type, string msg)
        {
            switch (type)
            {
                case MessageType.Info:
                    if (GlobalVars.ShowInfo)
                        MsgForm.Add("Info: " + msg);

                    GlobalVars.NedMain.PMsg.Text = msg;
                    GlobalVars.NedMain.PMsg.ForeColor = System.Drawing.Color.Black;
                    break;

                case MessageType.Warning:
                    if (GlobalVars.ShowWarnings)
                        MsgForm.Add("Warning: " + msg);

                    GlobalVars.NedMain.PMsg.Text = msg;
                    GlobalVars.NedMain.PMsg.ForeColor = System.Drawing.Color.Red;
                    //MsgForm.Show();
                    GlobalVars.MsgForm.Show();
                    break;

                case MessageType.Error:
                    System.Windows.Forms.MessageBox.Show(msg, "Error", System.Windows.Forms.MessageBoxButtons.OK);
                    MsgForm.Add("Error: " + msg);
                    break;
            }
        }
    }
}
